Type: Package
Package: MDP2
Title: Markov Decision Processes (MDPs)
Version: 2.2.2.0
Authors@R: c(
    person("Lars", "Relund Nielsen", , "lars@relund.dk", role = c("aut", "cre"),
           comment = c(ORCID = "0000-0002-4802-3071")))
Author: Lars Relund Nielsen [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-4802-3071>)
Maintainer: Lars Relund Nielsen <lars@relund.dk>
Description: Create and optimize (semi) MDPs with discrete time steps and
    state space. Both hierarchical and ordinary-traditional MDPs can be
    modeled.
License: GPL (>= 3)
URL: https://relund.github.io/mdp/, https://github.com/relund/mdp/,
        http://relund.github.io/mdp/
BugReports: https://github.com/relund/mdp/issues
Depends: R (>= 4.1.0)
Imports: diagram, dplyr, stringr, tidyr, magrittr, methods, purrr,
        rlang, tibble, Rcpp (>= 0.11.5)
Suggests: knitr, Matrix, rmarkdown, testthat (>= 3.0.0), readr, xml2,
        covr, roxygen2
LinkingTo: Rcpp, RcppArmadillo
VignetteBuilder: knitr
Encoding: UTF-8
Language: en-US
Config/roxygen2/version: 8.0.0
Config/testthat/edition: 3
NeedsCompilation: yes
Packaged: 2026-06-04 09:51:52 UTC; au15463
Repository: CRAN
Date/Publication: 2026-06-12 19:30:14 UTC
Built: R 4.5.2; aarch64-apple-darwin20; 2026-06-12 20:15:05 UTC; unix
Archs: MDP2.so.dSYM

# Generated by roxygen2: do not edit by hand

S3method(plot,HMDP)
export(binaryActionWriter)
export(binaryMDPWriter)
export(convertBinary2HMP)
export(convertHMP2Binary)
export(getBinInfoActions)
export(getBinInfoStates)
export(getHypergraph)
export(getInfo)
export(getPolicy)
export(getRPO)
export(getSteadyStatePr)
export(getWIdx)
export(hmpMDPWriter)
export(loadMDP)
export(memoryMDPWriter)
export(plotHypergraph)
export(randomHMDP)
export(runCalcWeights)
export(runPolicyIteAve)
export(runPolicyIteDiscount)
export(runValueIte)
export(saveMDP)
export(setPolicy)
import(diagram)
importClassesFrom(Rcpp,"C++Object")
importFrom(magrittr,"%>%")
importFrom(rlang,.data)
useDynLib(MDP2)

<!-- README.md is generated from README.Rmd. Please edit that file -->
<!-- badges: start -->

[![R-CMD-check](https://github.com/relund/mdp/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/relund/mdp/actions/workflows/R-CMD-check.yaml)
[![Codecov test
coverage](https://codecov.io/gh/relund/mdp/branch/master/graph/badge.svg)](https://app.codecov.io/gh/relund/mdp?branch=master)
<!-- badges: end -->

# Markov Decision Processes (MDPs) in R

The `MDP2` package in R is a package for solving Markov decision
processes (MDPs) with discrete time-steps, states and actions. Both
traditional MDPs (Puterman 1994), semi-Markov decision processes
(semi-MDPs) (Tijms 2003) and hierarchical-MDPs (HMDPs) (Kristensen and
Jørgensen 2000) can be solved under a finite and infinite time-horizon.

Building and solving an MDP is done in two steps. First, the MDP is
built and saved in a set of binary files. Next, you load the MDP into
memory from the binary files and apply various algorithms to the model.

The package implement well-known algorithms such as policy iteration and
value iteration under different criteria e.g. average reward per time
unit and expected total discounted reward. The model is stored using an
underlying data structure based on the *state-expanded directed
hypergraph* of the MDP (Nielsen and Kristensen (2006)) implemented in
`C++` for fast running times.

## Installation

Install the latest stable release from CRAN:

``` r
install.packages("MDP2")
```

Alternatively, install the latest development version from GitHub
(recommended):

``` r
remotes::install_github("relund/mdp")
```

We load the package using

``` r
library(MDP2)
```

Help about the package can be seen by writing

``` r
?MDP2
```

To illustrate the package capabilities, we use a few examples, namely,
an infinite and finite-horizon semi-MDP and a HMDP. Before each example
a short introduction to these models are given.

## Learning more

To get started, first read `vignette("MDP2")`.

For more examples see `example("MDP2")`.

## References

<div id="refs" class="references csl-bib-body hanging-indent">

<div id="ref-Kristensen00" class="csl-entry">

Kristensen, A. R., and E. Jørgensen. 2000. “Multi-Level Hierarchic
Markov Processes as a Framework for Herd Management Support.” *Annals of
Operations Research* 94: 69–89.
<https://doi.org/10.1023/A:1018921201113>.

</div>

<div id="ref-Relund06" class="csl-entry">

Nielsen, L. R., and A. R. Kristensen. 2006. “Finding the $K$ Best
Policies in a Finite-Horizon Markov Decision Process.” *European Journal
of Operational Research* 175 (2): 1164–79.
<https://doi.org/10.1016/j.ejor.2005.06.011>.

</div>

<div id="ref-Puterman94" class="csl-entry">

Puterman, M. L. 1994. *Markov Decision Processes*. Wiley Series in
Probability and Mathematical Statistics. Wiley-Interscience.

</div>

<div id="ref-Tijms03" class="csl-entry">

Tijms, Henk. C. 2003. *A First Course in Stochastic Models*. John Wiley
& Sons Ltd.

</div>

</div>

## ----setup, include=FALSE---------------------------------------------------------------
library(knitr)
options(rmarkdown.html_vignette.check_title = FALSE,
        # formatR.arrow = TRUE, 
        # scipen=999, 
        # digits=5,
        width=90) 
#thm <- knit_theme$get("edit-kwrite")   # whitengrey, bright, print, edit-flashdevelop, edit-kwrite
#knit_theme$set(thm)
knitr::opts_chunk$set(
   # collapse = TRUE,
   comment = "#>",
   fig.align = 'center',
   fig.width = 9,
   fig.height = 5,
   fig.show = 'hold',
   out.extra = 'style="max-width:100%;"',
   # tidy = TRUE,
   # prompt=T,
   # comment=NA,
   cache = F
   # background = "red"
)
library(magrittr)
library(dplyr)

## ----setup, include=FALSE---------------------------------------------------------------
library(knitr)
options(rmarkdown.html_vignette.check_title = FALSE,
        # formatR.arrow = TRUE, 
        # scipen=999, 
        # digits=5,
        width=90) 
#thm <- knit_theme$get("edit-kwrite")   # whitengrey, bright, print, edit-flashdevelop, edit-kwrite
#knit_theme$set(thm)
knit_hooks$set(
   par = function(before, options, envir) {
      if (before && options$fig.show != 'none')
         par(mar = c(0, 0, 0, 0), # bottom, left, top, and right
             oma = c(0, 0, 0, 0))}
)
knitr::opts_chunk$set(
   # collapse = TRUE,
   comment = "#>",
   fig.align = 'center',
   fig.width = 9,
   fig.height = 5,
   fig.show = 'hold',
   out.extra = 'style="max-width:100%;"',
   # tidy = TRUE,
   # prompt=T,
   # comment=NA,
   cache = F
   # background = "red"
)
library(magrittr)
library(dplyr)

## ---------------------------------------------------------------------------------------
library(MDP2)

## ----echo=FALSE, fig.cap="The state-expanded hypergraph of the first stage of a hierarchical MDP. Level 0 indicate the founder level, and the nodes indicates states at the different levels. A child process (oval box) is represented using its state-expanded hypergraph (hyperarcs not shown) and is uniquely defined by a given state and action of its parent process."----
knitr::include_graphics("vignette_files/hmdp_index.png")

## ----par=TRUE---------------------------------------------------------------------------
prefix <- paste0(system.file("models", package = "MDP2"), "/cow_")
mdp <- loadMDP(prefix)
mdp 

## ----plotHMDP, message=FALSE, par=TRUE--------------------------------------------------
hgf <- getHypergraph(mdp)
## Rename labels
dat <- hgf$nodes %>% 
   dplyr::mutate(label = dplyr::case_when(
      label == "Low yield" ~ "L",
      label == "Avg yield" ~ "A",
      label == "High yield" ~ "H",
      label == "Dummy" ~ "D",
      label == "Bad genetic level" ~ "Bad",
      label == "Avg genetic level" ~ "Avg",
      label == "Good genetic level" ~ "Good",
      TRUE ~ "Error"
   ))
## Set grid id
dat$gId[1:3]<-85:87
dat$gId[43:45]<-1:3
getGId<-function(process,stage,state) {
   if (process==0) start=18
   if (process==1) start=22
   if (process==2) start=26
   return(start + 14 * stage + state)
}
idx<-43
for (process in 0:2)
   for (stage in 0:4)
      for (state in 0:2) {
         if (stage==0 & state>0) break
         idx<-idx-1
         #cat(idx,process,stage,state,getGId(process,stage,state),"\n")
         dat$gId[idx]<-getGId(process,stage,state)
      }
hgf$nodes <- dat
## Rename labels
dat <- hgf$hyperarcs %>% 
   dplyr::mutate(label = dplyr::case_when(
      label == "Replace" ~ "R",
      label == "Keep" ~ "K",
      label == "Dummy" ~ "D",
      TRUE ~ "Error"
      ),
      col = dplyr::case_when(
         label == "R" ~ "deepskyblue3",
         label == "K" ~ "darkorange1",
         label == "D" ~ "black",
         TRUE ~ "Error"
      ),
      lwd = 0.5,
      label = ""
   ) 
hgf$hyperarcs <- dat
## Make the plot
plotHypergraph(hgf, gridDim = c(14, 7), cex = 0.8, radx = 0.02, rady = 0.03)

## ----Optimize (cow)---------------------------------------------------------------------
wLbl<-"Net reward"         # the weight we want to optimize (net reward)
durLbl<-"Duration"         # the duration/time label
runPolicyIteDiscount(mdp, wLbl, durLbl, rate = 0.1)

## ----plotPolicy, results='hide', message=FALSE, par=TRUE--------------------------------
hgf$hyperarcs <- right_join(hgf$hyperarcs, getPolicy(mdp), by = c("sId", "aIdx"))
plotHypergraph(hgf, gridDim = c(14, 7), cex = 0.8, radx = 0.02, rady = 0.03)

## ----eval=FALSE, include=FALSE----------------------------------------------------------
# # getPolicy(mdp)
# # rpo<-calcRPO(mdp, wLbl, iA=rep(0,42), criterion="discount", dur=durLbl, rate=rate, rateBase=rateBase)
# # policy<-merge(policy,rpo)
# # policy

## ----avePerLac, tidy.opts=list(comment=FALSE)-------------------------------------------
wLbl<-"Net reward"         # the weight we want to optimize (net reward)
durLbl<-"Duration"         # the duration/time label
runPolicyIteAve(mdp, wLbl, durLbl)
getPolicy(mdp)

## ----echo=TRUE--------------------------------------------------------------------------
runCalcWeights(mdp, w=wLbl, criterion="average", dur = "Yield")

## ----Reward/piglet (sow rep), echo=TRUE-------------------------------------------------
runCalcWeights(mdp, w="Yield", criterion="average", dur = durLbl)

## ----Delete bin, include=FALSE----------------------------------------------------------
do.call(file.remove,list(list.files(pattern = ".bin")))

## ----setup, include=FALSE---------------------------------------------------------------
library(knitr)
options(rmarkdown.html_vignette.check_title = FALSE,
        # formatR.arrow = TRUE, 
        # scipen=999, 
        # digits=5,
        width=90) 
#thm <- knit_theme$get("edit-kwrite")   # whitengrey, bright, print, edit-flashdevelop, edit-kwrite
#knit_theme$set(thm)
knit_hooks$set(
   par = function(before, options, envir) {
      if (before && options$fig.show != 'none')
         par(mar = c(0, 0, 0, 0), # bottom, left, top, and right
             oma = c(0, 0, 0, 0))}
)
knitr::opts_chunk$set(
   # collapse = TRUE,
   comment = "#>",
   fig.align = 'center',
   fig.width = 9,
   fig.height = 5,
   fig.show = 'hold',
   out.extra = 'style="max-width:100%;"',
   # tidy = TRUE,
   # prompt=T,
   # comment=NA,
   cache = F
   # background = "red"
)
library(magrittr)
library(dplyr)

## ---------------------------------------------------------------------------------------
library(MDP2)

## ----par=TRUE---------------------------------------------------------------------------
prefix <- paste0(system.file("models", package = "MDP2"), "/machine1_")
mdp <- loadMDP(prefix)
getInfo(mdp, withList = F, dfLevel = "action", asStringsActions = TRUE)  

## ----par=TRUE---------------------------------------------------------------------------
plot(mdp, actionColor = "label", radx = 0.06, marX = 0.065, marY = 0.055)

## ----solve3-----------------------------------------------------------------------------
scrapValues <- c(30, 10, 5, 0)   # scrap values (the values of the 4 states at the last stage)
runValueIte(mdp, "Net reward", termValues = scrapValues)

## ----par=TRUE---------------------------------------------------------------------------
pol <- getPolicy(mdp)
tail(pol)
plot(mdp, actionsVisible = "policy", stateLabel = "weight", 
     radx = 0.06, marX = 0.065, marY = 0.055)

## ----Set policy (machine rep),echo=TRUE,eval=TRUE---------------------------------------
policy<-data.frame(sId=c(8,11), aIdx=c(0,0)) # set the policy for sId 8 and 11 to mt
setPolicy(mdp, policy)
getPolicy(mdp)

## ----Calc reward (machine rep),echo=TRUE------------------------------------------------
runCalcWeights(mdp, "Net reward", termValues = scrapValues)
tail(getPolicy(mdp))    

## Use temp dir
wd <- setwd(tempdir())

# Create a small HMDP with two levels
w<-binaryMDPWriter()
w$setWeights(c("Duration","Net reward","Items"))
w$process()
   w$stage()
      w$state(label="M0")
         w$action(label="A0",weights=c(0,0,0),prob=c(2,0,1))
            w$process()
               w$stage()
                  w$state(label="D")
                     w$action(label="A0",weights=c(0,0,1),prob=c(1,0,0.5,1,1,0.5))
                     w$endAction()
                  w$endState()
               w$endStage()
               w$stage()
                  w$state(label="C0")
                     w$action(label="A0",weights=c(0,0,0),prob=c(1,0,1))
                     w$endAction()
                     w$action(label="A1",weights=c(1,2,1),prob=c(1,0,0.5,1,1,0.5))
                     w$endAction()
                  w$endState()
                  w$state(label="C1")
                     w$action(label="A0",weights=c(0,0,0),prob=c(1,0,1))
                     w$endAction()
                     w$action(label="A1",weights=c(1,2,1),prob=c(1,0,0.5,1,1,0.5))
                     w$endAction()
                  w$endState()
               w$endStage()
               w$stage()
                  w$state(label="C0")
                     w$action(label="A0",weights=c(1,4,0),prob=c(0,0,1))
                     w$endAction()
                  w$endState()
                  w$state(label="C1")
                     w$action(label="A0",weights=c(1,4,0),prob=c(0,0,1))
                     w$endAction()
                  w$endState()
               w$endStage()
            w$endProcess()
         w$endAction()
         w$action(label="A1",weights=c(0,0,0),prob=c(2,0,1))
            w$process()
               w$stage()
                  w$state(label="D")
                     w$action(label="A0",weights=c(0,0,1),prob=c(1,0,1))
                     w$endAction()
                  w$endState()
               w$endStage()
               w$stage()
                  w$state(label="C0")
                     w$action(label="A0",weights=c(0,0,0),prob=c(1,0,1))
                     w$endAction()
                     w$action(label="A1",weights=c(1,2,1),prob=c(1,0,0.5,1,1,0.5))
                     w$endAction()
                  w$endState()
               w$endStage()
               w$stage()
                  w$state(label="C0")
                     w$action(label="A0",weights=c(1,4,0),prob=c(0,0,1))
                     w$endAction()
                  w$endState()
                  w$state(label="C1")
                     w$action(label="A0",weights=c(1,4,0),prob=c(0,0,1))
                     w$endAction()
                     w$action(label="A1",weights=c(0,10,5),prob=c(0,0,0.5,0,1,0.5))
                     w$endAction()
                  w$endState()
               w$endStage()
            w$endProcess()
         w$endAction()
      w$endState()
      w$state(label="M1")
         w$action(label="A0",weights=c(0,0,0),prob=c(2,0,1))
            w$process()
               w$stage()
                  w$state(label="D")
                     w$action(label="A0",weights=c(0,0,1),prob=c(1,0,0.5,1,1,0.5))
                     w$endAction()
                  w$endState()
               w$endStage()
               w$stage()
                  w$state(label="C0")
                     w$action(label="A0",weights=c(0,0,0),prob=c(1,0,1))
                     w$endAction()
                  w$endState()
                  w$state(label="C1")
                     w$action(label="A0",weights=c(0,0,0),prob=c(1,0,1))
                     w$endAction()
                  w$endState()
               w$endStage()
               w$stage()
                  w$state(label="C0")
                     w$action(label="A0",weights=c(1,4,0),prob=c(0,0,1))
                     w$endAction()
                  w$endState()
                  w$state(label="C1")
                     w$action(label="A0",weights=c(1,4,0),prob=c(0,0,1))
                     w$endAction()
                  w$endState()
               w$endStage()
            w$endProcess()
         w$endAction()
      w$endState()
   w$endStage()
w$endProcess()
w$closeWriter()

## Info about the binary files (don't have to load the model first)
getBinInfoStates()
getBinInfoActions()

## reset working dir
setwd(wd)

## Use temp dir
wd <- setwd(tempdir())

## Create a small HMDP with two levels
w<-hmpMDPWriter()
w$setWeights(c("Duration","Net reward","Items"), duration=1)
w$process()
  w$stage()
   w$state(label="M0")
     w$action(label="A0",weights=c(0,0,0),prob=c(2,0,1))
      w$process()
        w$stage()
         w$state(label="D")
           w$action(label="A0",weights=c(0,0,1),prob=c(1,0,0.5,1,1,0.5))
           w$endAction()
         w$endState()
        w$endStage()
        w$stage()
         w$state(label="C0")
           w$action(label="A0",weights=c(0,0,0),prob=c(1,0,1))
           w$endAction()
           w$action(label="A1",weights=c(1,2,1),prob=c(1,0,0.5,1,1,0.5))
           w$endAction()
         w$endState()
         w$state(label="C1")
           w$action(label="A0",weights=c(0,0,0),prob=c(1,0,1))
           w$endAction()
           w$action(label="A1",weights=c(1,2,1),prob=c(1,0,0.5,1,1,0.5))
           w$endAction()
         w$endState()
        w$endStage()
        w$stage()
         w$state(label="C0")
           w$action(label="A0",weights=c(1,4,0),prob=c(0,0,1), statesNext=0)
           w$endAction()
         w$endState()
         w$state(label="C1")
           w$action(label="A0",weights=c(1,4,0),prob=c(0,0,1), statesNext=0)
           w$endAction()
         w$endState()
        w$endStage()
      w$endProcess()
     w$endAction()
     w$action(label="A1",weights=c(0,0,0),prob=c(2,0,1))
      w$process()
        w$stage()
         w$state(label="D")
           w$action(label="A0",weights=c(0,0,1),prob=c(1,0,1))
           w$endAction()
         w$endState()
        w$endStage()
        w$stage()
         w$state(label="C0")
           w$action(label="A0",weights=c(0,0,0),prob=c(1,0,1))
           w$endAction()
           w$action(label="A1",weights=c(1,2,1),prob=c(1,0,0.5,1,1,0.5))
           w$endAction()
         w$endState()
        w$endStage()
        w$stage()
         w$state(label="C0")
           w$action(label="A0",weights=c(1,4,0),prob=c(0,0,1), statesNext=0)
           w$endAction()
         w$endState()
         w$state(label="C1")
           w$action(label="A0",weights=c(1,4,0),prob=c(0,0,1), statesNext=0)
           w$endAction()
           w$action(label="A1",weights=c(0,10,5),prob=c(0,0,0.5,0,1,0.5), statesNext=0)
           w$endAction()
         w$endState()
        w$endStage()
      w$endProcess()
     w$endAction()
   w$endState()
   w$state(label="M1")
     w$action(label="A0",weights=c(0,0,0),prob=c(2,0,1))
      w$process()
        w$stage()
         w$state(label="D")
           w$action(label="A0",weights=c(0,0,1),prob=c(1,0,0.5,1,1,0.5))
           w$endAction()
         w$endState()
        w$endStage()
        w$stage()
         w$state(label="C0")
           w$action(label="A0",weights=c(0,0,0),prob=c(1,0,1))
           w$endAction()
         w$endState()
         w$state(label="C1")
           w$action(label="A0",weights=c(0,0,0),prob=c(1,0,1))
           w$endAction()
         w$endState()
        w$endStage()
        w$stage()
         w$state(label="C0")
           w$action(label="A0",weights=c(1,4,0),prob=c(0,0,1), statesNext=0)
           w$endAction()
         w$endState()
         w$state(label="C1")
           w$action(label="A0",weights=c(1,4,0),prob=c(0,0,1), statesNext=0)
           w$endAction()
         w$endState()
        w$endStage()
      w$endProcess()
     w$endAction()
   w$endState()
  w$endStage()
w$endProcess()
w$closeWriter()

## Have a look at the hmp file
cat(readr::read_file("r.hmp"))

## Reset working dir
setwd(wd)

## Set working dir
wd <- setwd(tempdir())

# Create the small machine repleacement problem used as an example in L.R. Nielsen and A.R.
# Kristensen. Finding the K best policies in a finite-horizon Markov decision process. European
# Journal of Operational Research, 175(2):1164-1179, 2006. doi:10.1016/j.ejor.2005.06.011.

## Create the MDP using a dummy replacement node
prefix<-"machine1_"
w <- binaryMDPWriter(prefix)
w$setWeights(c("Net reward"))
w$process()
   w$stage()   # stage n=0
      w$state(label="Dummy")          # v=(0,0)
         w$action(label="buy", weights=-100, prob=c(1,0,0.7, 1,1,0.3), end=TRUE)
      w$endState()
   w$endStage()
   w$stage()   # stage n=1
      w$state(label="good")           # v=(1,0)
         w$action(label="mt", weights=55, prob=c(1,0,1), end=TRUE)
         w$action(label="nmt", weights=70, prob=c(1,0,0.6, 1,1,0.4), end=TRUE)
      w$endState()
      w$state(label="average")        # v=(1,1)
         w$action(label="mt", weights=40, prob=c(1,0,1), end=TRUE)
         w$action(label="nmt", weights=50, prob=c(1,1,0.6, 1,2,0.4), end=TRUE)
      w$endState()
   w$endStage()
   w$stage()   # stage n=2
      w$state(label="good")           # v=(2,0)
         w$action(label="mt", weights=55, prob=c(1,0,1), end=TRUE)
         w$action(label="nmt", weights=70, prob=c(1,0,0.5, 1,1,0.5), end=TRUE)
      w$endState()
      w$state(label="average")        # v=(2,1)
         w$action(label="mt", weights=40, prob=c(1,0,1), end=TRUE)
         w$action(label="nmt", weights=50, prob=c(1,1,0.5, 1,2,0.5), end=TRUE)
      w$endState()
      w$state(label="not working")    # v=(2,2)
         w$action(label="mt", weights=30, prob=c(1,0,1), end=TRUE)
         w$action(label="rep", weights=5, prob=c(1,3,1), end=TRUE)
      w$endState()
   w$endStage()
   w$stage()   # stage n=3
      w$state(label="good")           # v=(3,0)
         w$action(label="mt", weights=55, prob=c(1,0,1), end=TRUE)
         w$action(label="nmt", weights=70, prob=c(1,0,0.2, 1,1,0.8), end=TRUE)
      w$endState()
      w$state(label="average")        # v=(3,1)
         w$action(label="mt", weights=40, prob=c(1,0,1), end=TRUE)
         w$action(label="nmt", weights=50, prob=c(1,1,0.2, 1,2,0.8), end=TRUE)
      w$endState()
      w$state(label="not working")    # v=(3,2)
         w$action(label="mt", weights=30, prob=c(1,0,1), end=TRUE)
         w$action(label="rep", weights=5, prob=c(1,3,1), end=TRUE)
      w$endState()
      w$state(label="replaced")       # v=(3,3)
         w$action(label="Dummy", weights=0, prob=c(1,3,1), end=TRUE)
      w$endState()
   w$endStage()
   w$stage()   # stage n=4
      w$state(label="good", end=TRUE)        # v=(4,0)
      w$state(label="average", end=TRUE)     # v=(4,1)
      w$state(label="not working", end=TRUE) # v=(4,2)
      w$state(label="replaced", end=TRUE)    # v=(4,3)
   w$endStage()
w$endProcess()
w$closeWriter()

## Load the model into memory
mdp<-loadMDP(prefix)
mdp
plot(mdp)

getInfo(mdp, withList = FALSE)
getInfo(mdp, withList = FALSE, dfLevel = "action", asStringsActions = TRUE)
getInfo(mdp, withList = FALSE, dfLevel = "action", asStringsActions = FALSE)

## Perform value iteration
w<-"Net reward"             # label of the weight we want to optimize
scrapValues<-c(30,10,5,0)   # scrap values (the values of the 4 states at stage 4)
runValueIte(mdp, w, termValues=scrapValues)
getPolicy(mdp)     # optimal policy

## Calculate the weights of the policy always to maintain
library(magrittr)
policy <- getInfo(mdp, withList = FALSE, dfLevel = "action")$df %>% 
   dplyr::filter(label_action == "mt") %>% 
   dplyr::select(sId, aIdx)
setPolicy(mdp, policy)
runCalcWeights(mdp, w, termValues=scrapValues)
getPolicy(mdp)  



# The example given in L.R. Nielsen and A.R. Kristensen. Finding the K best
# policies in a finite-horizon Markov decision process. European Journal of
# Operational Research, 175(2):1164-1179, 2006. doi:10.1016/j.ejor.2005.06.011,
# does actually not have any dummy replacement node as in the MDP above. The same
# model can be created using a single dummy node at the end of the process.

## Create the MDP using a single dummy node
prefix<-"machine2_"
w <- binaryMDPWriter(prefix)
w$setWeights(c("Net reward"))
w$process()
   w$stage()   # stage n=0
      w$state(label="Dummy")          # v=(0,0)
         w$action(label="buy", weights=-100, prob=c(1,0,0.7, 1,1,0.3), end=TRUE)
      w$endState()
   w$endStage()
   w$stage()   # stage n=1
      w$state(label="good")           # v=(1,0)
         w$action(label="mt", weights=55, prob=c(1,0,1), end=TRUE)
         w$action(label="nmt", weights=70, prob=c(1,0,0.6, 1,1,0.4), end=TRUE)
      w$endState()
      w$state(label="average")        # v=(1,1)
         w$action(label="mt", weights=40, prob=c(1,0,1), end=TRUE)
         w$action(label="nmt", weights=50, prob=c(1,1,0.6, 1,2,0.4), end=TRUE)
      w$endState()
   w$endStage()
   w$stage()   # stage n=2
      w$state(label="good")           # v=(2,0)
         w$action(label="mt", weights=55, prob=c(1,0,1), end=TRUE)
         w$action(label="nmt", weights=70, prob=c(1,0,0.5, 1,1,0.5), end=TRUE)
      w$endState()
      w$state(label="average")        # v=(2,1)
         w$action(label="mt", weights=40, prob=c(1,0,1), end=TRUE)
         w$action(label="nmt", weights=50, prob=c(1,1,0.5, 1,2,0.5), end=TRUE)
      w$endState()
      w$state(label="not working")    # v=(2,2)
         w$action(label="mt", weights=30, prob=c(1,0,1), end=TRUE)
         w$action(label="rep", weights=5, prob=c(3,12,1), end=TRUE) # transition to sId=12 (Dummy)
      w$endState()
   w$endStage()
   w$stage()   # stage n=3
      w$state(label="good")           # v=(3,0)
         w$action(label="mt", weights=55, prob=c(1,0,1), end=TRUE)
         w$action(label="nmt", weights=70, prob=c(1,0,0.2, 1,1,0.8), end=TRUE)
      w$endState()
      w$state(label="average")        # v=(3,1)
         w$action(label="mt", weights=40, prob=c(1,0,1), end=TRUE)
         w$action(label="nmt", weights=50, prob=c(1,1,0.2, 1,2,0.8), end=TRUE)
      w$endState()
      w$state(label="not working")    # v=(3,2)
         w$action(label="mt", weights=30, prob=c(1,0,1), end=TRUE)
         w$action(label="rep", weights=5, prob=c(3,12,1), end=TRUE)
      w$endState()
   w$endStage()
   w$stage()   # stage n=4
      w$state(label="good")        # v=(4,0)
         w$action(label="rep", weights=30, prob=c(1,0,1), end=TRUE)
      w$endState()
      w$state(label="average")     # v=(4,1)
         w$action(label="rep", weights=10, prob=c(1,0,1), end=TRUE)
      w$endState()
      w$state(label="not working") # v=(4,2)
         w$action(label="rep", weights=5, prob=c(1,0,1), end=TRUE)
      w$endState()
   w$endStage()
   w$stage()   # stage n=5
      w$state(label="Dummy", end=TRUE)        # v=(5,0)
   w$endStage()
w$endProcess()
w$closeWriter()

## Have a look at the state-expanded hypergraph
mdp<-loadMDP(prefix)
mdp
plot(mdp)

getInfo(mdp, withList = FALSE)
getInfo(mdp, withList = FALSE, dfLevel = "action", asStringsActions = TRUE)
getInfo(mdp, withList = FALSE, dfLevel = "action", asStringsActions = FALSE)

## Perform value iteration
w<-"Net reward"             # label of the weight we want to optimize
runValueIte(mdp, w, termValues = 0)
getPolicy(mdp)     # optimal policy

## Calculate the weights of the policy always to maintain
library(magrittr)
policy <- getInfo(mdp, withList = FALSE, dfLevel = "action")$df %>% 
   dplyr::filter(label_action == "mt") %>% 
   dplyr::select(sId, aIdx)
setPolicy(mdp, policy)
runCalcWeights(mdp, w, termValues=scrapValues)
getPolicy(mdp)  


## Reset working dir
setwd(wd)

## Set working dir
wd <- setwd(system.file("models", package = "MDP2"))

#### A finite-horizon replacement problem ####
mdp<-loadMDP("machine1_")
plot(mdp)
plot(mdp, actionColor = "label")  # colors based on labels
plot(mdp, transLabels = "state")  # label transitions with target state labels
plot(mdp, transLabels = "prob")  # label transitions with transition probabilities
plot(mdp, actionColor = "label", stateLabel = "sId|label")  # state labels are 'sId | label'
plot(mdp, stateLabel = "sIdx|label", radx = 0.01)  # adjust radx in states
plot(mdp, stateLabel = "label", actionWLabel = "none", actionLabel = "label", 
     transLabels = "sId", radx = 0.01)

scrapValues <- c(30, 10, 5, 0)  # scrap values (the values of the 4 states at stage 4)
runValueIte(mdp, "Net reward" , termValues = scrapValues)
plot(mdp, actionColor = "policy")  # highlight optimal policy
plot(mdp, actionsVisible = "policy", stateLabel = "weight")  # show only optimal policy


#### An infinite-horizon maintenance problem ####
mdp<-loadMDP("hct611-1_")
plot(mdp)  # plot the first two stages
plot(mdp, actionColor = "label")  # colors based on labels
plot(mdp, actionColor = "label", stateLabel = "sId|label")  # state labels are 'sId | label'
runPolicyIteAve(mdp,"Net reward","Duration")
plot(mdp, actionColor = "policy")  # highlight optimal policy
plot(mdp, actionsVisible = "policy")  # show only optimal policy


#### An infinite-horizon hierarchical replacement problem ####
library(magrittr)
mdp<-loadMDP("cow_")
hgf <- getHypergraph(mdp)
# modify labels
dat <- hgf$nodes %>% 
   dplyr::mutate(label = dplyr::case_when(
      label == "Low yield" ~ "L",
      label == "Avg yield" ~ "A",
      label == "High yield" ~ "H",
      label == "Dummy" ~ "D",
      label == "Bad genetic level" ~ "Bad",
      label == "Avg genetic level" ~ "Avg",
      label == "Good genetic level" ~ "Good",
      TRUE ~ "Error"
   ))
# assign nodes to grid ids
dat$gId[1:3]<-85:87
dat$gId[43:45]<-1:3
getGId<-function(process,stage,state) {
   if (process==0) start=18
   if (process==1) start=22
   if (process==2) start=26
   return(start + 14 * stage + state)
}
idx<-43
for (process in 0:2)
   for (stage in 0:4)
      for (state in 0:2) {
         if (stage==0 & state>0) break
         idx<-idx-1
         #cat(idx,process,stage,state,getGId(process,stage,state),"\n")
         dat$gId[idx]<-getGId(process,stage,state)
      }
hgf$nodes <- dat
# modify labels
dat <- hgf$hyperarcs %>% 
   dplyr::mutate(label = dplyr::case_when(
      label == "Replace" ~ "R",
      label == "Keep" ~ "K",
      label == "Dummy" ~ "D",
      TRUE ~ "Error"
   ),
   col = dplyr::case_when(
      label == "R" ~ "deepskyblue3",
      label == "K" ~ "darkorange1",
      label == "D" ~ "black",
      TRUE ~ "Error"
   ),
   lwd = 0.5,
   label = ""
   ) 
hgf$hyperarcs <- dat
# plot hypergraph
oldpar <- par(mai = c(0, 0, 0, 0))
plotHypergraph(gridDim = c(14, 7), hgf, cex = 0.8, radx = 0.02, rady = 0.03)
par(oldpar)


## A simple finite-horizon MDP with action and transition weights
prefix <- file.path(tempdir(), "plot_transition_rewards_")
w <- binaryMDPWriter(prefix)
w$setWeights("Cost")
w$setTransWeights(c("Reward", "Disease"))
w$process()
   w$stage()
      w$state(label = "S1")
         w$action(
            label = "A1", weights = 2, id = c(1), pr = c(1),
            transWeights = c(20, 0.3), end = TRUE
         )
         w$action(
            label = "A2", weights = 1, id = c(0, 1), pr = c(0.3, 0.7),
            transWeights = c(25, 0.4, 15, 0.2), end = TRUE
         )
      w$endState()
   w$endStage()
   w$stage()
      w$state(label = "S2")
         w$action(
            label = "A3", weights = 3, id = c(0, 1, 2), pr = c(0.5, 0.3, 0.2),
            transWeights = c(0, 0.05, 12, 0.2, 30, 0.8), end = TRUE
         )
         w$action(
            label = "A4", weights = 2, id = c(1, 2), pr = c(0.6, 0.4),
            transWeights = c(22, 0.35, 27, 0.7), end = TRUE
         )
      w$endState()
      w$state(label = "S3")
         w$action(
            label = "A5", weights = 1, id = c(0, 1), pr = c(0.4, 0.6),
            transWeights = c(5, 0, 16, 0.25), end = TRUE
         )
         w$action(
            label = "A6", weights = 4, id = c(0, 1, 2), pr = c(0.1, 0.3, 0.6),
            transWeights = c(14, 0.15, 21, 0.45, 29, 1), end = TRUE
         )
      w$endState()
   w$endStage()
   w$stage()
      w$state(label = "S4", end = TRUE)
      w$state(label = "S5", end = TRUE)
      w$state(label = "S6", end = TRUE)
   w$endStage()
w$endProcess()
w$closeWriter()

mdp <- loadMDP(prefix, getLog = FALSE)
plot(mdp, actionColor = "label", transLabels = "weights", actionWLabel = "weight", 
     radx = 0.005, rady = 0.01)

## Reset working dir
setwd(wd)

## Set working dir
fDir <- system.file("models", package = "MDP2")
wd <- setwd(tempdir())
## Convert the machine example to a hmp file
prefix1 <- paste0(fDir,"/machine1_")
getBinInfoStates(prefix1)
convertBinary2HMP(prefix1, duration = NULL, out = "machine1_converted.hmp")
# have a look at the hmp file
cat(readr::read_file("machine1_converted.hmp"))

## Convert the machine example hmp file to binary files
convertHMP2Binary(file = paste0(fDir,"/machine1.hmp"), prefix = "machine_cov_")
getBinInfoStates(prefix = "machine_cov_")
## Convert the machine example with a single dummy node to a hmp file
#convertBinary2HMP("machine2_")  # error since using scope = 3 not supported in hmp files

## Reset working dir
setwd(wd)

## ----setup, include=FALSE---------------------------------------------------------------
library(knitr)
options(rmarkdown.html_vignette.check_title = FALSE,
        # formatR.arrow = TRUE, 
        # scipen=999, 
        # digits=5,
        width=90) # 
#thm <- knit_theme$get("edit-kwrite")   # whitengrey, bright, print, edit-flashdevelop, edit-kwrite
#knit_theme$set(thm)
knit_hooks$set(
   par = function(before, options, envir) {
      if (before && options$fig.show != 'none')
         par(mar = c(0, 0, 0, 0), # bottom, left, top, and right
             oma = c(0, 0, 0, 0))}
)
knitr::opts_chunk$set(
   # collapse = TRUE,
   comment = "#>",
   fig.align = 'center',
   fig.width = 9,
   fig.height = 5,
   fig.show = 'hold',
   out.extra = 'style="max-width:100%;"',
   # tidy = TRUE,
   # prompt=T,
   # comment=NA,
   cache = F
   # background = "red"
)

## ---------------------------------------------------------------------------------------
library(MDP2)

## ----states-----------------------------------------------------------------------------
   N <- 5
   states <- tibble::tibble(
         idx = 1:N - 1,
         label = paste0("i = ", 1:N)
      )
   states

## ---------------------------------------------------------------------------------------
Cf <- -10
Cp <- c(0, -7, -7, -5)

## ---------------------------------------------------------------------------------------
Q <- matrix(c(
   0.90, 0.10, 0, 0, 0,
   0, 0.80, 0.10, 0.05, 0.05,
   0, 0, 0.70, 0.10, 0.20,
   0, 0, 0, 0.50, 0.50), 
   nrow=4, byrow=T) 

## ----transPr----------------------------------------------------------------------------
#' Transition probabilities
#' @param i State (1 <= i <= N). 
#' @param a Action (`nr`, `pr` or `fr`).
#' @return A list with non-zero transition probabilities and state index of the transitions.
transPr <- function(i, a) {
   pr <- NULL
   idx <- NULL
   if (a == "nr") {
      pr <- Q[i, ]
      idx <- which(pr > 0)  # only consider trans pr > 0
      pr <- pr[idx]
      idx <- idx - 1  # since state index is state-1
   }
   if (a == "pr" | a == "fr") {
      pr <- 1
      idx <- 0
   }
   return(list(pr = pr, idx = idx))
}

## ---------------------------------------------------------------------------------------
transPr(1, "nr") 

## ----buildMDP1, include=FALSE-----------------------------------------------------------
labels <- states$label
w<-binaryMDPWriter("hct611-1_") # use prefix hct611-1_ to the files
w$setWeights(c("Duration", "Net reward"))
w$process() # founder process
   w$stage() # a stage with states
      w$state(label = labels[1]) # state 1
         lst <- transPr(1, "nr")
         w$action(label = "nr", weights = c(1, 0), pr = lst$pr, id = lst$id, end = TRUE)
      w$endState() # end state 1
      for (i in 2:(N-1)) { # states 2 to N-1
         w$state(label = labels[i])
            lst <- transPr(i, "nr")
            w$action(label = "nr", weights = c(1, 0), pr = lst$pr, id = lst$id, end = TRUE)
            lst<-transPr(i, "pr")
            w$action(label = "pr", weights = c(1, Cp[i]), pr = lst$pr, id = lst$id, end = TRUE)
         w$endState()
      }
      w$state(label = labels[N])
         lst<-transPr(N, "fr")
         w$action(label = "fr", weights = c(2, Cf), pr = lst$pr, id = lst$id, end = TRUE)
      w$endState()
   w$endStage() # end stage
w$endProcess() # end process
w$closeWriter() # close the binary files

## ----semi-mdp, echo=FALSE, results='hide', message=FALSE, fig.cap="Figure 1: The state-expanded hypergraph for the semi-MDP.", par=TRUE----
mdp<-loadMDP("hct611-1_")
plot(mdp, actionColor = "label", marY = 0.06)

## ----view_buldMDP1, ref.label='buildMDP1'-----------------------------------------------
labels <- states$label
w<-binaryMDPWriter("hct611-1_") # use prefix hct611-1_ to the files
w$setWeights(c("Duration", "Net reward"))
w$process() # founder process
   w$stage() # a stage with states
      w$state(label = labels[1]) # state 1
         lst <- transPr(1, "nr")
         w$action(label = "nr", weights = c(1, 0), pr = lst$pr, id = lst$id, end = TRUE)
      w$endState() # end state 1
      for (i in 2:(N-1)) { # states 2 to N-1
         w$state(label = labels[i])
            lst <- transPr(i, "nr")
            w$action(label = "nr", weights = c(1, 0), pr = lst$pr, id = lst$id, end = TRUE)
            lst<-transPr(i, "pr")
            w$action(label = "pr", weights = c(1, Cp[i]), pr = lst$pr, id = lst$id, end = TRUE)
         w$endState()
      }
      w$state(label = labels[N])
         lst<-transPr(N, "fr")
         w$action(label = "fr", weights = c(2, Cf), pr = lst$pr, id = lst$id, end = TRUE)
      w$endState()
   w$endStage() # end stage
w$endProcess() # end process
w$closeWriter() # close the binary files

## ---------------------------------------------------------------------------------------
getBinInfoStates("hct611-1_")
getBinInfoActions("hct611-1_")

## ----buildMDP2--------------------------------------------------------------------------
## Define probability matrices
P <- list()
# a = nr (no repair)
P[[1]] <- as.matrix(rbind(Q, 0))

# a = pr (preventive repair)
Z <- matrix(0, nrow = N, ncol = N)
Z[2, 1] <- Z[3, 1] <- Z[4, 1] <- 1
P[[2]] <- Z

# a = fr (forced repair)
Z <- matrix(0, nrow = N, ncol = N)
Z[5, 1] <- 1
P[[3]] <- Z

## Rewards, a 5x3 matrix with one column for each action
R <- matrix(0, nrow = N, ncol = 3)
R[2:4, 2] <- Cp[2:4]
R[5, 3] <- Cf

## State lengths, a 5x3 matrix with one column for each action
D <- matrix(1, nrow = N, ncol = 3)
D[5, 3] <- 2

## Build model using the matrix specification
w <- binaryMDPWriter("hct611-2_")
w$setWeights(c("Duration", "Net reward"))
w$process(P, R, D)
w$closeWriter()

## ----buildMDP3--------------------------------------------------------------------------
prefix<-"machine1_"
w <- binaryMDPWriter(prefix)
w$setWeights(c("Net reward"))
w$process()
	w$stage()   # stage n=0
		w$state(label="dummy")       
			w$action(label="buy", weights=-100, pr=c(0.7,0.3), id=c(0,1), end=TRUE)
		w$endState()
	w$endStage()
	w$stage()   # stage n=1
		w$state(label="good")           
			w$action(label="mt", weights=55, pr=1, id=0, end=TRUE)
			w$action(label="nmt", weights=70, pr=c(0.6,0.4), id=c(0,1), end=TRUE)
		w$endState()
		w$state(label="average")        
			w$action(label="mt", weights=40, pr=1, id=0, end=TRUE)
			w$action(label="nmt", weights=50, pr=c(0.6,0.4), id=c(1,2), end=TRUE)
		w$endState()
	w$endStage()
	w$stage()   # stage n=2
		w$state(label="good")          
			w$action(label="mt", weights=55, pr=1, id=0, end=TRUE)
			w$action(label="nmt", weights=70, pr=c(0.5,0.5), id=c(0,1), end=TRUE)
		w$endState()
		w$state(label="average")       
			w$action(label="mt", weights=40, pr=1, id=0, end=TRUE)
			w$action(label="nmt", weights=50, pr=c(0.5,0.5), id=c(1,2), end=TRUE)
		w$endState()
		w$state(label="not working")    
			w$action(label="mt", weights=30, pr=1, id=0, end=TRUE)
			w$action(label="rep", weights=5, pr=1, id=3, end=TRUE)
		w$endState()
	w$endStage()
	w$stage()   # stage n=3
		w$state(label="good")           
			w$action(label="mt", weights=55, pr=1, id=0, end=TRUE)
			w$action(label="nmt", weights=70, pr=c(0.2,0.8), id=c(0,1), end=TRUE)
		w$endState()
		w$state(label="average")       
			w$action(label="mt", weights=40, pr=1, id=0, end=TRUE)
			w$action(label="nmt", weights=50, pr=c(0.2,0.8), id=c(1,2), end=TRUE)
		w$endState()
		w$state(label="not working")    
			w$action(label="mt", weights=30, pr=1, id=0, end=TRUE)
			w$action(label="rep", weights=5, pr=1, id=3, end=TRUE)
		w$endState()
		w$state(label="replaced")       
			w$action(label="dummy", weights=0, pr=1, id=3, end=TRUE)
		w$endState()
	w$endStage()
	w$stage()   # stage n=4
		w$state(label="good", end=TRUE)        
		w$state(label="average", end=TRUE)     
		w$state(label="not working", end=TRUE) 
		w$state(label="replaced", end=TRUE)   
	w$endStage()
w$endProcess()
w$closeWriter()

## ----plotHgf3, echo=FALSE, fig.cap="Figure 2: A finite-horizon MDP", par=TRUE-----------
scrapValues <- c(30, 10, 5, 0)   # scrap values (the values of the 4 states at stage 4)
mdp <- loadMDP("machine1_", getLog = FALSE)
plot(mdp, actionColor = "label", radx = 0.06, marX = 0.065, marY = 0.055, stateLabel = "sIdx|label")

## ----echo=FALSE, fig.cap="Figure 3: The state-expanded hypergraph of the first stage of a hierarchical MDP. Level 0 indicate the founder level, and the nodes indicates states at the different levels. A child process (oval box) is represented using its state-expanded hypergraph (hyperarcs not shown) and is uniquely defined by a given state and action of its parent process."----
knitr::include_graphics("vignette_files/hmdp_index.png")

## ----Generate cow MDP functions,echo=TRUE-----------------------------------------------
library(magrittr)
cowDf <- readr::read_csv("vignette_files/cow.csv")
cowDf

# Weights given a state at level 2
lev1W <- function(s0Idx, n1Idx, s1Idx, a1Lbl) {
   return(cowDf %>% 
      dplyr::filter(s0 == s0Idx & n1 == n1Idx & s1 == s1Idx & label == a1Lbl) %>% 
      dplyr::select(Duration, Reward, Output) %>% as.numeric()
   )
}
lev1W(2, 2, 1, 'Keep')     # good genetic merit, lactation 2, avg yield, keep action

# Trans pr given a state at level 2
lev1Pr <- function(s0Idx, n1Idx, s1Idx, a1Lbl) {
   return(cowDf %>% 
      dplyr::filter(s0 == s0Idx & n1 == n1Idx & s1 == s1Idx & label == a1Lbl) %>% 
      dplyr::select(scp0:last_col()) %>% as.numeric()
   )
}
lev1Pr(2, 2, 1, 'Replace') # good genetic merit, lactation 2, avg yield, replace action

## ----Generate cow MDP,echo=TRUE, tidy=FALSE---------------------------------------------
lblS0 <- c('Bad genetic level', 'Avg genetic level', 'Good genetic level')
lblS1 <- c('Low yield', 'Avg yield', 'High yield')
prefix<-"cow_"
w<-binaryMDPWriter(prefix)
w$setWeights(c("Duration", "Net reward", "Yield"))
w$process()
  w$stage()   # stage 0 at founder level
    for (s0 in 0:2) {
      w$state(label=lblS0[s0+1])   # state at founder
        w$action(label="Keep", weights=c(0,0,0), prob=c(2,0,1))   # action at founder
          w$process()
            w$stage()   # dummy stage at level 1
               w$state(label="Dummy")
                 w$action(label="Dummy", weights=c(0,0,0), 
                          prob=c(1,0,1/3, 1,1,1/3, 1,2,1/3), end=TRUE)
               w$endState()
            w$endStage()
            for (d1 in 1:4) {
              w$stage()   # stage at level 1
                for (s1 in 0:2) {
                  w$state(label=lblS1[s1+1])
                    if (d1!=4) {
                      w$action(label="Keep", weights=lev1W(s0,d1,s1,"Keep"), 
                               prob=lev1Pr(s0,d1,s1,"Keep"), end=TRUE)
                    }
                    w$action(label="Replace", weights=lev1W(s0,d1,s1,"Replace"), 
                             prob=lev1Pr(s0,d1,s1,"Replace"), end=TRUE)
                  w$endState()
                }
              w$endStage()
            }
          w$endProcess()
        w$endAction()
      w$endState()
    }
  w$endStage()
w$endProcess()
w$closeWriter()

## ----plotHMDP, echo=FALSE, message=FALSE, par=TRUE--------------------------------------
mdp<-loadMDP(prefix)
# hgf <- list(nodes = NULL, hyperarcs = NULL)
# hgf %>% plotHypergraph(gridDim=c(14,7), cex = 0.8, showGrid = T)
hgf <- getHypergraph(mdp)
dat <- hgf$nodes %>% 
   dplyr::mutate(label = dplyr::case_when(
      label == "Low yield" ~ "L",
      label == "Avg yield" ~ "A",
      label == "High yield" ~ "H",
      label == "Dummy" ~ "D",
      label == "Bad genetic level" ~ "Bad",
      label == "Avg genetic level" ~ "Avg",
      label == "Good genetic level" ~ "Good",
      TRUE ~ "Error"
   ))
dat$gId[1:3]<-85:87
dat$gId[43:45]<-1:3
getGId<-function(process,stage,state) {
   if (process==0) start=18
   if (process==1) start=22
   if (process==2) start=26
   return(start + 14 * stage + state)
}
idx<-43
for (process in 0:2)
   for (stage in 0:4)
      for (state in 0:2) {
         if (stage==0 & state>0) break
         idx<-idx-1
         #cat(idx,process,stage,state,getGId(process,stage,state),"\n")
         dat$gId[idx]<-getGId(process,stage,state)
      }
hgf$nodes <- dat

dat <- hgf$hyperarcs %>% 
   dplyr::mutate(label = dplyr::case_when(
      label == "Replace" ~ "R",
      label == "Keep" ~ "K",
      label == "Dummy" ~ "D",
      TRUE ~ "Error"
      ),
      col = dplyr::case_when(
         label == "R" ~ "deepskyblue3",
         label == "K" ~ "darkorange1",
         label == "D" ~ "black",
         TRUE ~ "Error"
      ),
      lwd = 0.5,
      label = ""
   ) 
hgf$hyperarcs <- dat
plotHypergraph(gridDim = c(14, 7), hgf, cex = 0.8, radx = 0.02, rady = 0.03)

## ----Delete bin, include=FALSE----------------------------------------------------------
do.call(file.remove,list(list.files(pattern = ".bin")))

## For R 2.15.1 and later this works. Note that calling loadModule() triggers
## a load action, so this does not have to be placed in .onLoad() or evalqOnLoad().
Rcpp::loadModule("HMDPModule", TRUE)

## Excercise 6.4 in Tijms, H.C., "A first course in stochastic models", John Wiley & Sons Ltd, 2003.
## The semi-MDP is specified using binaryMDPWriter and actions with id and pr

p<-c(1/8, 1/2, 1/4, 1/8)
states<-0:4
a0 <- "nothing"
a1 <- "empty"

# return the cost
cost<-function(i,a) {
   if (a==0) {
      if (i<2) return(0)   # no excess waste
      k <- (4-i+1):3
      return(30*sum( (i+k-4)*p[k+1] ) )
   }
   if (a==1) {
      return(25 + 5*i)
   }
   return(NULL)
}

# return the trans pr vector (scp=1,sId,pr, scp=1, sId, pr, ...) 
transPr<-function(i,a) {
   pr<-NULL
   if (a==0) {
      if (i<4) for (j in i:3) pr<-c(pr,p[j-i+1])
      if (i>0) pr<-c(pr,sum(p[(4-i):3+1]))
   }
   if (a==1) {
      for (j in 0:3) pr<-c(pr,p[j+1])
   }
   return(pr)
} 

idx<-function(i,a) {
   pr<-NULL
   if (a==0) {
      if (i<4) for (j in i:3) pr<-c(pr,j)
      if (i>0) pr<-c(pr,4)
   }
   if (a==1) {
      for (j in 0:3) pr<-c(pr,j)
   }
   return(pr)
} 

# Build the model which is stored in a set of binary files
w<-binaryMDPWriter("hct64v2_", getLog = FALSE)
w$setWeights(c("Duration","Net reward"))
w$process()
   w$stage()
      w$state(label=0)
         w$action(label=a0, weights=c(1,-cost(0,0)), pr=transPr(0,0), id=idx(0,0), end=T)
      w$endState()
      for (i in 1:4 ) {
      w$state(label=i)
         w$action(label=a0, weights=c(1,-cost(i,0)), pr=transPr(i,0), id=idx(i,0), end=T)
         w$action(label=a1, weights=c(1,-cost(i,1)), pr=transPr(i,1), id=idx(i,1), end=T)
      w$endState()
      }
   w$endStage()
w$endProcess()
w$closeWriter()

#' Generate a "random" HMDP stored in a set of binary files.
#'
#' @param prefix A character string with the prefix added to the file(s).
#' @param levels Maximum number of levels. Set `childProcessPr = 1` if want exact this number of levels.
#' @param timeHorizon The time horizon for each level (vector). For the founder the time-horizon can be Inf.
#' @param states Number of states at each stage at a given level (vector of length levels)
#' @param actions Min and max number of actions at a state.
#' @param childProcessPr Probability of creating a child process when define action.
#' @param externalProcessPr Probability of creating an external process given that we create a child process. Only works if levels>2 and and currently does not generate external processes which include external processes.
#' @param rewards Min and max reward used.
#' @param durations Min and max duration used.
#' @param rewardName Weight name used for reward.
#' @param durationName Weight name used for duration.
#'
#' @return The file prefix (character).
#'
#' @export
randomHMDP<-function(prefix="", levels=3, timeHorizon=c(Inf,3,4), states=c(2,4,5), actions=c(1,2),
                     childProcessPr = 0.5, externalProcessPr=0, rewards=c(0,100), durations=c(1,10), 
                     rewardName="Reward", durationName="Duration" ) 
{
   # gen finite time-horizon process function
   genProcess<-function(levels, timeHorizon, states, actions, childProcessPr, rewards, durations, statesFather=NULL) {
      w$process()
         for(l1 in 1:timeHorizon[1]-1 ) {
            w$stage()
               for (s1 in 1:states[1]-1) {
                  w$state(s1)
                     aSize = sample(actions[1]:actions[2],1)
                     for (a1 in 1:aSize-1) {
                        if (levels>1) isChild = stats::rbinom(1,1,childProcessPr)==1 else isChild = FALSE
                        if (isChild) {
                           idx<-sample(1:states[2]-1,states[2]/2) 
                           pr<-rep(1/length(idx),length(idx))
                           scp<-rep(2,length(idx))
                           isExt = stats::rbinom(1,1,externalProcessPr)==1
                           if (isExt) {
                              message("\n External: ", appendLF = FALSE)
                              pfx<-paste(prefix,l1,"-",s1,"-",a1,"_",sub("\\.","-",format(Sys.time(), "%H-%M-%OS4")),"_",sep="")
                              randomHMDP(pfx, levels-1, timeHorizon[2:length(timeHorizon)], states[2:length(states)], actions, childProcessPr, rewards, durations) 
                              #stop("tst")
                              w$includeProcess(pfx, label=a1, weights=c(sample(rewards[1]:rewards[2],1), sample(durations[1]:durations[2],1)), prob = as.vector( t(matrix(c(scp,idx,pr), ncol=3)) ), termStates=states[2])
                                 w$stage()   # jump actions of last stage in the external process
                                 for (s2 in 1:states[2]-1) {
                                    w$state(s2)
                                       idx<-sample(1:states[1]-1,states[1]/2) 
                                       pr<-rep(1/length(idx),length(idx))
                                       scp<-rep(0,length(idx))
                                       w$action(label="rep", weights=c(sample(rewards[1]:rewards[2],1), sample(durations[1]:durations[2],1)), prob = as.vector( t(matrix(c(scp,idx,pr), ncol=3)) ), end=TRUE)
                                    w$endState()
                                 }
                                 w$endStage()
                              w$endIncludeProcess()
                           } else {
                              w$action(label=a1, weights=c(sample(rewards[1]:rewards[2],1), sample(durations[1]:durations[2],1)), prob = as.vector( t(matrix(c(scp,idx,pr), ncol=3)) ))
                                 genProcess(levels-1, timeHorizon[2:length(timeHorizon)], states[2:length(states)], actions, childProcessPr, rewards, durations, states[1])
                              w$endAction()
                           }
                           
                        } else {
                           idx<-sample(1:states[1]-1,states[1]/2)
                           pr<-rep(1/length(idx),length(idx))
                           scp<-rep(1,length(idx))
                           w$action(label=a1, weights=c(sample(rewards[1]:rewards[2],1), sample(durations[1]:durations[2],1)), prob = as.vector( t(matrix(c(scp,idx,pr), ncol=3)) ), end=TRUE)
                        }
                     }
                  w$endState()
               }
            w$endStage()
         }
         w$stage()   # last stage
            for (s1 in 1:states[1]-1) {
               w$state(s1)
                  if (!is.null(statesFather)) {
                     idx<-sample(1:statesFather-1,statesFather/2)
                     pr<-rep(1/length(idx),length(idx))
                     scp<-rep(0,length(idx))
                     w$action(label=a1, weights=c(sample(rewards[1]:rewards[2],1), sample(durations[1]:durations[2],1)), 
                              prob = as.vector( t(matrix(c(scp,idx,pr), ncol=3)) ))
                     w$endAction()
                  }   
               w$endState()
            }
         w$endStage()
      w$endProcess()
   }
   
   message("Create random HMDP '", prefix, "' with at most ", levels, " levels ... ", appendLF = FALSE)
   w<-binaryMDPWriter(prefix)
   w$setWeights(c(rewardName,durationName))
   if (!is.infinite(timeHorizon[1])) genProcess(levels, timeHorizon, states, actions, childProcessPr, rewards, durations)
   else {
      w$process()
         w$stage()
            for (s1 in 1:states[1]) {
               w$state(s1)
                  aSize = sample(actions[1]:actions[2],1)
                  for (a1 in 1:aSize-1) {
                     if (levels>1) isChild = stats::rbinom(1,1,childProcessPr)==1 else isChild = FALSE
                     if (isChild) idx<-sample(1:states[2]-1,states[2]/2) else idx<-1:states[1]-1 
                     pr<-rep(1/length(idx),length(idx))
                     if (sum(pr)!=1) stop("Pr don't sum to one!")
                     if (isChild) scp<-rep(2,length(idx)) else scp<-rep(1,length(idx))
                     #print(as.vector( t(matrix(c(scp,idx,pr), ncol=3)) ))
                     w$action(label=a1, weights=c(sample(rewards[1]:rewards[2],1), sample(durations[1]:durations[2],1)), 
                              prob = as.vector( t(matrix(c(scp,idx,pr), ncol=3)) ))
                        if (isChild) genProcess(levels-1, timeHorizon[2:length(timeHorizon)], 
                                                states[2:length(states)], actions, childProcessPr, rewards, durations, states[1])
                     w$endAction()
                  }
                  
               w$endState()
            }
         w$endStage()
      w$endProcess()
   }
   w$closeWriter()
   message(" finished.")
   return(prefix)
}

## Example 6.1.1 in Tijms, H.C., "A first course in stochastic models", John Wiley & Sons Ltd, 2003.
## The semi-MDP is specified using binaryMDPWriter and actions with prob

N<-5; Cf<- -10; Cp<-c(0,-7,-7,-5) # use negative numbers since optimize reward
Q <- matrix(c(
   0.90, 0.10, 0, 0, 0,
   0, 0.80, 0.10, 0.05, 0.05,
   0, 0, 0.70, 0.10, 0.20,
   0, 0, 0, 0.50, 0.50), nrow=4, byrow=T) 
# note that we index from 0, i.e. state N correspond to index N-1 (in j col)
transPr0<-data.frame(i=rep(1:(N-1), times=N), scope=1, j=rep(1:N,each=4)-1, pr=as.vector(Q))
transPr0<-transPr0[transPr0$pr>0,]  # remove transitions with zero pr
transPr1<-data.frame(i=2:(N-1), scope=1, j=1-1, pr=rep(1, times=N-2))
getTransPr<-function(a,i) {
   if (a==0) return(as.vector(as.matrix(t(transPr0[transPr0$i==i,c(2,3,4)]))) )
   if (a==1) return(as.vector(as.matrix(t(transPr1[transPr1$i==i,c(2,3,4)]))))
   return(stop("Error"))
}

w<-binaryMDPWriter("hct611_", getLog = FALSE)
w$setWeights(c("Duration","Net reward"))
w$process()
   w$stage()
      w$state(label="i=1")
         w$action(label="no repair", weights=c(1,0), prob=getTransPr(a=0,i=1), end=T)
      w$endState()
      for (ii in 2:(N-1) ) {
      w$state(label=paste("i=",ii,sep=""))
         w$action(label="no repair", weights=c(1,0), prob=getTransPr(a=0,i=ii), end=T)
         w$action(label="preventive repair", weights=c(1,Cp[ii]), prob=getTransPr(a=1,i=ii), end=T)
      w$endState()
      }
      w$state(label=paste("i=",N,sep=""))
         w$action(label="forced repair", weights=c(2,Cf), prob=c(1,0,1), end=T)
      w$endState()
   w$endStage()
w$endProcess()
w$closeWriter()

#' Load the HMDP model defined in the binary files. The model are created in memory
#' using the external C++ library.
#'
#' @param prefix A character string with the prefix added to `binNames`. Used to identify a 
#'   specific model.
#' @param binNames A character vector of length 7 giving the names of the binary
#'     files storing the model.
#' @param eps The sum of the transition probabilities must at most differ `eps` from one.
#' @param check Check if the MDP seems correct.
#' @param verbose More output when running algorithms.
#' @param getLog Output the log messages.
#' 
#' @return A list containing relevant information about the model such as model file names 
#'    (`binNames`), time horizon (`timeHorizon`), number of states (`states`), number of states at 
#'    last stage of the founder process (`founderStatesLast`), number of actions (`actions`), 
#'    number of levels (`levels`), names of the weights associated to each action (`weightNames`)
#'    and a pointer `ptr` to the model object in memory. Note for models with an infinite 
#'    time-horizon the states at the founder level is repeated at stage two so have something aka
#'    a double array representation. 
#' @example inst/examples/machine-ex.R
#' @export
loadMDP <-
   function(prefix = "",
            binNames = c(
               "stateIdx.bin",
               "stateIdxLbl.bin",
               "actionIdx.bin",
               "actionIdxLbl.bin",
               "actionWeight.bin",
               "actionWeightLbl.bin",
               "transProb.bin",
               "externalProcesses.bin",
               "transWeight.bin",
               "transWeightLbl.bin"
            ),
            eps = 0.00001,
            check = TRUE,
            verbose = FALSE,
            getLog = TRUE
   ) {
	binNames<-paste(prefix,binNames,sep="")
	if (!is.logical(verbose)) verbose = FALSE
	mdp<-methods::new(HMDP, binNames, verbose)
   .makeMDPList(mdp, binNames, eps = eps, check = check, getLog = getLog)
}

.makeMDPList <- function(mdp,
                         binNames = character(),
                         eps = 0.00001,
                         check = TRUE,
                         getLog = TRUE) {
	if (!mdp$okay) {
		message(mdp$getLog())
		rm(mdp)
		return(invisible(NULL))
	}
	else if (getLog) message(mdp$getLog())

	if (check) {
	   msg<-mdp$checkHMDP(as.numeric(eps))
      if (msg==2) {
         stop(mdp$getLog(), call. = FALSE)
         return(invisible(NULL))
      }
	   else if (getLog) message(mdp$getLog())
	}

	timeHorizon = mdp$timeHorizon
	if (timeHorizon>=1000000000) timeHorizon = Inf
	if (timeHorizon>=Inf) {
	   founderStatesLast<-mdp$getStateSizeStage("1")
		states<-mdp$getStateSize()-founderStatesLast
	} else {
	   founderStatesLast<-mdp$getStateSizeStage(as.character(timeHorizon-1))
		states<-mdp$getStateSize()
	}
	actions <- mdp$getActionSize()
	levels <- mdp$levels
	weightActionNames <- mdp$wActionNames
	weightTransNames <- mdp$wTransNames
	weightNames <- mdp$wNames
	v<-list(binNames=binNames, timeHorizon=timeHorizon, states=states, 
	     founderStatesLast=founderStatesLast, actions=actions, levels=levels, 
	     weightNames=weightNames, weightActionNames=weightActionNames,
	     weightTransNames=weightTransNames, ptr=mdp)
	if (mdp$externalProc) {
	   v$external <- as.data.frame(matrix(mdp$getExternalInfo(),ncol = 2, byrow = TRUE), stringsAsFactors=FALSE)
	   colnames(v$external) <- c("stageStr","prefix")
	}
	class(v)<-c("HMDP", "list")
	return(v)
}



#' Internal function. Check if the indexes given are okay. Should not be used except you know what you are doing
#'
#' @aliases .checkWDurIdx
#' @param iW Index of the weight we want to optimize.
#' @param iDur Index of the duration/time.
#' @param wLth Number of weights in the model.
#' @return Nothing.
#' @name checkWDurIdx
#' @keywords internal
.checkWDurIdx<-function(iW, iDur, wLth) {
	if (length(iW)!=1) stop("Index iW must be of length one!",call. = FALSE)
	if (iW>wLth-1) stop("Global weight index iW must be less than ",wLth,"!",call. = FALSE)
	if (iW<0) stop("Global weight index iW must be greater or equal zero!",call. = FALSE)
	if (!is.null(iDur)) {
		if (length(iDur)!=1) stop("Index iDur must be of length one!",call. = FALSE)
		if (iW==iDur) stop("Indices iW and iDur must not be the same!",call. = FALSE)
		if (iDur>wLth-1) stop("Global duration weight index iDur must be less than ",wLth,"!",call. = FALSE)
		if (iDur<0) stop("Global duration weight index iDur must be greater or equal zero!",call. = FALSE)
	}
	invisible()
}


#' Internal function. Check if the index of the weight is okay. Should not be used except you know what you are doing
#'
#' @aliases .checkWIdx
#' @param iW Index of the weight we want to optimize.
#' @param wLth Number of weights in the model.
#' @return Nothing.
#' @name checkWIdx
#' @keywords internal
.checkWIdx<-function(iW, wLth) {
	if (max(iW)>wLth-1) stop("Global weight index iW must be less than ",wLth,"!",call. = FALSE)
	if (min(iW)<0) stop("Global weight index iW must be greater or equal zero!",call. = FALSE)
	invisible()
}


#' Return the index of a weight in the model. Note that index always start from zero (C++ style), i.e. the first weight, the first state at a stage etc has index 0.
#'
#' @param mdp The MDP loaded using [loadMDP()].
#' @param wLbl The label/string of the weight.
#' @return The index (integer).
#' @export
getWIdx <- function(mdp, wLbl) {
   idx <- which(mdp$weightNames == wLbl)
   if (length(idx) == 0) idx <- which(grepl(wLbl, mdp$weightNames, fixed = TRUE))
   if (length(idx) == 0)
      stop("The weight name does not seem to exist in the global weight namespace.", call. = FALSE)
   if (length(idx) > 1)
      stop("The weight name is ambiguous in the global weight namespace.", call. = FALSE)
   return(idx - 1)
}

.optSenseIdx <- function(objective) {
   objective <- match.arg(objective, c("max", "min"))
   if (objective == "max") return(0)
   1
}

.bellmanOpIdx <- function(bellmanOp, includeVariance = TRUE) {
   if (length(bellmanOp) > 1) bellmanOp <- bellmanOp[1]
   choices <- c("auto", "discount", "average", "expected", "min", "max", "secondMoment")
   if (includeVariance) choices <- c(choices, "variance")
   bellmanOp <- match.arg(bellmanOp, choices)
   switch(bellmanOp,
          discount = 0,
          average = 1,
          expected = 2,
          min = 5,
          max = 6,
          secondMoment = 7,
          variance = 8,
          auto = NA_integer_)
}


#' Perform policy iteration using the average expected-weight Bellman operator on the MDP.
#'
#' The policy can afterwards be received using functions `getPolicy` and `getPolicyW`.
#'
#' @param mdp The MDP loaded using [loadMDP()].
#' @param w The label of the weight we optimize.
#' @param dur The label of the duration/time such that discount rates can be calculated.
#' @param maxIte Max number of iterations. If the model does not satisfy the unichain assumption the algorithm may loop.
#' @param objective Optimize by maximizing (`"max"`) or minimizing (`"min"`) the Bellman value.
#' @param getLog Output the log messages.
#' 
#' @return The optimal gain (g) calculated.
#' @seealso [getPolicy()].
#' @export
runPolicyIteAve<-function(mdp, w, dur, maxIte=100, objective = c("max", "min"), getLog = TRUE) {
	iW<-getWIdx(mdp,w)
	iDur<-getWIdx(mdp,dur)
	sense <- .optSenseIdx(objective)
	.checkWDurIdx(iW,iDur,length(mdp$weightNames))
	g<-mdp$ptr$policyIte(1, sense, as.integer(maxIte), as.integer(iW), as.integer(iDur), discountFactor=1)
	#message(mdp$ptr$getLog())
	if (getLog) cat(mdp$ptr$getLog())
	return(g)
}


#' Perform policy iteration using the discounted expected-weight Bellman operator on the MDP.
#'
#' The policy can afterwards be received using functions `getPolicy` and `getPolicyW`.
#'
#' @param mdp The MDP loaded using [loadMDP()].
#' @param w The label of the weight we optimize.
#' @param dur The label of the duration/time such that discount rates can be calculated.
#' @param rate The interest rate.
#' @param rateBase The time-horizon the rate is valid over.
#' @param discountFactor The discount rate for one time unit. If specified `rate` and `rateBase` are not used to calculate the discount rate.
#' @param maxIte Max number of iterations. If the model does not satisfy the unichain assumption the algorithm may loop.
#' @param discountMethod Either 'continuous' or 'discrete', corresponding to discount factor `exp(-rate/rateBase)` or `1/(1 + rate/rateBase)`, respectively. Only used if `discountFactor` is `NULL`.
#' @param objective Optimize by maximizing (`"max"`) or minimizing (`"min"`) the Bellman value.
#' @param getLog Output the log messages.
#' 
#' @return Nothing.
#' @seealso [getPolicy()].
#' @export
runPolicyIteDiscount<-function(mdp, w, dur, rate = 0, rateBase = 1, discountFactor = NULL, maxIte = 100, 
                            discountMethod="continuous", objective = c("max", "min"), getLog = TRUE) {
	iW<-getWIdx(mdp,w)
	iDur<-getWIdx(mdp,dur)
	sense <- .optSenseIdx(objective)
	.checkWDurIdx(iW,iDur,length(mdp$weightNames))
	if (is.null(discountFactor)) {
	   if (discountMethod=="continuous") discountFactor<-exp(-rate/rateBase)
	   if (discountMethod=="discrete") discountFactor<-1/(1 + rate/rateBase)
	}
	g<-mdp$ptr$policyIte(0, sense, as.integer(maxIte), as.integer(iW), as.integer(iDur), discountFactor)
	if (getLog) cat(mdp$ptr$getLog())
	invisible()
}


#' Perform value iteration on the MDP.
#'
#' If the MDP has a finite time-horizon then arguments `times` and `eps`
#' are ignored.
#'
#' @param mdp The MDP loaded using [loadMDP()].
#' @param w The label of the weight we optimize.
#' @param dur The label of the duration/time such that discount rates can be calculated.
#' @param rate Interest rate.
#' @param rateBase The time-horizon the rate is valid over.
#' @param discountFactor The discount rate for one time unit. If specified `rate` and `rateBase` are not used to calculate the discount rate.
#' @param maxIte The max number of iterations value iteration is performed.
#' @param eps Stopping tolerance. If $max(w(t)-w(t+1)) < `eps`$ then stop the algorithm, i.e the policy becomes epsilon optimal (see Puterman p161).
#' @param termValues The terminal values used (values of the last stage in the MDP).
#' @param g Average weight. If specified then do a single iteration using the update equations under the average expected-weight Bellman operator with the specified g value.
#' @param objective Optimize by maximizing (`"max"`) or minimizing (`"min"`) the Bellman value.
#' @param bellmanOp Bellman operator. Use `"auto"` for existing behavior, `"min"` for the minimum-successor operator, `"max"` for the maximum-successor operator, or `"secondMoment"` for the second moment of total accumulated weight.
#' @param getLog Output the log messages.
#' @param discountMethod Either 'continuous' or 'discrete', corresponding to discount factor `exp(-rate/rateBase)` or `1/(1 + rate/rateBase)`, respectively. Only used if `discountFactor` is `NULL`.
#' 
#' @return NULL (invisible)
#' @references Puterman, M. Markov Decision Processes, Wiley-Interscience, 1994.
#' @example inst/examples/machine-ex.R
#' @export
runValueIte<-function(mdp, w, dur = NULL, rate = 0, rateBase = 1, discountFactor = NULL, maxIte = 100, 
                   eps = 1e-05, termValues = NULL, g=NULL, objective = c("max", "min"),
                   bellmanOp = c("auto", "expected", "discount", "average", "min", "max", "secondMoment"),
                   getLog = TRUE, discountMethod="continuous") {
	iW<-getWIdx(mdp,w)
	iDur<-NULL
	sense <- .optSenseIdx(objective)
	op <- .bellmanOpIdx(bellmanOp, includeVariance = FALSE)
	if (!is.null(dur)) iDur<-getWIdx(mdp,dur)
	.checkWDurIdx(iW,iDur,length(mdp$weightNames))
	if (is.null(discountFactor)) {
	   if (discountMethod=="continuous") discountFactor<-exp(-rate/rateBase)
	   if (discountMethod=="discrete") discountFactor<-1/(1 + rate/rateBase)
	}
	if (is.null(termValues)) termValues<-rep(0,mdp$founderStatesLast)
	if (!is.na(op)) {
	   if (op %in% c(0L, 1L) && is.null(iDur))
	      stop("A duration index must be specified for this Bellman operator.", call. = FALSE)
	   if (op == 1L && is.null(g))
	      stop("The average weight g must be specified for the average Bellman operator.", call. = FALSE)
	   mdp$ptr$valueIte(op, sense, as.integer(ifelse(mdp$timeHorizon>=Inf, maxIte, 1)),
            as.numeric(eps), as.integer(iW), as.integer(ifelse(is.null(iDur), 0, iDur)),
            as.numeric(termValues), as.numeric(ifelse(is.null(g), 0, g)), as.numeric(discountFactor))
	} else if (is.null(g)) {
   	if (mdp$timeHorizon>=Inf) {
   		if (is.null(iDur)) stop("A duration index must be specified under infinite time-horizon!")
	      mdp$ptr$valueIte(0, sense, as.integer(maxIte),
   			as.numeric(eps), as.integer(iW), as.integer(iDur), as.numeric(termValues),
   			as.numeric(0),	as.numeric(discountFactor) )
   	} else {
	      if (!is.null(iDur)) mdp$ptr$valueIte(0, sense, as.integer(1),
                as.numeric(0), as.integer(iW), as.integer(iDur), as.numeric(termValues),
                as.numeric(0), as.numeric(discountFactor) )
	      if (is.null(iDur)) mdp$ptr$valueIte(2, sense, as.integer(1),
               as.numeric(0), as.integer(iW), as.integer(0), as.numeric(termValues),
               as.numeric(0), as.numeric(1) )
   	}
	} else {  # value ite under average expected-weight Bellman operator
	   if (is.null(iDur)) stop("A duration index must be specified under the average expected-weight Bellman operator!")
	   mdp$ptr$valueIte(1, sense, as.integer(1),
           as.numeric(eps), as.integer(iW), as.integer(iDur), as.numeric(termValues),
           as.numeric(g), as.numeric(1) )
	}
	if (getLog) cat(mdp$ptr$getLog())
	invisible(NULL)
}


#' Get parts of the optimal policy.
#'
#' @param mdp The MDP loaded using [loadMDP()].
#' @param sId Vector of id's of the states we want to retrieve.
#' @param stageStr Stage string. If specified then find `sId` based on the stage string.
#' @param stateLabels Add state labels.
#' @param actionLabels Add action labels of policy.
#' @param actionIdx Add action index.
#' @param rewards Add weights calculated for each state.
#' @param stateStr Add the state string for each state.
#' @param external A vector of stage strings corresponding to external processes we want the optimal policy of.
#' @param ... Parameters passed on when find the optimal policy of the external processes.
#' 
#' Note if external is specified then it must contain stage strings from `mdp$external`. Moreover you 
#' must specify further arguments passed on to [runValueIte()] used for recreating the optimal policy e.g. 
#' the g value and the label for weight and duration. See the vignette about external processes.
#' 
#' @return The policy (data frame).
#' @example inst/examples/machine-ex.R
#' @export
getPolicy<-function(mdp, sId = ifelse(mdp$timeHorizon>=Inf, mdp$founderStatesLast+1,1):
                       ifelse(mdp$timeHorizon>=Inf, mdp$states + mdp$founderStatesLast,mdp$states)-1, 
                    stageStr = NULL, stateLabels = TRUE, actionLabels = TRUE, actionIdx = TRUE, 
                    rewards = TRUE, stateStr = TRUE, external = NULL, ...) {
	if (!is.null(stageStr)) sId = mdp$ptr$getStateIdsStages(stageStr)
   maxS<-ifelse(mdp$timeHorizon>=Inf, mdp$states + mdp$founderStatesLast,mdp$states)
	if (max(sId)>=maxS | min(sId)<0)
		stop("Out of range (sId). Need to be a subset of 0,...,",maxS-1,"!")
   cols <- 1 + stateLabels + actionIdx + actionLabels + rewards + stateStr
   policy<-data.frame(matrix(NA, nrow = length(sId), ncol = cols))
   cols<-1
   policy[,cols] <- sId
   colNames = "sId"; cols = cols + 1
   if (stateStr) {
      policy[,cols]<-mdp$ptr$getStateStr(sId)
      colNames = c(colNames, "stateStr"); cols = cols + 1
   }
   if (stateLabels) {
      policy[,cols] = mdp$ptr$getStateLabel(sId); 
      colNames = c(colNames, "stateLabel"); cols = cols + 1
   }
   if (actionIdx) {
      policy[,cols] = mdp$ptr$getPolicy(sId) 
      colNames = c(colNames, "aIdx"); cols = cols + 1       
   }
   if (actionLabels) {
      policy[,cols] = mdp$ptr$getPolicyLabel(sId)
      colNames = c(colNames, "actionLabel"); cols = cols + 1
   }
   if (rewards) {
      policy[,cols] = mdp$ptr$getPolicyW(sId) 
      colNames = c(colNames, "weight"); cols = cols + 1
   }
   colnames(policy) <- colNames
   
   if (!is.null(external)) {
      policy <- list(main=policy)
      for (s in external) {
         prefix <- subset(mdp$external, stageStr == s, select = "prefix", drop = TRUE)
         lastStage<-mdp$ptr$getNextStageStr(s)
         termValues <- getPolicy(mdp, stageStr = lastStage)$weight
         extMDP<-loadMDP(prefix, getLog = FALSE)
         valueIte(extMDP, termValues=termValues, getLog = FALSE, ...)
         extPolicy<-getPolicy(extMDP)
         policy[[s]]<-extPolicy
         rm(extMDP)
      }
   }
	return(dplyr::as_tibble(policy))
}




#' Information about the MDP
#' 
#' @param mdp The MDP loaded using [loadMDP()].
#' @param sId The id of the state(s) considered.
#' @param stateStr A character vector containing the index of the state(s) (e.g. "n0,s0,a0,n1,s1"). 
#'   Parameter `sId` are ignored if not NULL.
#' @param stageStr A character vector containing the index of the stage(s) (e.g. "n0,s0,a0,n1"). 
#'   Parameter `sId` and `idxS` are ignored if not NULL.
#' @param withList Output info as a list `lst`.  
#' @param withDF Output the info as a data frame.
#' @param dfLevel If `withDF` and equal `"state"` the data frame contains a row for each state. If equal `"action"` the data frame contains a row for each action.
#' @param asStringsState Write state vector as a string; otherwise, output it as columns.
#' @param asStringsActions Write action vectors (weights, transitions and probabilities) as strings; otherwise, output it as columns.
#' @param withHarc Output a hyperarcs data frame. Each row contains a hyperarc with the first column denoting the
#'   head (`sId`), the tails (`sId`) and the label.
#'   
#' @return A list containing the list, data frame(s).
#' @example inst/examples/machine-ex.R
#' @importFrom magrittr %>%
#' @importFrom rlang .data
#' @export
getInfo<-function(mdp, 
                  sId=1:ifelse(mdp$timeHorizon<Inf, mdp$states, mdp$states+mdp$founderStatesLast)-1,
                  stateStr=NULL, stageStr=NULL, 
                  withList = TRUE, 
                  withDF = TRUE, dfLevel = "state", asStringsState = TRUE, asStringsActions = FALSE,
                  withHarc = FALSE
                  ) {
   if (!is.null(stageStr)) {
      sId<-mdp$ptr$getStateIdsStages(stageStr)
      stateStr<-mdp$ptr$getStateStr(sId)
   }else {
      if (!is.null(stateStr)) sId<-mdp$ptr$getStateIdsStates(stateStr)
      else stateStr<-mdp$ptr$getStateStr(sId)
   }
   maxS<-ifelse(mdp$timeHorizon>=Inf, mdp$states + mdp$founderStatesLast,mdp$states)
   if (max(sId)>=maxS | min(sId)<0)
      stop("Out of range (sId). Need to be a subset of 0,...,",maxS-1,"!")
   l<-vector("list", length(sId))
   lapply(l,function(x) x<-list(sId=NULL,stateStr=NULL,label=NULL,actions=NULL))

   labels<-mdp$ptr$getStateLabel(sId)
   for (i in 1:length(l)) {
      l[[i]]$sId <- sId[i]
      l[[i]]$stateStr <- stateStr[i]
      l[[i]]$label <- labels[i]
      l[[i]]$actions <- mdp$ptr$getActionInfo(sId[i])
   }
   names(l) <- sId
   lst <- list()
   if (withList) lst$lst <- l
   if (withDF) {
      df <- dplyr::tibble(sId = l) # add list 
      df <- df %>% tidyr::unnest_wider(sId)  # convert states to columns
      if (dfLevel == "action") {
         df <- df %>% 
            tidyr::unnest_longer("actions") %>% # convert actions (one row for each action)
            tidyr::unnest_wider("actions", names_repair = tidyr::tidyr_legacy) # convert action to columns
         df <- df %>% 
            dplyr::rename(label_action = "label1")
         if (asStringsActions) {
            df <- df %>% 
               dplyr::mutate(weights = sapply(.data$weights, function(x) paste0(x, collapse = ",")),
                      trans = sapply(.data$trans, function(x) paste0(x, collapse = ",")),
                      pr = sapply(.data$pr, function(x) paste0(x, collapse = ","))) %>% 
               dplyr::mutate(weights = dplyr::na_if(.data$weights, ""),
                      trans = dplyr::na_if(.data$trans, ""),
                      pr = dplyr::na_if(.data$pr, ""))
         }
      } else {
         if (asStringsActions) {
            df <- df %>% 
               tidyr::unnest_longer("actions") %>% # convert actions (one row for each action)
               tidyr::unnest_wider("actions", names_repair = tidyr::tidyr_legacy) # convert action to columns
            df <- df %>% 
               dplyr::rename(label_action = .data$label1)
            df <- df %>% 
               dplyr::mutate(weights = sapply(.data$weights, function(x) paste0(x, collapse = ",")),
                      trans = sapply(.data$trans, function(x) paste0(x, collapse = ",")),
                      pr = sapply(.data$pr, function(x) paste0(x, collapse = ","))) %>% 
               dplyr::mutate(weights = dplyr::na_if(.data$weights, ""),
                      trans = dplyr::na_if(.data$trans, ""),
                      pr = dplyr::na_if(.data$pr, ""))
            df <- df %>% 
               dplyr::group_by(sId, .data$stateStr, .data$label) %>% 
               tidyr::nest() %>% 
               dplyr::mutate(data = lapply(.data$data, function(x) {if (all(is.na(x$aIdx))) return(NULL) else return(x)})) %>% 
               dplyr::rename(actions = .data$data)
         }
      }
      if (!asStringsState) {
         levels <- (max(stringr::str_count(df$stateStr, ",")) + 1) %/% 3 + 1
         if (levels == 1)
            nm <- paste(c("n", "s"), levels - 1, sep = "")
         if (levels > 1)
            nm <-
               c(paste(c("n", "s", "a"), rep(0:(levels - 2), each = 3), sep = ""),
                 paste(c("n", "s"), levels - 1, sep = ""))
         df <- df %>% 
            tidyr::separate(.data$stateStr, into = nm, sep = ",", remove = FALSE, fill = "right")
      }
      lst$df <- df
   }
   if (withHarc) {
      df <- dplyr::tibble(sId = l)  %>% 
         tidyr::unnest_wider(sId) %>% 
         tidyr::unnest_longer("actions") %>% # convert actions (one row for each action)
         tidyr::unnest_wider("actions", names_repair = tidyr::tidyr_legacy) %>% 
         tidyr::unnest_wider(.data$trans, names_sep = "") %>% 
         dplyr::filter(!is.na(.data$aIdx)) %>% 
         dplyr::select(.data$sId, tidyr::contains("trans"), label = .data$label1)
      colnames(df) <- stringr::str_replace(colnames(df), "trans", "tail")
      colnames(df)[1] <- "head"
      lst$harcDF <- df
   }
   return(lst)
}


#' Modify the current policy by setting policy action of states. 
#' 
#' If the policy does not contain all states then the actions from the previous optimal 
#' policy are used.
#'
#' @param mdp The MDP loaded using [loadMDP()].
#' @param policy A data frame with two columns state id `sId` and action index `aIdx`.
#' @return NULL (invisible)
#' @example inst/examples/machine-ex.R
#' @export
setPolicy<-function(mdp, policy) {
   if (!all(c("sId", "aIdx") %in% colnames(policy))) stop("You must specify `sId` and action index `aIdx`.")
   #if (dim(policy)[2]!=2) stop("You must specify two columns in policy.")
   mdp$ptr$setPolicy(as.integer(policy$sId),as.integer(policy$aIdx))
	return(invisible(NULL))
}


#' Calculate weights based on current policy. Normally run after an optimal policy has been found.
#'
#' @param mdp The MDP loaded using [loadMDP()].
#' @param wLbl The label of the weight we consider.
#' @param criterion The Bellman operator shortcut. If `expected` use expected weights, if `discount` use discounted expected weights, if `average` use average expected weights, if `min` use minimum-successor weights, if `max` use maximum-successor weights, if `secondMoment` use the second moment of total accumulated weight, and if `variance` use the law-of-total-variance recursion under the current policy.
#' @param durLbl The label of the duration/time such that discount rates can be calculated.
#' @param rate The interest rate.
#' @param rateBase The time-horizon the rate is valid over.
#' @param discountFactor The discount rate for one time unit. If specified `rate` and `rateBase` are not used to calculate the discount rate.
#' @param termValues The terminal values used (values of the last stage in the MDP).
#' @param discountMethod Either 'continuous' or 'discrete', corresponding to discount factor `exp(-rate/rateBase)` or `1/(1 + rate/rateBase)`, respectively. Only used if `discountFactor` is `NULL`.
#' 
#' @return Nothing.
#' @example inst/examples/machine-ex.R
#' @export
runCalcWeights<-function(mdp, wLbl, criterion="expected", durLbl = NULL, rate = 0, rateBase = 1, 
                      discountFactor = NULL, termValues = NULL, discountMethod = "continuous") {
	iW<-getWIdx(mdp,wLbl)
	if (!is.null(durLbl)) iDur<-getWIdx(mdp,durLbl)
	.checkWIdx(iW,length(mdp$weightNames))
	if (is.null(discountFactor)) {
	   if (discountMethod=="continuous") discountFactor<-exp(-rate/rateBase)
	   if (discountMethod=="discrete") discountFactor<-1/(1 + rate/rateBase)
	}
	if (mdp$timeHorizon<Inf) {
		if (is.null(termValues)) stop("Terminal values must be specified under finite time-horizon!")
		if (criterion=="expected") mdp$ptr$calcPolicy(2,iW,0,1,discountFactor)
		if (criterion=="min") mdp$ptr$calcPolicy(5,iW,0,1,discountFactor)
		if (criterion=="max") mdp$ptr$calcPolicy(6,iW,0,1,discountFactor)
		if (criterion=="secondMoment") mdp$ptr$calcPolicy(7,iW,0,1,discountFactor)
		if (criterion=="variance") {
		   mdp$ptr$setTerminalW(as.numeric(termValues))
		   mdp$ptr$calcPolicy(8,iW,0,1,discountFactor)
		}
		if (criterion=="discount") mdp$ptr$calcPolicy(0,iW,0,iDur,discountFactor)
	} else {
		if (criterion=="discount") mdp$ptr$policyIteFixedPolicy(0,iW,iDur,discountFactor)
		if (criterion=="average") return( mdp$ptr$policyIteFixedPolicy(1,iW,iDur,discountFactor) )
		#if (criterion=="expected") .Call("MDP_CalcWeightsFinite", mdp$ptr, as.integer(iW), as.numeric(termValues), PACKAGE="MDP")
	}
	invisible(NULL)
}


#' Calculate the retention pay-off (RPO) or opportunity cost for some states.
#'
#' The RPO is defined as the difference between
#' the weight of the state when using action `iA` and the maximum
#' weight of the node when using another predecessor different from `iA`.
#'
#' @param mdp The MDP loaded using [loadMDP()].
#' @param w The label of the weight we calculate RPO for.
#' @param iA  The action index we calculate the RPO with respect to (same size as `sId`).
#' @param sId Vector of id's of the states we want to retrieve.
#' @param criterion The Bellman operator shortcut. If `expected` use expected weights, if `discount` use discounted expected weights, if `average` use average expected weights.
#' @param dur The label of the duration/time such that discount rates can be calculated.
#' @param rate The interest rate.
#' @param rateBase The time-horizon the rate is valid over.
#' @param discountFactor The discount rate for one time unit. If specified `rate` and `rateBase` are not used to calculate the discount rate.
#' @param g The optimal gain (g) calculated (used if `criterion = "average"`).
#' @param objective Optimize by maximizing (`"max"`) or minimizing (`"min"`) the Bellman value.
#' @param discountMethod Either 'continuous' or 'discrete', corresponding to discount factor `exp(-rate/rateBase)` or `1/(1 + rate/rateBase)`, respectively. Only used if `discountFactor` is `NULL`.
#' @param stateStr Output the state string. 
#' 
#' @return The RPO (matrix/data frame).
#' @importFrom magrittr %>%
#' @export
getRPO<-function(mdp, w, iA, sId = ifelse(mdp$timeHorizon>=Inf, mdp$founderStatesLast+1,1):
                    ifelse(mdp$timeHorizon>=Inf, mdp$states + mdp$founderStatesLast,mdp$states)-1, 
                  criterion="expected", dur = "", rate = 0, rateBase = 1, discountFactor = NULL, 
                  g = 0, objective = c("max", "min"), discountMethod="continuous", stateStr = TRUE) {
   iW<-getWIdx(mdp,w)
   sense <- .optSenseIdx(objective)
   if (criterion != "expected" && !nzchar(dur))
      stop("A duration weight must be specified for this Bellman operator.", call. = FALSE)
   iDur<-if (nzchar(dur)) getWIdx(mdp,dur) else 0
   .checkWIdx(iW,length(mdp$weightNames))
   if (is.null(discountFactor)) {
      if (discountMethod=="continuous") discountFactor<-exp(-rate/rateBase)
      if (discountMethod=="discrete") discountFactor<-1/(1 + rate/rateBase)
   }
   maxS<-ifelse(mdp$timeHorizon>=Inf, mdp$states + mdp$founderStatesLast,mdp$states)
   if (max(sId)>=maxS | min(sId)<0)
      stop("Out of range (sId). Need to be a subset of 0,...,",maxS-1,"!")
   if (length(sId)!=length(iA))
      stop("Vectors sId and iA must have same length!")
   rpo<-NA
   if (criterion=="expected") rpo<-mdp$ptr$calcRPO(2, sense, as.integer(sId), iW, as.integer(iA), g, iDur, discountFactor)
   if (criterion=="discount") rpo<-mdp$ptr$calcRPO(0, sense, as.integer(sId), iW, as.integer(iA), g, iDur, discountFactor)
   if (criterion=="average")  rpo<-mdp$ptr$calcRPO(1, sense, as.integer(sId), iW, as.integer(iA), g, iDur, discountFactor)
   rpo[rpo <= -1.8e+16]<-NA # less than 2 actions
   rpo <- dplyr::tibble(sId=sId, rpo=rpo)
   if (stateStr) {
      rpo <- rpo %>% 
         dplyr::transmute(sId, stateStr = mdp$ptr$getStateStr(sId), rpo)
   }
   return(rpo)
}

#' Save the MDP to binary files
#' 
#' Currently do not save external files.
#'
#' @param mdp The MDP loaded using [loadMDP()].
#' @param prefix A character string with the prefix added to `binNames`. Used to identify a specific model.
#' @param getLog Output the log as a message.
#' 
#' @return ???
#' @export
saveMDP<-function(mdp,prefix="", getLog=TRUE) {
   mdp$ptr$save2Binary(prefix)
   if (getLog) message(mdp$ptr$getLog())
}


#' Calculate the steady state transition probabilities for the founder process (level 0).
#'
#' Assume that we consider an ergodic/irreducible time-homogeneous Markov chain specified using a policy in the MDP.
#'
#' @param mdp The MDP loaded using [loadMDP()].
#' @param getLog Output log text.
#' 
#' @return A vector with steady state probabilities for all the states at the founder level.
#' @export
getSteadyStatePr<-function(mdp, getLog=FALSE) {
   pr<-mdp$ptr$steadyStatePr()
   if (getLog) message(mdp$ptr$getLog())
	return(pr)
}



# #' Set the weight of an action.
# #'
# #' @param mdp The MDP loaded using \link{loadMDP}.
# #' @param w The weight.
# #' @param sId The state id of the state.
# #' @param idxA The action index.
# #' @param wLbl The label of the weight we consider.
# #' @return Nothing.
# #' @example inst/examples/machine.R
# #' @export
# setActionWeight<-function(mdp, w, sId, iA, wLbl) {
# 	iW<-getWIdx(mdp,wLbl)
# 	
# 	.Call("MDP_SetActionW", mdp$ptr, as.numeric(w), as.integer(sId), as.integer(iA), as.integer(iW), PACKAGE="MDP")
# 	invisible(NULL)
# }
# 




#
#
#
# #' Fix the action of a state. That is, the other actions are removed from the HMDP.
# #'
# #' The actions can be reset using \code{resetActions}.
# #'
# #' @param mdp The MDP loaded using \link{loadMDP}.
# #' @param sId The state id of the state we want to fix the action for.
# #' @param iA  The action index of the state.
# #' @return Nothing.
# #' @author Lars Relund \email{lars@@relund.dk}
# #' @seealso \code{\link{resetActions}}, \code{\link{removeAction}}.
# #' @export
# fixAction<-function(mdp, sId, iA) {
# 	.Call("MDP_FixAction", mdp$ptr, as.integer(sId), as.integer(iA), PACKAGE="MDP")
# 	invisible(NULL)
# }
#
#
# #' Remove the action of a state from the HMDP.
# #'
# #' The actions can be reset using \code{resetActions}.
# #'
# #' @param mdp The MDP loaded using \link{loadMDP}.
# #' @param sId The state id of the state we want to remove the action for.
# #' @param iA  The action index of the state.
# #' @return Nothing.
# #' @author Lars Relund \email{lars@@relund.dk}
# #' @seealso \code{\link{resetActions}}, \code{\link{fixAction}}.
# #' @example inst/examples/machine.R
# #' @export
# removeAction<-function(mdp, sId, iA) {
# 	.Call("MDP_RemoveAction", mdp$ptr, as.integer(sId), as.integer(iA), PACKAGE="MDP")
# 	invisible(NULL)
# }
#
#
# #' Reset the actions of a state.
# #'
# #' @param mdp The MDP loaded using \link{loadMDP}.
# #' @return Nothing.
# #' @author Lars Relund \email{lars@@relund.dk}
# #' @seealso \code{\link{resetActions}}, \code{\link{fixAction}}.
# #' @example inst/examples/machine.R
# #' @export
# resetActions<-function(mdp) {
# 	.Call("MDP_ResetActions", mdp$ptr, PACKAGE="MDP")
# 	invisible(NULL)
# }
#

#
#
# #' Set the weight of a state.
# #'
# #' @param mdp The MDP loaded using \link{loadMDP}.
# #' @param w The weight.
# #' @param sId The state id of the state.
# #' @param wLbl The label of the weight we consider.
# #' @return Nothing.
# #' @author Lars Relund \email{lars@@relund.dk}
# #' @export
# setStateWeight<-function(mdp, w, sId, wLbl) {
# 	iW<-getWIdx(mdp,wLbl)
# 	.Call("MDP_SetStateW", mdp$ptr, as.numeric(w), as.integer(sId), as.integer(iW), PACKAGE="MDP")
# 	invisible(NULL)
# }
#
#
# #' Return ids for states in a stage.
# #'
# #' @param mdp The MDP loaded using \link{loadMDP}.
# #' @param stages A char vector of index in the form "n0,s0,a0,n1", i.e. 3*level+1 elements in the string.
# #' @return A vector of ids for the states.
# #' @author Lars Relund \email{lars@@relund.dk}
# #' @example inst/examples/machine.R
# #' @export
# getIdSStages<-function(mdp, stages) {
# 	v<-.Call("MDP_GetIdSStage", mdp$ptr, as.character(stages), PACKAGE="MDP")
# 	return(v)
# }
#
#
# #' Return the index strings for states having id idS.
# #'
# #' @param mdp The MDP loaded using \link{loadMDP}.
# #' @param idS A vector of state ids.
# #' @return A vector of index for the states.
# #' @author Lars Relund \email{lars@@relund.dk}
# #' @example inst/examples/machine.R
# #' @export
# getStrIdxS<-function(mdp, idS) {
# 	n<- mdp$states + ifelse(mdp$timeHorizon>=Inf,mdp$founderStatesLast,0)
# 	idS <- idS[idS<n & idS>=0]
# 	v<-.Call("MDP_GetIdxS", mdp$ptr, as.integer(idS), PACKAGE="MDP")
# 	return(v)
# }
#
#
# #' Return the label of states having id idS.
# #'
# #' @param mdp The MDP loaded using \link{loadMDP}.
# #' @param idS A vector of state ids.
# #' @return A vector of labels for the states.
# #' @author Lars Relund \email{lars@@relund.dk}
# #' @example inst/examples/machine.R
# #' @export
# getLabel<-function(mdp, idS) {
# 	n<- mdp$states + ifelse(mdp$timeHorizon>=Inf,mdp$founderStatesLast,0)
# 	idS <- idS[idS<n & idS>=0]
# 	v<-.Call("MDP_GetLabel", mdp$ptr, as.integer(idS), PACKAGE="MDP")
# 	return(v)
# }
#
#
# #' Get the weights of an action.
# #'
# #' @param mdp The MDP loaded using \link{loadMDP}.
# #' @param idS The state id.
# #' @param idxA The action index.
# #' @return A vector of weights for the action.
# #' @author Lars Relund \email{lars@@relund.dk}
# #' @example inst/examples/machine.R
# #' @export
# getActionW<-function(mdp, idS, idxA) {
# 	l<-info(mdp, idS[1])
# 	l<-l[[1]]$actions[idxA+1]
# 	l<-substring(l,regexpr("w",l)+3)
# 	l<-gsub(").*","",l)
# 	zz<-textConnection(l)
# 	l<-scan(zz, sep=",", quiet = TRUE)
# 	close(zz)
# 	return(l)
# }
#
#
# #' Get the ids of the transition states of an action.
# #'
# #' @param mdp The MDP loaded using \link{loadMDP}.
# #' @param idS The state id.
# #' @param idxA The action index.
# #' @return A vector of weights for the action.
# #' @author Lars Relund \email{lars@@relund.dk}
# #' @example inst/examples/machine.R
# #' @export
# getActionTransIdS<-function(mdp, idS, idxA) {
# 	l<-info(mdp, idS[1])
# 	l<-l[[1]]$actions[idxA+1]
# 	l<-substring(l,regexpr("trans",l)+7)
# 	l<-gsub(").*","",l)
# 	zz<-textConnection(l)
# 	l<-scan(zz, sep=",", quiet = TRUE)
# 	close(zz)
# 	return(l)
# }
#
#
# #' Get the transition probabilities of the transition states of an action.
# #'
# #' @param mdp The MDP loaded using \link{loadMDP}.
# #' @param idS The state id.
# #' @param idxA The action index (c++ style starting from zero).
# #' @return A vector of weights for the action.
# #' @author Lars Relund \email{lars@@relund.dk}
# #' @example inst/examples/machine.R
# #' @export
# getActionTransPr<-function(mdp, idS, idxA) {
#   return(.Call("MDP_GetActionTransPr", mdp$ptr, as.integer(idS), as.integer(idxA), PACKAGE="MDP") )
# }
# # getActionTransPr<-function(mdp, idS, idxA) {
# # 	l<-info(mdp, idS[1])
# # 	l<-l[[1]]$actions[idxA+1]
# # 	l<-substring(l,regexpr("pr",l)+4)
# # 	l<-gsub(").*","",l)
# # 	zz<-textConnection(l)
# # 	l<-scan(zz, sep=",", quiet = TRUE)
# # 	close(zz)
# # 	return(l)
# # }
#
#

#
# #' Get the transition probability matrix P for the founder process (level 0).
# #'
# #' @param mdp The MDP loaded using \link{loadMDP}.
# #' @return The state probability matrix.
# #' @author Lars Relund \email{lars@@relund.dk}
# #' @export
# getTransPr<-function(mdp) {
# 	v<-.Call("MDP_GetTransPr", mdp$ptr, PACKAGE="MDP")
# 	v<-matrix(v,nrow=mdp$states)
# 	return(v)
# }

library(MDP2)
cleanUp<-function() unlink("*.bin")

test_that("steadyStatePr",{
   #sprintf("%.100f",g)
   source("files/HCT_ex6.1.1.R")
   mdp<-loadMDP("hct611_", getLog = FALSE)
   runPolicyIteAve(mdp,"Net reward","Duration", getLog = FALSE)
   pr<-getSteadyStatePr(mdp)
   expect_equal(pr,c(0.56603774, 0.28301887, 0.09433962, 0.02358491, 0.03301887), tolerance = 1.5e-7)
})

cleanUp()

# Create and optimize MDPs or hierarchical MDPs with discrete time steps and state space.
#
# @section To do:
#
# Nested loading in memory (specify a HMDP with special actions containing child + father jump actions)
# Idea when define the proc with an external nested process use
# w$includeProcess(prefix, transPr, index) (specify the child jump action)
#   w$stage()
#     w$state()
#       w$action(...) (specify father jump action scope must be zeor)
#       w$action(...) (specify father jump action scope must be zeor)
#     w$endState()
#   w$endStage()
# w$endIncludeProcess()
#
# The hgf then is formed with a subprocess mimic the 1. and last stage of the external proc, i.e. we include the jump pr in the hgf
# We need a new binary file "externalProcess.bin" for storing the nested process in the format "n0 s0 a0 n1 s1 prefix -1 ..." which specify which stage contain the states corresponding to the 1. stage of the nested process (it is here the nested hfg must be loaded and calculated
#
#
# @name MDP2
# @aliases MDP-package MDP
# @docType package
# @title Markov Decision Processes (MDPs) in R
# @useDynLib MDP2
# @importClassesFrom Rcpp "C++Object"
# NA



#' @seealso [loadMDP()].
#' @keywords internal
#' @useDynLib MDP2
#' @importClassesFrom Rcpp "C++Object"
"_PACKAGE"

## functions related to the binary files

#' Function for writing an HMDP model to binary files. The function defines
#' sub-functions which can be used to define an HMDP model saved in a set of binary
#' files.
#'
#' Binary files are efficient for storing large models. Compared to the HMP (XML)
#' format the binary files use less storage space and loads the model faster.
#'
#' The returned writer exposes these functions:
#'
#' * `setWeights(labels, ...)`: sets the labels of the weights used in the actions.
#'   `labels` is a vector of label names. `...` is currently ignored. Call this
#'   before building the model.
#' * `process()`: starts a (sub)process. It may also be used to specify a
#'   traditional MDP using matrices in `MDPtoolbox` style. In that style, `P` is
#'   a list of matrices, one per action, each of size `$S x S$` where `$S$` is
#'   the number of states. Each used row must sum to one, or all entries in a
#'   row must be zero if unused. `R` is a matrix of size `$S x A$`, where `$A$`
#'   is the number of actions, and `D` is a matrix of size `$S x A$` with
#'   durations. If `D` is omitted, all durations are assumed to be 1.
#' * `endProcess()`: ends a (sub)process.
#' * `stage(label = NULL)`: starts a stage. `label` is currently unused in the
#'   binary format.
#' * `endStage()`: ends a stage.
#' * `state(label = NULL)`: starts a state and returns, invisibly, the state id.
#'   That id can later be referenced with scope 3.
#' * `endState()`: ends a state.
#' * `action(scope = NULL, id = NULL, pr = NULL, prob = NULL, weights,
#'   transWeights = NULL, label = NULL, end = FALSE, ...)`: starts an action.
#'   `weights` must be a vector of action weights. `transWeights` must contain
#'   transition weights ordered by transition, with all transition weight labels
#'   for the first transition followed by all labels for the second transition,
#'   and so on. Transition probabilities can be entered in two ways:
#'
#'   1. `prob` contains triples `(scope, id, pr)`.
#'   2. `id` and `pr` are vectors of equal length. If `scope` is omitted, all
#'      scopes default to 1.
#'
#'   See the description of `actionIdx.bin` below. If `end = TRUE`, calling
#'   `endAction()` is not necessary. `...` is currently ignored.
#' * `endAction()`: ends an action. Do not use this if `end = TRUE` was used
#'   when the action was specified.
#' * `includeProcess(prefix, label = NULL, weights, prob, termStates,
#'   transWeights = NULL)`: includes an external process. External processes are
#'   loaded into memory only when needed, which helps with large models. `prefix`
#'   is the external process prefix. `weights` must be a vector of action
#'   weights, and `prob` must contain triples `(scope, idx, pr)`; see the
#'   description of `actionIdx.bin` below. `termStates` must specify the number
#'   of states at the last stage in the external process. Inside an
#'   `includeProcess ... endIncludeProcess`
#'   block, you must specify the father jump actions of the last stage in the
#'   external process. The external process is represented by its first and last
#'   stage together with its jump actions. The function returns, invisibly, the
#'   state ids of the first stage in the external process, which can later be
#'   referenced with scope 3.
#' * `endIncludeProcess()`: ends an `includeProcess` block.
#' * `closeWriter()`: closes the writer. Call this when the model description is
#'   finished.
#'
#' Ten binary files are created:
#'
#' * `stateIdx.bin`: integers defining all states in the format
#'   `"n0 s0 -1 n0 s0 a0 n1 s1 -1 n0 s0 a0 n1 s1 a1 n2 s2 -1 n0 s0 ..."`.
#'   Here `-1` indicates that a new state is considered.
#' * `stateIdxLbl.bin`: character data in the format `sIdx label sIdx label ...`.
#'   Here `sIdx` corresponds to the index or line number in `stateIdxLbl.bin`,
#'   starting from 0. No delimiter is used.
#' * `actionIdx.bin`: integers defining all actions in the format
#'   `sIdx scope idx scope idx scope idx -1 sIdx scope idx scope idx -1 sIdx scope -1 ...`.
#'   `sIdx` corresponds to the index or line number in `stateIdx.bin`, starting
#'   from 0. The following `(scope, idx)` pairs indicate possible transitions.
#'   Scope can take four values:
#'
#'   * `2`: a transition to a child process, at stage zero in the child process.
#'   * `1`: a transition to the next stage in the current process.
#'   * `0`: a transition to the next stage in the father process.
#'   * `3`: a transition to a state specified by its state `sIdx`.
#'
#'   For example, if `scope = 1` and `idx = 2`, the transition is to state
#'   number 3 at the next stage in the current process. If `scope = 3` and
#'   `idx = 5`, the transition is to the state specified at line 6 in
#'   `stateIdxLbl.bin`. This is useful when considering shared child processes.
#' * `actionIdxLbl.bin`: character data in the format `aIdx label aIdx label ...`.
#'   Here `aIdx` corresponds to the index or line number in `actionIdx.bin`,
#'   starting from 0. No delimiter is used.
#' * `actionWeight.bin`: doubles containing action weights in the format
#'   `"c1 c2 c3 c1 c2 c3 ..."`, assuming three weights for each action.
#' * `actionWeightLbl.bin`: character data containing the weight labels in the
#'   format `label1 label2 label3`, assuming three weights for each action.
#' * `transProb.bin`: doubles containing transition probabilities defined in
#'   `actionIdx.bin`. The format is `"p1 p2 p3 -1 p1 -1 p1 p2 -1 ..."`. Here
#'   `-1` indicates that a new action is considered.
#' * `externalProcesses.bin`: character data containing links to external
#'   processes in the format `stageStr prefix stageStr prefix ...`. Here
#'   `stageStr` corresponds to the stage index, for example `n0 s0 a0 n1`, of
#'   the stage corresponding to the first stage in the external process, and
#'   `prefix` is the external process prefix. No delimiter is used.
#' * `transWeight.bin`: doubles containing transition weights in the format
#'   `"t11 t12 t21 t22 -1 ..."`, assuming two transition weights for each
#'   transition and two transitions in the first action.
#' * `transWeightLbl.bin`: character data containing the transition weight
#'   labels.
#'
#' @param prefix A character string with the prefix added to `binNames`.
#' @param binNames A character vector giving the names of the binary files storing the model.
#' @param getLog Output log text.
#' 
#' @return A list of functions.
#' @note Note all indexes are starting from zero (C/C++ style).
#' @example inst/examples/binaryMDPWriter-ex.R
#' @export
binaryMDPWriter <-
   function(prefix = "",
            binNames = c(
               "stateIdx.bin",
               "stateIdxLbl.bin",
               "actionIdx.bin",
               "actionIdxLbl.bin",
               "actionWeight.bin",
               "actionWeightLbl.bin",
               "transProb.bin",
               "externalProcesses.bin",
               "transWeight.bin",
               "transWeightLbl.bin"
            ),
            getLog = TRUE
   )
{
   pushContext <- function(value) {
      writerContext <<- c(writerContext, value)
      invisible(NULL)
   }

   popContext <- function() {
      writerContext <<- writerContext[-length(writerContext)]
      invisible(NULL)
   }

   currentContext <- function() {
      if (length(writerContext) == 0) return(NULL)
      writerContext[length(writerContext)]
   }

   requireContext <- function(expected, message) {
      if (!identical(currentContext(), expected)) stop(message, call. = FALSE)
      invisible(NULL)
   }

   setWeights<-function(labels,...){
      if (wFixed) stop("Weights already added!")
      wCtr<<-length(labels)
      writeBin(as.character(labels), fACostLbl)
      wFixed<<-TRUE
      invisible(NULL)
   }

   setTransWeights<-function(labels,...){
      if (tWFixed) stop("Transition weights already added!")
      tWCtr<<-length(labels)
      writeBin(as.character(labels), fTransWLbl)
      tWFixed<<-TRUE
      invisible(NULL)
   }
   
   process<-function(P=NULL, R=NULL, D=NULL, .fromInclude = FALSE){
      if (!wFixed)
         stop("Weights must be added using 'setWeights' before starting building the HMDP!")
      if (length(writerContext)>0 && !identical(currentContext(), "action")) {
         stop("Cannot start a process before closing the current writer block.", call. = FALSE)
      }
      if (.fromInclude) {
         requireContext("action", "Cannot start an included process unless an include-process action is open.")
      }
      pushContext("process")
      dCtr<<- -1  # reset stage ctr
      sIdx<<-c(sIdx,NA)
      if (!is.null(P) & !is.null(R)) { # MDP specified using MDPtoolbox style
         if (is.null(D)) D<-matrix(1,nrow = nrow(R), ncol = ncol(R))
         stage()
         for (i in 1:nrow(R)) {
            state(label=i)
            for (j in 1:ncol(R)) {
               jIdx<-which(P[[j]][i,]>0)
               if (length(jIdx)==0) next
               action(label=j, pr=P[[j]][i,jIdx], id = jIdx-1, weights = c(D[i,j],R[i,j]), end = TRUE)
            }
            endState()
         }
         endStage()
         endProcess()
      }
      invisible(NULL)
   }
   
   endProcess<-function(){
      requireContext("process", "Cannot end a process unless a process is open.")
      if (length(sIdx)>1) sIdx<<-sIdx[1:(length(sIdx)-1)] else sIdx<<-NULL
      # set ctr's for current level
      dCtr<<-idx[length(idx)-2]
      sCtr<<-idx[length(idx)-1]
      aCtr<<-idx[length(idx)]
      popContext()
      invisible(NULL)
   }
   
   stage<-function(label=NULL){
      requireContext("process", "Cannot start a stage outside an open process.")
      pushContext("stage")
      dCtr<<-dCtr+1
      sCtr<<- -1  # reset state ctr
      idx<<-c(idx,dCtr)   # add stage idx
      #cat(paste("d:(",paste(c(idx),collapse=","),"),",dCtr,"|",sep=""))
      invisible(NULL)
   }
   
   endStage<-function(){
      requireContext("stage", "Cannot end a stage unless a stage is open.")
      if (length(idx)>1) idx<<-idx[1:(length(idx)-1)] else idx<<-NULL     # remove stage index
      #cat(paste("-d:(",paste(c(idx),collapse=","),"),",dCtr,"|",sep=""))
      popContext()
      invisible(NULL)
   }
   
   state<-function(label=NULL, end=FALSE){
      requireContext("stage", "Cannot start a state outside an open stage.")
      lastAutoClosedAction <<- FALSE
      pushContext("state")
      #cat("(",label,") ",sep="")
      sCtr<<-sCtr+1
      aCtr<<- -1  # reset action ctr
      idx<<-c(idx,sCtr)   # add state idx
      writeBin(as.integer(c(idx,-1)), fS)
      sRowId<<- sRowId+1
      sIdx[length(sIdx)]<<-sRowId
      #cat(paste("s:(",paste(c(idx),collapse=","),")",sRowId,"|",sep=""))
      if (!is.null(label)) writeBin(c(as.character(sRowId),label), fSLbl)   # sRowId added before label
      if (end) endState()
      invisible(sRowId)
   }
   
   endState<-function(){
      requireContext("state", "Cannot end a state while another writer block is open. Call endAction() or use action(..., end = TRUE) before endState().")
      idx<<-idx[1:(length(idx)-1)]    # remove state index
      #cat(paste("-s:(",paste(c(idx),collapse=","),")|",sep=""))
      popContext()
      invisible(NULL)
   }
   
   action <-
      function(scope = NULL,
               id = NULL,
               pr = NULL,
               prob = NULL,
               weights,
               transWeights = NULL,
               label = NULL,
               end = FALSE,
               ...) {
         requireContext("state", "Cannot start an action outside an open state.")
         lastAutoClosedAction <<- FALSE
         pushContext("action")
         # prop contain tripeles (scope,idx,prob)
	      #cat("action:\n")
      #print(weights)
      #print(prob)
      #if (is.null(label) | label=="") stop("label = null");
      #if (length(weights)!=wCtr) stop("Weight length must be ",wCtr,"!")
      aCtr<<-aCtr+1
      idx<<-c(idx,aCtr)   # add action idx
      #cat(paste("a:(",paste(c(idx),collapse=","),")|",sep=""))
      #cat(paste("a: sId=",sIdx[length(sIdx)],"|",sep=""))
      scpIdx<-NULL
      aRowId<<- aRowId+1
      if (!is.null(prob)) {
         for (i in 0:(length(prob)/3-1)) scpIdx<-c(scpIdx,prob[1:2+3*i])
         probs<-prob[1:(length(prob)/3)*3]
         writeBin(as.integer(c(sIdx[length(sIdx)],scpIdx,-1)), fA)
         writeBin(as.numeric(c(probs,-1)), fTransP)
         #cat("end action\n")
      } else if (!is.null(pr)) {
         #cat("pr:",paste0(pr,collapse = ",")," id:",paste0(id,collapse = ",")," w:",paste0(weights,collapse = ","),"\n"); cat
         if (is.null(scope)) scope<-rep(1,length(pr))
         i<-1:length(pr)-1
         scpIdx[1+i*2]<-scope
         scpIdx[2+i*2]<-id
         writeBin(as.integer(c(sIdx[length(sIdx)],scpIdx,-1)), fA)
         writeBin(as.numeric(c(pr,-1)), fTransP)
      }
      nTrans <- length(scpIdx)/2
      if (tWCtr>0) {
         if (is.null(transWeights)) transWeights <- rep(0, nTrans * tWCtr)
         if (length(transWeights) != nTrans * tWCtr)
            stop("transWeights must have length number of transitions times number of transition weights.")
         writeBin(as.numeric(c(transWeights,-1)), fTransW)
      }
      writeBin(as.numeric(weights), fACost)
      if (!is.null(label)) writeBin(c(as.character(aRowId),label), fALbl)   # aRowId added before label
      if (end) {
         endAction()
         lastAutoClosedAction <<- TRUE
      }
      invisible(NULL)
   }
   
   endAction<-function(){
      if (!identical(currentContext(), "action") && identical(currentContext(), "state") && lastAutoClosedAction) {
         lastAutoClosedAction <<- FALSE
         return(invisible(NULL))
      }
      requireContext("action", "Cannot end an action unless an action is open.")
      idx<<-idx[1:(length(idx)-1)]    # remove action index
      #cat(paste("-a:(",paste(c(idx),collapse=","),")|",sep=""))
      popContext()
      lastAutoClosedAction <<- FALSE
      invisible(NULL)
   }
   
   includeProcess<-function(prefix, label=NULL, weights, prob, termStates, transWeights = NULL){     # prop contain tripeles (scope,idx,prob) - Here all scope must be 2!!
      requireContext("state", "Cannot include a process outside an open state.")
      pushContext("action")
      stateId<-NULL # to store state id's
      #cat("action:\n")
      #print(weights)
      #print(prob)
      #if (is.null(label) | label=="") stop("label = null");
      #if (length(weights)!=wCtr) stop("Weight length must be ",wCtr,"!")
      aCtr<<-aCtr+1
      idx<<-c(idx,aCtr)   # add action idx
      #cat(paste("a:(",paste(c(idx),collapse=","),")|",sep=""))
      #cat(paste("a: sId=",sIdx[length(sIdx)],"|",sep=""))
      scpIdx<-NULL
      aRowId<<- aRowId+1
      for (i in 0:(length(prob)/3-1)) scpIdx<-c(scpIdx,prob[1:2+3*i])
      probs<-prob[1:(length(prob)/3)*3]
      #        if (any(scpIdx<0) | any(probs<0)) {
      #            print(label)
      #            print(prob)
      #            print(scpIdx)
      #            print(probs)
      #            stop()
      #        }
      writeBin(as.integer(c(sIdx[length(sIdx)],scpIdx,-1)), fA)
      if (!is.null(label)) writeBin(c(as.character(aRowId),label), fALbl)   # aRowId added before label
      writeBin(as.numeric(c(probs,-1)), fTransP)
      if (tWCtr>0) {
         nTrans <- length(scpIdx)/2
         if (is.null(transWeights)) transWeights <- rep(0, nTrans * tWCtr)
         if (length(transWeights) != nTrans * tWCtr)
            stop("transWeights must have length number of transitions times number of transition weights.")
         writeBin(as.numeric(c(transWeights,-1)), fTransW)
      }
      writeBin(as.numeric(weights), fACost)
      #cat("end action\n")
      maxId<-max(scpIdx[2*(1:(length(scpIdx)/2))])  # number of states to create at the first stage of the child
      process(.fromInclude = TRUE)  # start external subprocess
      stage()  # first stage of the external process
      writeBin(c(paste(idx,collapse=","), prefix), fExt)  # store the external process' name
      pr<-as.numeric( t(matrix(c(rep(1,termStates), 1:termStates-1, rep(1/termStates,termStates)), ncol=3)) )
      for (i in 0:maxId) {
         # create the states in the first stage (with no actions)
         stateId <- c(stateId, state())
         action(weights = rep(0, length(weights)),
                prob = pr,
                end = TRUE) # dummy action of external process with transition to all terminal states
         endState()
      }
      endStage()
      # now the user has to include the last stage using the normal syntax
      invisible(stateId)
   }
   
   endIncludeProcess<-function() {
      endProcess()   # end external subprocess
      requireContext("action", "Cannot end an included process unless an include-process action is open.")
      idx<<-idx[1:(length(idx)-1)]    # remove action index
      #cat(paste("-a:(",paste(c(idx),collapse=","),")|",sep=""))
      popContext()
      lastAutoClosedAction <<- FALSE
      invisible(NULL)
   }
   
   closeWriter<-function(){
      if (length(writerContext)>0) {
         stop(
            paste0("Cannot close writer while a ", currentContext(), " is still open."),
            call. = FALSE
         )
      }
      if (getLog) {
         cat("\n  Statistics:\n")
         cat("    states :",sRowId+1,"\n")
         cat("    actions:",aRowId+1,"\n")
         cat("    weights:",wCtr,"\n\n")
         cat("  Closing binary MDP writer.\n\n") 
      }
      close(fS)
      close(fSLbl)
      close(fA)
      close(fALbl)
      close(fACost)
      close(fACostLbl)
      close(fTransP)
      close(fExt)
      close(fTransW)
      close(fTransWLbl)
      invisible(NULL)
   }
   
   binNames<-paste(prefix,binNames,sep="")
   fS <- file(binNames[1], "wb")
   fSLbl <- file(binNames[2], "wb")
   fA <- file(binNames[3], "wb")
   fALbl <- file(binNames[4], "wb")
   fACost <- file(binNames[5], "wb")
   fACostLbl <- file(binNames[6], "wb")
   fTransP <- file(binNames[7], "wb")
   fExt <- file(binNames[8], "wb")
   fTransW <- file(binNames[9], "wb")
   fTransWLbl <- file(binNames[10], "wb")
   idx<-NULL  # containing the stage, state or action idx's
   sIdx<-NULL # containing the state row id's (used to find the state id the action is defined under)
   dCtr<- -1   # current stage at current level
   sCtr<- -1   # current state at current stage
   aCtr<- -1   # current action at current state
   wCtr<- 0    # number of weights in the model
   tWCtr<- 0    # number of transition weights in the model
   sRowId<- -1    # current row/line of state in stateIdx file
   aRowId<- -1    # current row/line of action in actionIdx file
   wFixed<-FALSE  # TRUE if size of weights are fixed
   tWFixed<-FALSE  # TRUE if size of transition weights are fixed
   writerContext<-character()
   lastAutoClosedAction<-FALSE
   v <-
      list(
         setWeights = setWeights,
         setTransWeights = setTransWeights,
         stage = stage,
         endStage = endStage,
         state = state,
         endState = endState,
         action = action,
         endAction = endAction,
         includeProcess = includeProcess,
         endIncludeProcess = endIncludeProcess,
         process = process,
         endProcess = endProcess,
         closeWriter = closeWriter
      )
   class(v) <- c("binaryMDPWriter")
   return(v)
}

#' Function for building an HMDP model directly in memory.
#'
#' `memoryMDPWriter()` defines the same main sub-functions as
#' [binaryMDPWriter()], but stores states and actions directly in C++ memory
#' instead of writing intermediate binary files. `closeWriter()` compiles the
#' model and returns the loaded `"HMDP"` object.
#'
#' External or included processes are not supported by `memoryMDPWriter()`.
#'
#' @param prefix A character string kept for compatibility and stored in the
#'   returned object metadata.
#' @param eps The sum of transition probabilities must at most differ `eps`
#'   from one when `check = TRUE`.
#' @param check Check if the MDP seems correct before returning it.
#' @param verbose More output when compiling and running algorithms.
#' @param getLog Output the log messages.
#' @return A list of functions. Calling `closeWriter()` returns an `"HMDP"`
#'   object.
#' @note Note all indexes are starting from zero (C/C++ style).
#' @example inst/examples/memoryMDPWriter-ex.R
#' @export
memoryMDPWriter <- function(prefix = "",
                            eps = 0.00001,
                            check = TRUE,
                            verbose = FALSE,
                            getLog = TRUE) {
   if (!is.logical(verbose)) verbose <- FALSE
   builder <- methods::new(HMDPBuilder, verbose)
   closed <- FALSE

   assertOpen <- function() {
      if (closed) stop("memoryMDPWriter is closed.", call. = FALSE)
      invisible(NULL)
   }

   pushContext <- function(value) {
      writerContext <<- c(writerContext, value)
      invisible(NULL)
   }

   popContext <- function() {
      writerContext <<- writerContext[-length(writerContext)]
      invisible(NULL)
   }

   currentContext <- function() {
      if (length(writerContext) == 0) return(NULL)
      writerContext[length(writerContext)]
   }

   requireContext <- function(expected, message) {
      if (!identical(currentContext(), expected)) stop(message, call. = FALSE)
      invisible(NULL)
   }

   setWeights <- function(labels, ...) {
      assertOpen()
      if (wFixed) stop("Weights already added!")
      wCtr <<- length(labels)
      builder$setWeights(as.character(labels))
      wFixed <<- TRUE
      invisible(NULL)
   }

   setTransWeights <- function(labels, ...) {
      assertOpen()
      if (tWFixed) stop("Transition weights already added!")
      tWCtr <<- length(labels)
      builder$setTransWeights(as.character(labels))
      tWFixed <<- TRUE
      invisible(NULL)
   }

   process <- function(P = NULL, R = NULL, D = NULL, .fromInclude = FALSE) {
      assertOpen()
      if (!wFixed)
         stop("Weights must be added using 'setWeights' before starting building the HMDP!")
      if (.fromInclude) {
         stop("memoryMDPWriter() does not support external processes.", call. = FALSE)
      }
      if (length(writerContext)>0 && !identical(currentContext(), "action")) {
         stop("Cannot start a process before closing the current writer block.", call. = FALSE)
      }
      pushContext("process")
      dCtr <<- -1
      sIdx <<- c(sIdx, NA)
      if (!is.null(P) & !is.null(R)) {
         if (is.null(D)) D <- matrix(1, nrow = nrow(R), ncol = ncol(R))
         stage()
         for (i in 1:nrow(R)) {
            state(label = i)
            for (j in 1:ncol(R)) {
               jIdx <- which(P[[j]][i,]>0)
               if (length(jIdx)==0) next
               action(label = j, pr = P[[j]][i,jIdx], id = jIdx-1,
                      weights = c(D[i,j], R[i,j]), end = TRUE)
            }
            endState()
         }
         endStage()
         endProcess()
      }
      invisible(NULL)
   }

   endProcess <- function() {
      assertOpen()
      requireContext("process", "Cannot end a process unless a process is open.")
      if (length(sIdx)>1) sIdx <<- sIdx[1:(length(sIdx)-1)] else sIdx <<- NULL
      dCtr <<- idx[length(idx)-2]
      sCtr <<- idx[length(idx)-1]
      aCtr <<- idx[length(idx)]
      popContext()
      invisible(NULL)
   }

   stage <- function(label = NULL) {
      assertOpen()
      requireContext("process", "Cannot start a stage outside an open process.")
      pushContext("stage")
      dCtr <<- dCtr+1
      sCtr <<- -1
      idx <<- c(idx, dCtr)
      invisible(NULL)
   }

   endStage <- function() {
      assertOpen()
      requireContext("stage", "Cannot end a stage unless a stage is open.")
      if (length(idx)>1) idx <<- idx[1:(length(idx)-1)] else idx <<- NULL
      popContext()
      invisible(NULL)
   }

   state <- function(label = NULL, end = FALSE) {
      assertOpen()
      requireContext("stage", "Cannot start a state outside an open stage.")
      lastAutoClosedAction <<- FALSE
      pushContext("state")
      sCtr <<- sCtr+1
      aCtr <<- -1
      idx <<- c(idx, sCtr)
      sRowId <<- builder$addState(as.integer(idx), if (is.null(label)) "" else as.character(label))
      sIdx[length(sIdx)] <<- sRowId
      if (end) endState()
      invisible(sRowId)
   }

   endState <- function() {
      assertOpen()
      requireContext("state", "Cannot end a state while another writer block is open. Call endAction() or use action(..., end = TRUE) before endState().")
      idx <<- idx[1:(length(idx)-1)]
      popContext()
      invisible(NULL)
   }

   action <- function(scope = NULL,
                      id = NULL,
                      pr = NULL,
                      prob = NULL,
                      weights,
                      transWeights = NULL,
                      label = NULL,
                      end = FALSE,
                      ...) {
      assertOpen()
      requireContext("state", "Cannot start an action outside an open state.")
      lastAutoClosedAction <<- FALSE
      pushContext("action")
      aCtr <<- aCtr+1
      idx <<- c(idx, aCtr)
      aRowId <<- aRowId+1
      scpIdx <- NULL
      probs <- NULL
      if (!is.null(prob)) {
         for (i in 0:(length(prob)/3-1)) scpIdx <- c(scpIdx, prob[1:2+3*i])
         probs <- prob[1:(length(prob)/3)*3]
      } else if (!is.null(pr)) {
         if (is.null(scope)) scope <- rep(1, length(pr))
         i <- 1:length(pr)-1
         scpIdx[1+i*2] <- scope
         scpIdx[2+i*2] <- id
         probs <- pr
      } else {
         stop("Either 'pr' or 'prob' must be provided.", call. = FALSE)
      }
      nTrans <- length(scpIdx)/2
      if (nTrans == 0) stop("An action must define at least one transition.", call. = FALSE)
      if (tWCtr>0) {
         if (is.null(transWeights)) transWeights <- rep(0, nTrans * tWCtr)
         if (length(transWeights) != nTrans * tWCtr)
            stop("transWeights must have length number of transitions times number of transition weights.")
      } else {
         transWeights <- numeric()
      }
      builder$addAction(
         as.integer(sIdx[length(sIdx)]),
         as.integer(scpIdx[2*(1:nTrans)-1]),
         as.integer(scpIdx[2*(1:nTrans)]),
         as.numeric(probs),
         as.numeric(weights),
         as.numeric(transWeights),
         if (is.null(label)) "" else as.character(label)
      )
      if (end) {
         endAction()
         lastAutoClosedAction <<- TRUE
      }
      invisible(NULL)
   }

   endAction <- function() {
      assertOpen()
      if (!identical(currentContext(), "action") && identical(currentContext(), "state") && lastAutoClosedAction) {
         lastAutoClosedAction <<- FALSE
         return(invisible(NULL))
      }
      requireContext("action", "Cannot end an action unless an action is open.")
      idx <<- idx[1:(length(idx)-1)]
      popContext()
      lastAutoClosedAction <<- FALSE
      invisible(NULL)
   }

   includeProcess <- function(...) {
      assertOpen()
      stop("memoryMDPWriter() does not support external processes.", call. = FALSE)
   }

   endIncludeProcess <- function(...) {
      assertOpen()
      stop("memoryMDPWriter() does not support external processes.", call. = FALSE)
   }

   closeWriter <- function() {
      assertOpen()
      if (length(writerContext)>0) {
         stop(
            paste0("Cannot close writer while a ", currentContext(), " is still open."),
            call. = FALSE
         )
      }
      mdpPtr <- builder$close()
      closed <<- TRUE
      builder <<- NULL
      if (getLog) {
         cat("\n  Statistics:\n")
         cat("    states :", sRowId+1, "\n")
         cat("    actions:", aRowId+1, "\n")
         cat("    weights:", wCtr, "\n\n")
         cat("  Closing memory MDP writer.\n\n")
      }
      .makeMDPList(
         mdpPtr,
         binNames = paste0(prefix, "<memory>"),
         eps = eps,
         check = check,
         getLog = getLog
      )
   }

   idx <- NULL
   sIdx <- NULL
   dCtr <- -1
   sCtr <- -1
   aCtr <- -1
   wCtr <- 0
   tWCtr <- 0
   sRowId <- -1
   aRowId <- -1
   wFixed <- FALSE
   tWFixed <- FALSE
   writerContext <- character()
   lastAutoClosedAction <- FALSE
   v <- list(
      setWeights = setWeights,
      setTransWeights = setTransWeights,
      stage = stage,
      endStage = endStage,
      state = state,
      endState = endState,
      action = action,
      endAction = endAction,
      includeProcess = includeProcess,
      endIncludeProcess = endIncludeProcess,
      process = process,
      endProcess = endProcess,
      closeWriter = closeWriter
   )
   class(v) <- c("memoryMDPWriter")
   return(v)
}


#' Function for writing actions of a HMDP model to binary files. The function defines
#' sub-functions which can be used to define actions saved in a set of binary
#' files. It is assumed that the states have been defined using `binaryMDPWriter` 
#' and that the id of the states is known (can be retrieved using e.g. `stateIdxDf`). 
#'
#' Binary files are efficient for storing large models. Compared to the HMP (XML)
#' format the binary files use less storage space and loading the model is faster.
#'
#' The returned writer exposes these functions:
#'
#' * `setWeights(labels, ...)`: sets the labels of the weights used in the
#'   actions. `labels` is a vector of label names. `...` is currently ignored.
#'   Call this before building the model.
#' * `addAction(label = NULL, sIdx, weights, prob, ...)`: adds an action. `sIdx`
#'   is the id of the state defining the action. `weights` must be a vector of
#'   action weights. `prob` is a matrix `(sIdx, pr)` where the first column
#'   contains the id of the transition state; see the description of
#'   `actionIdx.bin` below, where scope is assumed to be 3. `...` is currently
#'   ignored.
#' * `endAction()`: ends an action.
#' * `closeWriter()`: closes the writer. Call this when the model description is
#'   finished.
#'
#' Five binary files are created:
#'
#' * `actionIdx.bin`: integers defining all actions in the format
#'   `sIdx scope idx scope idx scope idx -1 sIdx scope idx scope idx -1 sIdx scope -1 ...`.
#'   `sIdx` corresponds to the index or line number in `stateIdx.bin`, starting
#'   from 0. The following `(scope, idx)` pairs indicate possible transitions.
#'   Scope can take four values:
#'
#'   * `2`: a transition to a child process, at stage zero in the child process.
#'   * `1`: a transition to the next stage in the current process.
#'   * `0`: a transition to the next stage in the father process.
#'   * `3`: a transition to a state specified by its state `sIdx`.
#'
#'   For example, if `scope = 1` and `idx = 2`, the transition is to state
#'   number 3 at the next stage in the current process. If `scope = 3` and
#'   `idx = 5`, the transition is to the state specified at line 6 in
#'   `stateIdxLbl.bin`. This is useful when considering shared child processes.
#' * `actionIdxLbl.bin`: character data in the format `aIdx label aIdx label ...`.
#'   Here `aIdx` corresponds to the index or line number in `actionIdx.bin`,
#'   starting from 0. No delimiter is used.
#' * `actionWeight.bin`: doubles containing action weights in the format
#'   `"c1 c2 c3 c1 c2 c3 ..."`, assuming three weights for each action.
#' * `actionWeightLbl.bin`: character data containing the weight labels in the
#'   format `label1 label2 label3`, assuming three weights for each action.
#' * `transProb.bin`: doubles containing the transition probabilities defined in
#'   `actionIdx.bin`. The format is `"p1 p2 p3 -1 p1 -1 p1 p2 -1 ..."`. Here
#'   `-1` indicates that a new action is considered.
#'
#' @param prefix A character string with the prefix added to `binNames`.
#' @param binNames A character vector of length 5 giving the names of the binary
#'     files storing the model.
#' @param append Logical indicating whether should keep the currents actions (default - TRUE) 
#' defined or delete them and start over (FALSE).
#' @return A list of functions.
#' @note Note all indexes are starting from zero (C/C++ style).
#' @example inst/examples/binaryMDPWriter-ex.R
#' @export
binaryActionWriter <- function(prefix = "",
                               binNames = c(
                                  "actionIdx.bin",
                                  "actionIdxLbl.bin",
                                  "actionWeight.bin",
                                  "actionWeightLbl.bin",
                                  "transProb.bin",
                                  "transWeight.bin",
                                  "transWeightLbl.bin"
                               ),
                               append = TRUE
)
{
   setWeights<-function(labels,...){
      if (wFixed) stop("Weights already added!")
      wCtr<<-length(labels)
      writeBin(as.character(labels), fACostLbl)
      wFixed<<-TRUE
      invisible(NULL)
   }

   setTransWeights<-function(labels,...){
      if (tWFixed) stop("Transition weights already added!")
      tWCtr<<-length(labels)
      writeBin(as.character(labels), fTransWLbl)
      tWFixed<<-TRUE
      invisible(NULL)
   }
   
   addAction<-function(label=NULL, sIdx, weights, prob, transWeights = NULL, ...){     # do not hold now: prop is a matrix with columns (idS,prob)
      # 		cat("action:\n")
      # 		print(weights)
      # 		print(prob)
      #		if (length(weights)!=wCtr) stop("Weight length must be ",wCtr,"!")
      #cat(paste("a:(",paste(c(idx),collapse=","),")|",sep=""))
      #cat(paste("a: sId=",sIdx[length(sIdx)],"|",sep=""))
      aRowId<<- aRowId+1
      scpIdx<-NULL
      for (i in 0:(length(prob)/3-1)) scpIdx<-c(scpIdx,prob[1:2+3*i])
      probs<-prob[1:(length(prob)/3)*3]    
      writeBin(as.integer(c(sIdx,scpIdx,-1)), fA)
      if (!is.null(label)) writeBin(c(as.character(aRowId),label), fALbl)   # aRowId added before label
      writeBin(as.numeric(c(probs,-1)), fTransP)
      if (tWCtr>0) {
         nTrans <- length(scpIdx)/2
         if (is.null(transWeights)) transWeights <- rep(0, nTrans * tWCtr)
         if (length(transWeights) != nTrans * tWCtr)
            stop("transWeights must have length number of transitions times number of transition weights.")
         writeBin(as.numeric(c(transWeights,-1)), fTransW)
      }
      writeBin(as.numeric(weights), fACost)
      #cat("end action\n")
      invisible(NULL)
   }
   
   closeWriter<-function(){
      if (!wFixed) stop("Weights must be added using 'setWeights'!")
      cat("\n  Statistics:\n")
      cat("    actions:",aRowId+1,"\n")
      cat("  Closing binary Action writer.\n\n")
      close(fA)
      close(fALbl)
      close(fACost)
      close(fACostLbl)
      close(fTransP)
      close(fTransW)
      close(fTransWLbl)
      invisible(NULL)
   }
   
   binNames<-paste(prefix,binNames,sep="")
   if (append) {
      # find number of actions already written
      tmp<-readBin(binNames[1], integer(),n=file.info(binNames[1])$size/4)
      aRowId<-length(tmp[tmp==-1])-1 # current number of actions defined
      wFixed<-TRUE  # TRUE if size of weights are fixed
      tWFixed<-TRUE
      tWCtr<-0
   } else {
      aRowId<- -1    # current row/line of action in actionIdx file
      wCtr<- 0    # number of weights in the model
      tWCtr<- 0
      wFixed<-FALSE  # TRUE if size of weights are fixed
      tWFixed<-FALSE
   }
   mode <- ifelse(append,"ab","wb")
   fA <- file(binNames[1], mode)
   fALbl <- file(binNames[2], mode)
   fACost <- file(binNames[3], mode)
   fACostLbl <- file(binNames[4], mode)
   fTransP <- file(binNames[5], mode)
   fTransW <- file(binNames[6], mode)
   fTransWLbl <- file(binNames[7], mode)
   v <- list(setWeights = setWeights, setTransWeights = setTransWeights, addAction = addAction, closeWriter = closeWriter)
   class(v) <- c("binaryActionWriter")
   return(v)
}



#' Info about the states in the binary files of the HMDP model under consideration.
#'
#' @param prefix A character string with the prefix added to til binary files.
#' @param labels Should labels be extracted.
#' @param stateStr Should state strings be extracted. If false then add columns (n0, s0, a0, ...)
#'   where n0 the index of the stage at level 0, s0 the index of the state and a0 the index of the
#'   action. If the HMDP has more than one level columns index (d1, s1, a1, ...) are added.
#' @param fileS The binary file containing the description of states.
#' @param labelS The binary file containing the state labels.
#'
#' @note The model don't have to be loaded, i.e only read the binary files. The state id (`sId`) will
#' not be the same as in the loaded model!
#'
#' @return A data frame with the information.
#' @export
getBinInfoStates <-
   function(prefix = "",
            labels = TRUE,
            stateStr = TRUE,
            fileS = "stateIdx.bin",
            labelS = "stateIdxLbl.bin"
   ) {
   fileS <- paste(prefix, fileS, sep = "")
   tmp <- readBin(fileS, integer(), n = file.info(fileS)$size / 4)
   rows <- length(tmp[tmp == -1])
   if (!stateStr) {
      cols <- max(rle(tmp != -1)$length)
      mat <- as.data.frame(matrix(NA, nrow = rows, ncol = cols + 1))
      idx <- c(0, which(tmp == -1))
      for (i in 1:(length(idx) - 1))
         mat[i, 1:(idx[i + 1] - idx[i] - 1) + 1] <-
         tmp[(idx[i] + 1):(idx[i + 1] - 1)]
      levels <- cols %/% 3 + 1
      if (levels == 1)
         colnames(mat) <- c("sId", paste(c("n", "s"), levels - 1, sep = ""))
      if (levels > 1)
         colnames(mat) <-
         c("sId", paste(c("n", "s", "a"), rep(0:(levels - 2), each = 3), sep = ""), paste(c("n", "s"), levels -
                                                                                             1, sep = ""))
   } else {
      mat <- as.data.frame(matrix(NA, nrow = rows, ncol = 2))
      idx <- c(0, which(tmp == -1))
      for (i in 1:(length(idx) - 1))
         mat[i, 2] <- paste(tmp[(idx[i] + 1):(idx[i + 1] - 1)], collapse = ",")
      colnames(mat) <- c("sId", "stageStr")
   }
   mat[, 1] <- 1:nrow(mat) - 1
   if (labels) {
      labelS <- paste(prefix, labelS, sep = "")
      tmp <- readBin(labelS, character(), n = file.info(labelS)$size)
      tmp <-
         as.data.frame(matrix(tmp, ncol = 2, byrow = TRUE), stringsAsFactors = FALSE)
      colnames(tmp) <- c("sId", "label")
      mat <- merge(mat, tmp, all.x = TRUE)
   }
   return(dplyr::as_tibble(mat))
}



#' Info about the actions in the HMDP model under consideration.
#'
#' @param prefix A character string with the prefix added to til binary files.
#' @param labels Should labels be extracted.
#' @param fileA The binary file containing the description of actions.
#' @param filePr The binary file containing the description of transition probabilities.
#' @param fileW The binary file containing the description of weights.
#' @param fileLabelA The binary file containing the action labels.
#' @param fileLabelW The binary file containing the weight labels.
#'
#' @return A data frame with the information. Scope string contain the scope of the transitions and
#'   can be 4 values: 
#'   * 0: A transition to the next stage in the father process,
#'   * 1: A transition to next stage in the current process, 
#'   * 2: A transition to a child process (stage zero in the child process), 
#'   * 3: A transition to the state with `sId = idx` is considered. 
#'   
#'   The index string denote the index (id is scope = 3) of the state at the next stage.
#'
#' @note The model don't have to be loaded, i.e only read the binary files. The state id (`sId`) will
#'   not be the same as in the loaded model!
#' @example inst/examples/binaryMDPWriter-ex.R
#' @export
getBinInfoActions<-function(prefix="", labels = TRUE, fileA="actionIdx.bin",
                         filePr="transProb.bin", fileW="actionWeight.bin",
                         fileLabelW="actionWeightLbl.bin", fileLabelA="actionIdxLbl.bin")
{
   fileA<-paste(prefix,fileA,sep="")
   filePr<-paste(prefix,filePr,sep="")
   fileW<-paste(prefix,fileW,sep="")
   fileLabelW<-paste(prefix,fileLabelW,sep="")

   tmpA<-readBin(fileA, integer(),n=file.info(fileA)$size/4)
   tmpPr<-readBin(filePr, numeric(),n=file.info(filePr)$size/8)
   tmpW<-readBin(fileW, numeric(),n=file.info(fileW)$size/8)
   colNames<-readBin(fileLabelW, character(),n=file.info(fileLabelW)$size)
   rows<-length(tmpA[tmpA==-1])
   cols<-5+length(colNames)

   mat<-as.data.frame(matrix(NA,nrow=rows,ncol=cols))
   mat[,1]<-1:nrow(mat)-1
   idxA<-c(0,which(tmpA== -1))
   idxPr<-c(0,which(tmpPr== -1))
   for (i in 1:(length(idxA)-1)) {
      v<-tmpA[(idxA[i]+1):(idxA[i+1]-1)]
      mat[i,2]<-v[1]
      mat[i,3]<-paste(v[ seq(2,length(v),2) ],collapse = ",")
      mat[i,4]<-paste(v[ seq(3,length(v),2) ],collapse = ",")
      v<-tmpPr[(idxPr[i]+1):(idxPr[i+1]-1)]
      mat[i,5]<-paste(v,collapse = ",")
   }

   for (i in 1:rows) {
      mat[i,6:cols]<-tmpW[(length(colNames)*(i-1)+1):(length(colNames)*i)]
   }
   colnames(mat)<-c("aId","sId","scope","index","pr", colNames)

   if (labels) {
      fileLabelA<-paste(prefix,fileLabelA,sep="")
      tmp<-readBin(fileLabelA, character(),n=file.info(fileLabelA)$size)
      tmp<-as.data.frame(matrix(tmp,ncol=2,byrow=TRUE),stringsAsFactors = FALSE)
      colnames(tmp)<-c("aId","label")
      tmp$aId<-as.numeric(tmp$aId)
      mat<-merge(mat,tmp,all.x=TRUE)
   }
   return(dplyr::as_tibble(mat))
}



#' Function for writing an HMDP model to a hmp file (XML). The function define
#' sub-functions which can be used to define an HMDP model stored in a hmp file.
#'
#' HMP files are in XML format and human readable using e.g. a text editor.
#' HMP files are not suitable for storing large HMDP models since text files are very
#' verbose. Moreover, approximation of the weights and probabilities may occur since
#' the parser writing the hmp file may no output all digits. If you consider large
#' models then use the binary file format instead.
#'
#' The returned writer exposes these functions:
#'
#' * `setWeights(labels, duration)`: sets the labels of the weights used in the
#'   actions. `labels` is a vector of label names. `duration` identifies which
#'   label corresponds to duration or time. For example, if the first entry in
#'   `labels` is time, then `duration = 1`. Call this before building the model.
#' * `setTransWeights(labels)`: sets the labels of transition-level weights.
#' * `process()`: starts a (sub)process.
#' * `endProcess()`: ends a (sub)process.
#' * `stage(label = NULL)`: starts a stage.
#' * `endStage()`: ends a stage.
#' * `state(label = NULL)`: starts a state and returns the state index `sIdx`.
#' * `endState()`: ends a state.
#' * `action(label = NULL, weights, prob, statesNext = NULL, transWeights = NULL)`: starts an
#'   action. `weights` must be a vector of action weights, and `prob` must
#'   contain triples `(scope, idx, pr)`. `scope` can take three values:
#'
#'   * `0`: a transition to the next stage in the father process.
#'   * `1`: a transition to the next stage in the current process.
#'   * `2`: a transition to a child process, at stage zero in the child process.
#'
#'   The `idx` value denotes the index of the state at the stage considered. For
#'   example, if `scope = 1` and `idx = 2`, the transition is to state number 3
#'   at the next stage in the current process, counting from zero. `scope = 3`
#'   is not supported in the `hmp` file format. `statesNext` is the number of
#'   states in the next stage of the process and is only needed when there is a
#'   transition to the father.
#' * `endAction()`: ends an action.
#' * `closeWriter()`: closes the writer. Call this when the model description is
#'   finished.
#'
#' @param file The name of the file storing the model (e.g. `r.hmp`).
#' @param rate The interest rate (used if consider discounting).
#' @param rateBase The time where the `rate` is taken over, e.g. if the `rate` is 0.1 and `rateBase` is 365 days
#'   then we have an interest rate of 10 percent over the year.
#' @param precision The precision used when checking if probabilities sum to one.
#' @param desc Description of the model.
#' @param getLog Output log text.
#' @return A list of functions.
#' @note Note all indexes are starting from zero (C/C++ style).
#' @example inst/examples/hmpMDPWriter-ex.R
#' @export
hmpMDPWriter<-function(file="r.hmp", rate=0.1, rateBase=1, precision=0.00001, desc="HMP file created using hmpMDPWriter in R", getLog = TRUE) {
   # addLevelRates<-function(rates){
   # 	tr$addTag("i",paste(rates,collapse=" "))
   #    xml2::xml_add_child(doc, "i", paste(rates,collapse=" "))
   # 	invisible(NULL)
   # }
   
   # setSources<-function(s){
   # 	tr$addTag("sources",paste(s-1,collapse=" "))
   #    xml2::xml_add_child(doc, "sources", paste(s-1,collapse=" "))
   # 	invisible(NULL)
   # }
   
   setWeights<-function(labels, duration) {
      if (is.null(duration)) durIdx <<- -1  # no duration specified by negative number
      else durIdx<<-duration
      # tr$addTag("i",rate)
      xml2::xml_add_child(doc, "i", rate)
      
      if (wFixed) stop("Weights already added!")
      for (i in 1:length(labels)) {
         if (i!=durIdx) {
            # tr$addTag("quantities",attrs=c(l=labels[i]))
            xml2::xml_add_child(doc, "quantities", l = labels[i])
         }
      }
      wFixed<<-TRUE
      # tr$addTag("sources","0 1")
      xml2::xml_add_child(doc, "sources", "0 1")
      invisible(NULL)
   }
   
   setTransWeights<-function(labels) {
      for (i in seq_along(labels)) {
         xml2::xml_add_child(doc, "transQuantities", l = labels[i])
      }
      invisible(NULL)
   }
   
   process<-function(){
      if (!wFixed) stop("Weights must be added using 'setWeights' before starting building the HMDP!")
      # tr$addTag("proc",close=FALSE)
      n <<- xml2::xml_add_child(n, "proc")
      invisible(NULL)
   }
   
   endProcess<-function(){
      # tr$closeTag()
      n <<- xml2::xml_parent(n)
      invisible(NULL)
   }
   
   stage<-function(label=NULL){
      if (is.null(label)) {
         # tr$addTag("g",close=FALSE)
         n <<- xml2::xml_add_child(n, "g")
      } else {
         # tr$addTag("g",attrs=c(l=label),close=FALSE)
         n <<- xml2::xml_add_child(n, "g", l = label)
      }
      invisible(NULL)
   }
   
   endStage<-function(){
      # tr$closeTag()
      n <<- xml2::xml_parent(n)
      invisible(NULL)
   }
   
   state<-function(label=NULL){
      if (is.null(label)) {
         # tr$addTag("s",close=FALSE)
         n <<- xml2::xml_add_child(n, "s")
      } else {
         # tr$addTag("s",attrs=c(l=label),close=FALSE)
         n <<- xml2::xml_add_child(n, "s", l = label)
      }
      invisible(NULL)
   }
   
   endState<-function(){
      # tr$closeTag()
      n <<- xml2::xml_parent(n)
      invisible(NULL)
   }
   
   action<-function(label=NULL, weights, prob, statesNext=NULL, transWeights=NULL){  # prop contain tripeles (scope,idx,prob), statesNext: Number of states in the next stage of the process, only needed if have a transition to the father
      scope<-prob[3*0:(length(prob)/3-1)+1]   # scopes we consider
      if (any(scope==3)) {
         stop("Scope = 3 is not supported in hmp files!")
      }
      term <- FALSE
      if (any(scope==0)) {     # we have an prob that return to the father
         if (is.null(statesNext)) stop("Number of states at the next stage must be specified!")
         if (statesNext!=0) term<-TRUE
         idx<-3*(which(scope==0)-1)+1            # index of scope==0
         prob[idx+1]<-prob[idx+1]+statesNext     # add number of states at next stage to father idx
      }
      n <<- xml2::xml_add_child(n, "a")
      tags<-NULL
      if (!is.null(label)) tags<-c(tags,l=label)
      if (term) tags<-c(tags,term='t')
      if (is.null(tags)) {
         # tr$addTag("a",close=FALSE)
      } else {
         # tr$addTag("a",attrs=tags,close=FALSE)
         xml2::xml_attrs(n) <- tags
      }
      if (any(scope==2)) {    # we have an prob to a new child process
         if (!all(prob==c(2,0,1))) stop("Only a deterministic transition to the dummy stage in the child process allowed (prop=(2,0,1))!")
         return(invisible(NULL))   # only a deterministic transition with zero weights allowed in the hmp format
      }
      # tr$addTag("q",paste(weights[which(1:length(weights)!=durIdx)],collapse=" "))  # quantities
      xml2::xml_add_child(n, "q", paste(weights[which(1:length(weights)!=durIdx)],collapse=" "))
      if (!is.null(transWeights)) {
         xml2::xml_add_child(n, "qt", paste(transWeights, collapse=" "))
      }
      probs<-prob[which((1:length(prob)-1)%%3!=0)]   # probs contain pairs (idx,prob)
      if (length(probs)==2) { # deterministic transition
         # tr$addTag("p",probs[1],attrs=c(t='d'))
         xml2::xml_add_child(n, "p", probs[1], t='d')
      } else {
         # tr$addTag("p",paste(probs,collapse=" "),attrs=c(t='s'))
         xml2::xml_add_child(n, "p", paste(probs,collapse=" "), t='s')
      }
      if (durIdx<0) {
         # tr$addTag("d", 1)
         xml2::xml_add_child(n, "d", 1)
      }
      else {
         names(weights) <- NULL
         # tr$addTag("d", weights[durIdx])
         xml2::xml_add_child(n, "d", weights[durIdx])
      }
      invisible(NULL)
   }
   
   endAction<-function(){
      # tr$closeTag()
      n <<- xml2::xml_parent(n)
      invisible(NULL)
   }
   
   closeWriter<-function(){
      # saveXML(tr$value(),file="old.hmp",compression=0,prefix = NULL)
      xml2::write_xml(doc, file)
      if (getLog) cat("\nModel saved to file:",file,"\n")
   }
   
   wFixed<-FALSE  # have weights been added
   durIdx <- NULL   # index of weight storing the duration (number from 1)
   # tr<-xmlTree("mlhmp",dtd=NULL,attrs=c(l=desc,b=rate,dsl=rateBase,precision=precision,version="1.1"))
   doc <- xml2::xml_new_root("mlhmp", l=desc, b=rate, dsl=rateBase, precision=precision, version="1.1")
   n <- doc  # current node
   
   v <- list(setWeights = setWeights, setTransWeights = setTransWeights,
             stage = stage, endStage = endStage, state = state, endState = endState,
             action = action, endAction = endAction, process = process, endProcess = endProcess,
             closeWriter = closeWriter)
   class(v) <- c("hmpMDPWriter")
   return(v)
}

library(MDP2)
cleanUp<-function() {
   unlink("*.bin")
   unlink("*.hmp")
}

test_that("binaryMDPWriter",{
   #sprintf("%.100f",g)
   source("files/HCT_ex6.1.1.R")
   mdp<-loadMDP("hct611_", getLog = FALSE)
   gProb<-runPolicyIteAve(mdp,"Net reward","Duration", getLog = FALSE)
   rm(mdp)
   source("files/HCT_ex6.1.1_v2.R")
   mdp<-loadMDP("hct611v2_", getLog = FALSE)
   gPr<-runPolicyIteAve(mdp,"Net reward","Duration", getLog = FALSE)
   rm(mdp)
   expect_equal(gProb, gPr)
   
   source("files/HCT_exc6.4.R")
   mdp<-loadMDP("hct64_", getLog = FALSE)
   gProb<-runPolicyIteAve(mdp,"Net reward","Duration", getLog = FALSE)
   rm(mdp)
   source("files/HCT_exc6.4_v2.R")
   mdp<-loadMDP("hct64v2_", getLog = FALSE)
   gPr<-runPolicyIteAve(mdp,"Net reward","Duration", getLog = FALSE)
   rm(mdp)
   expect_equal(gProb, gPr)
   
   source("files/HCT_exc6.7.R")
   mdp<-loadMDP("hct67_", getLog = FALSE)
   gProb<-runPolicyIteAve(mdp,"Net reward","Duration", getLog = FALSE)
   rm(mdp)
   source("files/HCT_exc6.7_v2.R")
   mdp<-loadMDP("hct67v2_", getLog = FALSE)
   gPr<-runPolicyIteAve(mdp,"Net reward","Duration", getLog = FALSE)
   rm(mdp)
   expect_equal(gProb, gPr)
   
   source("files/HCT_exc7.3.R")
   mdp<-loadMDP("hct73_", getLog = FALSE)
   gProb<-runPolicyIteAve(mdp,"Net reward","Duration", getLog = FALSE)
   rm(mdp)
   source("files/HCT_exc7.3_v2.R")
   mdp<-loadMDP("hct73v2_", getLog = FALSE)
   gPr<-runPolicyIteAve(mdp,"Net reward","Duration", getLog = FALSE)
   rm(mdp)
   expect_equal(gProb, gPr)
   
   ## test loading with P, R and D
   source("files/HCT_ex6.1.1_v3.R")
   # test using L.R.A reward / t.unit
   mdp<-loadMDP("hct611v3-1_", getLog = FALSE)
   expect_equal(runPolicyIteAve(mdp,"Net reward","Duration", getLog = FALSE), -0.433789954337899297254210750907077454030513763427734375)
   rm(mdp)
   mdp<-loadMDP("hct611v3-2_", getLog = FALSE)
   expect_equal(runPolicyIteAve(mdp,"Net reward","Duration", getLog = FALSE), -0.433789954337899297254210750907077454030513763427734375)
   rm(mdp)
   mdp<-loadMDP("hct611v3-3_", getLog = FALSE)
   expect_equal(runPolicyIteAve(mdp,"Net reward","Duration", getLog = FALSE), -0.433789954337899297254210750907077454030513763427734375)
   rm(mdp)
   mdp<-loadMDP("hct611v3-4_", getLog = FALSE)
   expect_equal(runPolicyIteAve(mdp,"Net reward","Duration", getLog = FALSE), -0.433789954337899297254210750907077454030513763427734375)
   rm(mdp)
   # test using discounted reward
   # compare MDP formulations
   mdp<-loadMDP("hct611v3-1_", getLog = FALSE)
   runPolicyIteDiscount(mdp, "Net reward", "Duration", discountFactor = 0.9, getLog = FALSE)
   weightsPolicyIte<-getPolicy(mdp)$weight
   expect_equal(sum(weightsPolicyIte), -37.94871237074722358784129028208553791046142578125)
   rm(mdp)
   mdp<-loadMDP("hct611v3-2_", getLog = FALSE)
   runPolicyIteDiscount(mdp, "Net reward", "Duration", discountFactor = 0.9, getLog = FALSE)
   weightsPolicyIte<-getPolicy(mdp)$weight
   expect_equal(sum(weightsPolicyIte), -37.94871237074722358784129028208553791046142578125)
   rm(mdp)
   mdp<-loadMDP("hct611v3-3_", getLog = FALSE)
   runPolicyIteDiscount(mdp, "Net reward", "Duration", discountFactor = 0.9, getLog = FALSE)
   weightsPolicyIte<-getPolicy(mdp)$weight
   expect_equal(sum(weightsPolicyIte), -37.94871237074722358784129028208553791046142578125)
   rm(mdp)
   # compare semi-MDP formulations
   mdp<-loadMDP("hct611v2_", getLog = FALSE)
   runPolicyIteDiscount(mdp, "Net reward", "Duration", discountFactor = 0.9, getLog = FALSE)
   weightsPolicyIte1<-getPolicy(mdp)$weight
   rm(mdp)
   mdp<-loadMDP("hct611v3-4_", getLog = FALSE)
   runPolicyIteDiscount(mdp, "Net reward", "Duration", discountFactor = 0.9, getLog = FALSE)
   weightsPolicyIte2<-getPolicy(mdp)$weight
   rm(mdp)
   expect_equal(weightsPolicyIte1, weightsPolicyIte2)
})

test_that("loadMDP supports models with states and no actions",{
   prefix <- paste0(tempfile("no_actions_"), "_")
   w <- binaryMDPWriter(prefix, getLog = FALSE)
   w$setWeights(c("Duration", "Net reward"))
   w$process()
      w$stage()
         for (ii in 2:10) {
            w$state(label = "test")
            w$endState()
         }
      w$endStage()
   w$endProcess()
   w$closeWriter()

   mdp <- loadMDP(prefix, getLog = FALSE)

   expect_s3_class(mdp, "HMDP")
   expect_equal(mdp$states, 9)
   expect_equal(mdp$actions, 0)
   expect_equal(mdp$weightNames, c("Duration", "Net reward"))
})

test_that("loadMDP rejects transitions to non-existing states",{
   prefix <- paste0(tempfile("bad_transition_"), "_")
   w <- binaryMDPWriter(prefix, getLog = FALSE)
   w$setWeights(c("Duration", "Net reward"))
   w$process()
      w$stage()
         for (ii in 2:10) {
            w$state(label = "test")
            w$action(
               label = "a1",
               scope = 1,
               weights = c(1, 100),
               id = c(0, 99),
               pr = c(0.5, 0.5),
               end = TRUE
            )
            w$endState()
         }
      w$endStage()
   w$endProcess()
   w$closeWriter()

   expect_message(
      mdp <- loadMDP(prefix, getLog = FALSE),
      "transition to a non-existing state"
   )
   expect_null(mdp)
})

test_that("binaryMDPWriter detects unclosed writer blocks",{
   prefix <- paste0(tempfile("writer_stack_"), "_")
   w <- binaryMDPWriter(prefix, getLog = FALSE)
   w$setWeights(c("Duration", "Net reward"))
   w$process()
   w$stage()
   w$state()
   w$action(scope = 1, weights = c(1, 100), id = 0, pr = 1)

   expect_error(
      w$endState(),
      "Call endAction\\(\\) or use action\\(\\.\\.\\., end = TRUE\\)"
   )
   expect_error(
      w$closeWriter(),
      "action is still open"
   )

   w$endAction()
   expect_error(
      w$endStage(),
      "Cannot end a stage unless a stage is open"
   )
   expect_error(
      w$closeWriter(),
      "state is still open"
   )

   w$endState()
   w$endStage()
   w$endProcess()
   w$closeWriter()
})

test_that("binaryMDPWriter requires correct block nesting",{
   prefix <- paste0(tempfile("writer_nesting_"), "_")
   w <- binaryMDPWriter(prefix, getLog = FALSE)
   w$setWeights(c("Duration", "Net reward"))

   expect_error(w$stage(), "outside an open process")
   expect_error(w$state(), "outside an open stage")
   expect_error(
      w$action(scope = 1, weights = c(1, 100), id = 0, pr = 1),
      "outside an open state"
   )
   expect_error(w$endAction(), "unless an action is open")
   expect_error(w$endState(), "while another writer block is open")
   expect_error(w$endStage(), "unless a stage is open")
   expect_error(w$endProcess(), "unless a process is open")

   w$process()
   expect_error(w$process(), "before closing the current writer block")
   w$endProcess()
   w$closeWriter()
})

test_that("binaryMDPWriter allows one redundant endAction after end equals TRUE",{
   prefix <- paste0(tempfile("writer_redundant_end_"), "_")
   w <- binaryMDPWriter(prefix, getLog = FALSE)
   w$setWeights(c("Duration", "Net reward"))
   w$process()
   w$stage()
   w$state()
   w$action(scope = 1, weights = c(1, 100), id = 0, pr = 1, end = TRUE)
   expect_no_error(w$endAction())
   expect_error(w$endAction(), "unless an action is open")
   w$endState()
   w$endStage()
   w$endProcess()
   w$closeWriter()
})

test_that("binaryMDPWriter allows nested process inside an action",{
   prefix <- paste0(tempfile("writer_nested_"), "_")
   w <- binaryMDPWriter(prefix, getLog = FALSE)
   w$setWeights(c("Duration", "Net reward"))
   w$process()
      w$stage()
         w$state()
            w$action(weights = c(0, 0), prob = c(2, 0, 1))
               expect_no_error(w$process())
                  w$stage()
                     w$state()
                        w$action(weights = c(0, 0), prob = c(1, 0, 1))
                        w$endAction()
                     w$endState()
                  w$endStage()
               w$endProcess()
            w$endAction()
         w$endState()
      w$endStage()
   w$endProcess()
   w$closeWriter()
})

test_that("memoryMDPWriter builds the same model as binaryMDPWriter",{
   build_model <- function(w) {
      w$setWeights(c("Duration", "Net reward"))
      w$process()
         w$stage()
            w$state(label = "s0")
               w$action(label = "a0", scope = c(1, 1), id = c(0, 1),
                        pr = c(0.25, 0.75), weights = c(1, 10), end = TRUE)
            w$endState()
            w$state(label = "s1")
               w$action(label = "a1", scope = c(1, 1), id = c(0, 1),
                        pr = c(0.5, 0.5), weights = c(1, 20), end = TRUE)
            w$endState()
         w$endStage()
         w$stage()
            w$state(label = "t0", end = TRUE)
            w$state(label = "t1", end = TRUE)
         w$endStage()
      w$endProcess()
      w
   }

   prefix <- paste0(tempfile("memory_compare_"), "_")
   wb <- build_model(binaryMDPWriter(prefix, getLog = FALSE))
   wb$closeWriter()
   binary_mdp <- loadMDP(prefix, getLog = FALSE)

   wm <- build_model(memoryMDPWriter(getLog = FALSE))
   memory_mdp <- wm$closeWriter()

   expect_s3_class(memory_mdp, "HMDP")
   expect_equal(memory_mdp$states, binary_mdp$states)
   expect_equal(memory_mdp$actions, binary_mdp$actions)
   expect_equal(memory_mdp$weightNames, binary_mdp$weightNames)
   memory_ids <- memory_mdp$ptr$getIds("0")
   binary_ids <- binary_mdp$ptr$getIds("0")
   expect_equal(memory_mdp$ptr$getActionInfo(memory_ids[1]), binary_mdp$ptr$getActionInfo(binary_ids[1]))
   expect_equal(memory_mdp$ptr$getActionInfo(memory_ids[2]), binary_mdp$ptr$getActionInfo(binary_ids[2]))
})

test_that("memoryMDPWriter supports transition weights",{
   w <- memoryMDPWriter(getLog = FALSE)
   w$setWeights(c("Duration"))
   w$setTransWeights(c("Reward", "Risk"))
   w$process()
      w$stage()
         w$state(label = "s0")
            w$action(
               label = "a0",
               scope = c(1, 1),
               id = c(0, 1),
               pr = c(0.5, 0.5),
               weights = 1,
               transWeights = c(10, 1, 20, 2),
               end = TRUE
            )
         w$endState()
      w$endStage()
      w$stage()
         w$state(label = "t0", end = TRUE)
         w$state(label = "t1", end = TRUE)
      w$endStage()
   w$endProcess()

   mdp <- w$closeWriter()
   info <- mdp$ptr$getActionInfo(mdp$ptr$getIds("0")[1])[[1]]

   expect_equal(mdp$weightActionNames, "Duration")
   expect_equal(mdp$weightTransNames, c("Reward", "Risk"))
   expect_equal(info$transWeights, c(10, 1, 20, 2))
})

test_that("memoryMDPWriter supports models with states and no actions",{
   w <- memoryMDPWriter(getLog = FALSE)
   w$setWeights(c("Duration", "Net reward"))
   w$process()
      w$stage()
         for (ii in 1:4) w$state(label = paste0("s", ii), end = TRUE)
      w$endStage()
   w$endProcess()

   mdp <- w$closeWriter()

   expect_s3_class(mdp, "HMDP")
   expect_equal(mdp$states, 4)
   expect_equal(mdp$actions, 0)
   expect_equal(mdp$weightNames, c("Duration", "Net reward"))
})

test_that("memoryMDPWriter supports nested processes",{
   expect_no_error(source(system.file("examples/memoryMDPWriter-ex.R", package = "MDP2")))
})

test_that("memoryMDPWriter rejects external processes and use after close",{
   w <- memoryMDPWriter(getLog = FALSE)
   w$setWeights(c("Duration", "Net reward"))
   expect_error(w$includeProcess(), "does not support external processes")
   w$process()
      w$stage()
         w$state(label = "s0", end = TRUE)
      w$endStage()
   w$endProcess()
   mdp <- w$closeWriter()

   expect_s3_class(mdp, "HMDP")
   expect_error(w$state(), "memoryMDPWriter is closed")
   expect_error(w$closeWriter(), "memoryMDPWriter is closed")
})

build_with_writer <- function(build_model) {
   prefix <- paste0(tempfile("writer_compare_"), "_")
   wb <- build_model(binaryMDPWriter(prefix, getLog = FALSE))
   wb$closeWriter()
   binary_mdp <- loadMDP(prefix, getLog = FALSE)

   wm <- build_model(memoryMDPWriter(getLog = FALSE))
   memory_mdp <- wm$closeWriter()

   list(binary = binary_mdp, memory = memory_mdp)
}

expect_same_model_summary <- function(models) {
   expect_s3_class(models$memory, "HMDP")
   expect_equal(models$memory$states, models$binary$states)
   expect_equal(models$memory$actions, models$binary$actions)
   expect_equal(models$memory$levels, models$binary$levels)
   expect_equal(models$memory$weightNames, models$binary$weightNames)
   expect_equal(models$memory$weightActionNames, models$binary$weightActionNames)
   expect_equal(models$memory$weightTransNames, models$binary$weightTransNames)
}

build_vignette_hct_hierarchical <- function(w) {
   N <- 5
   labels <- paste0("i = ", 1:N)
   Cf <- -10
   Cp <- c(0, -7, -7, -5)
   Q <- matrix(c(
      0.90, 0.10, 0, 0, 0,
      0, 0.80, 0.10, 0.05, 0.05,
      0, 0, 0.70, 0.10, 0.20,
      0, 0, 0, 0.50, 0.50),
      nrow = 4, byrow = TRUE)
   transPr <- function(i, a) {
      pr <- NULL
      idx <- NULL
      if (a == "nr") {
         pr <- Q[i, ]
         idx <- which(pr > 0)
         pr <- pr[idx]
         idx <- idx - 1
      }
      if (a == "pr" | a == "fr") {
         pr <- 1
         idx <- 0
      }
      list(pr = pr, idx = idx)
   }

   w$setWeights(c("Duration", "Net reward"))
   w$process()
      w$stage()
         w$state(label = labels[1])
            lst <- transPr(1, "nr")
            w$action(label = "nr", weights = c(1, 0), pr = lst$pr, id = lst$idx, end = TRUE)
         w$endState()
         for (i in 2:(N - 1)) {
            w$state(label = labels[i])
               lst <- transPr(i, "nr")
               w$action(label = "nr", weights = c(1, 0), pr = lst$pr, id = lst$idx, end = TRUE)
               lst <- transPr(i, "pr")
               w$action(label = "pr", weights = c(1, Cp[i]), pr = lst$pr, id = lst$idx, end = TRUE)
            w$endState()
         }
         w$state(label = labels[N])
            lst <- transPr(N, "fr")
            w$action(label = "fr", weights = c(2, Cf), pr = lst$pr, id = lst$idx, end = TRUE)
         w$endState()
      w$endStage()
   w$endProcess()
   w
}

build_vignette_hct_matrices <- function(w) {
   N <- 5
   Cf <- -10
   Cp <- c(0, -7, -7, -5)
   Q <- matrix(c(
      0.90, 0.10, 0, 0, 0,
      0, 0.80, 0.10, 0.05, 0.05,
      0, 0, 0.70, 0.10, 0.20,
      0, 0, 0, 0.50, 0.50),
      nrow = 4, byrow = TRUE)

   P <- list()
   P[[1]] <- as.matrix(rbind(Q, 0))
   Z <- matrix(0, nrow = N, ncol = N)
   Z[2, 1] <- Z[3, 1] <- Z[4, 1] <- 1
   P[[2]] <- Z
   Z <- matrix(0, nrow = N, ncol = N)
   Z[5, 1] <- 1
   P[[3]] <- Z

   R <- matrix(0, nrow = N, ncol = 3)
   R[2:4, 2] <- Cp[2:4]
   R[5, 3] <- Cf
   D <- matrix(1, nrow = N, ncol = 3)
   D[5, 3] <- 2

   w$setWeights(c("Duration", "Net reward"))
   w$process(P, R, D)
   w
}

build_vignette_machine <- function(w) {
   w$setWeights(c("Net reward"))
   w$process()
      w$stage()
         w$state(label = "dummy")
            w$action(label = "buy", weights = -100, pr = c(0.7, 0.3), id = c(0, 1), end = TRUE)
         w$endState()
      w$endStage()
      w$stage()
         w$state(label = "good")
            w$action(label = "mt", weights = 55, pr = 1, id = 0, end = TRUE)
            w$action(label = "nmt", weights = 70, pr = c(0.6, 0.4), id = c(0, 1), end = TRUE)
         w$endState()
         w$state(label = "average")
            w$action(label = "mt", weights = 40, pr = 1, id = 0, end = TRUE)
            w$action(label = "nmt", weights = 50, pr = c(0.6, 0.4), id = c(1, 2), end = TRUE)
         w$endState()
      w$endStage()
      w$stage()
         w$state(label = "good")
            w$action(label = "mt", weights = 55, pr = 1, id = 0, end = TRUE)
            w$action(label = "nmt", weights = 70, pr = c(0.5, 0.5), id = c(0, 1), end = TRUE)
         w$endState()
         w$state(label = "average")
            w$action(label = "mt", weights = 40, pr = 1, id = 0, end = TRUE)
            w$action(label = "nmt", weights = 50, pr = c(0.5, 0.5), id = c(1, 2), end = TRUE)
         w$endState()
         w$state(label = "not working")
            w$action(label = "mt", weights = 30, pr = 1, id = 0, end = TRUE)
            w$action(label = "rep", weights = 5, pr = 1, id = 3, end = TRUE)
         w$endState()
      w$endStage()
      w$stage()
         w$state(label = "good")
            w$action(label = "mt", weights = 55, pr = 1, id = 0, end = TRUE)
            w$action(label = "nmt", weights = 70, pr = c(0.2, 0.8), id = c(0, 1), end = TRUE)
         w$endState()
         w$state(label = "average")
            w$action(label = "mt", weights = 40, pr = 1, id = 0, end = TRUE)
            w$action(label = "nmt", weights = 50, pr = c(0.2, 0.8), id = c(1, 2), end = TRUE)
         w$endState()
         w$state(label = "not working")
            w$action(label = "mt", weights = 30, pr = 1, id = 0, end = TRUE)
            w$action(label = "rep", weights = 5, pr = 1, id = 3, end = TRUE)
         w$endState()
         w$state(label = "replaced")
            w$action(label = "dummy", weights = 0, pr = 1, id = 3, end = TRUE)
         w$endState()
      w$endStage()
      w$stage()
         w$state(label = "good", end = TRUE)
         w$state(label = "average", end = TRUE)
         w$state(label = "not working", end = TRUE)
         w$state(label = "replaced", end = TRUE)
      w$endStage()
   w$endProcess()
   w
}

build_vignette_cow <- function(w) {
   cowFile <- c(
      file.path("files", "cow.csv"),
      file.path("tests", "testthat", "files", "cow.csv"),
      file.path("vignettes", "vignette_files", "cow.csv"),
      file.path("..", "..", "vignettes", "vignette_files", "cow.csv"),
      file.path("..", "vignettes", "vignette_files", "cow.csv")
   )
   cowFile <- cowFile[file.exists(cowFile)][1]
   if (is.na(cowFile)) stop("Could not find cow.csv test fixture.", call. = FALSE)
   cowDf <- utils::read.csv(cowFile)
   lev1W <- function(s0Idx, n1Idx, s1Idx, a1Lbl) {
      row <- cowDf[cowDf$s0 == s0Idx & cowDf$n1 == n1Idx & cowDf$s1 == s1Idx & cowDf$label == a1Lbl, ]
      as.numeric(row[c("Duration", "Reward", "Output")])
   }
   lev1Pr <- function(s0Idx, n1Idx, s1Idx, a1Lbl) {
      row <- cowDf[cowDf$s0 == s0Idx & cowDf$n1 == n1Idx & cowDf$s1 == s1Idx & cowDf$label == a1Lbl, ]
      as.numeric(row[paste0(rep(c("scp", "idx", "pr"), 3), rep(0:2, each = 3))])
   }

   lblS0 <- c("Bad genetic level", "Avg genetic level", "Good genetic level")
   lblS1 <- c("Low yield", "Avg yield", "High yield")
   w$setWeights(c("Duration", "Net reward", "Yield"))
   w$process()
      w$stage()
         for (s0 in 0:2) {
            w$state(label = lblS0[s0 + 1])
               w$action(label = "Keep", weights = c(0, 0, 0), prob = c(2, 0, 1))
                  w$process()
                     w$stage()
                        w$state(label = "Dummy")
                           w$action(label = "Dummy", weights = c(0, 0, 0),
                                    prob = c(1, 0, 1/3, 1, 1, 1/3, 1, 2, 1/3), end = TRUE)
                        w$endState()
                     w$endStage()
                     for (d1 in 1:4) {
                        w$stage()
                           for (s1 in 0:2) {
                              w$state(label = lblS1[s1 + 1])
                                 if (d1 != 4) {
                                    w$action(label = "Keep", weights = lev1W(s0, d1, s1, "Keep"),
                                             prob = lev1Pr(s0, d1, s1, "Keep"), end = TRUE)
                                 }
                                 w$action(label = "Replace", weights = lev1W(s0, d1, s1, "Replace"),
                                          prob = lev1Pr(s0, d1, s1, "Replace"), end = TRUE)
                              w$endState()
                           }
                        w$endStage()
                     }
                  w$endProcess()
               w$endAction()
            w$endState()
         }
      w$endStage()
   w$endProcess()
   w
}

test_that("building vignette HCT hierarchical model matches for binary and memory writers",{
   models <- build_with_writer(build_vignette_hct_hierarchical)
   expect_same_model_summary(models)
   expect_equal(
      runPolicyIteAve(models$memory, "Net reward", "Duration", getLog = FALSE),
      runPolicyIteAve(models$binary, "Net reward", "Duration", getLog = FALSE)
   )
})

test_that("building vignette HCT matrix model matches for binary and memory writers",{
   models <- build_with_writer(build_vignette_hct_matrices)
   expect_same_model_summary(models)
   expect_equal(
      runPolicyIteAve(models$memory, "Net reward", "Duration", getLog = FALSE),
      runPolicyIteAve(models$binary, "Net reward", "Duration", getLog = FALSE)
   )
})

test_that("building vignette machine model matches for binary and memory writers",{
   models <- build_with_writer(build_vignette_machine)
   expect_same_model_summary(models)
   runValueIte(models$binary, "Net reward", termValues = c(30, 10, 5, 0), getLog = FALSE)
   binary_policy <- getPolicy(models$binary)
   runValueIte(models$memory, "Net reward", termValues = c(30, 10, 5, 0), getLog = FALSE)
   memory_policy <- getPolicy(models$memory)
   expect_equal(memory_policy$weight, binary_policy$weight)
   expect_equal(memory_policy$actionLabel, binary_policy$actionLabel)
})

test_that("building vignette cow model matches for binary and memory writers",{
   models <- build_with_writer(build_vignette_cow)
   expect_same_model_summary(models)
   expect_equal(
      runPolicyIteAve(models$memory, "Net reward", "Duration", getLog = FALSE),
      runPolicyIteAve(models$binary, "Net reward", "Duration", getLog = FALSE)
   )
   expect_equal(
      runPolicyIteAve(models$memory, "Yield", "Duration", getLog = FALSE),
      runPolicyIteAve(models$binary, "Yield", "Duration", getLog = FALSE)
   )
})

cleanUp()

library(MDP2)

test_that("plotHypergraph draws optional trans labels", {
   hgf <- list(
      nodes = tibble::tibble(
         sId = c(0, 1, 2),
         gId = c(1, 3, 4),
         label = c("Head", "Tail A", "Tail B")
      ),
      hyperarcs = tibble::tibble(
         sId = 0,
         actionWeights = list(10),
         trans = list(c(1, 2)),
         pr = list(c(0.25, 0.75)),
         transLabels = list(c("left", "right")),
         transWeights = list(matrix(c(99, 101), ncol = 1, dimnames = list(NULL, "Reward"))),
         aIdx = 0,
         label = "A",
         lwd = 1,
         lty = 1,
         col = "black"
      )
   )
   grDevices::pdf(file = tempfile(fileext = ".pdf"))
   on.exit(grDevices::dev.off(), add = TRUE)

   expect_invisible(plotHypergraph(hgf, c(2, 2)))
   expect_invisible(plotHypergraph(hgf, c(2, 2), transLabels = "state"))
   expect_invisible(plotHypergraph(hgf, c(2, 2), transLabels = "sId"))
   expect_invisible(plotHypergraph(hgf, c(2, 2), transLabels = "prob"))
   expect_invisible(plotHypergraph(hgf, c(2, 2), transLabels = "weights"))
   expect_invisible(plotHypergraph(hgf, c(2, 2), transLabels = "label|prob|weights"))
   expect_invisible(plotHypergraph(hgf, c(2, 2), transLabels = "custom"))
})

test_that("plotHypergraph ignores missing trans and validates label columns", {
   hgf <- list(
      nodes = tibble::tibble(sId = c(0, 1), gId = c(1, 3), label = c("Head", "Tail")),
      hyperarcs = tibble::tibble(
         sId = 0,
         actionWeights = list(numeric(0)),
         trans = list(c(1, NA_real_)),
         pr = list(c(1, NA_real_)),
         aIdx = 0,
         label = "A",
         lwd = 1,
         lty = 1,
         col = "black"
      )
   )
   grDevices::pdf(file = tempfile(fileext = ".pdf"))
   on.exit(grDevices::dev.off(), add = TRUE)

   expect_invisible(plotHypergraph(hgf, c(2, 2), transLabels = "prob"))
   expect_error(
      plotHypergraph(hgf, c(2, 2), transLabels = "custom"),
      'transLabels = "custom" requires a transLabels list-column'
   )
   expect_error(
      plotHypergraph(hgf, c(2, 2), transLabels = "weights"),
      'transLabels containing "weights" requires a transWeights list-column'
   )
   expect_error(
      plotHypergraph(
         list(nodes = hgf$nodes, hyperarcs = dplyr::select(hgf$hyperarcs, -pr)),
         c(2, 2),
         transLabels = "prob"
      ),
      'transLabels containing "prob" requires a pr list-column'
   )
})

test_that("plotHypergraph supports zero, single, and multiple transition weights", {
   hgf <- list(
      nodes = tibble::tibble(sId = c(0, 1), gId = c(1, 3), label = c("Head", "Tail")),
      hyperarcs = tibble::tibble(
         sId = c(0, 0, 0),
         actionWeights = list(numeric(0), numeric(0), numeric(0)),
         trans = list(1, 1, 1),
         pr = list(1, 1, 1),
         transWeights = list(
            matrix(numeric(0), nrow = 1, ncol = 0),
            matrix(10, nrow = 1),
            matrix(c(10, 20), nrow = 1)
         ),
         aIdx = c(0, 1, 2),
         label = c("A", "B", "C"),
         lwd = c(1, 1, 1),
         lty = c(1, 1, 1),
         col = c("black", "black", "black")
      )
   )
   grDevices::pdf(file = tempfile(fileext = ".pdf"))
   on.exit(grDevices::dev.off(), add = TRUE)

   expect_invisible(plotHypergraph(hgf, c(2, 2), transLabels = "weights"))
})

test_that("getHypergraph returns nested weight and transition columns", {
   prefix <- file.path(tempdir(), "plot_trans_weight_")
   w <- binaryMDPWriter(prefix = prefix, getLog = FALSE)
   w$setWeights("Action weight")
   w$setTransWeights("Transition weight")
   w$process()
      w$stage()
         w$state()
            w$action(
               weights = 5,
               prob = c(1, 0, 0.25, 1, 1, 0.75),
               transWeights = c(10, 20)
            )
            w$endAction()
         w$endState()
      w$endStage()
      w$stage()
         w$state()
         w$endState()
         w$state()
         w$endState()
      w$endStage()
   w$endProcess()
   w$closeWriter()

   mdp <- loadMDP(prefix, getLog = FALSE)
   hgf <- getHypergraph(mdp)

   expect_true(all(c("actionWeights", "trans", "pr", "transWeights") %in% names(hgf$hyperarcs)))
   expect_false(any(grepl("^trans[0-9]+$|^pr[0-9]+$", names(hgf$hyperarcs))))
   expect_equal(hgf$hyperarcs$actionWeights[[1]], 5)
   expect_equal(hgf$hyperarcs$trans[[1]], c(0, 1))
   expect_equal(hgf$hyperarcs$pr[[1]], c(0.25, 0.75))
   expect_equal(
      hgf$hyperarcs$transWeights[[1]],
      matrix(c(10, 20), ncol = 1, dimnames = list(NULL, "Transition weight"))
   )
})

test_that("plotHypergraph supports stateLabel, actionLabel, and actionWLabel options", {
   prefix <- file.path(tempdir(), "plot_opts_")
   w <- binaryMDPWriter(prefix = prefix, getLog = FALSE)
   w$setWeights("Action weight")
   w$setTransWeights("Transition weight")
   w$process()
      w$stage()
         w$state(label = "S1")
            w$action(
               label = "A1",
               weights = 5,
               prob = c(1, 0, 0.25, 1, 1, 0.75),
               transWeights = c(10, 20)
            )
            w$endAction()
         w$endState()
      w$endStage()
      w$stage()
         w$state(label = "S2")
         w$endState()
         w$state(label = "S3")
         w$endState()
      w$endStage()
   w$endProcess()
   w$closeWriter()

   mdp <- loadMDP(prefix, getLog = FALSE)
   hgf <- getHypergraph(mdp)
   hgf$nodes$gId <- c(1, 3, 4)

   grDevices::pdf(file = tempfile(fileext = ".pdf"))
   on.exit(grDevices::dev.off(), add = TRUE)

   expect_invisible(plotHypergraph(hgf, c(2, 2)))
   expect_invisible(plotHypergraph(hgf, c(2, 2), stateLabel = "sId|label", actionLabel = "aIdx|label", actionWLabel = "none"))
   expect_invisible(plotHypergraph(hgf, c(2, 2), stateLabel = "sIdx|label"))
   expect_invisible(plotHypergraph(hgf, c(2, 2), stateLabel = "weight", mdp = mdp))
   expect_invisible(plotHypergraph(hgf, c(2, 2), stateLabel = "label|weight", mdp = mdp))
   expect_invisible(plotHypergraph(hgf, c(2, 2), stateLabel = "sId|weight|sIdx", mdp = mdp))
   expect_invisible(plotHypergraph(hgf, c(2, 2), actionLabel = "label|aIdx"))
   expect_error(plotHypergraph(hgf, c(2, 2), stateLabel = "weight"), "mdp model must be provided")
   expect_error(plotHypergraph(hgf, c(2, 2), stateLabel = "label|weight"), "mdp model must be provided")
   expect_invisible(plotHypergraph(hgf, c(2, 2), actionColor = "label"))
   expect_invisible(plotHypergraph(hgf, c(2, 2), actionColor = "policy", mdp = mdp))
   expect_error(plotHypergraph(hgf, c(2, 2), actionColor = "policy"), "mdp model must be provided")
   expect_invisible(plotHypergraph(hgf, c(2, 2), actionsVisible = "policy", mdp = mdp))
   expect_error(plotHypergraph(hgf, c(2, 2), actionsVisible = "policy"), "mdp model must be provided")
})

test_that("plotHypergraph supports custom state, action, and action weight labels", {
   hgf <- list(
      nodes = tibble::tibble(
         sId = c(0, 1),
         gId = c(1, 3),
         label = c("Head", "Tail"),
         stateLabel = c("Custom head", "Custom tail")
      ),
      hyperarcs = tibble::tibble(
         sId = 0,
         actionWeights = list(10),
         trans = list(1),
         pr = list(1),
         actionLabel = "Custom action",
         actionWLabel = "Custom weight",
         aIdx = 0,
         label = "A",
         lwd = 1,
         lty = 1,
         col = "black"
      )
   )
   grDevices::pdf(file = tempfile(fileext = ".pdf"))
   on.exit(grDevices::dev.off(), add = TRUE)

   expect_invisible(
      plotHypergraph(
         hgf,
         c(2, 2),
         stateLabel = "custom",
         actionLabel = "custom",
         actionWLabel = "custom"
      )
   )
   expect_error(
      plotHypergraph(list(nodes = dplyr::select(hgf$nodes, -stateLabel), hyperarcs = hgf$hyperarcs), c(2, 2), stateLabel = "custom"),
      'stateLabel = "custom" requires a stateLabel column'
   )
   expect_error(
      plotHypergraph(list(nodes = hgf$nodes, hyperarcs = dplyr::select(hgf$hyperarcs, -actionLabel)), c(2, 2), actionLabel = "custom"),
      'actionLabel = "custom" requires an actionLabel column'
   )
   expect_error(
      plotHypergraph(list(nodes = hgf$nodes, hyperarcs = dplyr::select(hgf$hyperarcs, -actionWLabel)), c(2, 2), actionWLabel = "custom"),
      'actionWLabel = "custom" requires an actionWLabel column'
   )
})

test_that("plotHypergraph supports actionWLabel and multiple action weights", {
   hgf <- list(
      nodes = tibble::tibble(sId = c(0, 1), gId = c(1, 3), label = c("Head", "Tail")),
      hyperarcs = tibble::tibble(
         sId = 0,
         actionWeights = list(c(1.5, 2.5)), # multiple weights
         trans = list(1),
         pr = list(1),
         transWeights = list(matrix(10, nrow = 1)),
         aIdx = 0,
         label = "A",
         lwd = 1,
         lty = 1,
         col = "black"
      )
   )
   grDevices::pdf(file = tempfile(fileext = ".pdf"))
   on.exit(grDevices::dev.off(), add = TRUE)

   expect_invisible(plotHypergraph(hgf, c(2, 2), actionWLabel = "weight"))
})

test_that("plotHypergraph filters to states connected forward to selected states", {
   hgf <- list(
      nodes = tibble::tibble(
         sId = c(0, 1, 2, 3, 4, 5, 6),
         gId = c(99, 2, 99, 4, 99, 99, 99),
         label = paste0("S", c(0, 1, 2, 3, 4, 5, 6))
      ),
      hyperarcs = tibble::tibble(
         sId = c(0, 1, 2, 5),
         actionWeights = list(numeric(0), numeric(0), numeric(0), numeric(0)),
         trans = list(c(1, 2), 3, 4, 6),
         pr = list(c(0.25, 0.75), 1, 1, 1),
         transLabels = list(c("keep", "drop"), "target", "other", "disconnected"),
         transWeights = list(
            matrix(c(10, 20), ncol = 1),
            matrix(30, ncol = 1),
            matrix(40, ncol = 1),
            matrix(50, ncol = 1)
         ),
         aIdx = c(0, 0, 0, 0),
         label = c("A", "B", "C", "D"),
         lwd = c(1, 1, 1, 1),
         lty = c(1, 1, 1, 1),
         col = c("black", "black", "black", "black")
      )
   )
   grDevices::pdf(file = tempfile(fileext = ".pdf"))
   on.exit(grDevices::dev.off(), add = TRUE)

   expect_invisible(plotHypergraph(hgf, c(2, 2), connectedTo = 1, transLabels = "prob"))
   expect_invisible(plotHypergraph(hgf, c(2, 2), connectedTo = 1, transLabels = "custom"))
   expect_invisible(plotHypergraph(hgf, c(2, 2), connectedTo = 1, transLabels = "weights"))
   expect_error(
      plotHypergraph(hgf, c(2, 2), connectedTo = 999),
      "connectedTo contains sId values not present in hgf\\$nodes"
   )
})

test_that("plotHypergraph connectedTo uses policy-visible actions", {
   hgf <- list(
      nodes = tibble::tibble(
         sId = c(0, 1, 2),
         gId = c(1, 99, 2),
         label = c("S0", "S1", "S2")
      ),
      hyperarcs = tibble::tibble(
         sId = c(0, 0),
         actionWeights = list(numeric(0), numeric(0)),
         trans = list(1, 2),
         pr = list(1, 1),
         aIdx = c(0, 1),
         label = c("to S1", "to S2"),
         lwd = c(1, 1),
         lty = c(1, 1),
         col = c("black", "black")
      )
   )
   testthat::local_mocked_bindings(
      getPolicy = function(mdp) tibble::tibble(sId = 0, aIdx = 1),
      .package = "MDP2"
   )

   grDevices::pdf(file = tempfile(fileext = ".pdf"))
   on.exit(grDevices::dev.off(), add = TRUE)

   expect_invisible(plotHypergraph(hgf, c(1, 2), actionsVisible = "policy", connectedTo = 0, mdp = list()))
})

test_that("plotHypergraph recalculates visible grid after connectedTo filtering", {
   hgf <- list(
      nodes = tibble::tibble(
         sId = c(0, 1, 2, 3, 4, 5, 6),
         gId = c(1, 5, 6, 10, 12, 99, 100),
         label = paste0("S", c(0, 1, 2, 3, 4, 5, 6))
      ),
      hyperarcs = tibble::tibble(
         sId = c(0, 1, 2, 5),
         actionWeights = list(numeric(0), numeric(0), numeric(0), numeric(0)),
         trans = list(c(1, 2), 3, 4, 6),
         pr = list(c(0.5, 0.5), 1, 1, 1),
         aIdx = c(0, 0, 0, 0),
         label = c("A", "B", "C", "D"),
         lwd = c(1, 1, 1, 1),
         lty = c(1, 1, 1, 1),
         col = c("black", "black", "black", "black")
      )
   )
   grDevices::pdf(file = tempfile(fileext = ".pdf"))
   on.exit(grDevices::dev.off(), add = TRUE)

   expect_invisible(plotHypergraph(hgf, c(6, 2), connectedTo = 0, recalcGrid = TRUE))
})

library(MDP2)
cleanUp<-function() unlink("*.bin")


test_that("Total reward",{
   source("files/machine_replacement_v1.R")
   mdp<-loadMDP("machine1_", getLog = FALSE)
   w<-"Net reward"             # label of the weight we want to optimize
   scrapValues<-c(30,10,5,0)   # scrap values (the values of the 4 states at stage 4)
   runValueIte(mdp, w, termValues=scrapValues, getLog = FALSE)
   expect_equal(getPolicy(mdp,13)$weight,102.2)
   rm(mdp)
   
   # same model with a single dummy node
   source("files/machine_replacement_v2.R")
   mdp<-loadMDP("machine2_", getLog = FALSE)
   w<-"Net reward"             # label of the weight we want to optimize
   runValueIte(mdp, w, termValues=0, getLog = FALSE)
   expect_equal(getPolicy(mdp,12)$weight,102.2)
   rm(mdp)
})

test_that("Transition-level total reward",{
   w <- binaryMDPWriter(prefix = "trans_reward_", getLog = FALSE)
   w$setWeights(character())
   w$setTransWeights("Transition reward")
   w$process()
      w$stage()
         w$state()
            w$action(weights = numeric(0),
                     prob = c(1, 0, 0.25, 1, 1, 0.75),
                     transWeights = c(10, 20))
            w$endAction()
         w$endState()
      w$endStage()
      w$stage()
         w$state()
         w$endState()
         w$state()
         w$endState()
      w$endStage()
   w$endProcess()
   w$closeWriter()

   mdp <- loadMDP("trans_reward_", getLog = FALSE)
   expect_equal(mdp$weightActionNames, character())
   expect_equal(mdp$weightTransNames, "Transition reward")
   expect_error(
      mdp$ptr$valueIte(0, 0, 1L, 0, 0L, 0L, c(0, 0), 0, 1),
      "Transition-level weights are not supported for BellmanOp::Discounted"
   )
   runValueIte(mdp, "Transition reward", termValues = c(100, 200), getLog = FALSE)
   policy <- getPolicy(mdp)
   expect_equal(policy$weight[policy$stateStr == "0,0"], 192.5)
   rm(mdp)
})

test_that("Global weight lookup rejects ambiguous names", {
   mdp <- list(weightNames = c("Net", "Net reward"))
   expect_equal(getWIdx(mdp, "Net"), 0)
   expect_error(getWIdx(mdp, "e"), "ambiguous")
})

test_that("Value iteration supports minimization objective", {
   w <- binaryMDPWriter(prefix = "sense_", getLog = FALSE)
   w$setWeights("Cost")
   w$process()
      w$stage()
         w$state()
            w$action(weights = 10, prob = c(1, 0, 1), end = TRUE)
            w$endAction()
            w$action(weights = 1, prob = c(1, 0, 1), end = TRUE)
            w$endAction()
         w$endState()
      w$endStage()
      w$stage()
         w$state()
         w$endState()
      w$endStage()
   w$endProcess()
   w$closeWriter()

   mdp <- loadMDP("sense_", getLog = FALSE)
   runValueIte(mdp, "Cost", termValues = 0, objective = "max", getLog = FALSE)
   policy <- getPolicy(mdp)
   expect_equal(policy$aIdx[policy$stateStr == "0,0"], 0)
   expect_equal(policy$weight[policy$stateStr == "0,0"], 10)
   expect_equal(getRPO(mdp, "Cost", iA = 0, sId = 1, objective = "max")$rpo, 9)

   runValueIte(mdp, "Cost", termValues = 0, objective = "min", getLog = FALSE)
   policy <- getPolicy(mdp)
   expect_equal(policy$aIdx[policy$stateStr == "0,0"], 1)
   expect_equal(policy$weight[policy$stateStr == "0,0"], 1)
   expect_equal(getRPO(mdp, "Cost", iA = 1, sId = 1, objective = "min")$rpo, 9)
})

test_that("Value iteration supports minimum and maximum successor Bellman operators", {
   w <- binaryMDPWriter(prefix = "minmax_", getLog = FALSE)
   w$setWeights("Weight")
   w$setTransWeights("Trans weight")
   w$process()
      w$stage()
         w$state()
            w$action(weights = 0,
                     prob = c(1, 0, 0.5, 1, 1, 0.5),
                     transWeights = c(100, 0))
            w$endAction()
            w$action(weights = 5,
                     prob = c(1, 0, 1),
                     transWeights = 5)
            w$endAction()
         w$endState()
      w$endStage()
      w$stage()
         w$state()
         w$endState()
         w$state()
         w$endState()
      w$endStage()
   w$endProcess()
   w$closeWriter()

   mdp <- loadMDP("minmax_", getLog = FALSE)

   runValueIte(mdp, "Weight", termValues = c(1, 10), bellmanOp = "min", objective = "max", getLog = FALSE)
   policy <- getPolicy(mdp)
   expect_equal(policy$aIdx[policy$stateStr == "0,0"], 1)
   expect_equal(policy$weight[policy$stateStr == "0,0"], 6)

   runValueIte(mdp, "Weight", termValues = c(1, 10), bellmanOp = "max", objective = "max", getLog = FALSE)
   policy <- getPolicy(mdp)
   expect_equal(policy$aIdx[policy$stateStr == "0,0"], 0)
   expect_equal(policy$weight[policy$stateStr == "0,0"], 10)

   runValueIte(mdp, "Weight", termValues = c(1, 10), bellmanOp = "min", objective = "min", getLog = FALSE)
   policy <- getPolicy(mdp)
   expect_equal(policy$aIdx[policy$stateStr == "0,0"], 0)
   expect_equal(policy$weight[policy$stateStr == "0,0"], 1)

   runValueIte(mdp, "Weight", termValues = c(1, 10), bellmanOp = "max", objective = "min", getLog = FALSE)
   policy <- getPolicy(mdp)
   expect_equal(policy$aIdx[policy$stateStr == "0,0"], 1)
   expect_equal(policy$weight[policy$stateStr == "0,0"], 6)

   runValueIte(mdp, "Trans weight", termValues = c(1, 10), bellmanOp = "min", objective = "max", getLog = FALSE)
   policy <- getPolicy(mdp)
   expect_equal(policy$aIdx[policy$stateStr == "0,0"], 0)
   expect_equal(policy$weight[policy$stateStr == "0,0"], 10)

   runValueIte(mdp, "Trans weight", termValues = c(1, 10), bellmanOp = "max", objective = "max", getLog = FALSE)
   policy <- getPolicy(mdp)
   expect_equal(policy$aIdx[policy$stateStr == "0,0"], 0)
   expect_equal(policy$weight[policy$stateStr == "0,0"], 101)
})

test_that("SecondMoment supports action-level weights", {
   w <- binaryMDPWriter(prefix = "second_moment_action_", getLog = FALSE)
   w$setWeights("Weight")
   w$process()
      w$stage()
         w$state()
            w$action(label = "high", weights = 2, prob = c(1, 0, 1), end = TRUE)
            w$endAction()
            w$action(label = "low", weights = 1, prob = c(1, 1, 1), end = TRUE)
            w$endAction()
         w$endState()
      w$endStage()
      w$stage()
         w$state()
            w$action(weights = 3, prob = c(1, 0, 1), end = TRUE)
            w$endAction()
         w$endState()
         w$state()
            w$action(weights = 0, prob = c(1, 0, 1), end = TRUE)
            w$endAction()
         w$endState()
      w$endStage()
      w$stage()
         w$state()
         w$endState()
      w$endStage()
   w$endProcess()
   w$closeWriter()

   mdp <- loadMDP("second_moment_action_", getLog = FALSE)

   runValueIte(mdp, "Weight", termValues = 0, getLog = FALSE)
   runCalcWeights(mdp, "Weight", criterion = "secondMoment", termValues = 0)
   policy <- getPolicy(mdp)
   expect_equal(policy$weight[policy$stateStr == "0,0"], 25)

   runValueIte(mdp, "Weight", termValues = 0, bellmanOp = "secondMoment", objective = "max", getLog = FALSE)
   policy <- getPolicy(mdp)
   expect_equal(policy$aIdx[policy$stateStr == "0,0"], 0)
   expect_equal(policy$weight[policy$stateStr == "0,0"], 25)

   runValueIte(mdp, "Weight", termValues = 0, bellmanOp = "secondMoment", objective = "min", getLog = FALSE)
   policy <- getPolicy(mdp)
   expect_equal(policy$aIdx[policy$stateStr == "0,0"], 1)
   expect_equal(policy$weight[policy$stateStr == "0,0"], 1)
})

test_that("SecondMoment supports transition-level weights", {
   w <- binaryMDPWriter(prefix = "second_moment_transition_", getLog = FALSE)
   w$setWeights(character())
   w$setTransWeights("Weight")
   w$process()
      w$stage()
         w$state()
            w$action(label = "high", weights = numeric(0), prob = c(1, 0, 1),
                     transWeights = 2, end = TRUE)
            w$endAction()
            w$action(label = "low", weights = numeric(0), prob = c(1, 1, 1),
                     transWeights = 1, end = TRUE)
            w$endAction()
         w$endState()
      w$endStage()
      w$stage()
         w$state()
            w$action(weights = numeric(0), prob = c(1, 0, 1),
                     transWeights = 3, end = TRUE)
            w$endAction()
         w$endState()
         w$state()
            w$action(weights = numeric(0), prob = c(1, 0, 1),
                     transWeights = 0, end = TRUE)
            w$endAction()
         w$endState()
      w$endStage()
      w$stage()
         w$state()
         w$endState()
      w$endStage()
   w$endProcess()
   w$closeWriter()

   mdp <- loadMDP("second_moment_transition_", getLog = FALSE)

   runValueIte(mdp, "Weight", termValues = 0, getLog = FALSE)
   runCalcWeights(mdp, "Weight", criterion = "secondMoment", termValues = 0)
   policy <- getPolicy(mdp)
   expect_equal(policy$weight[policy$stateStr == "0,0"], 25)

   runValueIte(mdp, "Weight", termValues = 0, bellmanOp = "secondMoment", objective = "max", getLog = FALSE)
   policy <- getPolicy(mdp)
   expect_equal(policy$aIdx[policy$stateStr == "0,0"], 0)
   expect_equal(policy$weight[policy$stateStr == "0,0"], 25)

   runValueIte(mdp, "Weight", termValues = 0, bellmanOp = "secondMoment", objective = "min", getLog = FALSE)
   policy <- getPolicy(mdp)
   expect_equal(policy$aIdx[policy$stateStr == "0,0"], 1)
   expect_equal(policy$weight[policy$stateStr == "0,0"], 1)
})

test_that("SecondMoment value iteration is finite-horizon only", {
   source("files/two_level_hmdp.R")
   mdp <- loadMDP("2lev_", getLog = FALSE)
   runValueIte(mdp, "Net reward", "Duration", bellmanOp = "secondMoment", termValues = rep(0, mdp$founderStatesLast), getLog = FALSE)
   expect_match(mdp$ptr$getLog(), "SecondMoment value iteration is currently only supported for finite time-horizon HMDPs")
})

test_that("Variance supports action-level fixed-policy evaluation", {
   w <- binaryMDPWriter(prefix = "variance_action_", getLog = FALSE)
   w$setWeights("Weight")
   w$process()
      w$stage()
         w$state()
            w$action(weights = 1, prob = c(1, 0, 0.5, 1, 1, 0.5), end = TRUE)
            w$endAction()
         w$endState()
      w$endStage()
      w$stage()
         w$state()
            w$action(weights = 0, prob = c(1, 0, 1), end = TRUE)
            w$endAction()
         w$endState()
         w$state()
            w$action(weights = 2, prob = c(1, 0, 1), end = TRUE)
            w$endAction()
         w$endState()
      w$endStage()
      w$stage()
         w$state()
         w$endState()
      w$endStage()
   w$endProcess()
   w$closeWriter()

   mdp <- loadMDP("variance_action_", getLog = FALSE)
   runValueIte(mdp, "Weight", termValues = 0, getLog = FALSE)

   mdp$ptr$setTerminalW(0)
   runCalcWeights(mdp, "Weight", criterion = "expected", termValues = 0)
   expected <- getPolicy(mdp)$weight

   mdp$ptr$setTerminalW(0)
   runCalcWeights(mdp, "Weight", criterion = "secondMoment", termValues = 0)
   second_moment <- getPolicy(mdp)$weight

   runCalcWeights(mdp, "Weight", criterion = "variance", termValues = 0)
   variance <- getPolicy(mdp)$weight

   s0 <- getPolicy(mdp)$stateStr == "0,0"
   expect_equal(variance[s0], 1)
   expect_equal(variance, second_moment - expected^2)
})

test_that("Variance supports transition-level fixed-policy evaluation", {
   w <- binaryMDPWriter(prefix = "variance_transition_", getLog = FALSE)
   w$setWeights(character())
   w$setTransWeights("Weight")
   w$process()
      w$stage()
         w$state()
            w$action(weights = numeric(0), prob = c(1, 0, 0.5, 1, 1, 0.5),
                     transWeights = c(1, 1), end = TRUE)
            w$endAction()
         w$endState()
      w$endStage()
      w$stage()
         w$state()
            w$action(weights = numeric(0), prob = c(1, 0, 1),
                     transWeights = 0, end = TRUE)
            w$endAction()
         w$endState()
         w$state()
            w$action(weights = numeric(0), prob = c(1, 0, 1),
                     transWeights = 2, end = TRUE)
            w$endAction()
         w$endState()
      w$endStage()
      w$stage()
         w$state()
         w$endState()
      w$endStage()
   w$endProcess()
   w$closeWriter()

   mdp <- loadMDP("variance_transition_", getLog = FALSE)
   runValueIte(mdp, "Weight", termValues = 0, getLog = FALSE)

   mdp$ptr$setTerminalW(0)
   runCalcWeights(mdp, "Weight", criterion = "expected", termValues = 0)
   expected <- getPolicy(mdp)$weight

   mdp$ptr$setTerminalW(0)
   runCalcWeights(mdp, "Weight", criterion = "secondMoment", termValues = 0)
   second_moment <- getPolicy(mdp)$weight

   runCalcWeights(mdp, "Weight", criterion = "variance", termValues = 0)
   variance <- getPolicy(mdp)$weight

   s0 <- getPolicy(mdp)$stateStr == "0,0"
   expect_equal(variance[s0], 1)
   expect_equal(variance, second_moment - expected^2)
})

test_that("Variance uses terminal values as means and terminal variance zero", {
   w <- binaryMDPWriter(prefix = "variance_terminal_", getLog = FALSE)
   w$setWeights("Weight")
   w$process()
      w$stage()
         w$state()
            w$action(weights = 0, prob = c(1, 0, 0.5, 1, 1, 0.5), end = TRUE)
            w$endAction()
         w$endState()
      w$endStage()
      w$stage()
         w$state()
         w$endState()
         w$state()
         w$endState()
      w$endStage()
   w$endProcess()
   w$closeWriter()

   mdp <- loadMDP("variance_terminal_", getLog = FALSE)
   runValueIte(mdp, "Weight", termValues = c(0, 2), getLog = FALSE)
   runCalcWeights(mdp, "Weight", criterion = "variance", termValues = c(0, 2))
   policy <- getPolicy(mdp)

   expect_equal(policy$weight[policy$stateStr == "0,0"], 1)
   expect_equal(policy$weight[policy$stateStr == "1,0"], 0)
   expect_equal(policy$weight[policy$stateStr == "1,1"], 0)
})

test_that("Variance is not a value-iteration Bellman operator", {
   w <- binaryMDPWriter(prefix = "variance_not_value_ite_", getLog = FALSE)
   w$setWeights("Weight")
   w$process()
      w$stage()
         w$state()
            w$action(weights = 1, prob = c(1, 0, 1), end = TRUE)
            w$endAction()
         w$endState()
      w$endStage()
      w$stage()
         w$state()
         w$endState()
      w$endStage()
   w$endProcess()
   w$closeWriter()

   mdp <- loadMDP("variance_not_value_ite_", getLog = FALSE)
   expect_error(
      runValueIte(mdp, "Weight", termValues = 0, bellmanOp = "variance", getLog = FALSE),
      "should be one of"
   )
   expect_error(
      mdp$ptr$calcRPO(8, 0, as.integer(c(0)), 0L, as.integer(c(0)), 0, 0L, 1),
      "Bellman operator not implemented"
   )
})


test_that("Long run average reward",{
   source("files/two_level_hmdp.R")
   mdp<-loadMDP("2lev_", getLog = FALSE)
   expect_lt(mdp$ptr$policyIteFixedPolicy(1L, 0L, 1L, 1), -1e+15)
   expect_match(mdp$ptr$getLog(), "valid fixed policy must be set")
   expect_equal(runPolicyIteAve(mdp,"Net reward","Duration", getLog = FALSE), 5.71428571428571441259691710001789033412933349609375)
   expect_equal(runPolicyIteAve(mdp,"Net reward","Items", getLog = FALSE), 4)
   expect_equal(runPolicyIteAve(mdp,"Items","Duration", getLog = FALSE), 2.71428571428571441259691710001789033412933349609375)
   rm(mdp)
   
   #sprintf("%.100f",g)
   source("files/HCT_ex6.1.1.R")
   mdp<-loadMDP("hct611_", getLog = FALSE)
   expect_equal(runPolicyIteAve(mdp,"Net reward","Duration", getLog = FALSE), -0.433789954337899297254210750907077454030513763427734375)
   rm(mdp)
   
   source("files/HCT_exc6.4.R")
   mdp<-loadMDP("hct64_", getLog = FALSE)
   expect_equal(runPolicyIteAve(mdp,"Net reward","Duration", getLog = FALSE), -17.7686915887850460649133310653269290924072265625)
   rm(mdp)
   
   source("files/HCT_exc6.7.R")
   mdp<-loadMDP("hct67_", getLog = FALSE)
   expect_equal(runPolicyIteAve(mdp,"Net reward","Duration", getLog = FALSE), -7.8132707659357780727304998436011373996734619140625)
   rm(mdp)
   
   source("files/HCT_exc7.3.R")
   mdp<-loadMDP("hct73_", getLog = FALSE)
   expect_equal(runPolicyIteAve(mdp,"Net reward","Duration", getLog = FALSE), 4)
   rm(mdp)
})


test_that("Discounted expected reward",{
   mdp<-loadMDP("2lev_", getLog = FALSE)
   rate<-0.1
   runPolicyIteDiscount(mdp, "Net reward", "Duration", rate, getLog = FALSE)
   weightsPolicyIte<-getPolicy(mdp)$weight
   expect_equal(weightsPolicyIte[16], 56.33119951473481279435873148031532764434814453125)
   runValueIte(mdp, "Net reward", "Duration", rate, eps = 1e-15, maxIte = 10000, getLog = FALSE)
   weightsrunValueIte<-getPolicy(mdp)$weight
   expect_equal(weightsPolicyIte, weightsrunValueIte)
   rate<-0.01
   runPolicyIteDiscount(mdp, "Net reward", "Duration", rate, getLog = FALSE)
   weightsPolicyIte<-getPolicy(mdp)$weight
   runValueIte(mdp, "Net reward", "Duration", rate, eps = 1e-15, maxIte = 10000, getLog = FALSE)
   weightsrunValueIte<-getPolicy(mdp)$weight
   expect_equal(weightsPolicyIte, weightsrunValueIte)
   rm(mdp)
   
   mdp<-loadMDP("hct611_", getLog = FALSE)
   rate<-0.1
   runPolicyIteDiscount(mdp, "Net reward", "Duration", rate, getLog = FALSE)
   weightsPolicyIte<-getPolicy(mdp)$weight
   expect_equal(weightsPolicyIte[5], -12.34541222517888314769152202643454074859619140625)
   expect_equal(sum(weightsPolicyIte), -36.6474287369603786146399215795099735260009765625)
   runValueIte(mdp, "Net reward", "Duration", rate, eps = 1e-15, maxIte = 10000, getLog = FALSE)
   weightsrunValueIte<-getPolicy(mdp)$weight
   expect_equal(weightsPolicyIte, weightsrunValueIte)
   rate<-0.01
   runPolicyIteDiscount(mdp, "Net reward", "Duration", rate, getLog = FALSE)
   weightsPolicyIte<-getPolicy(mdp)$weight
   runValueIte(mdp, "Net reward", "Duration", rate, eps = 1e-15, maxIte = 10000, getLog = FALSE)
   weightsrunValueIte<-getPolicy(mdp)$weight
   expect_equal(weightsPolicyIte, weightsrunValueIte)
   rm(mdp)
   
   mdp<-loadMDP("hct64_", getLog = FALSE)
   rate<-0.1
   runPolicyIteDiscount(mdp, "Net reward", "Duration", rate, getLog = FALSE)
   weightsPolicyIte<-getPolicy(mdp)$weight
   expect_equal(sum(weightsPolicyIte), -920.57968416603171135648153722286224365234375)
   runValueIte(mdp, "Net reward", "Duration", rate, eps = 1e-15, maxIte = 10000, getLog = FALSE)
   weightsrunValueIte<-getPolicy(mdp)$weight
   expect_equal(weightsPolicyIte, weightsrunValueIte)
   rate<-0.01
   runPolicyIteDiscount(mdp, "Net reward", "Duration", rate, getLog = FALSE)
   weightsPolicyIte<-getPolicy(mdp)$weight
   runValueIte(mdp, "Net reward", "Duration", rate, eps = 1e-15, maxIte = 10000, getLog = FALSE)
   weightsrunValueIte<-getPolicy(mdp)$weight
   expect_equal(weightsPolicyIte, weightsrunValueIte)
   rm(mdp)
   
   mdp<-loadMDP("hct67_", getLog = FALSE)
   rate<-0.1
   runPolicyIteDiscount(mdp, "Net reward", "Duration", rate, getLog = FALSE)
   weightsPolicyIte<-getPolicy(mdp)$weight
   expect_equal(sum(weightsPolicyIte), -2754.88121322102369958884082734584808349609375)
   runValueIte(mdp, "Net reward", "Duration", rate, eps = 1e-15, maxIte = 10000, getLog = FALSE)
   weightsrunValueIte<-getPolicy(mdp)$weight
   expect_equal(weightsPolicyIte, weightsrunValueIte)
   rate<-0.01
   runPolicyIteDiscount(mdp, "Net reward", "Duration", rate, getLog = FALSE)
   weightsPolicyIte<-getPolicy(mdp)$weight
   runValueIte(mdp, "Net reward", "Duration", rate, eps = 1e-15, maxIte = 10000, getLog = FALSE)
   weightsrunValueIte<-getPolicy(mdp)$weight
   expect_equal(weightsPolicyIte, weightsrunValueIte)
   rm(mdp)
   
   mdp<-loadMDP("hct73_", getLog = FALSE)
   rate<-0.1
   runPolicyIteDiscount(mdp, "Net reward", "Duration", rate, getLog = FALSE)
   weightsPolicyIte<-getPolicy(mdp)$weight
   expect_equal(sum(weightsPolicyIte), 240.1591230611287528518005274236202239990234375)
   runValueIte(mdp, "Net reward", "Duration", rate, eps = 1e-15, maxIte = 10000, getLog = FALSE)
   weightsrunValueIte<-getPolicy(mdp)$weight
   expect_equal(weightsPolicyIte, weightsrunValueIte)
   rate<-0.01
   runPolicyIteDiscount(mdp, "Net reward", "Duration", rate, getLog = FALSE)
   weightsPolicyIte<-getPolicy(mdp)$weight
   runValueIte(mdp, "Net reward", "Duration", rate, eps = 1e-15, maxIte = 10000, getLog = FALSE)
   weightsrunValueIte<-getPolicy(mdp)$weight
   expect_equal(weightsPolicyIte, weightsrunValueIte)
   rm(mdp)
   
   # test discount factor conversion
   mdp<-loadMDP("hct73_", getLog = FALSE)
   d<-0.9
   rate<- -log(d)
   runPolicyIteDiscount(mdp, "Net reward", "Duration", rate, getLog = FALSE)
   weightsPolicyIte1<-getPolicy(mdp)$weight
   runPolicyIteDiscount(mdp, "Net reward", "Duration", discountFactor = d, getLog = FALSE)
   weightsPolicyIte2<-getPolicy(mdp)$weight 
   expect_equal(weightsPolicyIte1, weightsPolicyIte2)
   rm(mdp)
})


cleanUp()

library(MDP2)
cleanUp<-function() {
   unlink("*.bin")
   unlink("*.hmp")
}

test_that("convertHMP2Binary",{
   source("files/two_level_hmdp.R")
   prefix <- "2lev_"
   mdp <- loadMDP(prefix, getLog = FALSE)
   info1 <- getInfo(mdp, dfLevel = "action", asStringsActions = TRUE)
   g1 <- runPolicyIteAve(mdp,"Net reward","Duration", getLog = FALSE)
   rm(mdp)
   convertBinary2HMP(prefix, getLog = FALSE)
   prefix1 <- "2lev-converted_"
   convertHMP2Binary(paste0(prefix, "converted.hmp"), prefix1, getLog = FALSE)
   mdp <- loadMDP(prefix1, getLog = FALSE)
   info2 <- getInfo(mdp, dfLevel = "action", asStringsActions = TRUE)
   g2<-runPolicyIteAve(mdp,"Net reward","Duration", getLog = FALSE)
   rm(mdp)
   expect_equal(g1, g2)
   expect_equal(info1$df, info2$df)
   
   source("files/machine_replacement_v1.R")
   prefix <- "machine1_"
   mdp <- loadMDP(prefix, getLog = FALSE)
   info1 <- getInfo(mdp, dfLevel = "action", asStringsActions = TRUE)
   runValueIte(mdp, "Net reward" , termValues=c(30,10,5,0), getLog = FALSE)
   g1 <- getPolicy(mdp)[9,5]
   rm(mdp)
   convertBinary2HMP(prefix, duration = NULL, getLog = FALSE)
   prefix1 <- "machine1-converted_"
   convertHMP2Binary(paste0(prefix, "converted.hmp"), prefix1, getLog = FALSE)
   mdp <- loadMDP(prefix1, getLog = FALSE)
   info2 <- getInfo(mdp, dfLevel = "action", asStringsActions = TRUE)
   runValueIte(mdp, "Net reward" , termValues=c(30,10,5,0), getLog = FALSE)
   g2 <- getPolicy(mdp)[9,5]
   rm(mdp)
   expect_equal(g1, g2)
   expect_equal(info1$df$trans, info2$df$trans)
   expect_equal(info1$df$pr, info2$df$pr)
   
   # Small MDP
   prefix <- "test_"
   w = binaryMDPWriter(prefix, getLog = FALSE)
   w$setWeights(c("Duration", "Net rewards"))
   w$process()
      w$stage("Lactation cycle")
         w$state("Low")
            w$action(label = "Keep", scope = 1, weights = c(1, 10000), id = c(0, 1), pr = c(0.6, 0.4))
            w$endAction()
         w$endState()
         w$state("Average")
            w$action(label = "Keep", scope = 1, weights = c(1, 12000), id = c(0, 1), pr= c(0.2, 0.8))
            w$endAction()
         w$endState()
      w$endStage()
   w$endProcess()
   w$closeWriter()
   mdp <- loadMDP(prefix, getLog = FALSE)
   info1 <- getInfo(mdp, dfLevel = "action", asStringsActions = TRUE)
   g1 <- runPolicyIteAve(mdp,"Net reward","Duration", getLog = FALSE)
   rm(mdp)
   convertBinary2HMP(prefix, getLog = FALSE)
   prefix1 <- "test-converted_"
   convertHMP2Binary(paste0(prefix, "converted.hmp"), prefix1, getLog = FALSE)
   mdp <- loadMDP(prefix1, getLog = FALSE)
   info2 <- getInfo(mdp, dfLevel = "action", asStringsActions = TRUE)
   g2<-runPolicyIteAve(mdp,"Net reward","Duration", getLog = FALSE)
   rm(mdp)
   expect_equal(g1, g2)
   expect_equal(info1$df, info2$df)
   
   # Small MDP with no labels
   prefix <- "test2_"
   w = binaryMDPWriter(prefix, getLog = FALSE)
   w$setWeights(c("D", "N"))
   w$process()
      w$stage()
         w$state()
            w$action(scope = 1, weights = c(1, 10000), id = c(0, 1), pr = c(0.6, 0.4), end = TRUE)
         w$endState()
         w$state()
            w$action(scope = 1, weights = c(1, 12000), id = c(0, 1), pr= c(0.2, 0.8), end = TRUE)
         w$endState()
      w$endStage()
   w$endProcess()
   w$closeWriter()
   mdp <- loadMDP(prefix, getLog = FALSE)
   info1 <- getInfo(mdp, dfLevel = "action", asStringsActions = TRUE)
   g1 <- runPolicyIteAve(mdp,"N","D", getLog = FALSE)
   rm(mdp)
   convertBinary2HMP(prefix, getLog = FALSE)
   prefix1 <- "test-converted_"
   convertHMP2Binary(paste0(prefix, "converted.hmp"), prefix1, getLog = FALSE)
   mdp <- loadMDP(prefix1, getLog = FALSE)
   info2 <- getInfo(mdp, dfLevel = "action", asStringsActions = TRUE)
   g2<-runPolicyIteAve(mdp,"N","D", getLog = FALSE)
   rm(mdp)
   expect_equal(g1, g2)
   info1$df$label <- NULL
   info2$df$label <- NULL
   info1$df$label_action <- NULL
   info2$df$label_action <- NULL
   expect_equal(info1$df, info2$df)
   
   ## Existing hmp files 
   n <- 3
   prefix <- paste0("cow", n, "_")
   convertHMP2Binary(paste0("files/cow", n, ".hmp"), prefix, getLog = FALSE)
   mdp <- loadMDP(prefix, getLog = FALSE)
   info1 <- getInfo(mdp, dfLevel = "action", asStringsActions = TRUE)
   expect_equal(mdp$states, n)
   rm(mdp)
   convertBinary2HMP(prefix, getLog = FALSE)
   prefix1 <- paste0("cow-converted", n, "_")
   convertHMP2Binary(paste0(prefix, "converted.hmp"), prefix1, getLog = FALSE)
   mdp <- loadMDP(prefix1, getLog = FALSE)
   info2 <- getInfo(mdp, dfLevel = "action", asStringsActions = TRUE)
   rm(mdp)
   expect_equal(info1$df, info2$df)
   # 12 states
   n <- 12
   prefix <- paste0("cow", n, "_")
   convertHMP2Binary(paste0("files/cow", n, ".hmp"), prefix, getLog = FALSE)
   mdp <- loadMDP(prefix, getLog = FALSE)
   info1 <- getInfo(mdp, dfLevel = "action", asStringsActions = TRUE)
   expect_equal(mdp$states, n)
   rm(mdp)
   convertBinary2HMP(prefix, getLog = FALSE)
   prefix1 <- paste0("cow-converted", n, "_")
   convertHMP2Binary(paste0(prefix, "converted.hmp"), prefix1, getLog = FALSE)
   mdp <- loadMDP(prefix1, getLog = FALSE)
   info2 <- getInfo(mdp, dfLevel = "action", asStringsActions = TRUE)
   rm(mdp)
   expect_equal(info1$df, info2$df)
   ## 36 states
   n <- 36
   prefix <- paste0("cow", n, "_")
   convertHMP2Binary(paste0("files/cow", n, ".hmp"), prefix, getLog = FALSE)
   mdp <- loadMDP(prefix, getLog = FALSE)
   info1 <- getInfo(mdp, dfLevel = "action", asStringsActions = TRUE)
   expect_equal(mdp$states, n)
   rm(mdp)
   convertBinary2HMP(prefix, getLog = FALSE)
   prefix1 <- paste0("cow-converted", n, "_")
   convertHMP2Binary(paste0(prefix, "converted.hmp"), prefix1, getLog = FALSE)
   mdp <- loadMDP(prefix1, getLog = FALSE)
   info2 <- getInfo(mdp, dfLevel = "action", asStringsActions = TRUE)
   rm(mdp)
   expect_equal(info1$df, info2$df)
})

cleanUp()

## Excercise 6.7 in Tijms, H.C., "A first course in stochastic models", John Wiley & Sons Ltd, 2003.
## The semi-MDP is specified using binaryMDPWriter and actions with prob

MVal<-5
K<-200
a<-c(60,30,50,25,100)
P<-matrix(c(
   0.5,0.25,0.375,0.3,0.50,
   0.3,0.50,0.375,0.5,0.25,
   0.2,0.25,0.250,0.2,0.25), ncol=MVal, byrow=T)
sumP<-matrix(0,nrow=MVal+2,ncol=MVal)    # sumP[stock-1,town] = sum of pr at and above stock-1
for (r in 1:3) {
   for (c in 1:MVal) {
      sumP[r,c] <- sum(P[r:3,c])
   }
}

states<-expand.grid(s=0:MVal,j=1:5)    # define states in a data frame
states$label<-paste("(",states$s,",",states$j,")",sep="")
# set costs for each action
states$a0cost<-states$a1cost<-NA
for (i in 1:length(states$s)) {
   s<-states$s[i]
   j<-states$j[i]
   states$a0cost[i]<-K*sumP[(s+1)+1,j]
   states$a1cost[i]<-a[j]
}
# set trans pr matrix for each action
a0TransPr<-a1TransPr<-matrix(0, nrow=length(states$s), ncol=length(states$s))
for (i in 1:length(states$s)) {
   s<-states$s[i]
   j<-states$j[i]
   for (k in 1:length(states$s)) {
      h<-states$s[k]
      nextTown<-ifelse(j==5, 1, j+1)
      if (states$j[k]!=nextTown) next
      a0TransPr[i,k]<- ifelse(h==0, sumP[s+1,nextTown], ifelse(h<=s & s-h<3, P[s-h+1,nextTown], 0) )
      a1TransPr[i,k]<- ifelse(MVal-h<3, P[MVal-h+1,nextTown], 0)
   }
}


transPr<-function(a,rowId) {
   if (a==0) {
      idx<-which(a0TransPr[rowId,]>0)
      pr<-rbind(1,idx-1,a0TransPr[rowId,idx])
      return(as.vector(pr))
   }
   if (a==1) {
      idx<-which(a1TransPr[rowId,]>0)
      pr<-rbind(1,idx-1,a1TransPr[rowId,idx])
      return(as.vector(pr))
   }
   return(NULL)
}
w<-binaryMDPWriter("hct67_", getLog = FALSE)
w$setWeights(c("Duration","Net reward"))
w$process()
   w$stage()
      for (ii in 1:length(states$j)) {
         j<-states$j[ii]; s<-states$s[ii]
         w$state(label=states$label[ii])
            w$action(label="0", weights=c(1, -states$a0cost[ii]), prob=transPr(0,ii), end=T)
            w$action(label="1", weights=c(1, -states$a1cost[ii]), prob=transPr(1,ii), end=T)
         w$endState()
      }
   w$endStage()
w$endProcess()
w$closeWriter()

# Functions for converting different file formats


#' Convert a HMDP model stored in a hmp (xml) file to binary file format.
#'
#' The function simply parse the hmp file and create binary files using
#' the [binaryMDPWriter()].
#'
#' @param file The name of the HMP file (e.g. `r.hmp`).
#' @param prefix A character string with the prefix which will be added to the binary files.
#' @param getLog Output log text.
#' 
#' @return NULL (invisible).
#' @note Note all indexes are starting from zero (C/C++ style).
#' @seealso [binaryMDPWriter()].
#' @example inst/examples/convert-ex.R
#' @export
convertHMP2Binary<-function(file, prefix="", getLog = TRUE) {
   setWeights <- function(labels) {
      ctrW <<- length(labels) + 1
      w$setWeights(c("Duration", labels))
   }
   
   # stateCtr<-function(g) {
   #    browser()
   #    #xml2::xml_length(g)
   #    length(xmlChildren(g))
   # }
   
   # @param p A process node
   process<-function(p) {
      w$process()
      states <- c(xml2::xml_length(xml2::xml_children(p)), 0)  # number of states in each stage (add 0 to indicate last stage)
      for (i in 1:(length(states)-1)) stage(xml2::xml_child(p, i), states[i+1])
      w$endProcess()
   }
   
   # @param g A stage node
   # @param states Number of states at next stage
   stage <- function(g, states) {
      w$stage()
      r <- xml2::xml_children(g)
      # cat("states:\n"); print(r)
      for (i in 1:length(r)) state(r[i], states)
      w$endStage()
   }
   
   # @param s A state node
   # @param states Number of states at next stage
   state <- function(s, states) {
      w$state(label = xml2::xml_attr(s, "l"))
      r <- xml2::xml_children(s)
      if (length(r) > 0) {
         # cat("actions:\n"); print(r)
         for (i in 1:length(r)) action(r[i], states)
      }      
      w$endState()
   }
   
   # trim spaces in both ends
   trim<-function(x)
   {
      sub("[ \t\n\r]*$", "", sub("^[ \t\n\r]*", "", x))
   }
   
   # @param a An action node
   # @param states Number of states at next stage
   action<-function(a, states) {
      if (length(xml2::xml_find_all(a, "proc")) > 0) {  # if subprocess
         w$action(label=xml2::xml_attr(a, "l"), weights=rep(0,ctrW), prob=c(2,0,1))
         process(xml2::xml_child(a))
      } else {    # normal action
         v<-paste("c(", gsub(" +", ",", trim(xml2::xml_text(xml2::xml_child(a, "q")))), ")",sep="")
         v<-eval(parse(text=v))
         d<-paste("c(", gsub(" +", ",", trim(xml2::xml_text(xml2::xml_child(a, "d")))), ")",sep="")
         d<-eval(parse(text=d))
         if (length(d)>1) warning("More than one duration number in the action (see hmp file)! \nOnly one duration for each action is supported in the binary file format. \nUse the first one.", call.=FALSE)
         v<-c(d[1],v)
         type<-xml2::xml_attr(xml2::xml_child(a, "p"), "t")
         pr<-paste("c(", gsub(" +", ",", trim(xml2::xml_text(xml2::xml_child(a, "p")))), ")",sep="")
         pr<-eval(parse(text=pr))
         if (type=="s") {
            idx<-pr[1:length(pr)%%2==1]
            pr<-pr[1:length(pr)%%2==0]
            scp<-rep(1, length(pr))    # set scp to 1 (default)
         }
         if (type=="d") {
            idx<-pr[1]
            pr<-1
            scp<-1
         }
         if (type=="e") {
            idx<-1:length(pr)-1
            scp<-rep(1,length(pr))    # set scp to 1 (default)
         }
         if (isHMDP) {
            for (i in 1:length(idx)) {
               if (idx[i]>=states) {
                  scp[i]<-0
                  idx[i]<-idx[i]-states
               }
            }
         }
         i<-which(pr!=0)
         scp<-scp[i]
         idx<-idx[i]
         pr<-pr[i]
         pr<-as.numeric(rbind(scp,idx,pr))
         w$action(label=xml2::xml_attr(a, "l"), weights=v, prob=pr)
      }
      w$endAction()
   }
   
   ptm <- proc.time()
   ctrW<-0
   doc <- xml2::read_xml(file)
   isHMDP <- xml2::xml_find_num(doc, xpath = "count(.//proc)") > 1   # ordinary MDP or HMDP
   w<-binaryMDPWriter(prefix, getLog = getLog)
   r <- xml2::xml_find_all(doc, "./quantities")
   setWeights(xml2::xml_attr(r, "l"))
   process(xml2::xml_child(doc, "proc"))
   w$closeWriter()
   if (getLog) {
      cat("Converted",file,"to binary format.\n\n")
      print(proc.time() - ptm)
   }
   invisible(NULL)
}

#' Convert a HMDP model stored in binary format to a `hmp` (XML) file.
#' The function simply parse the binary files and create `hmp` files using
#' the [hmpMDPWriter()].
#'
#' @param prefix A character string with the prefix which will be added to the binary files.
#' @param binNames A character vector of length 7 giving the names of the binary files storing the model.
#' @param out The name of the HMP file (e.g. `r.hmp`).
#' @param duration Weight number storing the duration (NULL if none).
#' @param getLog Output log text.
#' 
#' @return NULL (invisible).
#' 
#' @note Note all indexes are starting from zero (C/C++ style).
#' 
#' @seealso [convertHMP2Binary()].
#' @example inst/examples/convert-ex.R
#' @export
convertBinary2HMP<-function(prefix="", binNames=c("stateIdx.bin","stateIdxLbl.bin","actionIdx.bin",
	"actionIdxLbl.bin","actionWeight.bin","actionWeightLbl.bin","transProb.bin"),
	out=paste0(prefix,'converted.hmp'), duration=1, getLog = TRUE) {

	# mat: matrix of state index
	process<-function(mat) {
		#cat("process\n"); print(mat)
		stages<-length(unique(mat[,2]))
		w$process()
			for (i in 1:stages-1) {
				#print(i); print(nrow(mat[mat[,2]==i+1,]))
				stage(mat[mat[,2]==i,,drop=FALSE], statesNext=nrow(mat[mat[,2]==i+1,]))
			}
		w$endProcess()
	}

	stage<-function(mat, statesNext) {
		#cat("stage\n"); print(mat)
		states<-length(unique(mat[,3]))
		w$stage()
			for (i in 1:states-1) {
				#print(i); print(nrow(mat[mat[,3]==i+1,]))
				state(mat[mat[,3]==i,,drop=FALSE], statesNext=statesNext)
			}
		w$endStage()
	}

	state<-function(mat, statesNext) {
		#cat("state\n"); print(mat)
		#level1<-sum(!is.na(mat[1,2:ncol(mat)])) %/% 3    # level of the first state in mat
		matA<-aIdx[aIdx[,2]==mat[1,1],,drop=FALSE]  # actions to the first state in mat
		if (nrow(mat)>1) {
			aIdx<-unique(mat[!is.na(mat[,4]),4])   # actions that define child processes
			aCtr<-1
			#print(aIdx)
		}
		w$state(label=sLabels[sLabels[,1]==mat[1,1],2]) # create state in hmp
			if (nrow(matA)>0) {
				for (i in 1:nrow(matA)) {   # scan actions
					#matSA<-mat[2:nrow(mat),]
					#cat("action\n"); print(matA[i,])
					scp<-matA[i,3:ncol(matA)]
					scp<-idx<-scp[!is.na(scp)]
					scp<-scp[1:length(scp)%%2==1]
					idx<-idx[1:length(idx)%%2==0]
					weights<-aW[aW[,1]==matA[i,1],1:wLth+1]
					#print(weights)
					if (any(scp==2)) { # new process
						if (length(scp)>1)
							stop("Only a deterministic transition to sub process allowed for action (aId) ",matA[i,1],"!")
						if (idx[1]!=0)
							stop("Only a deterministic transition to state 0 in sub process allowed for action (aId)",matA[i,1],"!")
						if (any(weights!=0))
							stop("Only zero weights allowed for transition to sub process, action (aId)",matA[i,1],"!")
						w$action(label=aLabels[aLabels[,1]==matA[i,1],2], weights=c(0,0,0), prob=c(2,0,1))
							process(mat[mat[,4]==aIdx[aCtr] &  !is.na(mat[,4]), c(1,5:ncol(mat))])
							aCtr<-aCtr+1
						w$endAction()
					} else {
						pr<-prMat[prMat[,1]==matA[i,1],2:ncol(prMat)]
						pr<-pr[!is.na(pr)]
						pr<-as.numeric(rbind(scp,idx,pr))
						#print(pr); print(statesNext)
						w$action(label=aLabels[aLabels[,1]==matA[i,1],2], weights=weights, prob=pr, statesNext=statesNext)
						w$endAction()
					}
				}
			}
		w$endState()
	}

	ptm <- proc.time()
	sIdx<-stateIdxMat(prefix,binNames[1])
	sLabels<-stateIdxDf(prefix,binNames[1])
	sLabels<-sLabels[,c(1,ncol(sLabels))]
	aIdx<-actionIdxMat(prefix,binNames[3])
	aLabels<-actionIdxDf(prefix,binNames[3])
	aLabels<-aLabels[,c(1,ncol(aLabels))]
	aW<-actionWeightMat(prefix,binNames[5],binNames[6])
	prMat<-transProbMat(prefix,binNames[7])
	wNames<-weightNames(prefix,binNames[6])
	wLth<-length(wNames)
	# level<-sum(!is.na(sIdx[i,2:cols])) %/% 3
	w<-hmpMDPWriter(file=out, desc="HMP file created by converting binary files", getLog = getLog)
		w$setWeights(wNames, duration)
		process(sIdx)
	w$closeWriter()
	if (getLog) {
	   cat("Converted binary files to hmp format.\n")
	   print(proc.time() - ptm)
	}
	invisible(NULL)
}








#' Info about the states in the HMDP model under consideration.
#'
#' @param prefix A character string with the prefix added to til file(s).
#' @param file The HMDP binary file containing the description under consideration.
#'
#' @return A matrix with columns `(sId, n0, s0, a0, ...)` where
#' `sId` is the state row id, `n0` the index of the stage at level 0, `s0` the index
#' of the state and `a0` the index of the action. If the HMDP has more
#' than one level columns index `(d1, s1, a1, ...)` are added.
#' @keywords internal
stateIdxMat<-function(prefix="", file="stateIdx.bin") {
   file<-paste(prefix,file,sep="")
   tmp<-readBin(file, integer(),n=file.info(file)$size/4)
   rows<-length(tmp[tmp==-1])
   cols<-max(rle(tmp!=-1)$length)
   mat<-matrix(NA,nrow=rows,ncol=cols+1)
   idx<-c(0,which(tmp== -1))
   for (i in 1:(length(idx)-1)) mat[i,1:(idx[i+1]-idx[i]-1)+1]<-tmp[(idx[i]+1):(idx[i+1]-1)]
   levels<-cols %/% 3 + 1
   if (levels==1) colnames(mat)<-c("sId",paste(c("n","s"),levels-1,sep=""))
   if (levels>1) colnames(mat)<-c("sId",paste(c("n","s","a"),rep(0:(levels-2),each=3),sep=""),paste(c("n","s"),levels-1,sep=""))
   mat[,1]<-1:nrow(mat)-1
   return(mat)
}


#' Info about the states in the HMDP model under consideration.
#'
#' @param prefix A character string with the prefix added to the file(s).
#' @param file The HMDP binary file containing the description under consideration.
#' @param labels The HMDP binary file containing the labels under consideration.
#'
#' @return A data frame with the same columns as in
#' `stateIdxMat` plus another column containing the labels.
#' @keywords internal
stateIdxDf<-function(prefix="", file="stateIdx.bin", labels="stateIdxLbl.bin") {
   labels<-paste(prefix,labels,sep="")
   mat<-stateIdxMat(prefix, file)
   tmp<-readBin(labels, character(),n=file.info(labels)$size)
   tmp<-as.data.frame(matrix(tmp,ncol=2,byrow=TRUE),stringsAsFactors = FALSE)
   colnames(tmp)<-c("sId","label")
   mat<-merge(mat,tmp,all.x=TRUE)
   return(mat)
}


#' Info about the transition probabilities in the HMDP model under consideration.
#'
#' @param prefix A character string with the prefix added to til file(s).
#' @param file The HMDP binary file containing the description under consideration.
#'
#' @return A matrix with columns (`aId`, ...) where
#' `aId` is the action row id and ... are the probabilities of the action.
#' @keywords internal
transProbMat<-function(prefix="", file="transProb.bin") {
   file<-paste(prefix,file,sep="")
   tmp<-readBin(file, numeric(),n=file.info(file)$size/8)
   rows<-length(tmp[tmp==-1])
   cols<-max(rle(tmp!=-1)$length)
   mat<-matrix(NA,nrow=rows,ncol=cols+1)
   idx<-c(0,which(tmp== -1))
   for (i in 1:(length(idx)-1)) mat[i,1:(idx[i+1]-idx[i]-1)+1]<-tmp[(idx[i]+1):(idx[i+1]-1)]
   colnames(mat)<-c("aId",paste("pr",1:(ncol(mat)-1)-1,sep=""))
   mat[,1]<-1:nrow(mat)-1
   return(mat)
}


#' Info about the actions in the HMDP model under consideration.
#'
#' @param prefix A character string with the prefix added to til file(s).
#' @param file The HMDP binary file containing the description under consideration.
#'
#' @return A matrix with columns (`aId`, ...) where
#' `aId` is the action row id and `...` are alternating pairs `(scp, idx)`, one for each
#' possible transition where `scp` is the scope that can be 4 values:
#' 2 - A transition to a child process (stage zero in the child process), 1 - A transition
#' to next stage in the current process, 0 - A transition to the next stage in the father
#' process. the idx in the pair denote the index of the state at the stage considered.
#' Finally, if scope equals 3 then a transition to the state with `sId = idx` is considered.
#' @keywords internal
actionIdxMat<-function(prefix="", file="actionIdx.bin") {
   file<-paste(prefix,file,sep="")
   tmp<-readBin(file, integer(),n=file.info(file)$size/4)
   rows<-length(tmp[tmp==-1])
   cols<-max(rle(tmp!=-1)$length)
   mat<-matrix(NA,nrow=rows,ncol=cols+1)
   idx<-c(0,which(tmp== -1))
   for (i in 1:(length(idx)-1)) mat[i,1:(idx[i+1]-idx[i]-1)+1]<-tmp[(idx[i]+1):(idx[i+1]-1)]
   colnames(mat)<-c("aId","sId",paste(c("scp","idx"),rep(1:((ncol(mat)-2)/2)-1,each=2),sep=""))
   mat[,1]<-1:nrow(mat)-1
   return(mat)
}


#' Info about the actions in the HMDP model under consideration.
#'
#' @param prefix A character string with the prefix added to til file(s).
#' @param file The HMDP binary file containing the description under consideration.
#' @param labels The HMDP binary file containing the labels under consideration.
#'
#' @return A data frame with the same columns as in
#' `actionIdxMat` plus another column containing the labels.
#' @keywords internal
actionIdxDf<-function(prefix="", file="actionIdx.bin", labels="actionIdxLbl.bin") {
   labels<-paste(prefix,labels,sep="")
   mat<-actionIdxMat(prefix, file)
   tmp<-readBin(labels, character(),n=file.info(labels)$size)
   tmp<-as.data.frame(matrix(tmp,ncol=2,byrow=TRUE),stringsAsFactors = FALSE)
   colnames(tmp)<-c("aId","label")
   tmp$aId<-as.numeric(tmp$aId)
   mat<-merge(mat,tmp,all.x=TRUE)
   colnames(mat)<-c("aId","sId",paste(c("scp","idx"),rep(1:((ncol(mat)-2)/2)-1,each=2),sep=""),"label")
   return(mat)
}


#' Info about the actions in the HMDP model under consideration.
#'
#' @param prefix A character string with the prefix added to til file(s).
#' @param file The HMDP binary file containing the description under consideration.
#' @param labels The HMDP binary file containing the labels under consideration.
#' @param weightFile The HMDP binary file containing the action costs.
#' @param transPrFile The HMDP binary file containing the transition probabilities.
#'
#' @return A matrix with columns from `actionIdxMat`,
#' `actionCostMat` and `transProbMat` if labels is NULL. If labels
#' not are NULL then a data frame are returned with a label column too.
#' @keywords internal
actionInfo<-function(prefix="", file="actionIdx.bin" , weightFile="actionWeight.bin", transPrFile="transProb.bin", labels="actionIdxLbl.bin") {
   labels<-paste(prefix,labels,sep="")
   mat<-actionIdxMat(prefix, file)
   mat1<-actionWeightMat(prefix, weightFile)
   mat<-merge(mat,mat1,all.x=TRUE)
   mat2<-transProbMat(prefix, transPrFile)
   mat<-merge(mat,mat2,all.x=TRUE)
   i<-(ncol(mat)-2-ncol(mat1)+1)/3     # number of idx used for (scp, idx, pr) triple
   mat<-mat[,c("aId","sId",colnames(mat1[,2:ncol(mat1),drop=FALSE]),paste(c("scp","idx","pr"),rep(1:i-1,each=3),sep=""))]
   if (!is.null(labels)) {
      tmp<-readBin(labels, character(),n=file.info(labels)$size)
      tmp<-as.data.frame(matrix(tmp,ncol=2,byrow=TRUE))
      colnames(tmp)<-c("aId","label")
      mat<-merge(mat,tmp,all.x=TRUE)
      return(mat)
   }
   mat<-as.matrix(mat)
   return(mat)
}


#' Info about the weights of the actions in the HMDP model under consideration.
#'
#' @param prefix A character string with the prefix added to til file(s).
#' @param file The HMDP binary file containing the description under consideration.
#' @param labels The HMDP binary file containing the labels under consideration.
#'
#' @return A matrix with columns (`aId`, ...) where
#' `aId` is the action row id and ... are the weights of the action.
#' @keywords internal
actionWeightMat<-function(prefix="", file="actionWeight.bin",labels="actionWeightLbl.bin") {
   file<-paste(prefix,file,sep="")
   labels<-paste(prefix,labels,sep="")
   tmp<-readBin(file, numeric(),n=file.info(file)$size/8)
   colNames<-readBin(labels, character(),n=file.info(labels)$size)
   cols<-length(colNames)
   rows<-length(tmp)/cols
   mat<-matrix(NA,nrow=rows,ncol=cols+1)
   for (i in 1:rows) mat[i,1:cols+1]<-tmp[(cols*(i-1)+1):(cols*i)]
   #colnames(mat)<-c("aId",paste("w",1:(ncol(mat)-1)-1,sep=""))
   colnames(mat)<-c("aId",colNames)
   mat[,1]<-1:nrow(mat)-1
   return(mat)
}


#' Names of weights used in actions.
#'
#' @param prefix A character string with the prefix added to the binary file names.
#' @param labels The HMDP binary file containing the weight labels.
#' @return Vector of weight names.
#' @keywords internal
weightNames<-function(prefix="", labels="actionWeightLbl.bin") {
   labels<-paste(prefix,labels,sep="")
   colNames<-readBin(labels, character(),n=file.info(labels)$size)
   return(colNames)
}

# A small HMDP with two levels
w<-binaryMDPWriter("2lev_", getLog = FALSE)
w$setWeights(c("Duration","Net reward","Items"))
w$process()
	w$stage()
		w$state(label="M0")
			w$action(label="A0",weights=c(0,0,0),prob=c(2,0,1))
				w$process()
					w$stage()
						w$state(label="D")
							w$action(label="A0",weights=c(0,0,1),prob=c(1,0,0.5,1,1,0.5))
							w$endAction()
						w$endState()
					w$endStage()
					w$stage()
						w$state(label="C0")
							w$action(label="A0",weights=c(0,0,0),prob=c(1,0,1))
							w$endAction()
							w$action(label="A1",weights=c(1,2,1),prob=c(1,0,0.5,1,1,0.5))
							w$endAction()
						w$endState()
						w$state(label="C1")
							w$action(label="A0",weights=c(0,0,0),prob=c(1,0,1))
							w$endAction()
							w$action(label="A1",weights=c(1,2,1),prob=c(1,0,0.5,1,1,0.5))
							w$endAction()
						w$endState()
					w$endStage()
					w$stage()
						w$state(label="C0")
							w$action(label="A0",weights=c(1,4,0),prob=c(0,0,1))
							w$endAction()
						w$endState()
						w$state(label="C1")
							w$action(label="A0",weights=c(1,4,0),prob=c(0,0,1))
							w$endAction()
						w$endState()
					w$endStage()
				w$endProcess()
			w$endAction()
			w$action(label="A1",weights=c(0,0,0),prob=c(2,0,1))
				w$process()
					w$stage()
						w$state(label="D")
							w$action(label="A0",weights=c(0,0,1),prob=c(1,0,1))
							w$endAction()
						w$endState()
					w$endStage()
					w$stage()
						w$state(label="C0")
							w$action(label="A0",weights=c(0,0,0),prob=c(1,0,1))
							w$endAction()
							w$action(label="A1",weights=c(1,2,1),prob=c(1,0,0.5,1,1,0.5))
							w$endAction()
						w$endState()
					w$endStage()
					w$stage()
						w$state(label="C0")
							w$action(label="A0",weights=c(1,4,0),prob=c(0,0,1))
							w$endAction()
						w$endState()
						w$state(label="C1")
							w$action(label="A0",weights=c(1,4,0),prob=c(0,0,1))
							w$endAction()
							w$action(label="A1",weights=c(0,10,5),prob=c(0,0,0.5,0,1,0.5))
							w$endAction()
						w$endState()
					w$endStage()
				w$endProcess()
			w$endAction()
		w$endState()
		w$state(label="M1")
			w$action(label="A0",weights=c(0,0,0),prob=c(2,0,1))
				w$process()
					w$stage()
						w$state(label="D")
							w$action(label="A0",weights=c(0,0,1),prob=c(1,0,0.5,1,1,0.5))
							w$endAction()
						w$endState()
					w$endStage()
					w$stage()
						w$state(label="C0")
							w$action(label="A0",weights=c(0,0,0),prob=c(1,0,1))
							w$endAction()
						w$endState()
						w$state(label="C1")
							w$action(label="A0",weights=c(0,0,0),prob=c(1,0,1))
							w$endAction()
						w$endState()
					w$endStage()
					w$stage()
						w$state(label="C0")
							w$action(label="A0",weights=c(1,4,0),prob=c(0,0,1))
							w$endAction()
						w$endState()
						w$state(label="C1")
							w$action(label="A0",weights=c(1,4,0),prob=c(0,0,1))
							w$endAction()
						w$endState()
					w$endStage()
				w$endProcess()
			w$endAction()
		w$endState()
	w$endStage()
w$endProcess()
w$closeWriter()

#' Plot parts of the state expanded hypergraph.
#' 
#' The plot is created based on a grid `(rows, cols)`. Each grid point is numbered from bottom to
#' top and left to right (starting from 1), i.e. given grid point with coordinates `(r, c)` (where
#' `(1,1)` is the top left corner and `(rows, cols)` is the bottom right corner) the grid id is `(c
#' - 1) * rows + r`. You must assign a node to the hypergraph to a grid point (see below).
#' 
#' @param hgf A list with the hypergraph containing two data frames, normally found using
#'   [getHypergraph()]. The data frame `nodes` must have columns: `sId` (state id), `gId` (grid id)
#'   and `label` (node label). The data frame `hyperarcs` must have columns `sId` (head node),
#'   `trans` (a list-column of tail node ids), `pr` (a list-column of transition
#'   probabilities), `actionWeights` (a list-column of action weights),
#'   `transWeights` (a list-column of transition-by-weight matrices), `aIdx`
#'   (action index), `label` (action label), `lwd` (hyperarc line width), `lty`
#'   (hyperarc line type) and `col` (hyperarc color).
#' @param gridDim A 2-dim vector (rows, cols) representing the size of the grid.
#' @param showGrid If true show the grid points (good for debugging).
#' @param radx Horizontal radius of the box.
#' @param rady Vertical radius of the box.
#' @param cex Relative size of text.
#' @param marX Horizontal margin.
#' @param marY Vertical margin.
#' @param drawBorder If `TRUE`, draw a border around the plot region and report the
#'   outside and inside padding (good for debugging).
#' @param actionOffset Distance used to separate actions with the same
#'   start and trans states. Set to `0` to draw overlapping actions.
#' @param transLabels Transition-label mode. `"none"` draws no transition labels
#'   (the default); `"custom"` draws values from an optional `transLabels`
#'   list-column in `hgf$hyperarcs`; otherwise use a `|`-separated
#'   combination of `"label"`, `"sId"`, `"prob"`, and `"weights"`, for example
#'   `"prob|weights"`. The older `"state"` spelling is treated as `"label"`.
#' @param transLabelCex Relative size of transition-label text.
#' @param transLabelAdj Position adjustment passed to `textempty()` for transition
#'   labels, drawn at the middle of each split-to-transition branch.
#' @param stateLabel What to plot in states. `"custom"` uses a `stateLabel`
#'   column in `hgf$nodes`; otherwise use a `|`-separated combination of
#'   `"label"` (state label, default), `"sId"` (state id), `"sIdx"`
#'   (stage-based state index), and `"weight"` (optimal weight of the state).
#' @param actionLabel What to plot near the split. One of `"none"`, `"custom"`
#'   (uses an `actionLabel` column in `hgf$hyperarcs`), or a `|`-separated
#'   combination of `"label"` (action label, default) and `"aIdx"`.
#' @param actionWLabel What to plot from the start state to the split. One of
#'   `"none"` (default), `"weight"`, or `"custom"` (uses an `actionWLabel`
#'   column in `hgf$hyperarcs`).
#' @param actionColor Action coloring scheme. Default `""` uses black lines. `"label"` uses different colors based on the action labels. `"policy"` highlights the current policy.
#' @param actionsVisible Action visibility mode. `"all"` (default) shows all actions. `"policy"` only shows actions in the current policy.
#' @param connectedTo Optional vector of state ids. If supplied, plot only states
#'   reachable from these states by following visible hyperarcs forward,
#'   and trim hyperarcs and transition-level data to the remaining states.
#' @param recalcGrid If `TRUE` and `connectedTo` is supplied, recalculate the
#'   grid for the visible nodes. Nodes keep their original columns, but visible
#'   nodes within each column are placed consecutively from the top and the
#'   number of grid rows is reduced to the maximum number of visible nodes in
#'   any column.
#' @param mdp The MDP model. Required if `stateLabel` contains `"weight"`,
#'   `actionColor = "policy"`, or `actionsVisible = "policy"`.
#' @param ... Graphical parameters passed to `textempty`. 
#'   
#' @return No return value (NULL invisible), called for side effects (plotting).
#' @seealso [getHypergraph()] and [plot.HMDP()].
#' @example inst/examples/plot-ex.R
#' @import diagram
#' @export
plotHypergraph <-
   function(hgf,
            gridDim,
            showGrid = FALSE,
            radx = 0.03,
            rady = 0.05,
            cex = 1,
            marX = 0.035,
            marY = 0.15,
            drawBorder = FALSE,
            actionOffset = 0.025,
            transLabels = "none",
            transLabelCex = 0.8 * cex,
            transLabelAdj = c(0.5, -0.6),
            stateLabel = "label",
            actionLabel = "label",
            actionWLabel = "none",
            actionColor = c("", "label", "policy"),
            actionsVisible = c("all", "policy"),
            connectedTo = NULL,
            recalcGrid = FALSE,
            mdp = NULL,
            ...) {
   normalizeLabelArg <- function(x, default, arg) {
      if (missing(x) || is.null(x) || identical(x, "")) return(default)
      if (!is.character(x) || length(x) != 1) {
         stop(paste0(arg, " must be a single character string."), call. = FALSE)
      }
      x
   }
   parseLabelSpec <- function(x, allowed, arg, special = character(0), aliases = character(0)) {
      if (x %in% special) return(x)
      tokens <- trimws(strsplit(x, "|", fixed = TRUE)[[1]])
      tokens <- tokens[tokens != ""]
      tokens <- dplyr::recode(tokens, !!!as.list(aliases), .default = tokens)
      invalid <- setdiff(tokens, allowed)
      if (length(tokens) == 0 || length(invalid) > 0) {
         stop(
            paste0(arg, " must use ", paste(c(special, allowed), collapse = ", "), "."),
            call. = FALSE
         )
      }
      tokens
   }
   collapseLabelParts <- function(parts) {
      parts <- as.character(parts)
      parts <- parts[!is.na(parts) & parts != ""]
      if (length(parts) == 0) return(NA_character_)
      paste(parts, collapse = " | ")
   }
   formatNumericLabel <- function(x) {
      as.character(round(as.numeric(x), 2))
   }
   formatWeightVector <- function(w) {
      w <- w[!is.na(w)]
      if (length(w) == 0) {
         NA_character_
      } else if (length(w) == 1) {
         formatNumericLabel(w)
      } else {
         paste0("(", paste(formatNumericLabel(w), collapse = ", "), ")")
      }
   }

   transLabels <- normalizeLabelArg(transLabels, "none", "transLabels")
   stateLabel <- normalizeLabelArg(stateLabel, "label", "stateLabel")
   actionLabel <- normalizeLabelArg(actionLabel, "label", "actionLabel")
   actionWLabel <- normalizeLabelArg(actionWLabel, "none", "actionWLabel")
   transLabelSpec <- parseLabelSpec(
      transLabels,
      c("label", "sId", "prob", "weights"),
      "transLabels",
      special = c("none", "custom"),
      aliases = c(state = "label")
   )
   stateLabelSpec <- parseLabelSpec(
      stateLabel,
      c("label", "sId", "sIdx", "weight"),
      "stateLabel",
      special = "custom"
   )
   actionLabelSpec <- parseLabelSpec(
      actionLabel,
      c("label", "aIdx"),
      "actionLabel",
      special = c("none", "custom")
   )
   actionWLabelSpec <- parseLabelSpec(
      actionWLabel,
      "weight",
      "actionWLabel",
      special = c("none", "custom")
   )
   actionColor <- match.arg(actionColor)
   actionsVisible <- match.arg(actionsVisible)

   # Apply actionsVisible and actionColor logic to hgf$hyperarcs
   if (!is.null(hgf$hyperarcs)) {
      if (actionsVisible == "policy") {
         if (is.null(mdp)) {
            stop("mdp model must be provided to plotHypergraph when actionsVisible = \"policy\".", call. = FALSE)
         }
         policy_arcs <- getPolicy(mdp) %>% dplyr::select("sId", "aIdx") %>% dplyr::mutate(is_policy = TRUE)
         hgf$hyperarcs <- hgf$hyperarcs %>% 
            dplyr::inner_join(policy_arcs, by = c("sId", "aIdx"))
      }
      if (actionColor == "label") {
         colDF <- tibble::tibble(label = unique(hgf$hyperarcs$label), 
                                 col = grDevices::rainbow(length(unique(hgf$hyperarcs$label))))
         hgf$hyperarcs <- hgf$hyperarcs %>% 
            dplyr::select(-col) %>% 
            dplyr::left_join(colDF, by = "label") 
      }
      if (actionColor == "policy") {
         if (is.null(mdp)) {
            stop("mdp model must be provided to plotHypergraph when actionColor = \"policy\".", call. = FALSE)
         }
         policy_arcs <- getPolicy(mdp) %>% dplyr::select("sId", "aIdx") %>% dplyr::mutate(is_policy = TRUE)
         hgf$hyperarcs <- hgf$hyperarcs %>% 
            dplyr::left_join(policy_arcs, by = c("sId", "aIdx")) %>% 
            dplyr::mutate(col = ifelse(is.na(.data$is_policy), "black", "blue")) %>%
            dplyr::select(-"is_policy")
      }
   }

   filterConnectedHypergraph <- function(hgf, connectedTo) {
      if (is.null(connectedTo)) return(hgf)
      if (!is.numeric(connectedTo)) {
         stop("connectedTo must be a numeric vector of state ids.", call. = FALSE)
      }
      connectedTo <- unique(stats::na.omit(as.numeric(connectedTo)))
      if (length(connectedTo) == 0) return(hgf)
      if (is.null(hgf$nodes) || !"sId" %in% names(hgf$nodes)) {
         stop("connectedTo requires hgf$nodes to contain an sId column.", call. = FALSE)
      }
      missingSId <- setdiff(connectedTo, hgf$nodes$sId)
      if (length(missingSId) > 0) {
         stop(
            paste0("connectedTo contains sId values not present in hgf$nodes: ", paste(missingSId, collapse = ", ")),
            call. = FALSE
         )
      }
      if (is.null(hgf$hyperarcs) || nrow(hgf$hyperarcs) == 0) {
         hgf$nodes <- hgf$nodes[hgf$nodes$sId %in% connectedTo, , drop = FALSE]
         return(hgf)
      }
      if (!"trans" %in% names(hgf$hyperarcs)) {
         stop("connectedTo requires hgf$hyperarcs to contain a trans list-column.", call. = FALSE)
      }

      forwardEdges <- lapply(hgf$nodes$sId, function(x) numeric(0))
      names(forwardEdges) <- as.character(hgf$nodes$sId)
      for (i in seq_len(nrow(hgf$hyperarcs))) {
         trans <- as.numeric(hgf$hyperarcs$trans[[i]])
         trans <- trans[!is.na(trans)]
         key <- as.character(hgf$hyperarcs$sId[i])
         if (!is.null(forwardEdges[[key]])) {
            forwardEdges[[key]] <- unique(c(forwardEdges[[key]], trans))
         }
      }

      reachable <- connectedTo
      queue <- connectedTo
      while (length(queue) > 0) {
         current <- queue[1]
         queue <- queue[-1]
         nextStates <- forwardEdges[[as.character(current)]]
         nextStates <- nextStates[nextStates %in% hgf$nodes$sId]
         nextStates <- nextStates[!nextStates %in% reachable]
         if (length(nextStates) > 0) {
            reachable <- c(reachable, nextStates)
            queue <- c(queue, nextStates)
         }
      }
      reachable <- unique(reachable)

      keepArc <- logical(nrow(hgf$hyperarcs))
      transKeep <- vector("list", nrow(hgf$hyperarcs))
      for (i in seq_len(nrow(hgf$hyperarcs))) {
         trans <- as.numeric(hgf$hyperarcs$trans[[i]])
         transKeep[[i]] <- !is.na(trans) & trans %in% reachable
         keepArc[i] <- hgf$hyperarcs$sId[i] %in% reachable && any(transKeep[[i]])
      }
      hgf$hyperarcs <- hgf$hyperarcs[keepArc, , drop = FALSE]
      transKeep <- transKeep[keepArc]

      for (i in seq_len(nrow(hgf$hyperarcs))) {
         keep <- transKeep[[i]]
         hgf$hyperarcs$trans[[i]] <- hgf$hyperarcs$trans[[i]][keep]
         if ("pr" %in% names(hgf$hyperarcs) && length(hgf$hyperarcs$pr[[i]]) == length(keep)) {
            hgf$hyperarcs$pr[[i]] <- hgf$hyperarcs$pr[[i]][keep]
         }
         if ("transLabels" %in% names(hgf$hyperarcs) && length(hgf$hyperarcs$transLabels[[i]]) == length(keep)) {
            hgf$hyperarcs$transLabels[[i]] <- hgf$hyperarcs$transLabels[[i]][keep]
         }
         if ("transWeights" %in% names(hgf$hyperarcs) && is.matrix(hgf$hyperarcs$transWeights[[i]]) &&
             nrow(hgf$hyperarcs$transWeights[[i]]) == length(keep)) {
            hgf$hyperarcs$transWeights[[i]] <- hgf$hyperarcs$transWeights[[i]][keep, , drop = FALSE]
         }
      }
      hgf$nodes <- hgf$nodes[hgf$nodes$sId %in% reachable, , drop = FALSE]
      return(hgf)
   }

   hgf <- filterConnectedHypergraph(hgf, connectedTo)

   recalculateGrid <- function(hgf, gridDim) {
      if (is.null(hgf$nodes) || nrow(hgf$nodes) == 0) return(list(hgf = hgf, gridDim = gridDim))
      if (!"gId" %in% names(hgf$nodes)) {
         stop("recalcGrid requires hgf$nodes to contain a gId column.", call. = FALSE)
      }
      originalCols <- (hgf$nodes$gId - 1) %/% gridDim[1] + 1
      originalRows <- (hgf$nodes$gId - 1) %% gridDim[1] + 1
      if (any(is.na(originalCols) | originalCols < 1 | originalCols > gridDim[2])) {
         stop("recalcGrid requires node gId values to be inside gridDim.", call. = FALSE)
      }
      visibleCounts <- tabulate(originalCols, nbins = gridDim[2])
      rowsNew <- max(visibleCounts)
      if (rowsNew == 0) return(list(hgf = hgf, gridDim = c(1, gridDim[2])))
      newRows <- integer(length(originalRows))
      for (col in seq_len(gridDim[2])) {
         idx <- which(originalCols == col)
         if (length(idx) == 0) next
         idx <- idx[order(originalRows[idx], hgf$nodes$sId[idx])]
         newRows[idx] <- seq_along(idx)
      }
      hgf$nodes$gId <- (originalCols - 1) * rowsNew + newRows
      list(hgf = hgf, gridDim = c(rowsNew, gridDim[2]))
   }

   if (!is.null(connectedTo) && isTRUE(recalcGrid)) {
      gridRecalc <- recalculateGrid(hgf, gridDim)
      hgf <- gridRecalc$hgf
      gridDim <- gridRecalc$gridDim
   }

   # Apply stateLabel logic to hgf$nodes$label
   if (!is.null(hgf$nodes)) {
      if (identical(stateLabelSpec, "custom")) {
         if (!"stateLabel" %in% names(hgf$nodes)) {
            stop(
               "stateLabel = \"custom\" requires a stateLabel column in hgf$nodes.",
               call. = FALSE
            )
         }
         hgf$nodes$label <- as.character(hgf$nodes$stateLabel)
      } else {
         stateLabels <- as.character(hgf$nodes$label)
         stateSIdx <- NULL
         stateWeights <- NULL
         if ("sIdx" %in% stateLabelSpec) {
            if (!"stateStr" %in% names(hgf$nodes)) {
               stop("stateLabel containing \"sIdx\" requires a stateStr column in hgf$nodes.", call. = FALSE)
            }
            parts <- strsplit(hgf$nodes$stateStr, ",")
            stateSIdx <- vapply(parts, function(x) {
               if (length(x) > 0) x[length(x)] else ""
            }, character(1))
         }
         if ("weight" %in% stateLabelSpec) {
            if (is.null(mdp)) {
               stop(
                  "mdp model must be provided to plotHypergraph when stateLabel contains \"weight\".",
                  call. = FALSE
               )
            }
            policy_weights <- getPolicy(mdp) %>% dplyr::select("sId", "weight")
            hgf$nodes <- hgf$nodes %>% dplyr::left_join(policy_weights, by = "sId")
            stateWeights <- formatNumericLabel(hgf$nodes$weight)
         }
         hgf$nodes$label <- vapply(seq_len(nrow(hgf$nodes)), function(i) {
            collapseLabelParts(vapply(stateLabelSpec, function(part) {
               switch(
                  part,
                  label = stateLabels[i],
                  sId = as.character(hgf$nodes$sId[i]),
                  sIdx = stateSIdx[i],
                  weight = stateWeights[i],
                  NA_character_
               )
            }, character(1)))
         }, character(1))
         hgf$nodes <- hgf$nodes %>% dplyr::select(-dplyr::any_of("weight"))
      }
   }

   # Apply actionLabel and actionWLabel logic to hgf$hyperarcs
   if (!is.null(hgf$hyperarcs)) {
      orig_label <- hgf$hyperarcs$label
      if (identical(actionLabelSpec, "none")) {
         hgf$hyperarcs$label <- NA_character_
      } else if (identical(actionLabelSpec, "custom")) {
         if (!"actionLabel" %in% names(hgf$hyperarcs)) {
            stop(
               "actionLabel = \"custom\" requires an actionLabel column in hgf$hyperarcs.",
               call. = FALSE
            )
         }
         hgf$hyperarcs$label <- as.character(hgf$hyperarcs$actionLabel)
      } else {
         hgf$hyperarcs$label <- vapply(seq_len(nrow(hgf$hyperarcs)), function(i) {
            collapseLabelParts(vapply(actionLabelSpec, function(part) {
               switch(
                  part,
                  label = as.character(orig_label[i]),
                  aIdx = as.character(hgf$hyperarcs$aIdx[i]),
                  NA_character_
               )
            }, character(1)))
         }, character(1))
      }

      if (identical(actionWLabelSpec, "weight")) {
         hgf$hyperarcs$actionWLabel <- purrr::map_chr(hgf$hyperarcs$actionWeights, function(w) {
            formatWeightVector(w)
         })
      } else if (identical(actionWLabelSpec, "custom")) {
         if (!"actionWLabel" %in% names(hgf$hyperarcs)) {
            stop(
               "actionWLabel = \"custom\" requires an actionWLabel column in hgf$hyperarcs.",
               call. = FALSE
            )
         }
         hgf$hyperarcs$actionWLabel <- as.character(hgf$hyperarcs$actionWLabel)
      } else {
         hgf$hyperarcs$actionWLabel <- NA_character_
      }
   }

   # internal functions
   gMap<-function(sId) return(hgf$nodes$gId[hgf$nodes$sId %in% sId])		# return gId given sId
   sMap<-function(gId) return(hgf$nodes$sId[hgf$nodes$gId %in% gId])		# return sId given gId
   ellipseBoundaryPoint <- function(mid, toward) {
      direction <- toward - mid
      if (all(direction == 0)) return(mid)
      scale <- 1 / sqrt((direction[1] / radx)^2 + (direction[2] / rady)^2)
      mid + scale * direction
   }
   ellipseBoundaryPoints <- function(mid, toward) {
      t(vapply(seq_len(nrow(mid)), function(i) {
         ellipseBoundaryPoint(mid[i, ], toward)
      }, numeric(2)))
   }
   pos <- coordinates(rep(gridDim[2], gridDim[1]), hor = TRUE)  # coordinates of each point in the grid
   
   # reposition
   posN <- pos
   for (i in 1:nrow(pos)) {
      c <- (i-1) %% gridDim[2] + 1
      r <- (i-1) %/% gridDim[2] + 1
      id <- (c - 1) * gridDim[1] + r
      # cat(i, r, c, id, "\n", sep= " ")
      posN[id, ] <- pos[i, ]
   }
   pos <- posN

   xlim <- c(min(pos[,1])-marX,max(pos[,1])+marX)
   ylim <- c(min(pos[,2])-marY,max(pos[,2])+marY)
   openplotmat(xlim = xlim, ylim = ylim)  #main = "State expanded hypergraph"
   if (drawBorder) {
      outsidePadding <- stats::setNames(graphics::par("mai"), c("bottom", "left", "top", "right"))
      insidePadding <- c(
         bottom = min(pos[, 2])-ylim[1],
         left = min(pos[, 1])-xlim[1],
         top = ylim[2]-max(pos[, 2]),
         right = xlim[2]-max(pos[, 1])
      )
      message(
         "plotHypergraph padding: ",
         "outside figure margin in inches ",
         paste(names(outsidePadding), round(outsidePadding, 3), sep = "=", collapse = ", "),
         "; inside plot margin in user coordinates ",
         paste(names(insidePadding), round(insidePadding, 3), sep = "=", collapse = ", "),
         ". To remove outside spacing use par(mai = c(0, 0, 0, 0)); ",
         "to remove inside spacing use marX = 0 and marY = 0."
      )
   }
   
   # plot time index
   # if (addTime) {
   #    posT <- matrix(c(unique(pos[,1]), rep(0, gridDim[2])), ncol = 2)  # coordinates for time index
   #    colnames(posT) <- colnames(pos)
   #    for (i in 1:gridDim[2] - 1) textempty(posT[i+1, ], lab = parse(text = str_c("italic(t == ", i, ")")), cex=cex)
   # }
   

   # plot actions
   if (!is.null(hgf$hyperarcs)) {
      if (!"trans" %in% names(hgf$hyperarcs)) {
         stop("hgf$hyperarcs must contain a trans list-column.", call. = FALSE)
      }
      if (!identical(transLabelSpec, "none") && !identical(transLabelSpec, "custom") &&
          "prob" %in% transLabelSpec && !"pr" %in% names(hgf$hyperarcs)) {
         stop(
            "transLabels containing \"prob\" requires a pr list-column in hgf$hyperarcs.",
            call. = FALSE
         )
      }
      if (!identical(transLabelSpec, "none") && !identical(transLabelSpec, "custom") &&
          "weights" %in% transLabelSpec && !"transWeights" %in% names(hgf$hyperarcs)) {
         stop(
            "transLabels containing \"weights\" requires a transWeights list-column in hgf$hyperarcs.",
            call. = FALSE
         )
      }
      if (identical(transLabelSpec, "custom") && !"transLabels" %in% names(hgf$hyperarcs)) {
         stop(
            "transLabels = \"custom\" requires a transLabels list-column in hgf$hyperarcs.",
            call. = FALSE
         )
      }

      # Pre-populate transLabels based on transLabels option
      if (!identical(transLabelSpec, "none") && !identical(transLabelSpec, "custom")) {
         hgf$hyperarcs$transLabels <- purrr::pmap(
            list(
               trans = hgf$hyperarcs$trans,
               pr = if ("pr" %in% names(hgf$hyperarcs)) hgf$hyperarcs$pr else rep(list(NULL), nrow(hgf$hyperarcs)),
               transWeights = if ("transWeights" %in% names(hgf$hyperarcs)) hgf$hyperarcs$transWeights else rep(list(NULL), nrow(hgf$hyperarcs))
            ),
            function(trans, pr, transWeights) {
               trans <- as.numeric(trans)
               nTrans <- length(trans)
               weightLabels <- NULL
               if ("weights" %in% transLabelSpec) {
                  if (!is.matrix(transWeights)) {
                     stop(
                        "transLabels containing \"weights\" requires a transition-by-weight matrix in each transWeights row.",
                        call. = FALSE
                     )
                  }
                  if (ncol(transWeights) == 0) {
                     weightLabels <- rep(NA_character_, nrow(transWeights))
                  } else {
                     weightLabels <- apply(transWeights, 1, formatWeightVector)
                  }
               }
               vapply(seq_len(nTrans), function(j) {
                  collapseLabelParts(vapply(transLabelSpec, function(part) {
                     switch(
                        part,
                        label = hgf$nodes$label[match(trans[j], hgf$nodes$sId)],
                        sId = as.character(trans[j]),
                        prob = formatNumericLabel(pr[j]),
                        weights = weightLabels[j],
                        NA_character_
                     )
                  }, character(1)))
               }, character(1))
            }
         )
      }

      actionKeys <- vapply(
         seq_len(nrow(hgf$hyperarcs)),
         function(i) {
            trans <- as.numeric(hgf$hyperarcs$trans[[i]])
            trans <- trans[!is.na(trans)]
            paste(hgf$hyperarcs$sId[i], paste(trans, collapse = ","), sep = ":")
         },
         character(1)
      )
      actionOffsets <- stats::ave(
         seq_along(actionKeys),
         actionKeys,
         FUN = function(x) seq_along(x) - (length(x) + 1) / 2
      )
      for (i in seq_len(nrow(hgf$hyperarcs))) {
         trans <- as.numeric(hgf$hyperarcs$trans[[i]])
         validTrans <- !is.na(trans)
         trans <- trans[validTrans]
         if (length(trans) == 0) next
         fromPos <- pos[gMap(trans), , drop = FALSE]
         toPos <- pos[gMap(hgf$hyperarcs$sId[i]), ]
         centre <- NULL
         if (actionOffsets[i] != 0 && actionOffset != 0) {
            baseCentre <- colMeans(fromPos) + 0.5 * (toPos - colMeans(fromPos))
            direction <- toPos - colMeans(fromPos)
            directionLength <- sqrt(sum(direction^2))
            if (directionLength > 0) {
               perpendicular <- c(-direction[2], direction[1]) / directionLength
               centre <- baseCentre + actionOffsets[i] * actionOffset * perpendicular
            }
         }
         splitCentre <- if (is.null(centre)) {
            colMeans(fromPos) + 0.5 * (toPos - colMeans(fromPos))
         } else {
            centre
         }
         fromBoundary <- ellipseBoundaryPoints(fromPos, splitCentre)
         toBoundary <- ellipseBoundaryPoint(toPos, splitCentre)
         #cat("i:",i,"highlight:",hgf$hyperarcs$highlight[i],"\n")
         # if (hgf$hyperarcs$highlight[i]) splitarrow(from = pos[gMap(trans), ], to = pos[gMap(hgf$hyperarcs[i,1]),], arr.side = 2, arr.pos = 0.1, lwd=2, lty=1,
         #                                      arr.type="curved", arr.lwd = 0.5, arr.length = 0.1, arr.width = 0.08, lcol="gray")
         pt <-
            splitarrow(
               from = fromBoundary,
               to = toBoundary,
               centre = centre,
               arr.side = 2,
               arr.pos = 0.1,
               lwd = hgf$hyperarcs$lwd[i],
               lty = hgf$hyperarcs$lty[i],
               arr.type = "curved",
               arr.lwd = 0.5,
               arr.length = 0.1,
               arr.width = 0.08,
               lcol = hgf$hyperarcs$col[i]
            )
         textempty(
            pt,
            lab = hgf$hyperarcs$label[i],
            adj = c(-0.1, 0.1),
            cex = cex,
            ...
         )
         if ("actionWLabel" %in% names(hgf$hyperarcs) && !is.na(hgf$hyperarcs$actionWLabel[i]) && hgf$hyperarcs$actionWLabel[i] != "") {
            textempty(
               (toBoundary + pt) / 2,
               lab = hgf$hyperarcs$actionWLabel[i],
               adj = c(0.5, -0.6),
               cex = cex,
               ...
            )
         }
         if (!identical(transLabelSpec, "none")) {
            labs <- hgf$hyperarcs$transLabels[[i]][validTrans]
            if (length(labs) != length(trans)) {
               stop(
                  paste0("transLabels = \"", transLabels, "\" requires one label per plotted transition."),
                  call. = FALSE
               )
            }
            for (j in seq_along(trans)) {
               stateIndex <- match(trans[j], hgf$nodes$sId)
               if (is.na(stateIndex) || is.na(labs[j])) next
               transPos <- pos[hgf$nodes$gId[stateIndex], ]
               transBoundary <- ellipseBoundaryPoint(transPos, splitCentre)
               labelPos <- (pt + transBoundary) / 2
               textempty(
                  labelPos,
                  lab = labs[j],
                  adj = transLabelAdj,
                  cex = transLabelCex,
                  ...
               )
            }
         }
      }
   }	   
   
   # plot states
   if (!is.null(hgf$nodes)) {
      for (i in seq_len(nrow(hgf$nodes))) { 
         textellipse(pos[hgf$nodes$gId[i], ], lab = hgf$nodes$label[i], radx = radx, rady=rady, shadow.size = 0, lwd=0.5, cex=cex) 
      }
   }
   
   # visual view of the point numbers (for figuring out how to map stateId to gridId)
   if (showGrid) {
      for (i in seq_len(dim(pos)[1])) textrect(pos[i, ], lab = i, radx = 0.0, cex=cex)
   }
   if (drawBorder) {
      graphics::box(which = "plot")
   }
   return(invisible(NULL))
}



#' Plot the state-expanded hypergraph of the MDP. 
#'
#' @param x The MDP model. 
#' @param ... Arguments passed to [plotHypergraph()].
#'
#' @return No return value (NULL invisible), called for side effects (plotting).
#' @seealso [getHypergraph()] and [plotHypergraph()] for possible arguments.
#' @example inst/examples/plot-ex.R
#' @importFrom rlang .data
#' @export
plot.HMDP <- function(x, ...) {
   args <- list(...)
   mdp <- x
   
   if (mdp$levels != 1) {
      message("Cannot plot a hierarchical MDP without specifying the placement of\n", 
              "states and actions (use `plotHypergraph` instead).")
      return(invisible(NULL))
   }
   if (mdp$timeHorizon < Inf) {
      statesCt <- purrr::map_dbl(paste0(1:mdp$timeHorizon - 1), mdp$ptr$getStateSizeStage)
      gridDim <- c(max(statesCt), mdp$timeHorizon)
   } else {
      gridDim <- c(mdp$founderStatesLast, 2)
   }
   hgf <- getHypergraph(mdp)
   hgf$nodes <- hgf$nodes %>% 
      tidyr::separate(.data$stateStr, into = c("c", "r"), remove = FALSE, convert = TRUE) %>% 
      dplyr::mutate(c = .data$c + 1, 
                    r = .data$r + 1, 
                    gId = (.data$c - 1) * max(.data$r) + .data$r) 

   do.call(plotHypergraph, args = c(list(hgf, gridDim, mdp = mdp), args))
   return(invisible(NULL))
}


#' Return the (parts of) state-expanded hypergraph
#' 
#' The function is useful together with [plotHypergraph()].
#'
#' @param mdp The MDP loaded using [loadMDP()].
#' @param ... Arguments passed to [getInfo()].
#'
#' @return A list representing the hypergraph with two elements: a tibble
#'   `nodes` and a tibble `hyperarcs`. `hyperarcs` stores `actionWeights`,
#'   `trans`, and `pr` as list-columns of vectors. `transWeights` is a
#'   list-column of matrices with one row per transition and one column per
#'   transition-weight namespace.
#' @seealso [plotHypergraph()] and [plot.HMDP()].
#' @importFrom rlang .data
#' @export
#' @example inst/examples/plot-ex.R
getHypergraph <- function(mdp, ...) {
   dat <- getInfo(mdp, withList = FALSE, ...)
   nActionWeights <- length(mdp$weightActionNames)
   nTransWeights <- length(mdp$weightTransNames)
   normalizeActionWeights <- function(x) {
      if (nActionWeights == 0) return(numeric(0))
      x <- as.numeric(x)
      if (length(x) == 0 || all(is.na(x))) return(rep(NA_real_, nActionWeights))
      if (length(x) != nActionWeights) {
         stop("Action weights do not match action weight names.", call. = FALSE)
      }
      x
   }
   normalizeTransWeights <- function(x, nTrans) {
      if (nTransWeights == 0) {
         return(matrix(numeric(0), nrow = nTrans, ncol = 0))
      }
      x <- as.numeric(x)
      if (length(x) == 0 || all(is.na(x))) x <- rep(NA_real_, nTrans * nTransWeights)
      if (length(x) != nTrans * nTransWeights) {
         stop("Transition weights do not match transitions and weight names.", call. = FALSE)
      }
      matrix(
         x,
         nrow = nTrans,
         ncol = nTransWeights,
         byrow = TRUE,
         dimnames = list(NULL, mdp$weightTransNames)
      )
   }
   actions <- dat$df %>% 
      tidyr::unnest_longer("actions") %>% # convert actions (one row for each action)
      tidyr::unnest_wider("actions", names_repair = tidyr::tidyr_legacy) %>% 
      dplyr::rename(stateLabel = "label") %>%
      dplyr::filter(!is.na(.data$aIdx)) %>% 
      dplyr::mutate(
         actionWeights = purrr::map(.data$weights, normalizeActionWeights),
         trans = purrr::map(.data$trans, as.numeric),
         pr = purrr::map(.data$pr, as.numeric),
         transWeights = purrr::map2(
            .data$transWeights,
            .data$trans,
            ~ normalizeTransWeights(.x, length(.y))
         ),
         transLabels = purrr::map(.data$trans, ~ rep(NA_character_, length(.x))),
         actionWLabel = NA_character_,
         label = as.character(.data$label1)
      ) %>%
      dplyr::mutate(lwd = 1, lty = 1, col = "black") %>% 
      dplyr::select(-"stateStr", -"stateLabel", -"weights", -"label1")

   states <- dat$df %>% 
      dplyr::mutate(gId = NA_integer_) %>% 
      dplyr::select("sId", "stateStr", "label", "gId")
   if (mdp$timeHorizon == Inf) {
      states$label[1:mdp$founderStatesLast] <- states$label[(nrow(states) - mdp$founderStatesLast + 1):nrow(states)]
   }

   return(list(nodes = states, hyperarcs = actions))
}

## Example 6.1.1 in Tijms, H.C., "A first course in stochastic models", John Wiley & Sons Ltd, 2003.
## Specify the MDP using two different approaches
N<-5; Cf<- -10; Cp<-c(NA,-7,-7,-5) # use negative numbers since optimize reward
Q <- matrix(c(
   0.90, 0.10, 0, 0, 0,
   0, 0.80, 0.10, 0.05, 0.05,
   0, 0, 0.70, 0.10, 0.20,
   0, 0, 0, 0.50, 0.50), nrow=4, byrow=T)


## Using binaryMDPWriter nested style
# note that we index from 0, i.e. state N correspond to index N-1 (in j col). Or index i corresponds to state i+1
transPr0<-data.frame(i=rep(1:(N-1), times=N), scope=1, j=rep(1:N,each=4)-1, pr=as.vector(Q))
transPr0<-transPr0[transPr0$pr>0,]  # remove transitions with zero pr
transPr1<-data.frame(i=2:(N-1), scope=1, j=1-1, pr=rep(1, times=N-2))
transPr2<-data.frame(i=c(N,N+1), scope=1, j=c(N+1,1)-1, pr=c(1,1))
getTransPr<-function(a,i) {
   if (a==0) return(as.vector(as.matrix(t(transPr0[transPr0$i==i,c(2,3,4)]))) )
   if (a==1) return(as.vector(as.matrix(t(transPr1[transPr1$i==i,c(2,3,4)]))))
   if (a==2) return(as.vector(as.matrix(t(transPr2[transPr2$i==i,c(2,3,4)]))))
   return(stop("Error"))
}
# Build the model which is stored in a set of binary files
w<-binaryMDPWriter("hct611v3-1_", getLog = FALSE)
w$setWeights(c("Duration","Net reward"))
w$process()
   w$stage()
      w$state(label="i=1")
         w$action(label="no repair", weights=c(1,0), prob=getTransPr(a=0,i=1), end=T)
      w$endState()
      for (ii in 2:(N-1) ) {
         w$state(label=paste("i=",ii,sep=""))
            w$action(label="no repair", weights=c(1,0), prob=getTransPr(a=0,i=ii), end=T)
            w$action(label="preventive repair", weights=c(1,Cp[ii]), prob=getTransPr(a=1,i=ii), end=T)
         w$endState()
      }
      for (ii in N:(N+1) ) {
         w$state(label=paste("i=",ii,sep=""))
            cost<-ifelse(ii==N, Cf, 0)
            w$action(label="forced repair", weights=c(1,cost), prob=getTransPr(a=2,i=ii), end=T)
         w$endState()
      }
   w$endStage()
w$endProcess()
w$closeWriter()


## Using binaryMDPWriter where the MDP is specified using a list of matrices and a matrix of rewards
# specify transition pr for each action (one 6x6 matrix for each action)
P<-list()
# a=1 (no repair)
Z<-as.matrix(rbind(cbind(Q,0),0,0))
#Z[5,1]<-Z[6,1]<-1
P[[1]]<-Z
# a=2 (preventive repair)
Z <- matrix(0, nrow = N+1, ncol = N+1)
Z[2,1]<-Z[3,1]<-Z[4,1]<-1
P[[2]]<-Z
# a=3 (forced repair)
Z <- matrix(0, nrow = N+1, ncol = N+1)
Z[5,6]<-Z[6,1]<-1
P[[3]]<-Z
# reward 6x3 matrix with one column for each action
R <- matrix(0, nrow = N+1, ncol = 3)
R[2:4,2]<-Cp[2:4]
R[5,3]<-Cf

# make bin files
w<-binaryMDPWriter("hct611v3-2_", getLog = FALSE)
w$setWeights(c("Duration","Net reward"))
w$process(P,R)
w$closeWriter()


## Using binaryMDPWriter where the MDP is specified using a list of sparse matrices and a sparse matrix of rewards
# specify transition pr for each action (one sparse 6x6 matrix for each action)
#library(Matrix)
P<-list()
# a=1 (no repair)
Z<-Matrix::Matrix(rbind(cbind(Q,0),0,0))
#Z[5,1]<-Z[6,1]<-1
P[[1]]<-Z
# a=2 (preventive repair)
Z <- Matrix::Matrix(0, nrow = N+1, ncol = N+1)
Z[2,1]<-Z[3,1]<-Z[4,1]<-1
P[[2]]<-Z
# a=3 (forced repair)
Z <- Matrix::Matrix(0, nrow = N+1, ncol = N+1)
Z[5,6]<-Z[6,1]<-1
P[[3]]<-Z

# reward 6x3 matrix with one column for each action
R <- Matrix::Matrix(0, nrow = N+1, ncol = 3)
R[2:4,2]<-Cp[2:4]
R[5,3]<-Cf

w<-binaryMDPWriter("hct611v3-3_", getLog = FALSE)
w$setWeights(c("Duration","Net reward"))
w$process(P,R)
w$closeWriter()


## Using binaryMDPWriter where the semi-MDP is specified using a list of matrices, a matrix of 
## rewards and a matrix of durations.
# specify transition pr for each action (one 5x5 matrix for each action)
P<-list()
# a=1 (no repair)
Z<-as.matrix(rbind(Q,0))
#Z[5,1]<-Z[6,1]<-1
P[[1]]<-Z
# a=2 (preventive repair)
Z <- matrix(0, nrow = N, ncol = N)
Z[2,1]<-Z[3,1]<-Z[4,1]<-1
P[[2]]<-Z
# a=3 (forced repair)
Z <- matrix(0, nrow = N, ncol = N)
Z[5,1]<-1
P[[3]]<-Z
# reward 5x3 matrix with one column for each action
R <- matrix(NA, nrow = N, ncol = 3)
R[,1]<-c(0,0,0,0,NA)
R[2:4,2]<-Cp[2:4]
R[5,3]<-Cf
# duration 5x3 matrix with one column for each action
D<-matrix(1,nrow = nrow(R), ncol = ncol(R))
D[5,3]<-2
# make bin files
w<-binaryMDPWriter("hct611v3-4_", getLog = FALSE)
w$setWeights(c("Duration","Net reward"))
w$process(P,R,D)
w$closeWriter()

# The example given in L.R. Nielsen and A.R. Kristensen. Finding the K best
# policies in a finite-horizon Markov decision process. European Journal of
# Operational Research, 175(2):1164-1179, 2006. doi:10.1016/j.ejor.2005.06.011,
# does actually not have any dummy replacement node as in the MDP above. The same
# model can be created using a single dummy node at the end of the process.

## Create the MDP using a single dummy node
prefix<-"machine2_"
w <- binaryMDPWriter(prefix, getLog = FALSE)
w$setWeights(c("Net reward"))
w$process()
	w$stage()   # stage n=0
		w$state(label="Dummy")          # v=(0,0)
			w$action(label="buy", weights=-100, prob=c(1,0,0.7, 1,1,0.3), end=TRUE)
		w$endState()
	w$endStage()
	w$stage()   # stage n=1
		w$state(label="good")           # v=(1,0)
			w$action(label="mt", weights=55, prob=c(1,0,1), end=TRUE)
			w$action(label="nmt", weights=70, prob=c(1,0,0.6, 1,1,0.4), end=TRUE)
		w$endState()
		w$state(label="average")        # v=(1,1)
			w$action(label="mt", weights=40, prob=c(1,0,1), end=TRUE)
			w$action(label="nmt", weights=50, prob=c(1,1,0.6, 1,2,0.4), end=TRUE)
		w$endState()
	w$endStage()
	w$stage()   # stage n=2
		w$state(label="good")           # v=(2,0)
			w$action(label="mt", weights=55, prob=c(1,0,1), end=TRUE)
			w$action(label="nmt", weights=70, prob=c(1,0,0.5, 1,1,0.5), end=TRUE)
		w$endState()
		w$state(label="average")        # v=(2,1)
			w$action(label="mt", weights=40, prob=c(1,0,1), end=TRUE)
			w$action(label="nmt", weights=50, prob=c(1,1,0.5, 1,2,0.5), end=TRUE)
		w$endState()
		w$state(label="not working")    # v=(2,2)
			w$action(label="mt", weights=30, prob=c(1,0,1), end=TRUE)
			w$action(label="rep", weights=5, prob=c(3,12,1), end=TRUE)     # transition to the node with sId=12 (Dummy)
		w$endState()
	w$endStage()
	w$stage()   # stage n=3
		w$state(label="good")           # v=(3,0)
			w$action(label="mt", weights=55, prob=c(1,0,1), end=TRUE)
			w$action(label="nmt", weights=70, prob=c(1,0,0.2, 1,1,0.8), end=TRUE)
		w$endState()
		w$state(label="average")        # v=(3,1)
			w$action(label="mt", weights=40, prob=c(1,0,1), end=TRUE)
			w$action(label="nmt", weights=50, prob=c(1,1,0.2, 1,2,0.8), end=TRUE)
		w$endState()
		w$state(label="not working")    # v=(3,2)
			w$action(label="mt", weights=30, prob=c(1,0,1), end=TRUE)
			w$action(label="rep", weights=5, prob=c(3,12,1), end=TRUE)
		w$endState()
	w$endStage()
	w$stage()   # stage n=4
		w$state(label="good")        # v=(4,0)
			w$action(label="rep", weights=30, prob=c(1,0,1), end=TRUE)
		w$endState()
		w$state(label="average")     # v=(4,1)
			w$action(label="rep", weights=10, prob=c(1,0,1), end=TRUE)
		w$endState()
		w$state(label="not working") # v=(4,2)
			w$action(label="rep", weights=5, prob=c(1,0,1), end=TRUE)
		w$endState()
	w$endStage()
	w$stage()   # stage n=5
		w$state(label="Dummy", end=TRUE)        # v=(5,0)
	w$endStage()
w$endProcess()
w$closeWriter()

## Excercise 7.3 in Tijms, H.C., "A first course in stochastic models", John Wiley & Sons Ltd, 2003.
## The semi-MDP is specified using binaryMDPWriter and actions with id and pr

tau<-matrix(c(
   NA, 3, 6, 3, 2,
   4, NA, 1, 7, 5,
   5, 1, NA, 6, 8,
   3, 8, 5, NA, 2,
   2, 5, 9, 2, NA
   ), nrow=5, byrow=T)
xi<-matrix(c(
   NA, 8, 12, 6, 6,
   20, NA, 2, 14, 16,
   16, 2, NA, 18, 16,
   6, 10, 20, NA, 6,
   8, 16, 20, 8, NA
   ), nrow=5, byrow=T)


w<-binaryMDPWriter("hct73v2_", getLog = FALSE)
w$setWeights(c("Duration","Net reward"))
w$process()
   w$stage()
      for (i in 1:5) {
         w$state(label=i)
            for (a in 1:5) {
               if (a!=i) {
                  w$action(label=a, weights=c(tau[i,a], xi[i,a]), id=a-1, pr=1, end=T)
               }
            }
         w$endState()
      }
   w$endStage()
w$endProcess()
w$closeWriter()

---
title: "An infinite-horizon HMDP"
author: "Lars Relund <lars@relund.dk>"
date: "`r Sys.Date()`"
bibliography: litt.bib
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{infinite-hmdp}
  %\VignetteEncoding{UTF-8}
  %\VignetteEngine{knitr::rmarkdown}
editor_options: 
  chunk_output_type: console
---

<style> 
p {text-align: justify;} 
//.sourceCode {background-color: white;}
pre {
  // border-style: solid;
  // border-width: 1px;
  // border-color: grey;
  //background-color: grey !important;
}
img {
   //width: 100%;
   border: 0;
}
</style>

<!-- scale math down -->
<script type="text/x-mathjax-config"> 
    MathJax.Hub.Config({ 
        "HTML-CSS": { scale: 80 }
        });
</script>

```{r setup, include=FALSE}
library(knitr)
options(rmarkdown.html_vignette.check_title = FALSE,
        # formatR.arrow = TRUE, 
        # scipen=999, 
        # digits=5,
        width=90) 
#thm <- knit_theme$get("edit-kwrite")   # whitengrey, bright, print, edit-flashdevelop, edit-kwrite
#knit_theme$set(thm)
knit_hooks$set(
   par = function(before, options, envir) {
      if (before && options$fig.show != 'none')
         par(mar = c(0, 0, 0, 0), # bottom, left, top, and right
             oma = c(0, 0, 0, 0))}
)
knitr::opts_chunk$set(
   # collapse = TRUE,
   comment = "#>",
   fig.align = 'center',
   fig.width = 9,
   fig.height = 5,
   fig.show = 'hold',
   out.extra = 'style="max-width:100%;"',
   # tidy = TRUE,
   # prompt=T,
   # comment=NA,
   cache = F
   # background = "red"
)
library(magrittr)
library(dplyr)
```


The `MDP2` package in R is a package for solving Markov decision processes (MDPs) with discrete 
time-steps, states and actions. Both traditional MDPs [@Puterman94], semi-Markov decision processes
(semi-MDPs) [@Tijms03] and hierarchical-MDPs (HMDPs) [@Kristensen00] can be solved under a finite
and infinite time-horizon.

The package implement well-known algorithms such as policy iteration and value iteration
under different criteria e.g. average reward per time unit and expected total discounted reward. The
model is stored using an underlying data structure based on the *state-expanded directed hypergraph*
of the MDP (@Relund06) implemented in `C++` for fast running times. <!-- Under development is also
support for MLHMP which is a Java implementation of algorithms for solving MDPs (@Kristensen03). 
-->

Building and solving an MDP is done in two steps. First, the MDP is built and saved in a set 
of binary files. Next, you load the MDP into memory from the binary files and apply various
algorithms to the model.

For building the MDP models see `vignette("building")`. In this vignette we focus on the second step, i.e. finding the optimal policy. Here we consider an infinite-horizon HMDP. 

```{r}
library(MDP2)
```


## An infinite-horizon HMDP

A hierarchical MDP is an MDP with parameters defined in a special way, but nevertheless in 
accordance with all usual rules and conditions relating to such processes (@Kristensen00). The basic
idea of the hierarchical structure is that stages of the process can be expanded to a so-called 
*child process*, which again may expand stages further to new child processes leading to multiple 
levels. To illustrate consider the HMDP shown in the figure below. The process has three levels. At
`Level 2` we have a set of finite-horizon semi-MDPs (one for each oval box) which all can be 
represented using a state-expanded hypergraph (hyperarcs not shown, only hyperarcs connecting 
processes are shown). A semi-MDP at `Level 2` is uniquely defined by a given state $s$ and action
$a$ of its *parent process* at `Level 1` (illustrated by the arcs with head and tail node at `Level
1` and `Level 2`, respectively). Moreover, when a child process at `Level 2` terminates a transition
from a state $s\in \mathcal{S}_{N}$ of the child process to a state at the next stage of the parent
process occur (illustrated by the (hyper)arcs having head and tail at `Level 2` and `Level 1`, 
respectively). 

```{r, echo=FALSE, fig.cap="The state-expanded hypergraph of the first stage of a hierarchical MDP. Level 0 indicate the founder level, and the nodes indicates states at the different levels. A child process (oval box) is represented using its state-expanded hypergraph (hyperarcs not shown) and is uniquely defined by a given state and action of its parent process."}
knitr::include_graphics("vignette_files/hmdp_index.png")
```

Since a child process is always defined by a stage, state and action of the parent process we have
that for instance a state at Level 1 can be identified using an index vector 
$\nu=(n_{0},s_{0},a_{0},n_{1},s_{1})$ where $s_1$ is the state id at the given stage $n_1$ in the 
process defined by the action $a_0$ in state $s_0$ at stage $n_0$. Note all values are ids starting 
from zero, e.g. if $s_1=0$ it is the first state at the corresponding stage and if $a_0=2$ it is the
third action at the corresponding state. In general a state $s$ and action $a$ at level $l$ can be
uniquely identified using
$$
\begin{aligned}
\nu_{s}&=(n_{0},s_{0},a_{0},n_{1},s_{1},\ldots,n_{l},s_{l}) \\
\nu_{a}&=(n_{0},s_{0},a_{0},n_{1},s_{1},\ldots,n_{l},s_{l},a_{l}).
\end{aligned}
$$
The index vectors for state $v_0$, $v_1$ and $v_2$ are illustrated in the figure. As under a 
semi-MDP another way to identify a state in the state-expanded hypergraph is using an unique id. 

## Example

Let us try to solve a small problem from livestock farming, namely the cow replacement problem where
we want to represent the age of the cow, i.e. the lactation number of the cow. During a lactation a
cow may have a high, average or low yield. We assume that a cow is always replaced after 4 
lactations.

In addition to lactation and milk yield we also want to take the genetic merit into account which is
either bad, average or good. When a cow is replaced we assume that the probability of a bad, average
or good heifer is equal.

We formulate the problem as a HMDP with 2 levels. At level 0 the states are the genetic merit and
the length of a stage is a life of a cow. At level 1 a stage describe a lactation and states
describe the yield. Decisions at level 1 are `keep` or `replace`.

Note the MDP runs over an infinite time-horizon at the founder level where each state (genetic
merit) define a semi-MDP at level 1 with 4 lactations.

Let us try to load the model and get some info:

```{r, par=TRUE}
prefix <- paste0(system.file("models", package = "MDP2"), "/cow_")
mdp <- loadMDP(prefix)
mdp 
```

The state-expanded hypergraph representing the HMDP with infinite time-horizon can be plotted 
using

```{r plotHMDP, message=FALSE, par=TRUE}
hgf <- getHypergraph(mdp)
## Rename labels
dat <- hgf$nodes %>% 
   dplyr::mutate(label = dplyr::case_when(
      label == "Low yield" ~ "L",
      label == "Avg yield" ~ "A",
      label == "High yield" ~ "H",
      label == "Dummy" ~ "D",
      label == "Bad genetic level" ~ "Bad",
      label == "Avg genetic level" ~ "Avg",
      label == "Good genetic level" ~ "Good",
      TRUE ~ "Error"
   ))
## Set grid id
dat$gId[1:3]<-85:87
dat$gId[43:45]<-1:3
getGId<-function(process,stage,state) {
   if (process==0) start=18
   if (process==1) start=22
   if (process==2) start=26
   return(start + 14 * stage + state)
}
idx<-43
for (process in 0:2)
   for (stage in 0:4)
      for (state in 0:2) {
         if (stage==0 & state>0) break
         idx<-idx-1
         #cat(idx,process,stage,state,getGId(process,stage,state),"\n")
         dat$gId[idx]<-getGId(process,stage,state)
      }
hgf$nodes <- dat
## Rename labels
dat <- hgf$hyperarcs %>% 
   dplyr::mutate(label = dplyr::case_when(
      label == "Replace" ~ "R",
      label == "Keep" ~ "K",
      label == "Dummy" ~ "D",
      TRUE ~ "Error"
      ),
      col = dplyr::case_when(
         label == "R" ~ "deepskyblue3",
         label == "K" ~ "darkorange1",
         label == "D" ~ "black",
         TRUE ~ "Error"
      ),
      lwd = 0.5,
      label = ""
   ) 
hgf$hyperarcs <- dat
## Make the plot
plotHypergraph(hgf, gridDim = c(14, 7), cex = 0.8, radx = 0.02, rady = 0.03)
```

Note action `keep` is drawn with orange color and action `replace` with blue color.

We find the optimal policy under the expected discounted reward criterion using policy iteration 
with an interest rate of 10%:

```{r Optimize (cow)}
wLbl<-"Net reward"         # the weight we want to optimize (net reward)
durLbl<-"Duration"         # the duration/time label
runPolicyIteDiscount(mdp, wLbl, durLbl, rate = 0.1)
```

The optimal policy is:

```{r plotPolicy, results='hide', message=FALSE, par=TRUE}
hgf$hyperarcs <- right_join(hgf$hyperarcs, getPolicy(mdp), by = c("sId", "aIdx"))
plotHypergraph(hgf, gridDim = c(14, 7), cex = 0.8, radx = 0.02, rady = 0.03)
```




```{r eval=FALSE, include=FALSE}
# getPolicy(mdp)
# rpo<-calcRPO(mdp, wLbl, iA=rep(0,42), criterion="discount", dur=durLbl, rate=rate, rateBase=rateBase)
# policy<-merge(policy,rpo)
# policy
```



We may also find the policy which maximize the average reward per lactation: 

```{r avePerLac, tidy.opts=list(comment=FALSE)}
wLbl<-"Net reward"         # the weight we want to optimize (net reward)
durLbl<-"Duration"         # the duration/time label
runPolicyIteAve(mdp, wLbl, durLbl)
getPolicy(mdp)
```

Since other weights are defined for each action we can calculate the average
reward per litre milk under the optimal policy:

```{r, echo=TRUE}
runCalcWeights(mdp, w=wLbl, criterion="average", dur = "Yield")
```

or the average yield per lactation:

```{r Reward/piglet (sow rep), echo=TRUE}
runCalcWeights(mdp, w="Yield", criterion="average", dur = durLbl)
```




```{r Delete bin, include=FALSE}
do.call(file.remove,list(list.files(pattern = ".bin")))
```


## References

---
title: "Solving a finite-horizon semi-MDP"
author: "Lars Relund <lars@relund.dk>"
date: "`r Sys.Date()`"
bibliography: litt.bib
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{finite-mdp}
  %\VignetteEncoding{UTF-8}
  %\VignetteEngine{knitr::rmarkdown}
editor_options: 
  chunk_output_type: console
---

<style> 
p {text-align: justify;} 
//.sourceCode {background-color: white;}
pre {
  // border-style: solid;
  // border-width: 1px;
  // border-color: grey;
  //background-color: grey !important;
}
img {
   //width: 100%;
   border: 0;
}
</style>

<!-- scale math down -->
<script type="text/x-mathjax-config"> 
    MathJax.Hub.Config({ 
        "HTML-CSS": { scale: 80 }
        });
</script>

```{r setup, include=FALSE}
library(knitr)
options(rmarkdown.html_vignette.check_title = FALSE,
        # formatR.arrow = TRUE, 
        # scipen=999, 
        # digits=5,
        width=90) 
#thm <- knit_theme$get("edit-kwrite")   # whitengrey, bright, print, edit-flashdevelop, edit-kwrite
#knit_theme$set(thm)
knit_hooks$set(
   par = function(before, options, envir) {
      if (before && options$fig.show != 'none')
         par(mar = c(0, 0, 0, 0), # bottom, left, top, and right
             oma = c(0, 0, 0, 0))}
)
knitr::opts_chunk$set(
   # collapse = TRUE,
   comment = "#>",
   fig.align = 'center',
   fig.width = 9,
   fig.height = 5,
   fig.show = 'hold',
   out.extra = 'style="max-width:100%;"',
   # tidy = TRUE,
   # prompt=T,
   # comment=NA,
   cache = F
   # background = "red"
)
library(magrittr)
library(dplyr)
```


The `MDP2` package in R is a package for solving Markov decision processes (MDPs) with discrete 
time-steps, states and actions. Both traditional MDPs [@Puterman94], semi-Markov decision processes
(semi-MDPs) [@Tijms03] and hierarchical-MDPs (HMDPs) [@Kristensen00] can be solved under a finite
and infinite time-horizon.

The package implement well-known algorithms such as policy iteration and value iteration
under different criteria e.g. average reward per time unit and expected total discounted reward. The
model is stored using an underlying data structure based on the *state-expanded directed hypergraph*
of the MDP (@Relund06) implemented in `C++` for fast running times. <!-- Under development is also
support for MLHMP which is a Java implementation of algorithms for solving MDPs (@Kristensen03). 
-->

Building and solving an MDP is done in two steps. First, the MDP is built and saved in a set 
of binary files. Next, you load the MDP into memory from the binary files and apply various
algorithms to the model.

For building the MDP models see `vignette("building")`. In this vignette we focus on the second step, i.e. finding the optimal policy. Here we consider a finite-horizon semi-MDP. 


```{r}
library(MDP2)
```


## A finite-horizon semi-MDP

A *finite-horizon semi-MDP* considers a sequential decision problem over $N$ *stages*. Let $I_{n}$
denote the finite set of system states at stage $n$. When *state* $i \in I_{n}$ is observed, an
*action* $a$ from the finite set of allowable actions $A_n(i)$ must be chosen, and this decision
generates *reward* $r_{n}(i,a)$. Moreover, let $\tau_n(i,a)$ denote the *stage length* of action
$a$, i.e. the expected time until the next decision epoch (stage $n+1$) given action $a$ and state
$i$. Finally, let $p_{ij}(a,n)$ denote the *transition probability* of obtaining state $j\in I_{n+1}$
at stage $n+1$ given that action $a$ is chosen in state $i$ at stage $n$. 

## Example

Consider a small machine repair problem used as an example in @Relund06 where the machine is always
replaced after 4 years. The state of the machine may be: good, average, and not working. Given the
machine's state we may maintain the machine. In this case the machine's state will be good at the
next decision epoch. Otherwise, the machine's state will not be better at next decision epoch. When
the machine is bought it may be either in state good or average. Moreover, if the machine is not
working it must be replaced.

The problem of when to replace the machine can be modeled using a Markov decision process with 
$N=5$ decision epochs. We use system states `good`,  `average`, `not working` and dummy state
`replaced` together with actions buy (`buy`), maintain (`mt`), no maintenance (`nmt`), and replace
(`rep`). The set of states at stage zero $S_{0}$ contains a single dummy state `dummy` representing
the machine before knowing its initial state. The only possible action is `buy`.

The cost of buying the machine is 100 with transition probability of 0.7 to state `good` and 0.3 to
state `average`. The reward (scrap value) of replacing a machine is 30, 10, and 5 in state `good`,  `average` and `not working`, respectively. The reward of the machine given action `mt` are 55, 40, and 30 in state `good`,  `average` and `not working`, respectively. Moreover, the system enters state 0 with probability 1 at the next stage.
Finally, the reward, transition states and probabilities given action $a=$`nmt` are given by:

  $n:s$               $1:$ `good`        $1:$ `average`        $2:$ `good`           $2:$ `average`         $3:$ `good`           $3:$ `average`
  -------------     ---------------    ------------------    ------------------    -------------------    ------------------    ------------------   
  $r_n(i,a)$          70                 50                    70                    50                     70                    50
  $j$                 $\{0,1\}$          $\{1,2\}$             $\{0,1\}$             $\{1,2\}$              $\{0,1\}$             $\{1,2\}$
  $p_{ij}(a,n)$       $\{0.6,0.4\}$      $\{0.6,0.4\}$         $\{0.5,0.5\}$         $\{0.5,0.5\}$          $\{0.2,0.8\}$         $\{0.2,0.8\}$


Let us try to load the model and get some info:

```{r, par=TRUE}
prefix <- paste0(system.file("models", package = "MDP2"), "/machine1_")
mdp <- loadMDP(prefix)
getInfo(mdp, withList = F, dfLevel = "action", asStringsActions = TRUE)  
```

The state-expanded hypergraph representing the semi-MDP with finite time-horizon can be plotted 
using

```{r, par=TRUE}
plot(mdp, actionColor = "label", radx = 0.06, marX = 0.065, marY = 0.055)
```

Each node corresponds to a specific state
and a directed hyperarc is defined for each possible action. For instance, action `mt` (maintain)
corresponds to a deterministic transition to state `good` and action `nmt` (not maintain)
corresponds to a transition to a condition/state not better than the current condition/state. We buy
the machine in stage 1 and may choose to replace the machine.


Let us use value iteration to find the optimal policy maximizing the expected total reward:

```{r solve3}
scrapValues <- c(30, 10, 5, 0)   # scrap values (the values of the 4 states at the last stage)
runValueIte(mdp, "Net reward", termValues = scrapValues)
```

The optimal policy is:

```{r, par=TRUE}
pol <- getPolicy(mdp)
tail(pol)
plot(mdp, actionsVisible = "policy", stateLabel = "weight", 
     radx = 0.06, marX = 0.065, marY = 0.055)
```

Note given the optimal policy the total expected reward is 
`r pol %>% filter(stateStr == "0,0") %>% pull(weight)` and the machine will never make a transition 
to states `not working` and `replaced`.

We may evaluate a certain policy, e.g. the policy always to maintain the machine:

```{r Set policy (machine rep),echo=TRUE,eval=TRUE}
policy<-data.frame(sId=c(8,11), aIdx=c(0,0)) # set the policy for sId 8 and 11 to mt
setPolicy(mdp, policy)
getPolicy(mdp)
```

If the policy specified in `setPolicy` does not contain all states then the actions from the
previous optimal policy are used. In the output above we can see that the policy now is to maintain
always. However, the reward of the policy has not been updated. Let us calculate the expected
reward:

```{r Calc reward (machine rep),echo=TRUE}
runCalcWeights(mdp, "Net reward", termValues = scrapValues)
tail(getPolicy(mdp))    
```

That is, the expected reward is 90.5 compared to 102.2 which was the reward of the optimal policy.





## References

---
title: "The MDP2 package"
author: "Lars Relund <lars@relund.dk>"
date: "`r Sys.Date()`"
bibliography: litt.bib
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{MDP2}
  %\VignetteEncoding{UTF-8}
  %\VignetteEngine{knitr::rmarkdown}
editor_options: 
  chunk_output_type: console
---

<style> 
p {text-align: justify;} 
//.sourceCode {background-color: white;}
pre {
  // border-style: solid;
  // border-width: 1px;
  // border-color: grey;
  //background-color: grey !important;
}
img {
   //width: 100%;
   border: 0;
}
</style>

<!-- scale math down -->
<script type="text/x-mathjax-config"> 
    MathJax.Hub.Config({ 
        "HTML-CSS": { scale: 80 }
        });
</script>

```{r setup, include=FALSE}
library(knitr)
options(rmarkdown.html_vignette.check_title = FALSE,
        # formatR.arrow = TRUE, 
        # scipen=999, 
        # digits=5,
        width=90) 
#thm <- knit_theme$get("edit-kwrite")   # whitengrey, bright, print, edit-flashdevelop, edit-kwrite
#knit_theme$set(thm)
knitr::opts_chunk$set(
   # collapse = TRUE,
   comment = "#>",
   fig.align = 'center',
   fig.width = 9,
   fig.height = 5,
   fig.show = 'hold',
   out.extra = 'style="max-width:100%;"',
   # tidy = TRUE,
   # prompt=T,
   # comment=NA,
   cache = F
   # background = "red"
)
library(magrittr)
library(dplyr)
```


The `MDP2` package in R is a package for solving Markov decision processes (MDPs) with discrete 
time-steps, states and actions. Both traditional MDPs [@Puterman94], semi-Markov decision processes
(semi-MDPs) [@Tijms03] and hierarchical-MDPs (HMDPs) [@Kristensen00] can be solved under a finite
and infinite time-horizon.

Building and solving an MDP is done in two steps. First, the MDP is built and saved in a set 
of binary files. Next, you load the MDP into memory from the binary files and apply various
algorithms to the model.

The package implement well-known algorithms such as policy iteration and value iteration
under different criteria e.g. average reward per time unit and expected total discounted reward. The
model is stored using an underlying data structure based on the *state-expanded directed hypergraph*
of the MDP (@Relund06) implemented in `C++` for fast running times. <!-- Under development is also
support for MLHMP which is a Java implementation of algorithms for solving MDPs (@Kristensen03). 
-->

To illustrate the package capabilities have a look at the vignettes:

   * Building MDP models `vignette("building")`.
   * Solving an infinite-horizon MDP (semi-MDP) `vignette("infinite-mdp")`.
   * Solving a finite-horizon MDP (semi-MDP) `vignette("finite-mdp")`.
   * Solving an infinite-horizon HMDP `vignette("infinite-hmdp")`.


## References

---
title: "Building an MDP model"
author: "Lars Relund <lars@relund.dk>"
date: "`r Sys.Date()`"
bibliography: litt.bib
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{building}
  %\VignetteEncoding{UTF-8}
  %\VignetteEngine{knitr::rmarkdown}
editor_options: 
  chunk_output_type: console
---

<style> 
p {text-align: justify;} 
//.sourceCode {background-color: white;}
pre {
  // border-style: solid;
  // border-width: 1px;
  // border-color: grey;
  //background-color: grey !important;
}
img {
   //width: 100%;
   border: 0;
}
</style>

<!-- scale math down -->
<script type="text/x-mathjax-config"> 
    MathJax.Hub.Config({ 
        "HTML-CSS": { scale: 80 }
        });
</script>

```{r setup, include=FALSE}
library(knitr)
options(rmarkdown.html_vignette.check_title = FALSE,
        # formatR.arrow = TRUE, 
        # scipen=999, 
        # digits=5,
        width=90) # 
#thm <- knit_theme$get("edit-kwrite")   # whitengrey, bright, print, edit-flashdevelop, edit-kwrite
#knit_theme$set(thm)
knit_hooks$set(
   par = function(before, options, envir) {
      if (before && options$fig.show != 'none')
         par(mar = c(0, 0, 0, 0), # bottom, left, top, and right
             oma = c(0, 0, 0, 0))}
)
knitr::opts_chunk$set(
   # collapse = TRUE,
   comment = "#>",
   fig.align = 'center',
   fig.width = 9,
   fig.height = 5,
   fig.show = 'hold',
   out.extra = 'style="max-width:100%;"',
   # tidy = TRUE,
   # prompt=T,
   # comment=NA,
   cache = F
   # background = "red"
)
```


Building and solving an MDP is done in two steps. First, the MDP is built and saved in a set 
of binary files. Next, you load the MDP into memory from the binary files and apply various
algorithms to the model.

In this vignette we focus on the first step, i.e. building and saving the model to a set of binary 
files. 

```{r}
library(MDP2)
```



## An infinite semi-MDP 

An *infinite-horizon semi-MDP* considers a sequential decision problem over an infinite number of 
*stages*. Let $I$ denote the finite set of system states at stage $n$. Note we assume that the 
semi-MDP is *homogeneous*, i.e the state space is independent of stage number. When *state* $i \in 
I$ is observed, an *action* $a$ from the finite set of allowable actions $A(i)$ must be chosen which
generates *reward* $r(i,a)$. Moreover, let $\tau(i,a)$ denote the *stage length* of action $a$, i.e.
the expected time until the next decision epoch (stage $n+1$) given action $a$ and state $i$.
Finally, let $p_{ij}(a)$ denote the *transition probability* of obtaining state $j\in I$ at stage
$n+1$ given that action $a$ is chosen in state $i$ at stage $n$. A policy is a decision 
rule/function that assigns to each state in the process an action. Note if the stage length is the
same for all states and actions the we have a regular *infinite-horizon MDP*.

Let us consider Example 6.1.1 in @Tijms03. At the beginning of each day a piece of equipment is
inspected to reveal its actual working condition. The equipment will be found in one of the working
conditions $i = 1,\ldots, N$ where the working condition $i$ is better than the working condition 
$i+1$. The equipment deteriorates in time. If the present working condition is $i$ and no repair is
done, then at the beginning of the next day the equipment has working condition $j$ with probability
$q_{ij}$. It is assumed that $q_{ij}=0$ for $j<i$ and $\sum_{j\geq i}q_{ij}=1$. The working
condition $i=N$ represents a malfunction that requires an enforced repair taking two days. For the
intermediate states $i$ with $1<i<N$ there is a choice between preventively repairing the equipment
and letting the equipment operate for the present day. A preventive repair takes only one day. A
repaired system has the working condition $i=1$. The cost of an enforced repair upon failure is
$C_{f}$ and the cost of a preemptive repair in working condition $i$ is $C_{p}(i)$. We wish to
determine a maintenance rule which minimizes the long-run average repair cost per day.

To formulate this problem as an infinite horizon semi-MDP the set of possible states of the system
is chosen as
$$
I=\{1,2,\ldots,N\}.
$$
State $i$ corresponds to the situation in which an inspection reveals working condition $i$. Define actions
$$
a=\left\{\begin{array}{ll}
nr & \text{if no repair.}\\
pr & \text{if preventive repair.}\\
fr & \text{if forced repair.}\\
\end{array}\right.
$$
The set of possible actions in state $i$ is chosen as $A(1)=\{nr\},\ A(i)=\{nr,pr\}$ for $1<i<N, 
A(N)=\{fr\}$. The one-step transition probabilities $p_{ij}(a)$ are given by $p_{ij}(0) = q_{ij}$ for
$1\leq i<N$, $p_{i1}(1) = 1$ for $1<i<N$, $p_{N1}(2)=1$  and zero otherwise. The one-step costs
$c_{i}(a)$ are given by $c_{i}(0)=0,\ c_{i}(1)=C_{p}(i)$ and $c_{N}(2)=C_{f}$. The stage length
until next decision epoch are $\tau(i,a) = 1, 0\leq i < N$ and $\tau(N,a) = 2$.

Assume that the number of possible working conditions equals $N=5$, i.e. the number of states are:

```{r states}
   N <- 5
   states <- tibble::tibble(
         idx = 1:N - 1,
         label = paste0("i = ", 1:N)
      )
   states
```

Note each state are given an index and are always numbered from zero. The repair costs (negative numbers since the MDP optimize based on rewards) are: 

```{r}
Cf <- -10
Cp <- c(0, -7, -7, -5)
```

The deterioration probabilities $q_{ij}$ are 

```{r}
Q <- matrix(c(
   0.90, 0.10, 0, 0, 0,
   0, 0.80, 0.10, 0.05, 0.05,
   0, 0, 0.70, 0.10, 0.20,
   0, 0, 0, 0.50, 0.50), 
   nrow=4, byrow=T) 
```

To build the model we need transition probabilities and the state index for the corresponding transitions. We may define function:

```{r transPr}
#' Transition probabilities
#' @param i State (1 <= i <= N). 
#' @param a Action (`nr`, `pr` or `fr`).
#' @return A list with non-zero transition probabilities and state index of the transitions.
transPr <- function(i, a) {
   pr <- NULL
   idx <- NULL
   if (a == "nr") {
      pr <- Q[i, ]
      idx <- which(pr > 0)  # only consider trans pr > 0
      pr <- pr[idx]
      idx <- idx - 1  # since state index is state-1
   }
   if (a == "pr" | a == "fr") {
      pr <- 1
      idx <- 0
   }
   return(list(pr = pr, idx = idx))
}
```

The transition probabilities for state 1 and the "no repair" action is now:

```{r}
transPr(1, "nr") 
```

That is, 90% for a transition to the state with index 0 ($i = 0$) and 10% for a transition to the 
state with index 1 ($i = 2$). Note we always assume that the states at the next stage is numbered 
using an index starting from zero!

The state-expanded hypergraph representing the semi-MDP with infinite time-horizon is shown in
Figure 1. Here the first stage is visualized. Each node corresponds to a specific state in the MDP.
A directed hyperarc is defined for each possible action. For instance, the state/node corresponding
to working condition $i=2$ and the two hyperarcs with head in this node corresponds to the two
possible actions preventive and no repair. Note the tails of a hyperarc represent a possible 
transition ($p_{ij}(a)>0$).

```{r buildMDP1, include=FALSE}
labels <- states$label
w<-binaryMDPWriter("hct611-1_") # use prefix hct611-1_ to the files
w$setWeights(c("Duration", "Net reward"))
w$process() # founder process
   w$stage() # a stage with states
      w$state(label = labels[1]) # state 1
         lst <- transPr(1, "nr")
         w$action(label = "nr", weights = c(1, 0), pr = lst$pr, id = lst$id, end = TRUE)
      w$endState() # end state 1
      for (i in 2:(N-1)) { # states 2 to N-1
         w$state(label = labels[i])
            lst <- transPr(i, "nr")
            w$action(label = "nr", weights = c(1, 0), pr = lst$pr, id = lst$id, end = TRUE)
            lst<-transPr(i, "pr")
            w$action(label = "pr", weights = c(1, Cp[i]), pr = lst$pr, id = lst$id, end = TRUE)
         w$endState()
      }
      w$state(label = labels[N])
         lst<-transPr(N, "fr")
         w$action(label = "fr", weights = c(2, Cf), pr = lst$pr, id = lst$id, end = TRUE)
      w$endState()
   w$endStage() # end stage
w$endProcess() # end process
w$closeWriter() # close the binary files
```

```{r semi-mdp, echo=FALSE, results='hide', message=FALSE, fig.cap="Figure 1: The state-expanded hypergraph for the semi-MDP.", par=TRUE}
mdp<-loadMDP("hct611-1_")
plot(mdp, actionColor = "label", marY = 0.06)
```


To build the semi-MDP, we use the `binaryMDPWriter` where the model can be built using either
matrices or an hierarchical structure. We first illustrate how to use the hierarchical structure.

```{r view_buldMDP1, ref.label='buildMDP1'}
```

Observe that 

   - We build the model with two weights applied to each action "Duration" and "Net reward". That 
     is, when we specify an action, we must add two weights. "Duration" equals 1 day except in state 
     $i=N$ where a forced repair takes 2 days. 
   - The process is built using first a `process` which contains a `stage` (we only specify one 
     stage, since we have a homogeneous semi-MDP over an infinite horizon) which contains 
     `state`s which contains `action`s. 
   - Transitions of an `action` are specified using the `pr` and `id` parameter. 
   - The `end = TRUE` argument in an action specify that the action is a normal action that not 
     define a sub-process.
   - The model is saved in a set of binary files with prefix `hct611-1_`.

Building the model may be error prone and you may use `getBinInfoStates()` and `getBinInfoActions()`
to retrieve info about the content of the binary files:

```{r}
getBinInfoStates("hct611-1_")
getBinInfoActions("hct611-1_")
```

The model can be loaded using the `loadMDP()` function.


The model can also be built by specifying a set of matrices. Note this way of specifying **only 
work** for infinite-horizon semi-MDPs (and not finite-horizon or hierarchical models). Specify a
list of probability matrices  (one for each action) where each row/state contains the transition 
probabilities (all zero if the action is not used in a state), a matrix with rewards and a matrix
with stage lengths (row = state, column = action). 

```{r buildMDP2}
## Define probability matrices
P <- list()
# a = nr (no repair)
P[[1]] <- as.matrix(rbind(Q, 0))

# a = pr (preventive repair)
Z <- matrix(0, nrow = N, ncol = N)
Z[2, 1] <- Z[3, 1] <- Z[4, 1] <- 1
P[[2]] <- Z

# a = fr (forced repair)
Z <- matrix(0, nrow = N, ncol = N)
Z[5, 1] <- 1
P[[3]] <- Z

## Rewards, a 5x3 matrix with one column for each action
R <- matrix(0, nrow = N, ncol = 3)
R[2:4, 2] <- Cp[2:4]
R[5, 3] <- Cf

## State lengths, a 5x3 matrix with one column for each action
D <- matrix(1, nrow = N, ncol = 3)
D[5, 3] <- 2

## Build model using the matrix specification
w <- binaryMDPWriter("hct611-2_")
w$setWeights(c("Duration", "Net reward"))
w$process(P, R, D)
w$closeWriter()
```



## A finite-horizon semi-MDP

A *finite-horizon semi-MDP* considers a sequential decision problem over $N$ *stages*. Let $I_{n}$
denote the finite set of system states at stage $n$. When *state* $i \in I_{n}$ is observed, an
*action* $a$ from the finite set of allowable actions $A_n(i)$ must be chosen, and this decision
generates *reward* $r_{n}(i,a)$. Moreover, let $\tau_n(i,a)$ denote the *stage length* of action
$a$, i.e. the expected time until the next decision epoch (stage $n+1$) given action $a$ and state
$i$. Finally, let $p_{ij}(a,n)$ denote the *transition probability* of obtaining state $j\in I_{n+1}$
at stage $n+1$ given that action $a$ is chosen in state $i$ at stage $n$. 

Consider a small machine repair problem used as an example in @Relund06 where the machine is always
replaced after 4 years. The state of the machine may be: good, average, and not working. Given the
machine's state we may maintain the machine. In this case the machine's state will be good at the
next decision epoch. Otherwise, the machine's state will not be better at next decision epoch. When
the machine is bought it may be either in state good or average. Moreover, if the machine is not
working it must be replaced.

The problem of when to replace the machine can be modeled using a Markov decision process with 
$N=5$ decision epochs. We use system states `good`,  `average`, `not working` and dummy state
`replaced` together with actions buy (`buy`), maintain (`mt`), no maintenance (`nmt`), and replace
(`rep`). The set of states at stage zero $S_{0}$ contains a single dummy state `dummy` representing
the machine before knowing its initial state. The only possible action is `buy`.

The cost of buying the machine is 100 with transition probability of 0.7 to state `good` and 0.3 to
state `average`. The reward (scrap value) of replacing a machine is 30, 10, and 5 in state `good`,
`average` and `not working`, respectively. The reward of the machine given action `mt` are 55, 40,
and 30 in state `good`,  `average` and `not working`, respectively. Moreover, the system enters
state 0 with probability 1 at the next stage. Finally, the reward, transition states and
probabilities given action $a=$`nmt` are given by:

  $n:s$               $1:$ `good`        $1:$ `average`        $2:$ `good`           $2:$ `average`         $3:$ `good`           $3:$ `average`
  -------------     ---------------    ------------------    ------------------    -------------------    ------------------    ------------------   
  $r_n(i,a)$          70                 50                    70                    50                     70                    50
  $j$                 $\{0,1\}$          $\{1,2\}$             $\{0,1\}$             $\{1,2\}$              $\{0,1\}$             $\{1,2\}$
  $p_{ij}(a,n)$       $\{0.6,0.4\}$      $\{0.6,0.4\}$         $\{0.5,0.5\}$         $\{0.5,0.5\}$          $\{0.2,0.8\}$         $\{0.2,0.8\}$

We build the MDP (all actions have same length) using the `binaryMDPWriter()` function:

```{r buildMDP3}
prefix<-"machine1_"
w <- binaryMDPWriter(prefix)
w$setWeights(c("Net reward"))
w$process()
	w$stage()   # stage n=0
		w$state(label="dummy")       
			w$action(label="buy", weights=-100, pr=c(0.7,0.3), id=c(0,1), end=TRUE)
		w$endState()
	w$endStage()
	w$stage()   # stage n=1
		w$state(label="good")           
			w$action(label="mt", weights=55, pr=1, id=0, end=TRUE)
			w$action(label="nmt", weights=70, pr=c(0.6,0.4), id=c(0,1), end=TRUE)
		w$endState()
		w$state(label="average")        
			w$action(label="mt", weights=40, pr=1, id=0, end=TRUE)
			w$action(label="nmt", weights=50, pr=c(0.6,0.4), id=c(1,2), end=TRUE)
		w$endState()
	w$endStage()
	w$stage()   # stage n=2
		w$state(label="good")          
			w$action(label="mt", weights=55, pr=1, id=0, end=TRUE)
			w$action(label="nmt", weights=70, pr=c(0.5,0.5), id=c(0,1), end=TRUE)
		w$endState()
		w$state(label="average")       
			w$action(label="mt", weights=40, pr=1, id=0, end=TRUE)
			w$action(label="nmt", weights=50, pr=c(0.5,0.5), id=c(1,2), end=TRUE)
		w$endState()
		w$state(label="not working")    
			w$action(label="mt", weights=30, pr=1, id=0, end=TRUE)
			w$action(label="rep", weights=5, pr=1, id=3, end=TRUE)
		w$endState()
	w$endStage()
	w$stage()   # stage n=3
		w$state(label="good")           
			w$action(label="mt", weights=55, pr=1, id=0, end=TRUE)
			w$action(label="nmt", weights=70, pr=c(0.2,0.8), id=c(0,1), end=TRUE)
		w$endState()
		w$state(label="average")       
			w$action(label="mt", weights=40, pr=1, id=0, end=TRUE)
			w$action(label="nmt", weights=50, pr=c(0.2,0.8), id=c(1,2), end=TRUE)
		w$endState()
		w$state(label="not working")    
			w$action(label="mt", weights=30, pr=1, id=0, end=TRUE)
			w$action(label="rep", weights=5, pr=1, id=3, end=TRUE)
		w$endState()
		w$state(label="replaced")       
			w$action(label="dummy", weights=0, pr=1, id=3, end=TRUE)
		w$endState()
	w$endStage()
	w$stage()   # stage n=4
		w$state(label="good", end=TRUE)        
		w$state(label="average", end=TRUE)     
		w$state(label="not working", end=TRUE) 
		w$state(label="replaced", end=TRUE)   
	w$endStage()
w$endProcess()
w$closeWriter()
```

Note that at each stage the states are numbered using index starting from zero such that e.g.
`w$action(label="nmt", weights=50, pr=c(0.2,0.8), id=c(1,2), end=TRUE)` define an action with
transition to states with indices 1 and 2 at the next stage with probability 0.2 and 0.8, 
respectively.

The MDP is illustrated in Figure 2. Each node corresponds to a specific state. Node text are the
state index (given stage) and its label. A directed hyperarc is defined for each possible action.
For instance, action `mt` (maintain) corresponds to a deterministic transition to state `good` and
action `nmt` (not maintain) corresponds to a transition to a condition/state not better than the
current condition/state. We buy the machine in stage 1 and may choose to replace the machine during
its lifetime.

```{r plotHgf3, echo=FALSE, fig.cap="Figure 2: A finite-horizon MDP", par=TRUE}
scrapValues <- c(30, 10, 5, 0)   # scrap values (the values of the 4 states at stage 4)
mdp <- loadMDP("machine1_", getLog = FALSE)
plot(mdp, actionColor = "label", radx = 0.06, marX = 0.065, marY = 0.055, stateLabel = "sIdx|label")
```


## An infinite-horizon HMDP

A hierarchical MDP is an MDP with parameters defined in a special way, but nevertheless in 
accordance with all usual rules and conditions relating to such processes (@Kristensen00). The basic
idea of the hierarchical structure is that stages of the process can be expanded to a so-called 
*child process*, which again may expand stages further to new child processes leading to multiple 
levels. To illustrate consider the HMDP shown in Figure 3. The process has three levels. At
`Level 2` we have a set of finite-horizon semi-MDPs (one for each oval box) which all can be 
represented using a state-expanded hypergraph (hyperarcs not shown, only hyperarcs connecting 
processes are shown). A semi-MDP at `Level 2` is uniquely defined by a given state $s$ and action
$a$ of its *parent process* at `Level 1` (illustrated by the arcs with head and tail node at `Level
1` and `Level 2`, respectively). Moreover, when a child process at `Level 2` terminates a transition
from a state $s\in \mathcal{S}_{N}$ of the child process to a state at the next stage of the parent
process occur (illustrated by the (hyper)arcs having head and tail at `Level 2` and `Level 1`, 
respectively). 

```{r, echo=FALSE, fig.cap="Figure 3: The state-expanded hypergraph of the first stage of a hierarchical MDP. Level 0 indicate the founder level, and the nodes indicates states at the different levels. A child process (oval box) is represented using its state-expanded hypergraph (hyperarcs not shown) and is uniquely defined by a given state and action of its parent process."}
knitr::include_graphics("vignette_files/hmdp_index.png")
```

Since a child process is always defined by a stage, state and action of the parent process we have
that for instance a state at Level 1 can be identified using an index vector 
$\nu=(n_{0},s_{0},a_{0},n_{1},s_{1})$ where $s_1$ is the state id at the given stage $n_1$ in the 
process defined by the action $a_0$ in state $s_0$ at stage $n_0$. Note all values are ids starting 
from zero, e.g. if $s_1=0$ it is the first state at the corresponding stage and if $a_0=2$ it is the
third action at the corresponding state. In general a state $s$ and action $a$ at level $l$ can be
uniquely identified using
$$
\begin{aligned}
\nu_{s}&=(n_{0},s_{0},a_{0},n_{1},s_{1},\ldots,n_{l},s_{l}) \\
\nu_{a}&=(n_{0},s_{0},a_{0},n_{1},s_{1},\ldots,n_{l},s_{l},a_{l}).
\end{aligned}
$$
The index vectors for state $v_0$, $v_1$ and $v_2$ are illustrated in Figure 3. 
<!-- As under a semi-MDP another way to identify a state in the state-expanded hypergraph is using an unique id.  -->


Let us try to solve a small problem from livestock farming, namely the cow replacement problem where
we want to represent the age of the cow, i.e. the lactation number of the cow. During a lactation a
cow may have a high, average or low yield. We assume that a cow is always replaced after 4 
lactations.

In addition to lactation and milk yield we also want to take the genetic merit into account which is
either bad, average or good. When a cow is replaced we assume that the probability of a bad, average
or good heifer is the same.

We formulate the problem as a HMDP with 2 levels. At level 0 the states are the genetic merit and
the length of a stage is a life of a cow. At level 1 a stage describe a lactation and states
describe the yield. Decisions at level 1 are `keep` or `replace`.

Note the MDP runs over an infinite time-horizon at the founder level where each state (genetic
merit) define a semi-MDP at level 1 with 4 lactations.

To generate the MDP we need to know the weights and transition probabilities which are provided in a
CSV file. To ease the understanding we provide 2 functions for reading from the CSV:

```{r Generate cow MDP functions,echo=TRUE}
library(magrittr)
cowDf <- readr::read_csv("vignette_files/cow.csv")
cowDf

# Weights given a state at level 2
lev1W <- function(s0Idx, n1Idx, s1Idx, a1Lbl) {
   return(cowDf %>% 
      dplyr::filter(s0 == s0Idx & n1 == n1Idx & s1 == s1Idx & label == a1Lbl) %>% 
      dplyr::select(Duration, Reward, Output) %>% as.numeric()
   )
}
lev1W(2, 2, 1, 'Keep')     # good genetic merit, lactation 2, avg yield, keep action

# Trans pr given a state at level 2
lev1Pr <- function(s0Idx, n1Idx, s1Idx, a1Lbl) {
   return(cowDf %>% 
      dplyr::filter(s0 == s0Idx & n1 == n1Idx & s1 == s1Idx & label == a1Lbl) %>% 
      dplyr::select(scp0:last_col()) %>% as.numeric()
   )
}
lev1Pr(2, 2, 1, 'Replace') # good genetic merit, lactation 2, avg yield, replace action
```

We can now generate the model with three weights
```{r Generate cow MDP,echo=TRUE, tidy=FALSE}
lblS0 <- c('Bad genetic level', 'Avg genetic level', 'Good genetic level')
lblS1 <- c('Low yield', 'Avg yield', 'High yield')
prefix<-"cow_"
w<-binaryMDPWriter(prefix)
w$setWeights(c("Duration", "Net reward", "Yield"))
w$process()
  w$stage()   # stage 0 at founder level
    for (s0 in 0:2) {
      w$state(label=lblS0[s0+1])   # state at founder
        w$action(label="Keep", weights=c(0,0,0), prob=c(2,0,1))   # action at founder
          w$process()
            w$stage()   # dummy stage at level 1
               w$state(label="Dummy")
                 w$action(label="Dummy", weights=c(0,0,0), 
                          prob=c(1,0,1/3, 1,1,1/3, 1,2,1/3), end=TRUE)
               w$endState()
            w$endStage()
            for (d1 in 1:4) {
              w$stage()   # stage at level 1
                for (s1 in 0:2) {
                  w$state(label=lblS1[s1+1])
                    if (d1!=4) {
                      w$action(label="Keep", weights=lev1W(s0,d1,s1,"Keep"), 
                               prob=lev1Pr(s0,d1,s1,"Keep"), end=TRUE)
                    }
                    w$action(label="Replace", weights=lev1W(s0,d1,s1,"Replace"), 
                             prob=lev1Pr(s0,d1,s1,"Replace"), end=TRUE)
                  w$endState()
                }
              w$endStage()
            }
          w$endProcess()
        w$endAction()
      w$endState()
    }
  w$endStage()
w$endProcess()
w$closeWriter()
```

Note that the model is built using the `prob` parameter which contains triples of `(scope, id, pr)`.
Scope can be: 2 = a transition to a child process (stage zero in the child process), 1 = a
transition to next stage in the current process and 0 = a transition to the next stage in the father
process. For instance if `prob = c(1,2,0.3, 0,0,0.7)` we specify transitions to the third state (id
= 2) at the next stage of the process and to the first state (id = 0) at the next stage of the
father process with probabilities 0.3 and 0.7, respectively.

A plot of the state-expanded hypergraph is given below where action `keep` is drawn with orange
color and action `replace` with blue color.

```{r plotHMDP, echo=FALSE, message=FALSE, par=TRUE}
mdp<-loadMDP(prefix)
# hgf <- list(nodes = NULL, hyperarcs = NULL)
# hgf %>% plotHypergraph(gridDim=c(14,7), cex = 0.8, showGrid = T)
hgf <- getHypergraph(mdp)
dat <- hgf$nodes %>% 
   dplyr::mutate(label = dplyr::case_when(
      label == "Low yield" ~ "L",
      label == "Avg yield" ~ "A",
      label == "High yield" ~ "H",
      label == "Dummy" ~ "D",
      label == "Bad genetic level" ~ "Bad",
      label == "Avg genetic level" ~ "Avg",
      label == "Good genetic level" ~ "Good",
      TRUE ~ "Error"
   ))
dat$gId[1:3]<-85:87
dat$gId[43:45]<-1:3
getGId<-function(process,stage,state) {
   if (process==0) start=18
   if (process==1) start=22
   if (process==2) start=26
   return(start + 14 * stage + state)
}
idx<-43
for (process in 0:2)
   for (stage in 0:4)
      for (state in 0:2) {
         if (stage==0 & state>0) break
         idx<-idx-1
         #cat(idx,process,stage,state,getGId(process,stage,state),"\n")
         dat$gId[idx]<-getGId(process,stage,state)
      }
hgf$nodes <- dat

dat <- hgf$hyperarcs %>% 
   dplyr::mutate(label = dplyr::case_when(
      label == "Replace" ~ "R",
      label == "Keep" ~ "K",
      label == "Dummy" ~ "D",
      TRUE ~ "Error"
      ),
      col = dplyr::case_when(
         label == "R" ~ "deepskyblue3",
         label == "K" ~ "darkorange1",
         label == "D" ~ "black",
         TRUE ~ "Error"
      ),
      lwd = 0.5,
      label = ""
   ) 
hgf$hyperarcs <- dat
plotHypergraph(gridDim = c(14, 7), hgf, cex = 0.8, radx = 0.02, rady = 0.03)
```


```{r Delete bin, include=FALSE}
do.call(file.remove,list(list.files(pattern = ".bin")))
```




## References

---
title: "Solving an infinite-horizon semi-MDP"
author: "Lars Relund <lars@relund.dk>"
date: "`r Sys.Date()`"
bibliography: litt.bib
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{infinite-mdp}
  %\VignetteEncoding{UTF-8}
  %\VignetteEngine{knitr::rmarkdown}
editor_options: 
  chunk_output_type: console
---

<style> 
p {text-align: justify;} 
//.sourceCode {background-color: white;}
pre {
  // border-style: solid;
  // border-width: 1px;
  // border-color: grey;
  //background-color: grey !important;
}
img {
   //width: 100%;
   border: 0;
}
</style>

<!-- scale math down -->
<script type="text/x-mathjax-config"> 
    MathJax.Hub.Config({ 
        "HTML-CSS": { scale: 80 }
        });
</script>

```{r setup, include=FALSE}
library(knitr)
options(rmarkdown.html_vignette.check_title = FALSE,
        # formatR.arrow = TRUE, 
        # scipen=999, 
        # digits=5,
        width=90) 
#thm <- knit_theme$get("edit-kwrite")   # whitengrey, bright, print, edit-flashdevelop, edit-kwrite
#knit_theme$set(thm)
knit_hooks$set(
   par = function(before, options, envir) {
      if (before && options$fig.show != 'none')
         par(mar = c(0, 0, 0, 0), # bottom, left, top, and right
             oma = c(0, 0, 0, 0))}
)
knitr::opts_chunk$set(
   # collapse = TRUE,
   comment = "#>",
   fig.align = 'center',
   fig.width = 9,
   fig.height = 5,
   fig.show = 'hold',
   out.extra = 'style="max-width:100%;"',
   # tidy = TRUE,
   # prompt=T,
   # comment=NA,
   cache = F
   # background = "red"
)
library(magrittr)
library(dplyr)
```


The `MDP2` package in R is a package for solving Markov decision processes (MDPs) with discrete 
time-steps, states and actions. Both traditional MDPs [@Puterman94], semi-Markov decision processes
(semi-MDPs) [@Tijms03] and hierarchical-MDPs (HMDPs) [@Kristensen00] can be solved under a finite
and infinite time-horizon.

The package implement well-known algorithms such as policy iteration and value iteration
under different criteria e.g. average reward per time unit and expected total discounted reward. The
model is stored using an underlying data structure based on the *state-expanded directed hypergraph*
of the MDP (@Relund06) implemented in `C++` for fast running times. <!-- Under development is also
support for MLHMP which is a Java implementation of algorithms for solving MDPs (@Kristensen03). 
-->

Building and solving an MDP is done in two steps. First, the MDP is built and saved in a set 
of binary files. Next, you load the MDP into memory from the binary files and apply various
algorithms to the model.

For building the MDP models see `vignette("building")`. In this vignette we focus on the second step, i.e. finding the optimal policy. Here we consider an infinite semi-MDP. 

```{r}
library(MDP2)
```



## An infinite-horizon semi-MDP 

An *infinite-horizon semi-MDP* considers a sequential decision problem over an infinite number of 
*stages*. Let $I$ denote the finite set of system states at stage $n$. Note we assume that the 
semi-MDP is *homogeneous*, i.e the state space is independent of stage number. When *state* $i \in 
I$ is observed, an *action* $a$ from the finite set of allowable actions $A(i)$ must be chosen which
generates *reward* $r(i,a)$. Moreover, let $\tau(i,a)$ denote the *stage length* of action $a$, i.e.
the expected time until the next decision epoch (stage $n+1$) given action $a$ and state $i$.
Finally, let $p_{ij}(a)$ denote the *transition probability* of obtaining state $j\in I$ at stage
$n+1$ given that action $a$ is chosen in state $i$ at stage $n$. A policy is a decision 
rule/function that assigns to each state in the process an action.

## Example

Let us consider example 6.1.1 in @Tijms03. At the beginning of each day a piece of equipment is
inspected to reveal its actual working condition. The equipment will be found in one of the working
conditions $i = 1,\ldots, N$ where the working condition $i$ is better than the working condition 
$i+1$. The equipment deteriorates in time. If the present working condition is $i$ and no repair is
done, then at the beginning of the next day the equipment has working condition $j$ with probability
$q_{ij}$. It is assumed that $q_{ij}=0$ for $j<i$ and $\sum_{j\geq i}q_{ij}=1$. The working
condition $i=N$ represents a malfunction that requires an enforced repair taking two days. For the
intermediate states $i$ with $1<i<N$ there is a choice between preventively repairing the equipment
and letting the equipment operate for the present day. A preventive repair takes only one day. A
repaired system has the working condition $i=1$. The cost of an enforced repair upon failure is
$C_{f}$ and the cost of a preemptive repair in working condition $i$ is $C_{p}(i)$. We wish to
determine a maintenance rule which minimizes the long-run average repair cost per day.

To formulate this problem as an infinite horizon semi-MDP the set of possible states of the system
is chosen as
$$
I=\{1,2,\ldots,N\}.
$$
State $i$ corresponds to the situation in which an inspection reveals working condition $i$. Define actions
$$
a=\left\{\begin{array}{ll}
nr & \text{if no repair.}\\
pr & \text{if preventive repair.}\\
fr & \text{if forced repair.}\\
\end{array}\right.
$$
The set of possible actions in state $i$ is chosen as $A(1)=\{nr\},\ A(i)=\{nr,pr\}$ for $1<i<N, 
A(N)=\{fr\}$. The one-step transition probabilities $p_{ij}(a)$ are given by $p_{ij}(0) = q_{ij}$ for
$1\leq i<N$, $p_{i1}(1) = 1$ for $1<i<N$, $p_{N1}(2)=1$  and zero otherwise. The one-step costs
$c_{i}(a)$ are given by $c_{i}(0)=0,\ c_{i}(1)=C_{p}(i)$ and $c_{N}(2)=C_{f}$. The stage length
until next decision epoch are $\tau(i,a) = 1, 0\leq i < N$ and $\tau(N,a) = 2$.


Assume that the number of possible working conditions equals $N=5$. The repair costs are given by $C_{f}=10,\ C_{p}(2)=7,\ C_{p}(3)=7$ and $C_{p}(4)=5$. The deterioration probabilities $q_{ij}$ are given by

```{r parameters,  include=FALSE}
N<-5; Cf<- -10; Cp<-c(0,-7,-7,-5) # use negative numbers since the MDP optimize based on rewards
Q <- matrix(c(
   0.90, 0.10, 0, 0, 0,
   0, 0.80, 0.10, 0.05, 0.05,
   0, 0, 0.70, 0.10, 0.20,
   0, 0, 0, 0.50, 0.50), nrow=4, byrow=T) 
```

```{r Qtable, results='asis', echo=FALSE}
rownames(Q)<-1:4
colnames(Q)<-1:5
knitr::kable(Q, row.names = T)
```

For building and saving the model see the `vignette("building")`. We load the model using 

```{r}
prefix <- paste0(system.file("models", package = "MDP2"), "/hct611-1_")
mdp <- loadMDP(prefix)
```

The variable `mdp` is a list with a pointer to the MDP object stored in memory. 

```{r}
mdp
```

For instance the total number of actions is `r mdp$actions` and the model use two weights applied 
to each action "Duration" and "Net reward". Information about the MDP can be retrieved using 
`getInfo()`:

```{r}
getInfo(mdp, withList = F, dfLevel = "action", asStringsActions = TRUE)  
```

Here the tibble has a row for each state and action. For instance the weight "Duration" equals 1 day 
except in state $i=5$ where a forced repair takes 2 days (row 13). States with no actions are also 
given.  

The state-expanded hypergraph representing the semi-MDP with infinite time-horizon can be plotted 
using

```{r plotHgf, par=TRUE}
plot(mdp, actionColor = "label", stateLabel = "sId|label")
```

Each node corresponds to a specific state in the MDP and is a *unique id* (`sId`) such that you can
identify all the states (**id always start from zero**). These ids are not equal to the ids used
when you built the model, since the order of the nodes in the hypergraph data structure is
optimized! A directed hyperarc is defined for each possible action. For
instance, the state/node with `sId = 6` corresponds to working condition $i=2$ and the two hyperarcs
with head in this node corresponds to the two actions preventive and no repair. Note the tails of a
hyperarc represent a possible transition ($p_{ij}(a)>0$).

Given the model in memory, we now can find the optimal policy under various policies. Let us first
try to optimize the average reward per time unit.

```{r solve1_ave, par=TRUE}
runPolicyIteAve(mdp,"Net reward","Duration")
getPolicy(mdp)
plot(mdp, actionsVisible = "policy")
```

Note it is optimal to do a preventive repair in state $i=4$. Let us try to optimize the expected 
total discounted reward with a discount factor of 0.5 using policy iteration:

```{r, par=TRUE}
runPolicyIteDiscount(mdp,"Net reward","Duration", discountFactor = 0.5)
getPolicy(mdp)
plot(mdp, actionsVisible = "policy")
```

Note given a discount factor of 0.5, it is optimal to not do a preventive repair in state $i=4$. 
The same results can be found using value iteration:

```{r}
runValueIte(mdp,"Net reward","Duration", discountFactor = 0.5, eps = 1e-10, maxIte = 1000)
getPolicy(mdp)
```



## References

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mdp.R
\name{getWIdx}
\alias{getWIdx}
\title{Return the index of a weight in the model. Note that index always start from zero (C++ style), i.e. the first weight, the first state at a stage etc has index 0.}
\usage{
getWIdx(mdp, wLbl)
}
\arguments{
\item{mdp}{The MDP loaded using \code{\link[=loadMDP]{loadMDP()}}.}

\item{wLbl}{The label/string of the weight.}
}
\value{
The index (integer).
}
\description{
Return the index of a weight in the model. Note that index always start from zero (C++ style), i.e. the first weight, the first state at a stage etc has index 0.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mdp-package.R
\docType{package}
\name{MDP2-package}
\alias{MDP2}
\alias{MDP2-package}
\title{MDP2: Markov Decision Processes (MDPs)}
\description{
Create and optimize (semi) MDPs with discrete time steps and state space. Both hierarchical and ordinary-traditional MDPs can be modeled.
}
\seealso{
\code{\link[=loadMDP]{loadMDP()}}.
}
\author{
\strong{Maintainer}: Lars Relund Nielsen \email{lars@relund.dk} (\href{https://orcid.org/0000-0002-4802-3071}{ORCID})

Authors:
\itemize{
  \item Lars Relund Nielsen \email{lars@relund.dk} (\href{https://orcid.org/0000-0002-4802-3071}{ORCID})
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.R
\name{plot.HMDP}
\alias{plot.HMDP}
\title{Plot the state-expanded hypergraph of the MDP.}
\usage{
\method{plot}{HMDP}(x, ...)
}
\arguments{
\item{x}{The MDP model.}

\item{...}{Arguments passed to \code{\link[=plotHypergraph]{plotHypergraph()}}.}
}
\value{
No return value (NULL invisible), called for side effects (plotting).
}
\description{
Plot the state-expanded hypergraph of the MDP.
}
\examples{
## Set working dir
wd <- setwd(system.file("models", package = "MDP2"))

#### A finite-horizon replacement problem ####
mdp<-loadMDP("machine1_")
plot(mdp)
plot(mdp, actionColor = "label")  # colors based on labels
plot(mdp, transLabels = "state")  # label transitions with target state labels
plot(mdp, transLabels = "prob")  # label transitions with transition probabilities
plot(mdp, actionColor = "label", stateLabel = "sId|label")  # state labels are 'sId | label'
plot(mdp, stateLabel = "sIdx|label", radx = 0.01)  # adjust radx in states
plot(mdp, stateLabel = "label", actionWLabel = "none", actionLabel = "label", 
     transLabels = "sId", radx = 0.01)

scrapValues <- c(30, 10, 5, 0)  # scrap values (the values of the 4 states at stage 4)
runValueIte(mdp, "Net reward" , termValues = scrapValues)
plot(mdp, actionColor = "policy")  # highlight optimal policy
plot(mdp, actionsVisible = "policy", stateLabel = "weight")  # show only optimal policy


#### An infinite-horizon maintenance problem ####
mdp<-loadMDP("hct611-1_")
plot(mdp)  # plot the first two stages
plot(mdp, actionColor = "label")  # colors based on labels
plot(mdp, actionColor = "label", stateLabel = "sId|label")  # state labels are 'sId | label'
runPolicyIteAve(mdp,"Net reward","Duration")
plot(mdp, actionColor = "policy")  # highlight optimal policy
plot(mdp, actionsVisible = "policy")  # show only optimal policy


#### An infinite-horizon hierarchical replacement problem ####
library(magrittr)
mdp<-loadMDP("cow_")
hgf <- getHypergraph(mdp)
# modify labels
dat <- hgf$nodes \%>\% 
   dplyr::mutate(label = dplyr::case_when(
      label == "Low yield" ~ "L",
      label == "Avg yield" ~ "A",
      label == "High yield" ~ "H",
      label == "Dummy" ~ "D",
      label == "Bad genetic level" ~ "Bad",
      label == "Avg genetic level" ~ "Avg",
      label == "Good genetic level" ~ "Good",
      TRUE ~ "Error"
   ))
# assign nodes to grid ids
dat$gId[1:3]<-85:87
dat$gId[43:45]<-1:3
getGId<-function(process,stage,state) {
   if (process==0) start=18
   if (process==1) start=22
   if (process==2) start=26
   return(start + 14 * stage + state)
}
idx<-43
for (process in 0:2)
   for (stage in 0:4)
      for (state in 0:2) {
         if (stage==0 & state>0) break
         idx<-idx-1
         #cat(idx,process,stage,state,getGId(process,stage,state),"\n")
         dat$gId[idx]<-getGId(process,stage,state)
      }
hgf$nodes <- dat
# modify labels
dat <- hgf$hyperarcs \%>\% 
   dplyr::mutate(label = dplyr::case_when(
      label == "Replace" ~ "R",
      label == "Keep" ~ "K",
      label == "Dummy" ~ "D",
      TRUE ~ "Error"
   ),
   col = dplyr::case_when(
      label == "R" ~ "deepskyblue3",
      label == "K" ~ "darkorange1",
      label == "D" ~ "black",
      TRUE ~ "Error"
   ),
   lwd = 0.5,
   label = ""
   ) 
hgf$hyperarcs <- dat
# plot hypergraph
oldpar <- par(mai = c(0, 0, 0, 0))
plotHypergraph(gridDim = c(14, 7), hgf, cex = 0.8, radx = 0.02, rady = 0.03)
par(oldpar)


## A simple finite-horizon MDP with action and transition weights
prefix <- file.path(tempdir(), "plot_transition_rewards_")
w <- binaryMDPWriter(prefix)
w$setWeights("Cost")
w$setTransWeights(c("Reward", "Disease"))
w$process()
   w$stage()
      w$state(label = "S1")
         w$action(
            label = "A1", weights = 2, id = c(1), pr = c(1),
            transWeights = c(20, 0.3), end = TRUE
         )
         w$action(
            label = "A2", weights = 1, id = c(0, 1), pr = c(0.3, 0.7),
            transWeights = c(25, 0.4, 15, 0.2), end = TRUE
         )
      w$endState()
   w$endStage()
   w$stage()
      w$state(label = "S2")
         w$action(
            label = "A3", weights = 3, id = c(0, 1, 2), pr = c(0.5, 0.3, 0.2),
            transWeights = c(0, 0.05, 12, 0.2, 30, 0.8), end = TRUE
         )
         w$action(
            label = "A4", weights = 2, id = c(1, 2), pr = c(0.6, 0.4),
            transWeights = c(22, 0.35, 27, 0.7), end = TRUE
         )
      w$endState()
      w$state(label = "S3")
         w$action(
            label = "A5", weights = 1, id = c(0, 1), pr = c(0.4, 0.6),
            transWeights = c(5, 0, 16, 0.25), end = TRUE
         )
         w$action(
            label = "A6", weights = 4, id = c(0, 1, 2), pr = c(0.1, 0.3, 0.6),
            transWeights = c(14, 0.15, 21, 0.45, 29, 1), end = TRUE
         )
      w$endState()
   w$endStage()
   w$stage()
      w$state(label = "S4", end = TRUE)
      w$state(label = "S5", end = TRUE)
      w$state(label = "S6", end = TRUE)
   w$endStage()
w$endProcess()
w$closeWriter()

mdp <- loadMDP(prefix, getLog = FALSE)
plot(mdp, actionColor = "label", transLabels = "weights", actionWLabel = "weight", 
     radx = 0.005, rady = 0.01)

## Reset working dir
setwd(wd)
}
\seealso{
\code{\link[=getHypergraph]{getHypergraph()}} and \code{\link[=plotHypergraph]{plotHypergraph()}} for possible arguments.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/convert.R
\name{stateIdxMat}
\alias{stateIdxMat}
\title{Info about the states in the HMDP model under consideration.}
\usage{
stateIdxMat(prefix = "", file = "stateIdx.bin")
}
\arguments{
\item{prefix}{A character string with the prefix added to til file(s).}

\item{file}{The HMDP binary file containing the description under consideration.}
}
\value{
A matrix with columns \verb{(sId, n0, s0, a0, ...)} where
\code{sId} is the state row id, \code{n0} the index of the stage at level 0, \code{s0} the index
of the state and \code{a0} the index of the action. If the HMDP has more
than one level columns index \verb{(d1, s1, a1, ...)} are added.
}
\description{
Info about the states in the HMDP model under consideration.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mdp.R
\name{runPolicyIteAve}
\alias{runPolicyIteAve}
\title{Perform policy iteration using the average expected-weight Bellman operator on the MDP.}
\usage{
runPolicyIteAve(
  mdp,
  w,
  dur,
  maxIte = 100,
  objective = c("max", "min"),
  getLog = TRUE
)
}
\arguments{
\item{mdp}{The MDP loaded using \code{\link[=loadMDP]{loadMDP()}}.}

\item{w}{The label of the weight we optimize.}

\item{dur}{The label of the duration/time such that discount rates can be calculated.}

\item{maxIte}{Max number of iterations. If the model does not satisfy the unichain assumption the algorithm may loop.}

\item{objective}{Optimize by maximizing (\code{"max"}) or minimizing (\code{"min"}) the Bellman value.}

\item{getLog}{Output the log messages.}
}
\value{
The optimal gain (g) calculated.
}
\description{
The policy can afterwards be received using functions \code{getPolicy} and \code{getPolicyW}.
}
\seealso{
\code{\link[=getPolicy]{getPolicy()}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/writer.R
\name{hmpMDPWriter}
\alias{hmpMDPWriter}
\title{Function for writing an HMDP model to a hmp file (XML). The function define
sub-functions which can be used to define an HMDP model stored in a hmp file.}
\usage{
hmpMDPWriter(
  file = "r.hmp",
  rate = 0.1,
  rateBase = 1,
  precision = 1e-05,
  desc = "HMP file created using hmpMDPWriter in R",
  getLog = TRUE
)
}
\arguments{
\item{file}{The name of the file storing the model (e.g. \code{r.hmp}).}

\item{rate}{The interest rate (used if consider discounting).}

\item{rateBase}{The time where the \code{rate} is taken over, e.g. if the \code{rate} is 0.1 and \code{rateBase} is 365 days
then we have an interest rate of 10 percent over the year.}

\item{precision}{The precision used when checking if probabilities sum to one.}

\item{desc}{Description of the model.}

\item{getLog}{Output log text.}
}
\value{
A list of functions.
}
\description{
HMP files are in XML format and human readable using e.g. a text editor.
HMP files are not suitable for storing large HMDP models since text files are very
verbose. Moreover, approximation of the weights and probabilities may occur since
the parser writing the hmp file may no output all digits. If you consider large
models then use the binary file format instead.
}
\details{
The returned writer exposes these functions:
\itemize{
\item \code{setWeights(labels, duration)}: sets the labels of the weights used in the
actions. \code{labels} is a vector of label names. \code{duration} identifies which
label corresponds to duration or time. For example, if the first entry in
\code{labels} is time, then \code{duration = 1}. Call this before building the model.
\item \code{setTransWeights(labels)}: sets the labels of transition-level weights.
\item \code{process()}: starts a (sub)process.
\item \code{endProcess()}: ends a (sub)process.
\item \code{stage(label = NULL)}: starts a stage.
\item \code{endStage()}: ends a stage.
\item \code{state(label = NULL)}: starts a state and returns the state index \code{sIdx}.
\item \code{endState()}: ends a state.
\item \code{action(label = NULL, weights, prob, statesNext = NULL, transWeights = NULL)}: starts an
action. \code{weights} must be a vector of action weights, and \code{prob} must
contain triples \verb{(scope, idx, pr)}. \code{scope} can take three values:
\itemize{
\item \code{0}: a transition to the next stage in the father process.
\item \code{1}: a transition to the next stage in the current process.
\item \code{2}: a transition to a child process, at stage zero in the child process.
}

The \code{idx} value denotes the index of the state at the stage considered. For
example, if \code{scope = 1} and \code{idx = 2}, the transition is to state number 3
at the next stage in the current process, counting from zero. \code{scope = 3}
is not supported in the \code{hmp} file format. \code{statesNext} is the number of
states in the next stage of the process and is only needed when there is a
transition to the father.
\item \code{endAction()}: ends an action.
\item \code{closeWriter()}: closes the writer. Call this when the model description is
finished.
}
}
\note{
Note all indexes are starting from zero (C/C++ style).
}
\examples{
## Use temp dir
wd <- setwd(tempdir())

## Create a small HMDP with two levels
w<-hmpMDPWriter()
w$setWeights(c("Duration","Net reward","Items"), duration=1)
w$process()
  w$stage()
   w$state(label="M0")
     w$action(label="A0",weights=c(0,0,0),prob=c(2,0,1))
      w$process()
        w$stage()
         w$state(label="D")
           w$action(label="A0",weights=c(0,0,1),prob=c(1,0,0.5,1,1,0.5))
           w$endAction()
         w$endState()
        w$endStage()
        w$stage()
         w$state(label="C0")
           w$action(label="A0",weights=c(0,0,0),prob=c(1,0,1))
           w$endAction()
           w$action(label="A1",weights=c(1,2,1),prob=c(1,0,0.5,1,1,0.5))
           w$endAction()
         w$endState()
         w$state(label="C1")
           w$action(label="A0",weights=c(0,0,0),prob=c(1,0,1))
           w$endAction()
           w$action(label="A1",weights=c(1,2,1),prob=c(1,0,0.5,1,1,0.5))
           w$endAction()
         w$endState()
        w$endStage()
        w$stage()
         w$state(label="C0")
           w$action(label="A0",weights=c(1,4,0),prob=c(0,0,1), statesNext=0)
           w$endAction()
         w$endState()
         w$state(label="C1")
           w$action(label="A0",weights=c(1,4,0),prob=c(0,0,1), statesNext=0)
           w$endAction()
         w$endState()
        w$endStage()
      w$endProcess()
     w$endAction()
     w$action(label="A1",weights=c(0,0,0),prob=c(2,0,1))
      w$process()
        w$stage()
         w$state(label="D")
           w$action(label="A0",weights=c(0,0,1),prob=c(1,0,1))
           w$endAction()
         w$endState()
        w$endStage()
        w$stage()
         w$state(label="C0")
           w$action(label="A0",weights=c(0,0,0),prob=c(1,0,1))
           w$endAction()
           w$action(label="A1",weights=c(1,2,1),prob=c(1,0,0.5,1,1,0.5))
           w$endAction()
         w$endState()
        w$endStage()
        w$stage()
         w$state(label="C0")
           w$action(label="A0",weights=c(1,4,0),prob=c(0,0,1), statesNext=0)
           w$endAction()
         w$endState()
         w$state(label="C1")
           w$action(label="A0",weights=c(1,4,0),prob=c(0,0,1), statesNext=0)
           w$endAction()
           w$action(label="A1",weights=c(0,10,5),prob=c(0,0,0.5,0,1,0.5), statesNext=0)
           w$endAction()
         w$endState()
        w$endStage()
      w$endProcess()
     w$endAction()
   w$endState()
   w$state(label="M1")
     w$action(label="A0",weights=c(0,0,0),prob=c(2,0,1))
      w$process()
        w$stage()
         w$state(label="D")
           w$action(label="A0",weights=c(0,0,1),prob=c(1,0,0.5,1,1,0.5))
           w$endAction()
         w$endState()
        w$endStage()
        w$stage()
         w$state(label="C0")
           w$action(label="A0",weights=c(0,0,0),prob=c(1,0,1))
           w$endAction()
         w$endState()
         w$state(label="C1")
           w$action(label="A0",weights=c(0,0,0),prob=c(1,0,1))
           w$endAction()
         w$endState()
        w$endStage()
        w$stage()
         w$state(label="C0")
           w$action(label="A0",weights=c(1,4,0),prob=c(0,0,1), statesNext=0)
           w$endAction()
         w$endState()
         w$state(label="C1")
           w$action(label="A0",weights=c(1,4,0),prob=c(0,0,1), statesNext=0)
           w$endAction()
         w$endState()
        w$endStage()
      w$endProcess()
     w$endAction()
   w$endState()
  w$endStage()
w$endProcess()
w$closeWriter()

## Have a look at the hmp file
cat(readr::read_file("r.hmp"))

## Reset working dir
setwd(wd)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/writer.R
\name{getBinInfoActions}
\alias{getBinInfoActions}
\title{Info about the actions in the HMDP model under consideration.}
\usage{
getBinInfoActions(
  prefix = "",
  labels = TRUE,
  fileA = "actionIdx.bin",
  filePr = "transProb.bin",
  fileW = "actionWeight.bin",
  fileLabelW = "actionWeightLbl.bin",
  fileLabelA = "actionIdxLbl.bin"
)
}
\arguments{
\item{prefix}{A character string with the prefix added to til binary files.}

\item{labels}{Should labels be extracted.}

\item{fileA}{The binary file containing the description of actions.}

\item{filePr}{The binary file containing the description of transition probabilities.}

\item{fileW}{The binary file containing the description of weights.}

\item{fileLabelW}{The binary file containing the weight labels.}

\item{fileLabelA}{The binary file containing the action labels.}
}
\value{
A data frame with the information. Scope string contain the scope of the transitions and
can be 4 values:
\itemize{
\item 0: A transition to the next stage in the father process,
\item 1: A transition to next stage in the current process,
\item 2: A transition to a child process (stage zero in the child process),
\item 3: A transition to the state with \code{sId = idx} is considered.
}

The index string denote the index (id is scope = 3) of the state at the next stage.
}
\description{
Info about the actions in the HMDP model under consideration.
}
\note{
The model don't have to be loaded, i.e only read the binary files. The state id (\code{sId}) will
not be the same as in the loaded model!
}
\examples{
## Use temp dir
wd <- setwd(tempdir())

# Create a small HMDP with two levels
w<-binaryMDPWriter()
w$setWeights(c("Duration","Net reward","Items"))
w$process()
   w$stage()
      w$state(label="M0")
         w$action(label="A0",weights=c(0,0,0),prob=c(2,0,1))
            w$process()
               w$stage()
                  w$state(label="D")
                     w$action(label="A0",weights=c(0,0,1),prob=c(1,0,0.5,1,1,0.5))
                     w$endAction()
                  w$endState()
               w$endStage()
               w$stage()
                  w$state(label="C0")
                     w$action(label="A0",weights=c(0,0,0),prob=c(1,0,1))
                     w$endAction()
                     w$action(label="A1",weights=c(1,2,1),prob=c(1,0,0.5,1,1,0.5))
                     w$endAction()
                  w$endState()
                  w$state(label="C1")
                     w$action(label="A0",weights=c(0,0,0),prob=c(1,0,1))
                     w$endAction()
                     w$action(label="A1",weights=c(1,2,1),prob=c(1,0,0.5,1,1,0.5))
                     w$endAction()
                  w$endState()
               w$endStage()
               w$stage()
                  w$state(label="C0")
                     w$action(label="A0",weights=c(1,4,0),prob=c(0,0,1))
                     w$endAction()
                  w$endState()
                  w$state(label="C1")
                     w$action(label="A0",weights=c(1,4,0),prob=c(0,0,1))
                     w$endAction()
                  w$endState()
               w$endStage()
            w$endProcess()
         w$endAction()
         w$action(label="A1",weights=c(0,0,0),prob=c(2,0,1))
            w$process()
               w$stage()
                  w$state(label="D")
                     w$action(label="A0",weights=c(0,0,1),prob=c(1,0,1))
                     w$endAction()
                  w$endState()
               w$endStage()
               w$stage()
                  w$state(label="C0")
                     w$action(label="A0",weights=c(0,0,0),prob=c(1,0,1))
                     w$endAction()
                     w$action(label="A1",weights=c(1,2,1),prob=c(1,0,0.5,1,1,0.5))
                     w$endAction()
                  w$endState()
               w$endStage()
               w$stage()
                  w$state(label="C0")
                     w$action(label="A0",weights=c(1,4,0),prob=c(0,0,1))
                     w$endAction()
                  w$endState()
                  w$state(label="C1")
                     w$action(label="A0",weights=c(1,4,0),prob=c(0,0,1))
                     w$endAction()
                     w$action(label="A1",weights=c(0,10,5),prob=c(0,0,0.5,0,1,0.5))
                     w$endAction()
                  w$endState()
               w$endStage()
            w$endProcess()
         w$endAction()
      w$endState()
      w$state(label="M1")
         w$action(label="A0",weights=c(0,0,0),prob=c(2,0,1))
            w$process()
               w$stage()
                  w$state(label="D")
                     w$action(label="A0",weights=c(0,0,1),prob=c(1,0,0.5,1,1,0.5))
                     w$endAction()
                  w$endState()
               w$endStage()
               w$stage()
                  w$state(label="C0")
                     w$action(label="A0",weights=c(0,0,0),prob=c(1,0,1))
                     w$endAction()
                  w$endState()
                  w$state(label="C1")
                     w$action(label="A0",weights=c(0,0,0),prob=c(1,0,1))
                     w$endAction()
                  w$endState()
               w$endStage()
               w$stage()
                  w$state(label="C0")
                     w$action(label="A0",weights=c(1,4,0),prob=c(0,0,1))
                     w$endAction()
                  w$endState()
                  w$state(label="C1")
                     w$action(label="A0",weights=c(1,4,0),prob=c(0,0,1))
                     w$endAction()
                  w$endState()
               w$endStage()
            w$endProcess()
         w$endAction()
      w$endState()
   w$endStage()
w$endProcess()
w$closeWriter()

## Info about the binary files (don't have to load the model first)
getBinInfoStates()
getBinInfoActions()

## reset working dir
setwd(wd)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/writer.R
\name{memoryMDPWriter}
\alias{memoryMDPWriter}
\title{Function for building an HMDP model directly in memory.}
\usage{
memoryMDPWriter(
  prefix = "",
  eps = 1e-05,
  check = TRUE,
  verbose = FALSE,
  getLog = TRUE
)
}
\arguments{
\item{prefix}{A character string kept for compatibility and stored in the
returned object metadata.}

\item{eps}{The sum of transition probabilities must at most differ \code{eps}
from one when \code{check = TRUE}.}

\item{check}{Check if the MDP seems correct before returning it.}

\item{verbose}{More output when compiling and running algorithms.}

\item{getLog}{Output the log messages.}
}
\value{
A list of functions. Calling \code{closeWriter()} returns an \code{"HMDP"}
object.
}
\description{
\code{memoryMDPWriter()} defines the same main sub-functions as
\code{\link[=binaryMDPWriter]{binaryMDPWriter()}}, but stores states and actions directly in C++ memory
instead of writing intermediate binary files. \code{closeWriter()} compiles the
model and returns the loaded \code{"HMDP"} object.
}
\details{
External or included processes are not supported by \code{memoryMDPWriter()}.
}
\note{
Note all indexes are starting from zero (C/C++ style).
}
\examples{
## Use temp dir
wd <- setwd(tempdir())

# Create a small HMDP with two levels
w<-memoryMDPWriter()
w$setWeights(c("Duration","Net reward","Items"))
w$process()
   w$stage()
      w$state(label="M0")
         w$action(label="A0",weights=c(0,0,0),prob=c(2,0,1))
            w$process()
               w$stage()
                  w$state(label="D")
                     w$action(label="A0",weights=c(0,0,1),prob=c(1,0,0.5,1,1,0.5))
                     w$endAction()
                  w$endState()
               w$endStage()
               w$stage()
                  w$state(label="C0")
                     w$action(label="A0",weights=c(0,0,0),prob=c(1,0,1))
                     w$endAction()
                     w$action(label="A1",weights=c(1,2,1),prob=c(1,0,0.5,1,1,0.5))
                     w$endAction()
                  w$endState()
                  w$state(label="C1")
                     w$action(label="A0",weights=c(0,0,0),prob=c(1,0,1))
                     w$endAction()
                     w$action(label="A1",weights=c(1,2,1),prob=c(1,0,0.5,1,1,0.5))
                     w$endAction()
                  w$endState()
               w$endStage()
               w$stage()
                  w$state(label="C0")
                     w$action(label="A0",weights=c(1,4,0),prob=c(0,0,1))
                     w$endAction()
                  w$endState()
                  w$state(label="C1")
                     w$action(label="A0",weights=c(1,4,0),prob=c(0,0,1))
                     w$endAction()
                  w$endState()
               w$endStage()
            w$endProcess()
         w$endAction()
         w$action(label="A1",weights=c(0,0,0),prob=c(2,0,1))
            w$process()
               w$stage()
                  w$state(label="D")
                     w$action(label="A0",weights=c(0,0,1),prob=c(1,0,1))
                     w$endAction()
                  w$endState()
               w$endStage()
               w$stage()
                  w$state(label="C0")
                     w$action(label="A0",weights=c(0,0,0),prob=c(1,0,1))
                     w$endAction()
                     w$action(label="A1",weights=c(1,2,1),prob=c(1,0,0.5,1,1,0.5))
                     w$endAction()
                  w$endState()
               w$endStage()
               w$stage()
                  w$state(label="C0")
                     w$action(label="A0",weights=c(1,4,0),prob=c(0,0,1))
                     w$endAction()
                  w$endState()
                  w$state(label="C1")
                     w$action(label="A0",weights=c(1,4,0),prob=c(0,0,1))
                     w$endAction()
                     w$action(label="A1",weights=c(0,10,5),prob=c(0,0,0.5,0,1,0.5))
                     w$endAction()
                  w$endState()
               w$endStage()
            w$endProcess()
         w$endAction()
      w$endState()
      w$state(label="M1")
         w$action(label="A0",weights=c(0,0,0),prob=c(2,0,1))
            w$process()
               w$stage()
                  w$state(label="D")
                     w$action(label="A0",weights=c(0,0,1),prob=c(1,0,0.5,1,1,0.5))
                     w$endAction()
                  w$endState()
               w$endStage()
               w$stage()
                  w$state(label="C0")
                     w$action(label="A0",weights=c(0,0,0),prob=c(1,0,1))
                     w$endAction()
                  w$endState()
                  w$state(label="C1")
                     w$action(label="A0",weights=c(0,0,0),prob=c(1,0,1))
                     w$endAction()
                  w$endState()
               w$endStage()
               w$stage()
                  w$state(label="C0")
                     w$action(label="A0",weights=c(1,4,0),prob=c(0,0,1))
                     w$endAction()
                  w$endState()
                  w$state(label="C1")
                     w$action(label="A0",weights=c(1,4,0),prob=c(0,0,1))
                     w$endAction()
                  w$endState()
               w$endStage()
            w$endProcess()
         w$endAction()
      w$endState()
   w$endStage()
w$endProcess()
w$closeWriter()

## Info about the binary files (don't have to load the model first)
if (FALSE) {
   getBinInfoStates()
   getBinInfoActions()
}

## reset working dir
setwd(wd)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/convert.R
\name{transProbMat}
\alias{transProbMat}
\title{Info about the transition probabilities in the HMDP model under consideration.}
\usage{
transProbMat(prefix = "", file = "transProb.bin")
}
\arguments{
\item{prefix}{A character string with the prefix added to til file(s).}

\item{file}{The HMDP binary file containing the description under consideration.}
}
\value{
A matrix with columns (\code{aId}, ...) where
\code{aId} is the action row id and ... are the probabilities of the action.
}
\description{
Info about the transition probabilities in the HMDP model under consideration.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mdp.R
\name{getSteadyStatePr}
\alias{getSteadyStatePr}
\title{Calculate the steady state transition probabilities for the founder process (level 0).}
\usage{
getSteadyStatePr(mdp, getLog = FALSE)
}
\arguments{
\item{mdp}{The MDP loaded using \code{\link[=loadMDP]{loadMDP()}}.}

\item{getLog}{Output log text.}
}
\value{
A vector with steady state probabilities for all the states at the founder level.
}
\description{
Assume that we consider an ergodic/irreducible time-homogeneous Markov chain specified using a policy in the MDP.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mdp.R
\name{runPolicyIteDiscount}
\alias{runPolicyIteDiscount}
\title{Perform policy iteration using the discounted expected-weight Bellman operator on the MDP.}
\usage{
runPolicyIteDiscount(
  mdp,
  w,
  dur,
  rate = 0,
  rateBase = 1,
  discountFactor = NULL,
  maxIte = 100,
  discountMethod = "continuous",
  objective = c("max", "min"),
  getLog = TRUE
)
}
\arguments{
\item{mdp}{The MDP loaded using \code{\link[=loadMDP]{loadMDP()}}.}

\item{w}{The label of the weight we optimize.}

\item{dur}{The label of the duration/time such that discount rates can be calculated.}

\item{rate}{The interest rate.}

\item{rateBase}{The time-horizon the rate is valid over.}

\item{discountFactor}{The discount rate for one time unit. If specified \code{rate} and \code{rateBase} are not used to calculate the discount rate.}

\item{maxIte}{Max number of iterations. If the model does not satisfy the unichain assumption the algorithm may loop.}

\item{discountMethod}{Either 'continuous' or 'discrete', corresponding to discount factor \code{exp(-rate/rateBase)} or \code{1/(1 + rate/rateBase)}, respectively. Only used if \code{discountFactor} is \code{NULL}.}

\item{objective}{Optimize by maximizing (\code{"max"}) or minimizing (\code{"min"}) the Bellman value.}

\item{getLog}{Output the log messages.}
}
\value{
Nothing.
}
\description{
The policy can afterwards be received using functions \code{getPolicy} and \code{getPolicyW}.
}
\seealso{
\code{\link[=getPolicy]{getPolicy()}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.R
\name{plotHypergraph}
\alias{plotHypergraph}
\title{Plot parts of the state expanded hypergraph.}
\usage{
plotHypergraph(
  hgf,
  gridDim,
  showGrid = FALSE,
  radx = 0.03,
  rady = 0.05,
  cex = 1,
  marX = 0.035,
  marY = 0.15,
  drawBorder = FALSE,
  actionOffset = 0.025,
  transLabels = "none",
  transLabelCex = 0.8 * cex,
  transLabelAdj = c(0.5, -0.6),
  stateLabel = "label",
  actionLabel = "label",
  actionWLabel = "none",
  actionColor = c("", "label", "policy"),
  actionsVisible = c("all", "policy"),
  connectedTo = NULL,
  recalcGrid = FALSE,
  mdp = NULL,
  ...
)
}
\arguments{
\item{hgf}{A list with the hypergraph containing two data frames, normally found using
\code{\link[=getHypergraph]{getHypergraph()}}. The data frame \code{nodes} must have columns: \code{sId} (state id), \code{gId} (grid id)
and \code{label} (node label). The data frame \code{hyperarcs} must have columns \code{sId} (head node),
\code{trans} (a list-column of tail node ids), \code{pr} (a list-column of transition
probabilities), \code{actionWeights} (a list-column of action weights),
\code{transWeights} (a list-column of transition-by-weight matrices), \code{aIdx}
(action index), \code{label} (action label), \code{lwd} (hyperarc line width), \code{lty}
(hyperarc line type) and \code{col} (hyperarc color).}

\item{gridDim}{A 2-dim vector (rows, cols) representing the size of the grid.}

\item{showGrid}{If true show the grid points (good for debugging).}

\item{radx}{Horizontal radius of the box.}

\item{rady}{Vertical radius of the box.}

\item{cex}{Relative size of text.}

\item{marX}{Horizontal margin.}

\item{marY}{Vertical margin.}

\item{drawBorder}{If \code{TRUE}, draw a border around the plot region and report the
outside and inside padding (good for debugging).}

\item{actionOffset}{Distance used to separate actions with the same
start and trans states. Set to \code{0} to draw overlapping actions.}

\item{transLabels}{Transition-label mode. \code{"none"} draws no transition labels
(the default); \code{"custom"} draws values from an optional \code{transLabels}
list-column in \code{hgf$hyperarcs}; otherwise use a \code{|}-separated
combination of \code{"label"}, \code{"sId"}, \code{"prob"}, and \code{"weights"}, for example
\code{"prob|weights"}. The older \code{"state"} spelling is treated as \code{"label"}.}

\item{transLabelCex}{Relative size of transition-label text.}

\item{transLabelAdj}{Position adjustment passed to \code{textempty()} for transition
labels, drawn at the middle of each split-to-transition branch.}

\item{stateLabel}{What to plot in states. \code{"custom"} uses a \code{stateLabel}
column in \code{hgf$nodes}; otherwise use a \code{|}-separated combination of
\code{"label"} (state label, default), \code{"sId"} (state id), \code{"sIdx"}
(stage-based state index), and \code{"weight"} (optimal weight of the state).}

\item{actionLabel}{What to plot near the split. One of \code{"none"}, \code{"custom"}
(uses an \code{actionLabel} column in \code{hgf$hyperarcs}), or a \code{|}-separated
combination of \code{"label"} (action label, default) and \code{"aIdx"}.}

\item{actionWLabel}{What to plot from the start state to the split. One of
\code{"none"} (default), \code{"weight"}, or \code{"custom"} (uses an \code{actionWLabel}
column in \code{hgf$hyperarcs}).}

\item{actionColor}{Action coloring scheme. Default \code{""} uses black lines. \code{"label"} uses different colors based on the action labels. \code{"policy"} highlights the current policy.}

\item{actionsVisible}{Action visibility mode. \code{"all"} (default) shows all actions. \code{"policy"} only shows actions in the current policy.}

\item{connectedTo}{Optional vector of state ids. If supplied, plot only states
reachable from these states by following visible hyperarcs forward,
and trim hyperarcs and transition-level data to the remaining states.}

\item{recalcGrid}{If \code{TRUE} and \code{connectedTo} is supplied, recalculate the
grid for the visible nodes. Nodes keep their original columns, but visible
nodes within each column are placed consecutively from the top and the
number of grid rows is reduced to the maximum number of visible nodes in
any column.}

\item{mdp}{The MDP model. Required if \code{stateLabel} contains \code{"weight"},
\code{actionColor = "policy"}, or \code{actionsVisible = "policy"}.}

\item{...}{Graphical parameters passed to \code{textempty}.}
}
\value{
No return value (NULL invisible), called for side effects (plotting).
}
\description{
The plot is created based on a grid \verb{(rows, cols)}. Each grid point is numbered from bottom to
top and left to right (starting from 1), i.e. given grid point with coordinates \verb{(r, c)} (where
\verb{(1,1)} is the top left corner and \verb{(rows, cols)} is the bottom right corner) the grid id is `(c
\itemize{
\item 
\enumerate{
\item 
\itemize{
\item rows + r`. You must assign a node to the hypergraph to a grid point (see below).
}
}
}
}
\examples{
## Set working dir
wd <- setwd(system.file("models", package = "MDP2"))

#### A finite-horizon replacement problem ####
mdp<-loadMDP("machine1_")
plot(mdp)
plot(mdp, actionColor = "label")  # colors based on labels
plot(mdp, transLabels = "state")  # label transitions with target state labels
plot(mdp, transLabels = "prob")  # label transitions with transition probabilities
plot(mdp, actionColor = "label", stateLabel = "sId|label")  # state labels are 'sId | label'
plot(mdp, stateLabel = "sIdx|label", radx = 0.01)  # adjust radx in states
plot(mdp, stateLabel = "label", actionWLabel = "none", actionLabel = "label", 
     transLabels = "sId", radx = 0.01)

scrapValues <- c(30, 10, 5, 0)  # scrap values (the values of the 4 states at stage 4)
runValueIte(mdp, "Net reward" , termValues = scrapValues)
plot(mdp, actionColor = "policy")  # highlight optimal policy
plot(mdp, actionsVisible = "policy", stateLabel = "weight")  # show only optimal policy


#### An infinite-horizon maintenance problem ####
mdp<-loadMDP("hct611-1_")
plot(mdp)  # plot the first two stages
plot(mdp, actionColor = "label")  # colors based on labels
plot(mdp, actionColor = "label", stateLabel = "sId|label")  # state labels are 'sId | label'
runPolicyIteAve(mdp,"Net reward","Duration")
plot(mdp, actionColor = "policy")  # highlight optimal policy
plot(mdp, actionsVisible = "policy")  # show only optimal policy


#### An infinite-horizon hierarchical replacement problem ####
library(magrittr)
mdp<-loadMDP("cow_")
hgf <- getHypergraph(mdp)
# modify labels
dat <- hgf$nodes \%>\% 
   dplyr::mutate(label = dplyr::case_when(
      label == "Low yield" ~ "L",
      label == "Avg yield" ~ "A",
      label == "High yield" ~ "H",
      label == "Dummy" ~ "D",
      label == "Bad genetic level" ~ "Bad",
      label == "Avg genetic level" ~ "Avg",
      label == "Good genetic level" ~ "Good",
      TRUE ~ "Error"
   ))
# assign nodes to grid ids
dat$gId[1:3]<-85:87
dat$gId[43:45]<-1:3
getGId<-function(process,stage,state) {
   if (process==0) start=18
   if (process==1) start=22
   if (process==2) start=26
   return(start + 14 * stage + state)
}
idx<-43
for (process in 0:2)
   for (stage in 0:4)
      for (state in 0:2) {
         if (stage==0 & state>0) break
         idx<-idx-1
         #cat(idx,process,stage,state,getGId(process,stage,state),"\n")
         dat$gId[idx]<-getGId(process,stage,state)
      }
hgf$nodes <- dat
# modify labels
dat <- hgf$hyperarcs \%>\% 
   dplyr::mutate(label = dplyr::case_when(
      label == "Replace" ~ "R",
      label == "Keep" ~ "K",
      label == "Dummy" ~ "D",
      TRUE ~ "Error"
   ),
   col = dplyr::case_when(
      label == "R" ~ "deepskyblue3",
      label == "K" ~ "darkorange1",
      label == "D" ~ "black",
      TRUE ~ "Error"
   ),
   lwd = 0.5,
   label = ""
   ) 
hgf$hyperarcs <- dat
# plot hypergraph
oldpar <- par(mai = c(0, 0, 0, 0))
plotHypergraph(gridDim = c(14, 7), hgf, cex = 0.8, radx = 0.02, rady = 0.03)
par(oldpar)


## A simple finite-horizon MDP with action and transition weights
prefix <- file.path(tempdir(), "plot_transition_rewards_")
w <- binaryMDPWriter(prefix)
w$setWeights("Cost")
w$setTransWeights(c("Reward", "Disease"))
w$process()
   w$stage()
      w$state(label = "S1")
         w$action(
            label = "A1", weights = 2, id = c(1), pr = c(1),
            transWeights = c(20, 0.3), end = TRUE
         )
         w$action(
            label = "A2", weights = 1, id = c(0, 1), pr = c(0.3, 0.7),
            transWeights = c(25, 0.4, 15, 0.2), end = TRUE
         )
      w$endState()
   w$endStage()
   w$stage()
      w$state(label = "S2")
         w$action(
            label = "A3", weights = 3, id = c(0, 1, 2), pr = c(0.5, 0.3, 0.2),
            transWeights = c(0, 0.05, 12, 0.2, 30, 0.8), end = TRUE
         )
         w$action(
            label = "A4", weights = 2, id = c(1, 2), pr = c(0.6, 0.4),
            transWeights = c(22, 0.35, 27, 0.7), end = TRUE
         )
      w$endState()
      w$state(label = "S3")
         w$action(
            label = "A5", weights = 1, id = c(0, 1), pr = c(0.4, 0.6),
            transWeights = c(5, 0, 16, 0.25), end = TRUE
         )
         w$action(
            label = "A6", weights = 4, id = c(0, 1, 2), pr = c(0.1, 0.3, 0.6),
            transWeights = c(14, 0.15, 21, 0.45, 29, 1), end = TRUE
         )
      w$endState()
   w$endStage()
   w$stage()
      w$state(label = "S4", end = TRUE)
      w$state(label = "S5", end = TRUE)
      w$state(label = "S6", end = TRUE)
   w$endStage()
w$endProcess()
w$closeWriter()

mdp <- loadMDP(prefix, getLog = FALSE)
plot(mdp, actionColor = "label", transLabels = "weights", actionWLabel = "weight", 
     radx = 0.005, rady = 0.01)

## Reset working dir
setwd(wd)
}
\seealso{
\code{\link[=getHypergraph]{getHypergraph()}} and \code{\link[=plot.HMDP]{plot.HMDP()}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mdp.R
\name{getRPO}
\alias{getRPO}
\title{Calculate the retention pay-off (RPO) or opportunity cost for some states.}
\usage{
getRPO(
  mdp,
  w,
  iA,
  sId = ifelse(mdp$timeHorizon >= Inf, mdp$founderStatesLast + 1,
    1):ifelse(mdp$timeHorizon >= Inf, mdp$states + mdp$founderStatesLast, mdp$states) - 1,
  criterion = "expected",
  dur = "",
  rate = 0,
  rateBase = 1,
  discountFactor = NULL,
  g = 0,
  objective = c("max", "min"),
  discountMethod = "continuous",
  stateStr = TRUE
)
}
\arguments{
\item{mdp}{The MDP loaded using \code{\link[=loadMDP]{loadMDP()}}.}

\item{w}{The label of the weight we calculate RPO for.}

\item{iA}{The action index we calculate the RPO with respect to (same size as \code{sId}).}

\item{sId}{Vector of id's of the states we want to retrieve.}

\item{criterion}{The Bellman operator shortcut. If \code{expected} use expected weights, if \code{discount} use discounted expected weights, if \code{average} use average expected weights.}

\item{dur}{The label of the duration/time such that discount rates can be calculated.}

\item{rate}{The interest rate.}

\item{rateBase}{The time-horizon the rate is valid over.}

\item{discountFactor}{The discount rate for one time unit. If specified \code{rate} and \code{rateBase} are not used to calculate the discount rate.}

\item{g}{The optimal gain (g) calculated (used if \code{criterion = "average"}).}

\item{objective}{Optimize by maximizing (\code{"max"}) or minimizing (\code{"min"}) the Bellman value.}

\item{discountMethod}{Either 'continuous' or 'discrete', corresponding to discount factor \code{exp(-rate/rateBase)} or \code{1/(1 + rate/rateBase)}, respectively. Only used if \code{discountFactor} is \code{NULL}.}

\item{stateStr}{Output the state string.}
}
\value{
The RPO (matrix/data frame).
}
\description{
The RPO is defined as the difference between
the weight of the state when using action \code{iA} and the maximum
weight of the node when using another predecessor different from \code{iA}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mdp.R
\name{saveMDP}
\alias{saveMDP}
\title{Save the MDP to binary files}
\usage{
saveMDP(mdp, prefix = "", getLog = TRUE)
}
\arguments{
\item{mdp}{The MDP loaded using \code{\link[=loadMDP]{loadMDP()}}.}

\item{prefix}{A character string with the prefix added to \code{binNames}. Used to identify a specific model.}

\item{getLog}{Output the log as a message.}
}
\value{
???
}
\description{
Currently do not save external files.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/writer.R
\name{getBinInfoStates}
\alias{getBinInfoStates}
\title{Info about the states in the binary files of the HMDP model under consideration.}
\usage{
getBinInfoStates(
  prefix = "",
  labels = TRUE,
  stateStr = TRUE,
  fileS = "stateIdx.bin",
  labelS = "stateIdxLbl.bin"
)
}
\arguments{
\item{prefix}{A character string with the prefix added to til binary files.}

\item{labels}{Should labels be extracted.}

\item{stateStr}{Should state strings be extracted. If false then add columns (n0, s0, a0, ...)
where n0 the index of the stage at level 0, s0 the index of the state and a0 the index of the
action. If the HMDP has more than one level columns index (d1, s1, a1, ...) are added.}

\item{fileS}{The binary file containing the description of states.}

\item{labelS}{The binary file containing the state labels.}
}
\value{
A data frame with the information.
}
\description{
Info about the states in the binary files of the HMDP model under consideration.
}
\note{
The model don't have to be loaded, i.e only read the binary files. The state id (\code{sId}) will
not be the same as in the loaded model!
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/convert.R
\name{weightNames}
\alias{weightNames}
\title{Names of weights used in actions.}
\usage{
weightNames(prefix = "", labels = "actionWeightLbl.bin")
}
\arguments{
\item{prefix}{A character string with the prefix added to the binary file names.}

\item{labels}{The HMDP binary file containing the weight labels.}
}
\value{
Vector of weight names.
}
\description{
Names of weights used in actions.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mdp.R
\name{checkWIdx}
\alias{checkWIdx}
\alias{.checkWIdx}
\title{Internal function. Check if the index of the weight is okay. Should not be used except you know what you are doing}
\usage{
.checkWIdx(iW, wLth)
}
\arguments{
\item{iW}{Index of the weight we want to optimize.}

\item{wLth}{Number of weights in the model.}
}
\value{
Nothing.
}
\description{
Internal function. Check if the index of the weight is okay. Should not be used except you know what you are doing
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/convert.R
\name{convertHMP2Binary}
\alias{convertHMP2Binary}
\title{Convert a HMDP model stored in a hmp (xml) file to binary file format.}
\usage{
convertHMP2Binary(file, prefix = "", getLog = TRUE)
}
\arguments{
\item{file}{The name of the HMP file (e.g. \code{r.hmp}).}

\item{prefix}{A character string with the prefix which will be added to the binary files.}

\item{getLog}{Output log text.}
}
\value{
NULL (invisible).
}
\description{
The function simply parse the hmp file and create binary files using
the \code{\link[=binaryMDPWriter]{binaryMDPWriter()}}.
}
\note{
Note all indexes are starting from zero (C/C++ style).
}
\examples{
## Set working dir
fDir <- system.file("models", package = "MDP2")
wd <- setwd(tempdir())
## Convert the machine example to a hmp file
prefix1 <- paste0(fDir,"/machine1_")
getBinInfoStates(prefix1)
convertBinary2HMP(prefix1, duration = NULL, out = "machine1_converted.hmp")
# have a look at the hmp file
cat(readr::read_file("machine1_converted.hmp"))

## Convert the machine example hmp file to binary files
convertHMP2Binary(file = paste0(fDir,"/machine1.hmp"), prefix = "machine_cov_")
getBinInfoStates(prefix = "machine_cov_")
## Convert the machine example with a single dummy node to a hmp file
#convertBinary2HMP("machine2_")  # error since using scope = 3 not supported in hmp files

## Reset working dir
setwd(wd)
}
\seealso{
\code{\link[=binaryMDPWriter]{binaryMDPWriter()}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mdp.R
\name{runValueIte}
\alias{runValueIte}
\title{Perform value iteration on the MDP.}
\usage{
runValueIte(
  mdp,
  w,
  dur = NULL,
  rate = 0,
  rateBase = 1,
  discountFactor = NULL,
  maxIte = 100,
  eps = 1e-05,
  termValues = NULL,
  g = NULL,
  objective = c("max", "min"),
  bellmanOp = c("auto", "expected", "discount", "average", "min", "max", "secondMoment"),
  getLog = TRUE,
  discountMethod = "continuous"
)
}
\arguments{
\item{mdp}{The MDP loaded using \code{\link[=loadMDP]{loadMDP()}}.}

\item{w}{The label of the weight we optimize.}

\item{dur}{The label of the duration/time such that discount rates can be calculated.}

\item{rate}{Interest rate.}

\item{rateBase}{The time-horizon the rate is valid over.}

\item{discountFactor}{The discount rate for one time unit. If specified \code{rate} and \code{rateBase} are not used to calculate the discount rate.}

\item{maxIte}{The max number of iterations value iteration is performed.}

\item{eps}{Stopping tolerance. If $max(w(t)-w(t+1)) < \code{eps}$ then stop the algorithm, i.e the policy becomes epsilon optimal (see Puterman p161).}

\item{termValues}{The terminal values used (values of the last stage in the MDP).}

\item{g}{Average weight. If specified then do a single iteration using the update equations under the average expected-weight Bellman operator with the specified g value.}

\item{objective}{Optimize by maximizing (\code{"max"}) or minimizing (\code{"min"}) the Bellman value.}

\item{bellmanOp}{Bellman operator. Use \code{"auto"} for existing behavior, \code{"min"} for the minimum-successor operator, \code{"max"} for the maximum-successor operator, or \code{"secondMoment"} for the second moment of total accumulated weight.}

\item{getLog}{Output the log messages.}

\item{discountMethod}{Either 'continuous' or 'discrete', corresponding to discount factor \code{exp(-rate/rateBase)} or \code{1/(1 + rate/rateBase)}, respectively. Only used if \code{discountFactor} is \code{NULL}.}
}
\value{
NULL (invisible)
}
\description{
If the MDP has a finite time-horizon then arguments \code{times} and \code{eps}
are ignored.
}
\examples{
## Set working dir
wd <- setwd(tempdir())

# Create the small machine repleacement problem used as an example in L.R. Nielsen and A.R.
# Kristensen. Finding the K best policies in a finite-horizon Markov decision process. European
# Journal of Operational Research, 175(2):1164-1179, 2006. doi:10.1016/j.ejor.2005.06.011.

## Create the MDP using a dummy replacement node
prefix<-"machine1_"
w <- binaryMDPWriter(prefix)
w$setWeights(c("Net reward"))
w$process()
   w$stage()   # stage n=0
      w$state(label="Dummy")          # v=(0,0)
         w$action(label="buy", weights=-100, prob=c(1,0,0.7, 1,1,0.3), end=TRUE)
      w$endState()
   w$endStage()
   w$stage()   # stage n=1
      w$state(label="good")           # v=(1,0)
         w$action(label="mt", weights=55, prob=c(1,0,1), end=TRUE)
         w$action(label="nmt", weights=70, prob=c(1,0,0.6, 1,1,0.4), end=TRUE)
      w$endState()
      w$state(label="average")        # v=(1,1)
         w$action(label="mt", weights=40, prob=c(1,0,1), end=TRUE)
         w$action(label="nmt", weights=50, prob=c(1,1,0.6, 1,2,0.4), end=TRUE)
      w$endState()
   w$endStage()
   w$stage()   # stage n=2
      w$state(label="good")           # v=(2,0)
         w$action(label="mt", weights=55, prob=c(1,0,1), end=TRUE)
         w$action(label="nmt", weights=70, prob=c(1,0,0.5, 1,1,0.5), end=TRUE)
      w$endState()
      w$state(label="average")        # v=(2,1)
         w$action(label="mt", weights=40, prob=c(1,0,1), end=TRUE)
         w$action(label="nmt", weights=50, prob=c(1,1,0.5, 1,2,0.5), end=TRUE)
      w$endState()
      w$state(label="not working")    # v=(2,2)
         w$action(label="mt", weights=30, prob=c(1,0,1), end=TRUE)
         w$action(label="rep", weights=5, prob=c(1,3,1), end=TRUE)
      w$endState()
   w$endStage()
   w$stage()   # stage n=3
      w$state(label="good")           # v=(3,0)
         w$action(label="mt", weights=55, prob=c(1,0,1), end=TRUE)
         w$action(label="nmt", weights=70, prob=c(1,0,0.2, 1,1,0.8), end=TRUE)
      w$endState()
      w$state(label="average")        # v=(3,1)
         w$action(label="mt", weights=40, prob=c(1,0,1), end=TRUE)
         w$action(label="nmt", weights=50, prob=c(1,1,0.2, 1,2,0.8), end=TRUE)
      w$endState()
      w$state(label="not working")    # v=(3,2)
         w$action(label="mt", weights=30, prob=c(1,0,1), end=TRUE)
         w$action(label="rep", weights=5, prob=c(1,3,1), end=TRUE)
      w$endState()
      w$state(label="replaced")       # v=(3,3)
         w$action(label="Dummy", weights=0, prob=c(1,3,1), end=TRUE)
      w$endState()
   w$endStage()
   w$stage()   # stage n=4
      w$state(label="good", end=TRUE)        # v=(4,0)
      w$state(label="average", end=TRUE)     # v=(4,1)
      w$state(label="not working", end=TRUE) # v=(4,2)
      w$state(label="replaced", end=TRUE)    # v=(4,3)
   w$endStage()
w$endProcess()
w$closeWriter()

## Load the model into memory
mdp<-loadMDP(prefix)
mdp
plot(mdp)

getInfo(mdp, withList = FALSE)
getInfo(mdp, withList = FALSE, dfLevel = "action", asStringsActions = TRUE)
getInfo(mdp, withList = FALSE, dfLevel = "action", asStringsActions = FALSE)

## Perform value iteration
w<-"Net reward"             # label of the weight we want to optimize
scrapValues<-c(30,10,5,0)   # scrap values (the values of the 4 states at stage 4)
runValueIte(mdp, w, termValues=scrapValues)
getPolicy(mdp)     # optimal policy

## Calculate the weights of the policy always to maintain
library(magrittr)
policy <- getInfo(mdp, withList = FALSE, dfLevel = "action")$df \%>\% 
   dplyr::filter(label_action == "mt") \%>\% 
   dplyr::select(sId, aIdx)
setPolicy(mdp, policy)
runCalcWeights(mdp, w, termValues=scrapValues)
getPolicy(mdp)  



# The example given in L.R. Nielsen and A.R. Kristensen. Finding the K best
# policies in a finite-horizon Markov decision process. European Journal of
# Operational Research, 175(2):1164-1179, 2006. doi:10.1016/j.ejor.2005.06.011,
# does actually not have any dummy replacement node as in the MDP above. The same
# model can be created using a single dummy node at the end of the process.

## Create the MDP using a single dummy node
prefix<-"machine2_"
w <- binaryMDPWriter(prefix)
w$setWeights(c("Net reward"))
w$process()
   w$stage()   # stage n=0
      w$state(label="Dummy")          # v=(0,0)
         w$action(label="buy", weights=-100, prob=c(1,0,0.7, 1,1,0.3), end=TRUE)
      w$endState()
   w$endStage()
   w$stage()   # stage n=1
      w$state(label="good")           # v=(1,0)
         w$action(label="mt", weights=55, prob=c(1,0,1), end=TRUE)
         w$action(label="nmt", weights=70, prob=c(1,0,0.6, 1,1,0.4), end=TRUE)
      w$endState()
      w$state(label="average")        # v=(1,1)
         w$action(label="mt", weights=40, prob=c(1,0,1), end=TRUE)
         w$action(label="nmt", weights=50, prob=c(1,1,0.6, 1,2,0.4), end=TRUE)
      w$endState()
   w$endStage()
   w$stage()   # stage n=2
      w$state(label="good")           # v=(2,0)
         w$action(label="mt", weights=55, prob=c(1,0,1), end=TRUE)
         w$action(label="nmt", weights=70, prob=c(1,0,0.5, 1,1,0.5), end=TRUE)
      w$endState()
      w$state(label="average")        # v=(2,1)
         w$action(label="mt", weights=40, prob=c(1,0,1), end=TRUE)
         w$action(label="nmt", weights=50, prob=c(1,1,0.5, 1,2,0.5), end=TRUE)
      w$endState()
      w$state(label="not working")    # v=(2,2)
         w$action(label="mt", weights=30, prob=c(1,0,1), end=TRUE)
         w$action(label="rep", weights=5, prob=c(3,12,1), end=TRUE) # transition to sId=12 (Dummy)
      w$endState()
   w$endStage()
   w$stage()   # stage n=3
      w$state(label="good")           # v=(3,0)
         w$action(label="mt", weights=55, prob=c(1,0,1), end=TRUE)
         w$action(label="nmt", weights=70, prob=c(1,0,0.2, 1,1,0.8), end=TRUE)
      w$endState()
      w$state(label="average")        # v=(3,1)
         w$action(label="mt", weights=40, prob=c(1,0,1), end=TRUE)
         w$action(label="nmt", weights=50, prob=c(1,1,0.2, 1,2,0.8), end=TRUE)
      w$endState()
      w$state(label="not working")    # v=(3,2)
         w$action(label="mt", weights=30, prob=c(1,0,1), end=TRUE)
         w$action(label="rep", weights=5, prob=c(3,12,1), end=TRUE)
      w$endState()
   w$endStage()
   w$stage()   # stage n=4
      w$state(label="good")        # v=(4,0)
         w$action(label="rep", weights=30, prob=c(1,0,1), end=TRUE)
      w$endState()
      w$state(label="average")     # v=(4,1)
         w$action(label="rep", weights=10, prob=c(1,0,1), end=TRUE)
      w$endState()
      w$state(label="not working") # v=(4,2)
         w$action(label="rep", weights=5, prob=c(1,0,1), end=TRUE)
      w$endState()
   w$endStage()
   w$stage()   # stage n=5
      w$state(label="Dummy", end=TRUE)        # v=(5,0)
   w$endStage()
w$endProcess()
w$closeWriter()

## Have a look at the state-expanded hypergraph
mdp<-loadMDP(prefix)
mdp
plot(mdp)

getInfo(mdp, withList = FALSE)
getInfo(mdp, withList = FALSE, dfLevel = "action", asStringsActions = TRUE)
getInfo(mdp, withList = FALSE, dfLevel = "action", asStringsActions = FALSE)

## Perform value iteration
w<-"Net reward"             # label of the weight we want to optimize
runValueIte(mdp, w, termValues = 0)
getPolicy(mdp)     # optimal policy

## Calculate the weights of the policy always to maintain
library(magrittr)
policy <- getInfo(mdp, withList = FALSE, dfLevel = "action")$df \%>\% 
   dplyr::filter(label_action == "mt") \%>\% 
   dplyr::select(sId, aIdx)
setPolicy(mdp, policy)
runCalcWeights(mdp, w, termValues=scrapValues)
getPolicy(mdp)  


## Reset working dir
setwd(wd)
}
\references{
Puterman, M. Markov Decision Processes, Wiley-Interscience, 1994.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/randomHMDP.R
\name{randomHMDP}
\alias{randomHMDP}
\title{Generate a "random" HMDP stored in a set of binary files.}
\usage{
randomHMDP(
  prefix = "",
  levels = 3,
  timeHorizon = c(Inf, 3, 4),
  states = c(2, 4, 5),
  actions = c(1, 2),
  childProcessPr = 0.5,
  externalProcessPr = 0,
  rewards = c(0, 100),
  durations = c(1, 10),
  rewardName = "Reward",
  durationName = "Duration"
)
}
\arguments{
\item{prefix}{A character string with the prefix added to the file(s).}

\item{levels}{Maximum number of levels. Set \code{childProcessPr = 1} if want exact this number of levels.}

\item{timeHorizon}{The time horizon for each level (vector). For the founder the time-horizon can be Inf.}

\item{states}{Number of states at each stage at a given level (vector of length levels)}

\item{actions}{Min and max number of actions at a state.}

\item{childProcessPr}{Probability of creating a child process when define action.}

\item{externalProcessPr}{Probability of creating an external process given that we create a child process. Only works if levels>2 and and currently does not generate external processes which include external processes.}

\item{rewards}{Min and max reward used.}

\item{durations}{Min and max duration used.}

\item{rewardName}{Weight name used for reward.}

\item{durationName}{Weight name used for duration.}
}
\value{
The file prefix (character).
}
\description{
Generate a "random" HMDP stored in a set of binary files.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/writer.R
\name{binaryMDPWriter}
\alias{binaryMDPWriter}
\title{Function for writing an HMDP model to binary files. The function defines
sub-functions which can be used to define an HMDP model saved in a set of binary
files.}
\usage{
binaryMDPWriter(
  prefix = "",
  binNames = c("stateIdx.bin", "stateIdxLbl.bin", "actionIdx.bin", "actionIdxLbl.bin",
    "actionWeight.bin", "actionWeightLbl.bin", "transProb.bin", "externalProcesses.bin",
    "transWeight.bin", "transWeightLbl.bin"),
  getLog = TRUE
)
}
\arguments{
\item{prefix}{A character string with the prefix added to \code{binNames}.}

\item{binNames}{A character vector giving the names of the binary files storing the model.}

\item{getLog}{Output log text.}
}
\value{
A list of functions.
}
\description{
Binary files are efficient for storing large models. Compared to the HMP (XML)
format the binary files use less storage space and loads the model faster.
}
\details{
The returned writer exposes these functions:
\itemize{
\item \code{setWeights(labels, ...)}: sets the labels of the weights used in the actions.
\code{labels} is a vector of label names. \code{...} is currently ignored. Call this
before building the model.
\item \code{process()}: starts a (sub)process. It may also be used to specify a
traditional MDP using matrices in \code{MDPtoolbox} style. In that style, \code{P} is
a list of matrices, one per action, each of size \verb{$S x S$} where \verb{$S$} is
the number of states. Each used row must sum to one, or all entries in a
row must be zero if unused. \code{R} is a matrix of size \verb{$S x A$}, where \verb{$A$}
is the number of actions, and \code{D} is a matrix of size \verb{$S x A$} with
durations. If \code{D} is omitted, all durations are assumed to be 1.
\item \code{endProcess()}: ends a (sub)process.
\item \code{stage(label = NULL)}: starts a stage. \code{label} is currently unused in the
binary format.
\item \code{endStage()}: ends a stage.
\item \code{state(label = NULL)}: starts a state and returns, invisibly, the state id.
That id can later be referenced with scope 3.
\item \code{endState()}: ends a state.
\item \code{action(scope = NULL, id = NULL, pr = NULL, prob = NULL, weights, transWeights = NULL, label = NULL, end = FALSE, ...)}: starts an action.
\code{weights} must be a vector of action weights. \code{transWeights} must contain
transition weights ordered by transition, with all transition weight labels
for the first transition followed by all labels for the second transition,
and so on. Transition probabilities can be entered in two ways:
\enumerate{
\item \code{prob} contains triples \verb{(scope, id, pr)}.
\item \code{id} and \code{pr} are vectors of equal length. If \code{scope} is omitted, all
scopes default to 1.
}

See the description of \code{actionIdx.bin} below. If \code{end = TRUE}, calling
\code{endAction()} is not necessary. \code{...} is currently ignored.
\item \code{endAction()}: ends an action. Do not use this if \code{end = TRUE} was used
when the action was specified.
\item \code{includeProcess(prefix, label = NULL, weights, prob, termStates, transWeights = NULL)}: includes an external process. External processes are
loaded into memory only when needed, which helps with large models. \code{prefix}
is the external process prefix. \code{weights} must be a vector of action
weights, and \code{prob} must contain triples \verb{(scope, idx, pr)}; see the
description of \code{actionIdx.bin} below. \code{termStates} must specify the number
of states at the last stage in the external process. Inside an
\verb{includeProcess ... endIncludeProcess}
block, you must specify the father jump actions of the last stage in the
external process. The external process is represented by its first and last
stage together with its jump actions. The function returns, invisibly, the
state ids of the first stage in the external process, which can later be
referenced with scope 3.
\item \code{endIncludeProcess()}: ends an \code{includeProcess} block.
\item \code{closeWriter()}: closes the writer. Call this when the model description is
finished.
}

Ten binary files are created:
\itemize{
\item \code{stateIdx.bin}: integers defining all states in the format
\code{"n0 s0 -1 n0 s0 a0 n1 s1 -1 n0 s0 a0 n1 s1 a1 n2 s2 -1 n0 s0 ..."}.
Here \code{-1} indicates that a new state is considered.
\item \code{stateIdxLbl.bin}: character data in the format \verb{sIdx label sIdx label ...}.
Here \code{sIdx} corresponds to the index or line number in \code{stateIdxLbl.bin},
starting from 0. No delimiter is used.
\item \code{actionIdx.bin}: integers defining all actions in the format
\verb{sIdx scope idx scope idx scope idx -1 sIdx scope idx scope idx -1 sIdx scope -1 ...}.
\code{sIdx} corresponds to the index or line number in \code{stateIdx.bin}, starting
from 0. The following \verb{(scope, idx)} pairs indicate possible transitions.
Scope can take four values:
\itemize{
\item \code{2}: a transition to a child process, at stage zero in the child process.
\item \code{1}: a transition to the next stage in the current process.
\item \code{0}: a transition to the next stage in the father process.
\item \code{3}: a transition to a state specified by its state \code{sIdx}.
}

For example, if \code{scope = 1} and \code{idx = 2}, the transition is to state
number 3 at the next stage in the current process. If \code{scope = 3} and
\code{idx = 5}, the transition is to the state specified at line 6 in
\code{stateIdxLbl.bin}. This is useful when considering shared child processes.
\item \code{actionIdxLbl.bin}: character data in the format \verb{aIdx label aIdx label ...}.
Here \code{aIdx} corresponds to the index or line number in \code{actionIdx.bin},
starting from 0. No delimiter is used.
\item \code{actionWeight.bin}: doubles containing action weights in the format
\code{"c1 c2 c3 c1 c2 c3 ..."}, assuming three weights for each action.
\item \code{actionWeightLbl.bin}: character data containing the weight labels in the
format \verb{label1 label2 label3}, assuming three weights for each action.
\item \code{transProb.bin}: doubles containing transition probabilities defined in
\code{actionIdx.bin}. The format is \code{"p1 p2 p3 -1 p1 -1 p1 p2 -1 ..."}. Here
\code{-1} indicates that a new action is considered.
\item \code{externalProcesses.bin}: character data containing links to external
processes in the format \verb{stageStr prefix stageStr prefix ...}. Here
\code{stageStr} corresponds to the stage index, for example \verb{n0 s0 a0 n1}, of
the stage corresponding to the first stage in the external process, and
\code{prefix} is the external process prefix. No delimiter is used.
\item \code{transWeight.bin}: doubles containing transition weights in the format
\code{"t11 t12 t21 t22 -1 ..."}, assuming two transition weights for each
transition and two transitions in the first action.
\item \code{transWeightLbl.bin}: character data containing the transition weight
labels.
}
}
\note{
Note all indexes are starting from zero (C/C++ style).
}
\examples{
## Use temp dir
wd <- setwd(tempdir())

# Create a small HMDP with two levels
w<-binaryMDPWriter()
w$setWeights(c("Duration","Net reward","Items"))
w$process()
   w$stage()
      w$state(label="M0")
         w$action(label="A0",weights=c(0,0,0),prob=c(2,0,1))
            w$process()
               w$stage()
                  w$state(label="D")
                     w$action(label="A0",weights=c(0,0,1),prob=c(1,0,0.5,1,1,0.5))
                     w$endAction()
                  w$endState()
               w$endStage()
               w$stage()
                  w$state(label="C0")
                     w$action(label="A0",weights=c(0,0,0),prob=c(1,0,1))
                     w$endAction()
                     w$action(label="A1",weights=c(1,2,1),prob=c(1,0,0.5,1,1,0.5))
                     w$endAction()
                  w$endState()
                  w$state(label="C1")
                     w$action(label="A0",weights=c(0,0,0),prob=c(1,0,1))
                     w$endAction()
                     w$action(label="A1",weights=c(1,2,1),prob=c(1,0,0.5,1,1,0.5))
                     w$endAction()
                  w$endState()
               w$endStage()
               w$stage()
                  w$state(label="C0")
                     w$action(label="A0",weights=c(1,4,0),prob=c(0,0,1))
                     w$endAction()
                  w$endState()
                  w$state(label="C1")
                     w$action(label="A0",weights=c(1,4,0),prob=c(0,0,1))
                     w$endAction()
                  w$endState()
               w$endStage()
            w$endProcess()
         w$endAction()
         w$action(label="A1",weights=c(0,0,0),prob=c(2,0,1))
            w$process()
               w$stage()
                  w$state(label="D")
                     w$action(label="A0",weights=c(0,0,1),prob=c(1,0,1))
                     w$endAction()
                  w$endState()
               w$endStage()
               w$stage()
                  w$state(label="C0")
                     w$action(label="A0",weights=c(0,0,0),prob=c(1,0,1))
                     w$endAction()
                     w$action(label="A1",weights=c(1,2,1),prob=c(1,0,0.5,1,1,0.5))
                     w$endAction()
                  w$endState()
               w$endStage()
               w$stage()
                  w$state(label="C0")
                     w$action(label="A0",weights=c(1,4,0),prob=c(0,0,1))
                     w$endAction()
                  w$endState()
                  w$state(label="C1")
                     w$action(label="A0",weights=c(1,4,0),prob=c(0,0,1))
                     w$endAction()
                     w$action(label="A1",weights=c(0,10,5),prob=c(0,0,0.5,0,1,0.5))
                     w$endAction()
                  w$endState()
               w$endStage()
            w$endProcess()
         w$endAction()
      w$endState()
      w$state(label="M1")
         w$action(label="A0",weights=c(0,0,0),prob=c(2,0,1))
            w$process()
               w$stage()
                  w$state(label="D")
                     w$action(label="A0",weights=c(0,0,1),prob=c(1,0,0.5,1,1,0.5))
                     w$endAction()
                  w$endState()
               w$endStage()
               w$stage()
                  w$state(label="C0")
                     w$action(label="A0",weights=c(0,0,0),prob=c(1,0,1))
                     w$endAction()
                  w$endState()
                  w$state(label="C1")
                     w$action(label="A0",weights=c(0,0,0),prob=c(1,0,1))
                     w$endAction()
                  w$endState()
               w$endStage()
               w$stage()
                  w$state(label="C0")
                     w$action(label="A0",weights=c(1,4,0),prob=c(0,0,1))
                     w$endAction()
                  w$endState()
                  w$state(label="C1")
                     w$action(label="A0",weights=c(1,4,0),prob=c(0,0,1))
                     w$endAction()
                  w$endState()
               w$endStage()
            w$endProcess()
         w$endAction()
      w$endState()
   w$endStage()
w$endProcess()
w$closeWriter()

## Info about the binary files (don't have to load the model first)
getBinInfoStates()
getBinInfoActions()

## reset working dir
setwd(wd)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/convert.R
\name{actionIdxMat}
\alias{actionIdxMat}
\title{Info about the actions in the HMDP model under consideration.}
\usage{
actionIdxMat(prefix = "", file = "actionIdx.bin")
}
\arguments{
\item{prefix}{A character string with the prefix added to til file(s).}

\item{file}{The HMDP binary file containing the description under consideration.}
}
\value{
A matrix with columns (\code{aId}, ...) where
\code{aId} is the action row id and \code{...} are alternating pairs \verb{(scp, idx)}, one for each
possible transition where \code{scp} is the scope that can be 4 values:
2 - A transition to a child process (stage zero in the child process), 1 - A transition
to next stage in the current process, 0 - A transition to the next stage in the father
process. the idx in the pair denote the index of the state at the stage considered.
Finally, if scope equals 3 then a transition to the state with \code{sId = idx} is considered.
}
\description{
Info about the actions in the HMDP model under consideration.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/convert.R
\name{actionInfo}
\alias{actionInfo}
\title{Info about the actions in the HMDP model under consideration.}
\usage{
actionInfo(
  prefix = "",
  file = "actionIdx.bin",
  weightFile = "actionWeight.bin",
  transPrFile = "transProb.bin",
  labels = "actionIdxLbl.bin"
)
}
\arguments{
\item{prefix}{A character string with the prefix added to til file(s).}

\item{file}{The HMDP binary file containing the description under consideration.}

\item{weightFile}{The HMDP binary file containing the action costs.}

\item{transPrFile}{The HMDP binary file containing the transition probabilities.}

\item{labels}{The HMDP binary file containing the labels under consideration.}
}
\value{
A matrix with columns from \code{actionIdxMat},
\code{actionCostMat} and \code{transProbMat} if labels is NULL. If labels
not are NULL then a data frame are returned with a label column too.
}
\description{
Info about the actions in the HMDP model under consideration.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/convert.R
\name{actionWeightMat}
\alias{actionWeightMat}
\title{Info about the weights of the actions in the HMDP model under consideration.}
\usage{
actionWeightMat(
  prefix = "",
  file = "actionWeight.bin",
  labels = "actionWeightLbl.bin"
)
}
\arguments{
\item{prefix}{A character string with the prefix added to til file(s).}

\item{file}{The HMDP binary file containing the description under consideration.}

\item{labels}{The HMDP binary file containing the labels under consideration.}
}
\value{
A matrix with columns (\code{aId}, ...) where
\code{aId} is the action row id and ... are the weights of the action.
}
\description{
Info about the weights of the actions in the HMDP model under consideration.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/convert.R
\name{actionIdxDf}
\alias{actionIdxDf}
\title{Info about the actions in the HMDP model under consideration.}
\usage{
actionIdxDf(prefix = "", file = "actionIdx.bin", labels = "actionIdxLbl.bin")
}
\arguments{
\item{prefix}{A character string with the prefix added to til file(s).}

\item{file}{The HMDP binary file containing the description under consideration.}

\item{labels}{The HMDP binary file containing the labels under consideration.}
}
\value{
A data frame with the same columns as in
\code{actionIdxMat} plus another column containing the labels.
}
\description{
Info about the actions in the HMDP model under consideration.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/writer.R
\name{binaryActionWriter}
\alias{binaryActionWriter}
\title{Function for writing actions of a HMDP model to binary files. The function defines
sub-functions which can be used to define actions saved in a set of binary
files. It is assumed that the states have been defined using \code{binaryMDPWriter}
and that the id of the states is known (can be retrieved using e.g. \code{stateIdxDf}).}
\usage{
binaryActionWriter(
  prefix = "",
  binNames = c("actionIdx.bin", "actionIdxLbl.bin", "actionWeight.bin",
    "actionWeightLbl.bin", "transProb.bin", "transWeight.bin", "transWeightLbl.bin"),
  append = TRUE
)
}
\arguments{
\item{prefix}{A character string with the prefix added to \code{binNames}.}

\item{binNames}{A character vector of length 5 giving the names of the binary
files storing the model.}

\item{append}{Logical indicating whether should keep the currents actions (default - TRUE)
defined or delete them and start over (FALSE).}
}
\value{
A list of functions.
}
\description{
Binary files are efficient for storing large models. Compared to the HMP (XML)
format the binary files use less storage space and loading the model is faster.
}
\details{
The returned writer exposes these functions:
\itemize{
\item \code{setWeights(labels, ...)}: sets the labels of the weights used in the
actions. \code{labels} is a vector of label names. \code{...} is currently ignored.
Call this before building the model.
\item \code{addAction(label = NULL, sIdx, weights, prob, ...)}: adds an action. \code{sIdx}
is the id of the state defining the action. \code{weights} must be a vector of
action weights. \code{prob} is a matrix \verb{(sIdx, pr)} where the first column
contains the id of the transition state; see the description of
\code{actionIdx.bin} below, where scope is assumed to be 3. \code{...} is currently
ignored.
\item \code{endAction()}: ends an action.
\item \code{closeWriter()}: closes the writer. Call this when the model description is
finished.
}

Five binary files are created:
\itemize{
\item \code{actionIdx.bin}: integers defining all actions in the format
\verb{sIdx scope idx scope idx scope idx -1 sIdx scope idx scope idx -1 sIdx scope -1 ...}.
\code{sIdx} corresponds to the index or line number in \code{stateIdx.bin}, starting
from 0. The following \verb{(scope, idx)} pairs indicate possible transitions.
Scope can take four values:
\itemize{
\item \code{2}: a transition to a child process, at stage zero in the child process.
\item \code{1}: a transition to the next stage in the current process.
\item \code{0}: a transition to the next stage in the father process.
\item \code{3}: a transition to a state specified by its state \code{sIdx}.
}

For example, if \code{scope = 1} and \code{idx = 2}, the transition is to state
number 3 at the next stage in the current process. If \code{scope = 3} and
\code{idx = 5}, the transition is to the state specified at line 6 in
\code{stateIdxLbl.bin}. This is useful when considering shared child processes.
\item \code{actionIdxLbl.bin}: character data in the format \verb{aIdx label aIdx label ...}.
Here \code{aIdx} corresponds to the index or line number in \code{actionIdx.bin},
starting from 0. No delimiter is used.
\item \code{actionWeight.bin}: doubles containing action weights in the format
\code{"c1 c2 c3 c1 c2 c3 ..."}, assuming three weights for each action.
\item \code{actionWeightLbl.bin}: character data containing the weight labels in the
format \verb{label1 label2 label3}, assuming three weights for each action.
\item \code{transProb.bin}: doubles containing the transition probabilities defined in
\code{actionIdx.bin}. The format is \code{"p1 p2 p3 -1 p1 -1 p1 p2 -1 ..."}. Here
\code{-1} indicates that a new action is considered.
}
}
\note{
Note all indexes are starting from zero (C/C++ style).
}
\examples{
## Use temp dir
wd <- setwd(tempdir())

# Create a small HMDP with two levels
w<-binaryMDPWriter()
w$setWeights(c("Duration","Net reward","Items"))
w$process()
   w$stage()
      w$state(label="M0")
         w$action(label="A0",weights=c(0,0,0),prob=c(2,0,1))
            w$process()
               w$stage()
                  w$state(label="D")
                     w$action(label="A0",weights=c(0,0,1),prob=c(1,0,0.5,1,1,0.5))
                     w$endAction()
                  w$endState()
               w$endStage()
               w$stage()
                  w$state(label="C0")
                     w$action(label="A0",weights=c(0,0,0),prob=c(1,0,1))
                     w$endAction()
                     w$action(label="A1",weights=c(1,2,1),prob=c(1,0,0.5,1,1,0.5))
                     w$endAction()
                  w$endState()
                  w$state(label="C1")
                     w$action(label="A0",weights=c(0,0,0),prob=c(1,0,1))
                     w$endAction()
                     w$action(label="A1",weights=c(1,2,1),prob=c(1,0,0.5,1,1,0.5))
                     w$endAction()
                  w$endState()
               w$endStage()
               w$stage()
                  w$state(label="C0")
                     w$action(label="A0",weights=c(1,4,0),prob=c(0,0,1))
                     w$endAction()
                  w$endState()
                  w$state(label="C1")
                     w$action(label="A0",weights=c(1,4,0),prob=c(0,0,1))
                     w$endAction()
                  w$endState()
               w$endStage()
            w$endProcess()
         w$endAction()
         w$action(label="A1",weights=c(0,0,0),prob=c(2,0,1))
            w$process()
               w$stage()
                  w$state(label="D")
                     w$action(label="A0",weights=c(0,0,1),prob=c(1,0,1))
                     w$endAction()
                  w$endState()
               w$endStage()
               w$stage()
                  w$state(label="C0")
                     w$action(label="A0",weights=c(0,0,0),prob=c(1,0,1))
                     w$endAction()
                     w$action(label="A1",weights=c(1,2,1),prob=c(1,0,0.5,1,1,0.5))
                     w$endAction()
                  w$endState()
               w$endStage()
               w$stage()
                  w$state(label="C0")
                     w$action(label="A0",weights=c(1,4,0),prob=c(0,0,1))
                     w$endAction()
                  w$endState()
                  w$state(label="C1")
                     w$action(label="A0",weights=c(1,4,0),prob=c(0,0,1))
                     w$endAction()
                     w$action(label="A1",weights=c(0,10,5),prob=c(0,0,0.5,0,1,0.5))
                     w$endAction()
                  w$endState()
               w$endStage()
            w$endProcess()
         w$endAction()
      w$endState()
      w$state(label="M1")
         w$action(label="A0",weights=c(0,0,0),prob=c(2,0,1))
            w$process()
               w$stage()
                  w$state(label="D")
                     w$action(label="A0",weights=c(0,0,1),prob=c(1,0,0.5,1,1,0.5))
                     w$endAction()
                  w$endState()
               w$endStage()
               w$stage()
                  w$state(label="C0")
                     w$action(label="A0",weights=c(0,0,0),prob=c(1,0,1))
                     w$endAction()
                  w$endState()
                  w$state(label="C1")
                     w$action(label="A0",weights=c(0,0,0),prob=c(1,0,1))
                     w$endAction()
                  w$endState()
               w$endStage()
               w$stage()
                  w$state(label="C0")
                     w$action(label="A0",weights=c(1,4,0),prob=c(0,0,1))
                     w$endAction()
                  w$endState()
                  w$state(label="C1")
                     w$action(label="A0",weights=c(1,4,0),prob=c(0,0,1))
                     w$endAction()
                  w$endState()
               w$endStage()
            w$endProcess()
         w$endAction()
      w$endState()
   w$endStage()
w$endProcess()
w$closeWriter()

## Info about the binary files (don't have to load the model first)
getBinInfoStates()
getBinInfoActions()

## reset working dir
setwd(wd)
}
